/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/geometry/geometry.h,v 1.4 2001/12/08 23:30:31 wex Exp $
 *
 * $Log: geometry.h,v $
 * Revision 1.4  2001/12/08 23:30:31  wex
 * Pass viewport in for culling
 *
 * Revision 1.3  2001/11/12 09:08:17  wex
 * Added high level GEOMETRY structure
 *
 * Revision 1.2  2001/11/01 04:26:53  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.1  2001/10/26 15:48:30  wex
 * First working version
 *
 *
 */

#ifndef _GEOMETRY_H_
#define _GEOMETRY_H_

#include <Python.h>

#include <matrix.h>


#define GEOMETRY_ERR_NONE       0
#define GEOMETRY_ERR_MEMORY     1

/* contains the error code if a function returns an error */
extern int GEOMETRY_errno;

/* returns a user readable string for error messages */
char *GEOMETRY_error_string(void);


typedef struct vertex {
    PyObject_HEAD
    
    float p[3];
    float q[3];
    float u, v;
} VERTEX;

typedef struct subd SUBD;
typedef struct particle PARTICLE;
typedef struct patch PATCH;

typedef struct geometry GEOMETRY;


typedef struct grid {
    int ns, nt;
    double (*p)[3];
    double (*q)[3];
    double (*uv)[2];
} GRID;
    
typedef struct grid_info {
    GEOMETRY *geometry;                 /* either subd or vertex is valid */
    VERTEX *vertex;
    GRID *grid;
    void *user_data;
} GRID_INFO;

typedef int (*GRID_CAPTURE)(GRID_INFO *info);

VERTEX *VERTEX_create(double px, double py, double pz,
    double qx, double qy, double qz, double u, double v);
void VERTEX_initialize(VERTEX *vtx, double px, double py, double pz, double qx,
    double qy, double qz, double u, double v);
void VERTEX_destroy(VERTEX *v);
void VERTEX_print(VERTEX *v, FILE *fp, int flags);
void VERTEX_lerp(VERTEX *v0, VERTEX *v1, double t, VERTEX *dst);

PARTICLE *PARTICLE_create(void);
void PARTICLE_destroy(PARTICLE *particle);
void PARTICLE_point(PARTICLE *particle, VERTEX *v, double radius);
void PARTICLE_generate_grids(PARTICLE *particle, MATRIX *l2s, double area,
    GRID_CAPTURE *capture, void *user_data);

SUBD *SUBD_create(void);
void SUBD_destroy(SUBD *subd);
void SUBD_print(SUBD *subd, FILE *fp);
/* Edge weights go from 0=hard to 1=smooth */
void SUBD_tri(SUBD *subd, VERTEX *v0, VERTEX *v1, VERTEX *v2,
    double e0, double e1, double e2);
void SUBD_quad(SUBD *subd, VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3,
    double e0, double e1, double e2, double e3);

int SUBD_generate_grids(SUBD *subd, MATRIX *l2s, int viewport[4], double area,
    GRID_CAPTURE capture, void *user_data);

PATCH *PATCH_create(
    VERTEX *v00, VERTEX *v01, VERTEX *v02, VERTEX *v03,
    VERTEX *v10, VERTEX *v11, VERTEX *v12, VERTEX *v13,
    VERTEX *v20, VERTEX *v21, VERTEX *v22, VERTEX *v23,
    VERTEX *v30, VERTEX *v31, VERTEX *v32, VERTEX *v33);
void PATCH_destroy(PATCH *patch);
void PATCH_subdivide_u(PATCH *patch, double u, PATCH **lo, PATCH **hi);
void PATCH_subdivide_v(PATCH *patch, double v, PATCH **lo, PATCH **hi);
void PATCH_generate_grids(PATCH *patch, MATRIX *l2s, double area,
    GRID_CAPTURE capture, void *user_data);

void GRID_destroy(GRID *grid);
void GRID_get_local_bbox(GRID *grid, VECTOR **min, VECTOR **max);
void GRID_get_screen_bbox(GRID *grid, MATRIX *l2s, double *xmin, double *ymin,
    double *xmax, double *ymax);


void GEOMETRY_print(GEOMETRY *geometry, FILE *fd);
void GEOMETRY_split(GEOMETRY *geometry, int *n, GEOMETRY **child);
void GEOMETRY_generate_grids(GEOMETRY *geometry, MATRIX *l2s, int vp[4],
    double area, GRID_CAPTURE capture, void *user_data);


#endif /* _GEOMETRY_H_ */
