/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/geometry/geometry_glue.c,v 1.8 2002/01/03 05:39:13 wex Exp $
 *
 * $Log: geometry_glue.c,v $
 * Revision 1.8  2002/01/03 05:39:13  wex
 * Updated to geometrycmodule for package format
 *
 * Revision 1.7  2001/12/20 21:19:13  wex
 * Support for Win32
 *
 * Revision 1.6  2001/12/05 10:35:32  wex
 * Changed displacement default value
 *
 * Revision 1.5  2001/12/05 10:25:13  wex
 * Added displacement to vertex routines, modifications for dicing.
 *
 * Revision 1.4  2001/11/16 09:22:41  wex
 * Stats
 *
 * Revision 1.3  2001/11/15 06:58:14  wex
 * Removed extra INCREFs.
 *
 * Revision 1.2  2001/11/12 09:08:17  wex
 * Added high level GEOMETRY structure
 *
 * Revision 1.1  2001/11/01 04:26:53  wex
 * Working with Python scene description and geometry shaders
 *
 *
 */

#include <float.h>

#include "geometry_internal.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isGeometry(m) ((m)->ob_type == &Geometry)
#define isVertex(m) ((m)->ob_type == &Vertex)


static PyObject *geometry_new(PyObject *self, PyObject *args);
static GEOMETRY *geometry_alloc(void);
static void geometry_dealloc(GEOMETRY *self);
static int geometry_print(GEOMETRY *self, FILE *fp, int flags);
static PyObject *geometry_getattr(GEOMETRY *self, char *name);
static PyObject *geometry_quad(GEOMETRY *self, PyObject *args);

static PyObject *vertex_new(PyObject *self, PyObject *args);
static VERTEX *vertex_alloc(double x, double y, double z, double u, double v,
    double mx, double my, double mz);
static void vertex_dealloc(VERTEX *self);
static int vertex_print(VERTEX *self, FILE *fp, int flags);
static PyObject *vertex_getattr(VERTEX *self, char *name);
static PyObject *vertex_multiply(VERTEX *self, PyObject *args);


/** Type Descriptors **/

static struct PyMethodDef geometry_methods[] = {
    {"quad",  (PyCFunction)geometry_quad,  1, NULL},
    {NULL,            NULL}
};

static PyNumberMethods vertex_number = {
    (binaryfunc) NULL,                              /* nb_add       "x + y"  */
    (binaryfunc) NULL,                              /* nb_sub       "x - y"  */
    (binaryfunc) vertex_multiply,                   /* nb_multiply  "x * y"  */
    (binaryfunc) NULL,                              /* nb_divide    "x / y"  */
    (binaryfunc) NULL,                              /* nb_remainder "x % y"  */
    (binaryfunc) NULL,                              /* nb_divmod    "divmod" */
    (ternaryfunc) NULL,                             /* nb_power     "pow(x)" */
    (unaryfunc) NULL,                               /* nb_negative  "-x"     */
    (unaryfunc) NULL,                               /* nb_positive  "+x"     */
    (unaryfunc) NULL,                               /* nb_absolute  "abs(x)" */
    (inquiry) NULL,                                 /* nb_nonzero   "if x"   */
    (unaryfunc) NULL,                               /* nb_invert    "~x"     */
    (binaryfunc) NULL,                              /* nb_lshift    "x << y" */
    (binaryfunc) NULL,                              /* nb_rshift    "x >> y" */
    (binaryfunc) NULL,                              /* nb_and       "x & y"  */
    (binaryfunc) NULL,                              /* nb_xor       "x ^ y"  */
    (binaryfunc) NULL,                              /* nb_or        "x | y"  */
    (coercion) NULL,                                /* nb_coerce    (n/a)    */
    (unaryfunc) NULL,                               /* nb_int       "int(x)" */
    (unaryfunc) NULL,                               /* nb_long      "long(x)"*/
    (unaryfunc) NULL,                               /* nb_float     "float(x)*/
    (unaryfunc) NULL,                               /* nb_oct       "oct(x)" */
    (unaryfunc) NULL,                               /* nb_hex       "hex(x)" */
};

static PyTypeObject Geometry = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else    
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "geometry",                                     /* tp_name               */
    sizeof(GEOMETRY),                               /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)geometry_dealloc,                   /* tp_dealloc refcount=0 */
    (printfunc)geometry_print,                      /* tp_print   "print x"  */
    (getattrfunc)geometry_getattr,                  /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static PyTypeObject Vertex = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else    
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "vertex",                                       /* tp_name               */
    sizeof(VERTEX),                                 /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)vertex_dealloc,                     /* tp_dealloc refcount=0 */
    (printfunc)vertex_print,                        /* tp_print   "print x"  */
    (getattrfunc)vertex_getattr,                    /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    &vertex_number,                                 /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef geometrytype_methods[] = {
    {"Geometry", geometry_new, 1},
    {"Vertex", vertex_new, 1},
    {NULL, NULL}
};

int vertex_alloced = 0;
int vertex_destroyed = 0;
int vertex_max = 0;
int geometry_alloced = 0;
int geometry_destroyed = 0;
int geometry_max = 0;


void
initgeometrycmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32    
    Geometry.ob_type = &PyType_Type;
    Vertex.ob_type = &PyType_Type;
#endif    
    m = Py_InitModule("geometrycmodule", geometrytype_methods);
    d = PyModule_GetDict(m);                        /* with 'Geometry' func */
    ErrorObject = Py_BuildValue("s", "geometrycmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize geometry C module");
    }
}


/** Instance Methods **/
static PyObject *
geometry_quad(GEOMETRY *self, PyObject *args)
{
    VERTEX *v0, *v1, *v2, *v3;

    if (!PyArg_ParseTuple(args, "OOOO", &v0, &v1, &v2, &v3))
        fprintf(stderr, "quad required four vertex arguments");

    SUBD_quad(self->subd, v0, v1, v2, v3, 0, 0, 0,0);

    Py_INCREF(Py_None);
    
    return (PyObject *)Py_None;
}

        
/** Basic Type Methods **/

static PyObject *
geometry_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)geometry_alloc();
}

static GEOMETRY *
geometry_alloc(void)
{
    GEOMETRY *geometry;
    geometry_alloced++;
    if (geometry_alloced - geometry_destroyed > geometry_max) {
        geometry_max = geometry_alloced - geometry_destroyed;
    }
    geometry = PyObject_NEW(GEOMETRY, &Geometry);    /* malloc, init, incref */
    geometry_initialize(geometry);
    return geometry;
}

static void
geometry_dealloc(GEOMETRY *self)
{
    geometry_destroyed++;
    geometry_delete(self);
    PyMem_DEL(self);
}

static int
geometry_print(GEOMETRY *self, FILE *fp, int flags)
{
    GEOMETRY_print(self, fp);
    return 0;
}

static PyObject *
geometry_getattr(GEOMETRY *self, char *name)
{
#if 0    
    if (strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[s]", "data");
    } else if (strcmp(name, "l2w") == 0) {
        return Py_BuildValue("((dddd)(dddd)(dddd)(dddd))",
            self->m[0][0], self->m[0][1], self->m[0][2], self->m[0][3],
            self->m[1][0], self->m[1][1], self->m[1][2], self->m[1][3],
            self->m[2][0], self->m[2][1], self->m[2][2], self->m[2][3],
            self->m[3][0], self->m[3][1], self->m[3][2], self->m[3][3]);
    }
    else if (strcmp(name, "parent") == 0) {
        return Py_BuildValue("O", self->parent);
    } else if (strcmp(name, "child_count") == 0) {
        return Py_BuildValue("i", self->child_count);
    } else if (strcmp(name, "child") == 0) {
        return Py_BuildValue("???");
    }
#endif

    return Py_FindMethod(geometry_methods, (PyObject *)self, name);
}


/*
 *
 * Vertex
 *
 */


/** Basic Type Methods **/

static PyObject *
vertex_new(PyObject *self, PyObject *args)
{
    double x, y, z, u, v, mx, my, mz;

    mx = my = mz = -DBL_MAX;

    if (!PyArg_ParseTuple(args, "(ddd)(dd)|(ddd)", &x, &y, &z, &u, &v,
            &mx, &my, &mz))
        onError("You must pass a position 3-tuple and a uv 2-tuple when "
            "creating a vertex, with an optional 3-tuple for the motion "
            "position");

    if (mx == DBL_MAX || mx < -99999) {
        mx = x;
        my = y;
        mz = z;
    }
    
    return (PyObject *)vertex_alloc(x, y, z, u, v, mx, my, mz);
}

static VERTEX *
vertex_alloc(double x, double y, double z, double u, double v,
    double mx, double my, double mz)
{
    VERTEX *vtx;

    vertex_alloced++;
    if (vertex_alloced - vertex_destroyed > vertex_max) {
        vertex_max = vertex_alloced - vertex_destroyed;
    }
    
    vtx = PyObject_NEW(VERTEX, &Vertex);           /* malloc, init, incref */
    VERTEX_initialize(vtx, x,y,z, mx,my,mz, u,v);
    return vtx;
}

static void
vertex_dealloc(VERTEX *self)
{
    vertex_destroyed++;
    
    vertex_delete(self);
    PyMem_DEL(self);
}

static int
vertex_print(VERTEX *self, FILE *fp, int flags)
{
    VERTEX_print(self, fp, flags);
    return 0;
}

static PyObject *
vertex_getattr(VERTEX *self, char *name)
{
    if (strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[s]", "data");
    } else if (strcmp(name, "p") == 0) {
        return Py_BuildValue("(ddd)", self->p[0], self->p[1], self->p[2]);
    } else if (strcmp(name, "q") == 0) {
        return Py_BuildValue("(ddd)", self->q[0], self->q[1], self->q[2]);
    } else if (strcmp(name, "uv") == 0) {
        return Py_BuildValue("(dd)", self->u, self->v);
    } else if (strcmp(name, "x") == 0) {
        return Py_BuildValue("d", self->p[0]);
    } else if (strcmp(name, "y") == 0) {
        return Py_BuildValue("d", self->p[1]);
    } else if (strcmp(name, "z") == 0) {
        return Py_BuildValue("d", self->p[2]);
    } else if (strcmp(name, "u") == 0) {
        return Py_BuildValue("d", self->u);
    } else if (strcmp(name, "v") == 0) {
        return Py_BuildValue("d", self->v);
    }

    Py_INCREF(Py_None);
    
    return Py_None;
}

static PyObject *
vertex_multiply(VERTEX *self, PyObject *args)
{
    Py_INCREF(Py_None);
    
    return Py_None;
}
