/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/geometry/subd.c,v 1.9 2001/12/16 00:24:13 wex Exp $
 *
 * $Log: subd.c,v $
 * Revision 1.9  2001/12/16 00:24:13  wex
 * Conservative clipping outside frustum. Could be improved, but safe.
 *
 * Revision 1.8  2001/12/08 23:30:31  wex
 * Pass viewport in for culling
 *
 * Revision 1.7  2001/12/05 10:25:13  wex
 * Added displacement to vertex routines, modifications for dicing.
 *
 * Revision 1.6  2001/11/13 11:44:57  wex
 * Slightly improved handling of camera frustum issues.
 *
 * Revision 1.5  2001/11/12 18:52:04  wex
 * Cleanup for Python1.5
 *
 * Revision 1.4  2001/11/12 09:08:17  wex
 * Added high level GEOMETRY structure
 *
 * Revision 1.3  2001/11/01 04:26:53  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:22  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:30  wex
 * First working version
 *
 *
 */
#include <math.h>

#include "geometry_internal.h"
#include "vertex.h"


static SUBD_FACE *subd_face_create(SUBD *subd, VERTEX *v0, VERTEX *v1,
    VERTEX *v2, VERTEX *v3, double e0, double e1, double e2, double e3);
static void subd_face_destroy(SUBD_FACE *face);
static int compute_subdivisions(double d0, double area);
static int subd_generate_grid(VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3,
    MATRIX *l2s, int vp[4], double area, GRID_CAPTURE capture, void *user_data,
    int lv);
static int should_subdivide(int n0, int n1, int level);



SUBD *
SUBD_create(void)
{
    SUBD *subd;

    subd = (SUBD *)malloc(sizeof(SUBD));
    if (subd == NULL) {
        GEOMETRY_errno = GEOMETRY_ERR_MEMORY;
        return NULL;
    }

    subd_initialize(subd);
    
    return subd;
}

void
subd_initialize(SUBD *subd)
{
    subd->face_count = 0;
    subd->face = NULL;
}

void
SUBD_destroy(SUBD *subd)
{
    subd_delete(subd);
    free(subd);
}

void
subd_delete(SUBD *subd)
{
    int i;
    
    for (i = 0; i < subd->face_count; i++) {
        subd_face_destroy(subd->face[i]);
        subd->face[i] = NULL;
    }
    
    if (subd->face != NULL) {
        free(subd->face);
        subd->face = NULL;
        subd->face_count = 0;
    }
}

void
SUBD_print(SUBD *subd, FILE *fp)
{
    fprintf(fp, "Subdivision geometry at %p with %d faces\n", subd,
        subd->face_count);
}

void
SUBD_tri(SUBD *subd, VERTEX *v0, VERTEX *v1, VERTEX *v2,
    double e0, double e1, double e2)
{
    Py_INCREF(v0);
    Py_INCREF(v1);
    Py_INCREF(v2);
    subd_face_create(subd, v0, v1, v2, NULL, e0, e1, e2, 0);
}

void
SUBD_quad(SUBD *subd, VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3,
    double e0, double e1, double e2, double e3)
{
    Py_INCREF(v0);
    Py_INCREF(v1);
    Py_INCREF(v2);
    Py_INCREF(v3);
    subd_face_create(subd, v0, v1, v2, v3, e0, e1, e2, e3);
}

/* null v3 implies tri */
static SUBD_FACE *
subd_face_create(SUBD *subd, VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3,
    double e0, double e1, double e2, double e3)
{
    int i;
    SUBD_FACE *face;

    face = (SUBD_FACE *)malloc(sizeof(SUBD_FACE));
    if (face == NULL) {
        GEOMETRY_errno = GEOMETRY_ERR_MEMORY;
        return NULL;
    }

    face->v0 = v0;
    face->v1 = v1;
    face->v2 = v2;
    face->v3 = v3;
    face->weight[0] = e0;
    face->weight[1] = e1;
    face->weight[2] = e2;
    face->weight[3] = e3;

    i = subd->face_count;
    subd->face = realloc(subd->face, (i+1)*sizeof(SUBD_FACE));
    subd->face[i] = face;
    subd->face_count = i+1;
    
    return face;
}

static void
subd_face_destroy(SUBD_FACE *face)
{
    Py_DECREF(face->v0);
    Py_DECREF(face->v1);
    Py_DECREF(face->v2);
    if (face->v3 != NULL) {
        Py_DECREF(face->v3);
    }
    free(face);
}

int
SUBD_generate_grids(SUBD *subd, MATRIX *l2s, int viewport[4], double area,
    GRID_CAPTURE capture, void *user_data)
{
    int i;
    SUBD_FACE *face;
    
    for (i = 0; i < subd->face_count; i++) {
        face = subd->face[i];
        if (subd_generate_grid(face->v0, face->v1, face->v2, face->v3,
                l2s, viewport, area, capture, user_data, 0) < 0) {
            return -1;
        }
    }

    return 0;
}

static int
compute_subdivisions(double d, double area)
{
    int n;

    if (d < 0) return -1;
    
    n = sqrt(d / area);
    if (n < 2) n = 2;

    return n;
}

/* Recursive grid generation */

static int
subd_generate_grid(VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3, MATRIX *l2s,
    int vp[4], double area, GRID_CAPTURE capture, void *user_data, int lv)
{
    int i, u, v, ns, nt;
    int ns0, ns1, nt0, nt1, subs, subt;
    double du, dv;
    double t[3], b[3];
    GRID *grid;
    GRID_INFO info;
    double d0, d1, d2, d3;
    VERTEX m, e0, e1, e2, e3;  /* recursive vertices stored on stack */
        

    if (vertex_get_edge_lengths_squared(v0,v1,v2,v3,l2s,vp,&d0,&d1,&d2,&d3)<0){
        return -1; 
    } else if (d0 < 0 || d1 < 0 || d2 < 0 || d3 < 0) {
        /* crosses screen or outside viewport, return minimal tessellation */
        /* note: would be faster if we could skip these, but 2D cull is fast */
        ns0 = ns1 = 2;
        nt0 = nt1 = 2;
        subs = subt = 0;
    } else {
        ns0 = compute_subdivisions(d0, area);
        ns1 = compute_subdivisions(d2, area);
        subs = should_subdivide(ns0, ns1, lv);
        if (!subs && (ns0 < 0 || ns1 < 0)) return -1;
    
        nt0 = compute_subdivisions(d1, area);
        nt1 = compute_subdivisions(d3, area);
        subt = should_subdivide(nt0, nt1, lv);
        if (!subt && (nt0 < 0 || nt1 < 0)) return -1;
    }
    
    if (subs && subt) {
        /* compute new edge and midpoint and subivide into 4 quads */
        VERTEX_lerp(v0, v1, 0.5, &e0);
        VERTEX_lerp(v1, v2, 0.5, &e1);
        VERTEX_lerp(v2, v3, 0.5, &e2);
        VERTEX_lerp(v3, v0, 0.5, &e3);
        VERTEX_lerp(&e0, &e2, 0.5, &m);
        subd_generate_grid(v0, &e0, &m,&e3,l2s,vp,area,capture,user_data,lv+1);
        subd_generate_grid(&e0, v1, &e1,&m,l2s,vp,area,capture,user_data,lv+1);
        subd_generate_grid(&m, &e1, v2,&e2,l2s,vp,area,capture,user_data,lv+1);
        subd_generate_grid(&e3, &m, &e2,v3,l2s,vp,area,capture,user_data,lv+1);
        return 0;
    } else if (subs) {
        /* split in s */
        VERTEX_lerp(v0, v1, 0.5, &e0);
        VERTEX_lerp(v2, v3, 0.5, &e2);
        subd_generate_grid(v0, &e0, &e2,v3,l2s,vp,area,capture,user_data,lv+1);
        subd_generate_grid(&e0, v1, v2,&e2,l2s,vp,area,capture,user_data,lv+1);
        return 0;
    } else if (subt) {
        /* split in t */
        VERTEX_lerp(v1, v2, 0.5, &e1);
        VERTEX_lerp(v3, v0, 0.5, &e3);
        subd_generate_grid(v0, v1, &e1,&e3,l2s,vp,area,capture,user_data,lv+1);
        subd_generate_grid(&e3, &e1, v2,v3,l2s,vp,area,capture,user_data,lv+1);
        return 0;
    }

    if (ns0 > 128) ns0 = 128;
    if (ns1 > 128) ns1 = 128;
    if (nt0 > 128) nt0 = 128;
    if (nt1 > 128) nt1 = 128;

    
/*    ns = 0.5 * (ns0 + ns1);
      nt = 0.5 * (nt0 + nt1); */
    if (ns0 > ns1) ns = ns0;
    else ns = ns1;
    
    if (nt0 > nt1) nt = nt0;
    else nt = nt1;
    
    grid = grid_create(ns, nt);
        
    for (v = 0, i = 0; v < grid->nt; v++) {
        dv = (float)v / (grid->nt - 1);

        for (u = 0; u < grid->ns; u++, i++) {
            du = (float)u / (grid->ns - 1);

            /* bi-linear interpolation: */

            /** Frame open position **/
            V3_LERP(v0->p, v1->p, du, b);
            if (v3 == NULL) {
                V3_COPY(v2->p, t);
            } else {
                V3_LERP(v3->p, v2->p, du, t);
            }
            V3_LERP(b, t, dv, grid->p[i]);

            /** Frame Close Position **/
            V3_LERP(v0->q, v1->q, du, b);
            if (v3 == NULL) {
                V3_COPY(v2->q, t);
            } else {
                V3_LERP(v3->q, v2->q, du, t);
            }
            V3_LERP(b, t, dv, grid->q[i]);
                
            /** Texture coordinates **/

            /* interpolate along top and bottom edges */
            b[0] = v0->u + du * (v1->u - v0->u);
            b[1] = v0->v + du * (v1->v - v0->v);

            if (v3 == NULL) {
                t[0] = v2->u;
                t[1] = v2->v;
            } else {
                t[0] = v3->u + du*(v2->u-v3->u);
                t[1] = v3->v + du*(v2->v-v3->v);
            }
            /* interpolate between top and bottom samples */
            grid->uv[i][0] = b[0] + dv * (t[0] - b[0]);
            grid->uv[i][1] = b[1] + dv * (t[1] - b[1]);
        }
    }

    info.grid = grid;
    info.user_data = user_data;
    capture(&info);

    GRID_destroy(grid);
    
    return 0;
}

static int
should_subdivide(int n0, int n1, int level)
{
    if (level > 8) return 0;
    if (n0 < 0 || n1 < 0) return 1;
    
    
    /* don't subdivide if the grid will be small enough */
    if (n0 < 16 && n1 < 16) return 0;

    /* subdivide if either side is too long */
    if (n0 > 16 || n1 > 16) {
        return 1;
    }
    
    return 0;
}
