/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/geometry/vertex.c,v 1.6 2001/12/16 00:24:13 wex Exp $
 *
 * $Log: vertex.c,v $
 * Revision 1.6  2001/12/16 00:24:13  wex
 * Conservative clipping outside frustum. Could be improved, but safe.
 *
 * Revision 1.5  2001/12/08 23:30:31  wex
 * Pass viewport in for culling
 *
 * Revision 1.4  2001/11/13 11:44:58  wex
 * Slightly improved handling of camera frustum issues.
 *
 * Revision 1.3  2001/11/12 09:08:17  wex
 * Added high level GEOMETRY structure
 *
 * Revision 1.2  2001/11/01 04:26:53  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.1  2001/10/26 15:48:30  wex
 * First working version
 *
 *
 */

#include <math.h>

#include "geometry.h"

void vertex_delete(VERTEX *v);


void
vertex_delete(VERTEX *v)
{
    /* do nothing since python deallocates space in glue */
}

void
VERTEX_initialize(VERTEX *vtx, double px, double py, double pz, double qx,
    double qy, double qz, double u, double v)
{
    vtx->p[0] = px;
    vtx->p[1] = py;
    vtx->p[2] = pz;
    
    vtx->q[0] = qx;
    vtx->q[1] = qy;
    vtx->q[2] = qz;

    vtx->u = u;
    vtx->v = v;
}

void
VERTEX_print(VERTEX *v, FILE *fp, int flags)
{
    fprintf(fp, "Vertex at %p: p(%g, %g, %g)  q(%g, %g, %g)  uv(%g, %g)\n",
        v, v->p[0], v->p[1], v->p[2], v->q[0], v->q[1], v->q[2], v->u, v->v);
}

int
vertex_get_edge_lengths_squared(VERTEX *v0, VERTEX *v1, VERTEX *v2, VERTEX *v3,
    MATRIX *l2s, int vp[4], double *e0, double *e1, double *e2, double *e3)
{
    VECTOR p, s0, s1, s2, s3;
    double dx, dy;

    /* compute screen space positions (could be shared between faces :( ) */
    V3_SET(p.v, v0->p[0], v0->p[1], v0->p[2]);
    if (VECTOR_project(&p, l2s, &s0) < 0) s0.v[2] = -fabs(s0.v[2]);
    V3_SET(p.v, v1->p[0], v1->p[1], v1->p[2]);
    if (VECTOR_project(&p, l2s, &s1) < 0) s1.v[2] = -fabs(s1.v[2]);
    V3_SET(p.v, v2->p[0], v2->p[1], v2->p[2]);
    if (VECTOR_project(&p, l2s, &s2) < 0) s2.v[2] = -fabs(s2.v[2]);
    V3_SET(p.v, v3->p[0], v3->p[1], v3->p[2]);
    if (VECTOR_project(&p, l2s, &s3) < 0) s3.v[2] = -fabs(s3.v[2]);

    /* outside viewport? (does not include motion position!) */
    if (s0.v[0]<vp[0] && s1.v[0]<vp[0] && s2.v[0] < vp[0] && s3.v[0] < vp[0]) {
        *e0 = -1;
        return 0;
    }
    if (s0.v[0]>vp[2] && s1.v[0]>vp[2] && s2.v[0] > vp[2] && s3.v[0] > vp[2]) {
        *e0 = -1;
        return 0;
    }
    if (s0.v[1]<vp[1] && s1.v[1]<vp[1] && s2.v[1] < vp[1] && s3.v[1] < vp[1]) {
        *e0 = -1;
        return 0;
    }
    if (s0.v[1]>vp[3] && s1.v[1]>vp[3] && s2.v[1] > vp[3] && s3.v[1] > vp[3]) {
        *e0 = -1;
        return 0;
    }
    
    /* behind camera? */
    if (s0.v[2] < -1 && s1.v[2] < -1 && s2.v[2] < -1 && s3.v[2] < -1) {
        return -1;
    }
    
    /* crosses camera? */
    if (s0.v[2] * s1.v[2] < 0) {
        *e0 = -1;
    } else {
        dx = s0.v[0] - s1.v[0];
        dy = s0.v[1] - s1.v[1];
        *e0 = dx*dx + dy*dy;
    }
    
    if (s1.v[2] * s2.v[2] < 0) {
        *e1 = -1;
    } else {
        dx = s1.v[0] - s2.v[0];
        dy = s1.v[1] - s2.v[1];
        *e1 = dx*dx + dy*dy;
    }
    
    if (s2.v[2] * s3.v[2] < 0) {
        *e2 = -1;
    } else {
        dx = s2.v[0] - s3.v[0];
        dy = s2.v[1] - s3.v[1];
        *e2 = dx*dx + dy*dy;
    }
    
    if (s3.v[2] * s0.v[2] < 0) {
        *e3 = -1;
    } else {
        dx = s3.v[0] - s0.v[0];
        dy = s3.v[1] - s0.v[1];
        *e3 = dx*dx + dy*dy;
    }

    return 0;
}

void
VERTEX_lerp(VERTEX *v0, VERTEX *v1, double t, VERTEX *dst)
{
    V3_LERP(v0->p, v1->p, t, dst->p);
    V3_LERP(v0->q, v1->q, t, dst->q);
    dst->u = v0->u + t * (v1->u - v0->u);
    dst->v = v0->v + t * (v1->v - v0->v);
}

    
