/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image.c,v 1.14 2002/01/07 22:23:01 wex Exp $
 *
 * $Log: image.c,v $
 * Revision 1.14  2002/01/07 22:23:01  wex
 * Added IMAGE_empty and IMAGE_remove_notify_client
 *
 * Revision 1.13  2002/01/03 05:20:53  wex
 * Updated to imagecmodule for package format
 *
 * Revision 1.12  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.11  2001/12/19 00:59:08  wex
 * Win32 support (EXPORT), added image_rgba.c and VC++ project files.
 *
 * Revision 1.10  2001/12/16 00:34:26  wex
 * Fixed bad free'ing of client and event lists
 *
 * Revision 1.9  2001/12/08 23:25:23  wex
 * Moved client_list and event_list into IMAGE structure
 *
 * Revision 1.8  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 * Revision 1.7  2001/11/13 07:52:03  wex
 * Memory fixes
 *
 * Revision 1.6  2001/11/12 09:09:49  wex
 * Fixed up memory access errors.
 *
 * Revision 1.5  2001/11/10 05:57:12  wex
 * Allow viewport to be reset.
 *
 * Revision 1.4  2001/11/01 04:26:55  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.3  2001/10/29 00:50:24  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.2  2001/10/26 15:48:36  wex
 * First working version
 *
 * Revision 1.1  2001/07/18 04:58:51  wex
 * Added
 *
 *
 */

#include <assert.h>
#include <pthread.h>

#include "image_internal.h"

IMAGE *
IMAGE_create(void)
{
    IMAGE *image;

    image = image_alloc();
    if (image == NULL) {
        IMAGE_errno = IMAGE_ERR_MEMORY;
        return NULL;
    }

    
    image_initialize(image);

    return image;
}

void
image_initialize(IMAGE *image)
{
    pthread_mutex_init(&image->modify_mutex, NULL);
    pthread_mutex_lock(&image->modify_mutex);
    
    image->tile = NULL;
    image->x_tile_count = 0;
    image->y_tile_count = 0;
    image->viewport[0] = 0;
    image->viewport[1] = 0;
    image->viewport[2] = -1;
    image->viewport[3] = -1;
    image->is_color_defined = 0;
    image->is_alpha_defined = 0;
    image->is_depth_defined = 0;
    image->read_format = NULL;

    image->client_list = LIST_create();
    image->event_list = LIST_create();
    
    pthread_mutex_init(&image->notify_mutex, NULL);
    
#if 0    
    w2c = NULL;
    c2s = NULL;
#endif    

    pthread_mutex_unlock(&image->modify_mutex);
}

void
image_delete(IMAGE *image)
{
    IMAGE_free(image);

    LIST_destroy(image->event_list);
    LIST_destroy(image->client_list);
#if 0    
    if (image->w2c != NULL) {
        MATRIX_destroy(image->w2c);
        image->w2c = NULL;
    }

    if (image->c2s != NULL) {
        MATRIX_destroy(image->c2s);
        image->c2s = NULL;
    }
#endif    
}

void
IMAGE_destroy(IMAGE *image)
{
    pthread_mutex_lock(&image->modify_mutex);
    Py_DECREF(image);
    pthread_mutex_unlock(&image->modify_mutex);
}

void
IMAGE_free(IMAGE *image)
{
    if (image->tile != NULL) {
        image_destroy_tiles(image);
    }
}

void
image_destroy_tiles(IMAGE *image)
{
    int i, n;

    n = image->x_tile_count * image->y_tile_count;
    
    for (i = 0; i < n; i++) {
        if (image->tile[i] != NULL) {
            image_tile_destroy(image->tile[i]);
            image->tile[i] = NULL;
        }
    }

    image->x_tile_count = 0;
    image->y_tile_count = 0;

    free(image->tile);
    image->tile = NULL;
}

int
IMAGE_get_width(IMAGE *image)
{
    int w;
    pthread_mutex_lock(&image->modify_mutex);
    w = IMAGE_WIDTH(image);
    pthread_mutex_unlock(&image->modify_mutex);
    return w;
}

int
IMAGE_get_height(IMAGE *image)
{
    int h;
    
    pthread_mutex_lock(&image->modify_mutex);
    h = IMAGE_HEIGHT(image);
    pthread_mutex_unlock(&image->modify_mutex);
    return h;
}
    
int
IMAGE_set_viewport(IMAGE *image, int x0, int y0, int x1, int y1)
{
    pthread_mutex_lock(&image->modify_mutex);
#if 0    
    if (image->viewport[0] != 0 || image->viewport[1] != 0 ||
        image->viewport[2] != -1 || image->viewport[3] != -1) {
        IMAGE_free(image);
    }
#endif
    
    image->viewport[0] = x0;
    image->viewport[1] = y0;
    image->viewport[2] = x1;
    image->viewport[3] = y1;

    image_compute_tiling(image);

    pthread_mutex_unlock(&image->modify_mutex);

    return 0;
}

void
IMAGE_get_viewport(IMAGE *image, int *x0, int *y0, int *x1, int *y1)
{
    pthread_mutex_lock(&image->modify_mutex);
    *x0 = image->viewport[0];
    *y0 = image->viewport[1];
    *x1 = image->viewport[2];
    *y1 = image->viewport[3];
    pthread_mutex_unlock(&image->modify_mutex);
}

int
IMAGE_set_color_defined(IMAGE *image, int state)
{
    int old_state;
    pthread_mutex_lock(&image->modify_mutex);
    old_state = image->is_color_defined;
    image->is_color_defined = state;
    pthread_mutex_unlock(&image->modify_mutex);
    return old_state;
}

int
IMAGE_set_alpha_defined(IMAGE *image, int state)
{
    int old_state;
    pthread_mutex_lock(&image->modify_mutex);
    old_state = image->is_alpha_defined;
    image->is_alpha_defined = state;
    pthread_mutex_unlock(&image->modify_mutex);
    return old_state;
}

int
IMAGE_set_depth_defined(IMAGE *image, int state)
{
    int old_state;
    pthread_mutex_lock(&image->modify_mutex);
    old_state = image->is_depth_defined;
    image->is_depth_defined = state;
    pthread_mutex_unlock(&image->modify_mutex);
    return old_state;
}

int
IMAGE_is_color_defined(IMAGE *image)
{
    int state;
    pthread_mutex_lock(&image->modify_mutex);
    state = image->is_color_defined;
    pthread_mutex_unlock(&image->modify_mutex);
    return state;
}

int
IMAGE_is_alpha_defined(IMAGE *image)
{
    int state;
    pthread_mutex_lock(&image->modify_mutex);
    state = image->is_alpha_defined;
    pthread_mutex_unlock(&image->modify_mutex);
    return state;
}

int
IMAGE_is_depth_defined(IMAGE *image)
{
    int state;
    pthread_mutex_lock(&image->modify_mutex);
    state = image->is_depth_defined;
    pthread_mutex_unlock(&image->modify_mutex);
    return state;
}

double
IMAGE_allocated_kilobytes(IMAGE *image)
{
    double tile_kilobytes, total_kilobytes;
    int i, n;
    
    pthread_mutex_lock(&image->modify_mutex);

    tile_kilobytes = IMAGE_TILE_BYTES / 1024.0;
    total_kilobytes = sizeof(IMAGE) / 1024.0;
    
    n = image->x_tile_count * image->y_tile_count;
    for (i = 0; i < n; i++) {
        if (image->tile[i] != NULL) {
            total_kilobytes += tile_kilobytes;
        }
    }

    pthread_mutex_unlock(&image->modify_mutex);
    return total_kilobytes;
}

void
IMAGE_incref(IMAGE *image)
{
    pthread_mutex_lock(&image->modify_mutex);
    Py_INCREF(image);
    pthread_mutex_unlock(&image->modify_mutex);
}

int
IMAGE_empty(IMAGE *image)
{
    int i, n;

    n = image->x_tile_count * image->y_tile_count;
    for (i = 0; i < n; i++) {
        if (image->tile[i] != NULL) {
            return 0;
        }
    }

    return 1;
}
