/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image.h,v 1.10 2002/01/07 22:23:01 wex Exp $
 *
 * $Log: image.h,v $
 * Revision 1.10  2002/01/07 22:23:01  wex
 * Added IMAGE_empty and IMAGE_remove_notify_client
 *
 * Revision 1.9  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.8  2001/12/19 00:59:08  wex
 * Win32 support (EXPORT), added image_rgba.c and VC++ project files.
 *
 * Revision 1.7  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 * Revision 1.6  2001/11/26 21:24:41  wex
 * A few extra checks in pixel routines, accumulate handles alpha.
 *
 * Revision 1.5  2001/11/12 09:09:49  wex
 * Fixed up memory access errors.
 *
 * Revision 1.4  2001/11/10 05:57:12  wex
 * Allow viewport to be reset.
 *
 * Revision 1.3  2001/11/01 04:26:55  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/26 15:48:36  wex
 * First working version
 *
 * Revision 1.1  2001/07/18 04:58:51  wex
 * Added
 *
 *
 */

#ifndef _IMAGE_H_
#define _IMAGE_H_

/* error return codes */
#define IMAGE_ERR_NONE          0
#define IMAGE_ERR_MEMORY        1
#define IMAGE_ERR_NO_CHANNEL    2
#define IMAGE_ERR_VIEWPORT_SET  3
#define IMAGE_ERR_READ_FORMAT   4
#define IMAGE_ERR_WRITE_FORMAT  5
#define IMAGE_ERR_FILE_NOT_OPEN 6
#define IMAGE_ERR_LIBC          7
#define IMAGE_ERR_PPM           8

/* contains the error code if a function returns an error */
extern int IMAGE_errno;

/* returns a user readable string for error messages */
char *IMAGE_error_string(void);

typedef struct image IMAGE;

void IMAGE_support_all(void);

IMAGE *IMAGE_create(void);
void IMAGE_incref(IMAGE *image);
void IMAGE_destroy(IMAGE *image);
void IMAGE_free(IMAGE *image);

int IMAGE_read(IMAGE *image, char *filename);  /* data read on demand */
int IMAGE_write(IMAGE *image, char *filename);

int IMAGE_check(IMAGE *image);  /* returns TRUE if valid image */
int IMAGE_empty(IMAGE *image);  /* returns TRUE if no valid data */

/* Call these functions when allocating your own image */
int IMAGE_set_viewport(IMAGE *image, int x0, int y0, int x1, int y1);
int IMAGE_set_color_defined(IMAGE *image, int state);
int IMAGE_set_alpha_defined(IMAGE *image, int state);
int IMAGE_set_depth_defined(IMAGE *image, int state);

int IMAGE_is_color_defined(IMAGE *image);
int IMAGE_is_alpha_defined(IMAGE *image);
int IMAGE_is_depth_defined(IMAGE *image);

int IMAGE_get_width(IMAGE *image);
int IMAGE_get_height(IMAGE *image);
void IMAGE_get_viewport(IMAGE *image, int *x0,int *y0, int *x1,int *y1);

/* no range checking is done in the get and set pixel operations, use this: */
int IMAGE_is_valid_pixel(IMAGE *image, int x, int y);

/* use this to wrap or mirror pixels to the viewport */
void IMAGE_get_viewport_wrapped_pixel(IMAGE *image,
    double sx, double sy, double *wx, double *wy);

/* creates a tile on-demand */
int IMAGE_set_color(IMAGE *image, int x, int y,
    double r, double g, double b);
int IMAGE_set_alpha(IMAGE *image, int x, int y, double a);
int IMAGE_set_depth(IMAGE *image, int x, int y, double z);

void IMAGE_get_color(IMAGE *image, int x,int y,
    double *r, double *g, double *b);
double IMAGE_get_alpha(IMAGE *image, int x,int y);
double IMAGE_get_depth(IMAGE *image, int x,int y);

void IMAGE_get_bilerp_color(IMAGE *image, double x, double y, double *r,
    double *g, double *b);

int IMAGE_accumulate_color(IMAGE *image, int x, int y,
    double r, double g, double b, double a);

/* sets the specified Z value if it is less than current value,
   returns 1 if updated, 0 if not, and -1 on error */
int IMAGE_update_depth(IMAGE *image, int x, int y, double z);

/* returns an RGBA buffer suitable for OpenGL. Caller must free buffer. */
unsigned long *IMAGE_create_rgba(IMAGE *image);

/* update a portion of the rgba buffer */
void IMAGE_rgba_update_viewport(IMAGE *image, unsigned long *rgba, 
    int x0, int y0, int x1, int y1);


/* set a callback to be called when notify_update_viewport is called */
typedef enum {
    IMAGE_NOTIFY_VIEWPORT,
    IMAGE_NOTIFY_START,
    IMAGE_NOTIFY_STOP
} IMAGE_NOTIFY_TYPE;

typedef struct image_notify_info {
    IMAGE *image;
    IMAGE_NOTIFY_TYPE type;
    int viewport[4];
} IMAGE_NOTIFY_INFO;

/* add a client interested in image notify events */
void IMAGE_add_notify_client(IMAGE *image, int id);
void IMAGE_remove_notify_client(IMAGE *image, int id);

/* force a call to all the notify callbacks */
void IMAGE_notify_update_viewport(IMAGE *image, int viewport[4]);
void IMAGE_notify_start_processing(IMAGE *image);
void IMAGE_notify_stop_processing(IMAGE *image);

/* retrieve an event from the queue and return in passed in info structure */
int IMAGE_notify_get_event(IMAGE *image, int id,
    IMAGE_NOTIFY_INFO *info);

#if 0
/* Used depth maps */
int IMAGE_set_camera(IMAGE *image, MATRIX *w2c, MATRIX *c2s);
void IMAGE_get_camera(IMAGE *image, MATRIX *w2c, MATRIX *c2s);
#endif

double IMAGE_allocated_kilobytes(IMAGE *image);

#endif /* _IMAGE_H_ */
