/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image_glue.c,v 1.10 2002/01/03 05:20:53 wex Exp $
 *
 * $Log: image_glue.c,v $
 * Revision 1.10  2002/01/03 05:20:53  wex
 * Updated to imagecmodule for package format
 *
 * Revision 1.9  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.8  2001/12/19 00:59:08  wex
 * Win32 support (EXPORT), added image_rgba.c and VC++ project files.
 *
 * Revision 1.7  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 * Revision 1.6  2001/11/26 21:24:41  wex
 * A few extra checks in pixel routines, accumulate handles alpha.
 *
 * Revision 1.5  2001/11/13 07:52:03  wex
 * Memory fixes
 *
 * Revision 1.4  2001/11/12 09:09:49  wex
 * Fixed up memory access errors.
 *
 * Revision 1.3  2001/11/06 06:55:38  wex
 * Set/get color, tiling for ppm
 *
 * Revision 1.2  2001/11/01 04:26:55  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.1  2001/07/18 04:58:51  wex
 * Added
 *
 *
 */

#include "image_internal.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isImage(m) ((m)->ob_type == &Image)


static PyObject *image_new(PyObject *self, PyObject *args);
static void image_dealloc(IMAGE *self);
static int image_print(IMAGE *self, FILE *fp, int flags);
static PyObject *image_getattr(IMAGE *self, char *name);
static PyObject *image_read(IMAGE *self, PyObject *args);
static PyObject *image_write(IMAGE *self, PyObject *args);
static PyObject *image_set_viewport(IMAGE *self, PyObject *args);
static PyObject *image_set_color(IMAGE *self, PyObject *args);
static PyObject *image_get_color(IMAGE *self, PyObject *args);


/** Type Descriptors **/

static struct PyMethodDef image_methods[] = {
    {"read",        (PyCFunction)image_read,                1, NULL},
    {"write",       (PyCFunction)image_write,               1, NULL},
    {"viewport",    (PyCFunction)image_set_viewport,        1, NULL},
    {"set_color",       (PyCFunction)image_set_color,       1, NULL},
    {"get_color",       (PyCFunction)image_get_color,       1, NULL},
    {NULL,          NULL}
};

static PyTypeObject Image = {
    /* type header */
#ifndef WIN32    
    PyObject_HEAD_INIT(&PyType_Type)
#else
    PyObject_HEAD_INIT(NULL)
#endif
    0,                                              /* ob_size               */
    "image",                                        /* tp_name               */
    sizeof(IMAGE),                                  /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)image_dealloc,                      /* tp_dealloc refcount=0 */
    (printfunc)image_print,                         /* tp_print   "print x"  */
    (getattrfunc)image_getattr,                     /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef imagetype_methods[] = {
    {"Image", image_new, 1},
    {NULL, NULL}
};


void
initimagecmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32
#ifdef FREETYPE_ENABLED    
    Image.ob_type = &PyType_Type;
#endif    
#endif    
    m = Py_InitModule("imagecmodule", imagetype_methods);/* make the module */
    d = PyModule_GetDict(m);                        /* with 'Image' func */
    ErrorObject = Py_BuildValue("s", "imagecmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize image C module");
    }
    IMAGE_support_all();
}

static PyObject *
image_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)image_alloc();
}

IMAGE *
image_alloc(void)
{
    IMAGE *im;
    im = PyObject_NEW(IMAGE, &Image);              /* malloc, init, incref */
    image_initialize(im);
    IMAGE_set_color_defined(im, 1);
    return im;
}

static void
image_dealloc(IMAGE *self)
{
    image_delete(self);
    PyMem_DEL(self);
}

static int
image_print(IMAGE *self, FILE *fp, int flags)
{
    fprintf(fp, "[%d, %d, %d, %d] %s%s%s%s\n",
        self->viewport[0], self->viewport[1],
        self->viewport[2], self->viewport[3],
        self->is_color_defined ? "color " : "",
        self->is_alpha_defined ? "alpha " : "",
        self->is_depth_defined ? "depth " : "",
        self->is_color_defined || self->is_alpha_defined ||
        self->is_depth_defined ? "" : "no data");
    return 0;
}

static PyObject *
image_getattr(IMAGE *self, char *name)
{
    if (strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[ssssss]", "width", "height", "origin",
            "has_color", "has_alpha", "has_depth");
    } else if (strcmp(name, "width") == 0) {
        return Py_BuildValue("d", IMAGE_WIDTH(self));
    } else if (strcmp(name, "height") == 0) {
        return Py_BuildValue("d", IMAGE_HEIGHT(self));
    } else if (strcmp(name, "origin") == 0) {
        return Py_BuildValue("(dd)", self->viewport[0], self->viewport[1]);
    } else if (strcmp(name, "has_color") == 0) {
        return Py_BuildValue("d", self->is_color_defined);
    } else if (strcmp(name, "has_alpha") == 0) {
        return Py_BuildValue("d", self->is_alpha_defined);
    } else if (strcmp(name, "has_depth") == 0) {
        return Py_BuildValue("d", self->is_depth_defined);
    }

    return Py_FindMethod(image_methods, (PyObject *)self, name);
}

static PyObject *
image_read(IMAGE *self, PyObject *args)
{
    int i;
    char *filename;
    
    if (!PyArg_ParseTuple(args, "s", &filename)) 
        onError("read requires a filename string");
        
    i = IMAGE_read(self, filename);
    
    return Py_BuildValue("i", i);
}

static PyObject *
image_write(IMAGE *self, PyObject *args)
{
    int i;
    char *filename;
    
    if (!PyArg_ParseTuple(args, "s", &filename)) 
        onError("write requires a filename string");
        
    i = IMAGE_write(self, filename);
    return Py_BuildValue("d", i);
}

static PyObject *
image_set_viewport(IMAGE *self, PyObject *args)
{
    int i, x0, y0, x1, y1;
    
    if (!PyArg_ParseTuple(args, "(iiii)", &x0, &y0, &x1, &y1)) 
        onError("set viewport requires a 4-element integer tuple");

    i = IMAGE_set_viewport(self, x0, y0, x1, y1);
    return Py_BuildValue("d", i);
}

static PyObject *
image_set_color(IMAGE *self, PyObject *args)
{
    int i, x, y;
    double r, g, b;
    
    if (!PyArg_ParseTuple(args, "(ii)(ddd)", &x, &y, &r, &g, &b)) 
        onError("set color requires (x y) (r g b)");

    i = IMAGE_set_color(self, x, y, r, g, b);
    return Py_BuildValue("d", i);
}

static PyObject *
image_get_color(IMAGE *self, PyObject *args)
{
    double sx, sy;
    double x, y;
    double r, g, b;
    
    if (!PyArg_ParseTuple(args, "dd", &x, &y)) 
        onError("get_color requires (x y)");

    sx = x * IMAGE_get_width(self);
    sy = y * IMAGE_get_height(self);

    IMAGE_get_viewport_wrapped_pixel(self, sx, sy, &sx, &sy);
/*    IMAGE_get_color(self, sx, sy, &r, &g, &b); */
    IMAGE_get_bilerp_color(self, sx, sy, &r, &g, &b);
        
    return Py_BuildValue("(ddd)", r, g, b);
}

int
IMAGE_check(IMAGE *image)
{
    return isImage(image);
}

