/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image_notify.c,v 1.6 2002/01/07 22:23:02 wex Exp $
 *
 * $Log: image_notify.c,v $
 * Revision 1.6  2002/01/07 22:23:02  wex
 * Added IMAGE_empty and IMAGE_remove_notify_client
 *
 * Revision 1.5  2002/01/03 05:20:53  wex
 * Updated to imagecmodule for package format
 *
 * Revision 1.4  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.3  2001/12/19 00:59:08  wex
 * Win32 support (EXPORT), added image_rgba.c and VC++ project files.
 *
 * Revision 1.2  2001/12/08 23:25:24  wex
 * Moved client_list and event_list into IMAGE structure
 *
 * Revision 1.1  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 *
 */

#include <assert.h>
#include <pthread.h>
#include <list.h>

#include "image_internal.h"

typedef struct image_notify_client {
    LIST_ELEMENT list_element;
    int id;
    void *user_data;
} IMAGE_NOTIFY_CLIENT;

typedef struct image_notify_event {
    LIST_ELEMENT list_element;
    int viewport[4];
    int *id;
    int client_count;
    IMAGE_NOTIFY_TYPE type;
}  IMAGE_NOTIFY_EVENT;

void
IMAGE_add_notify_client(IMAGE *image, int id)
{
    IMAGE_NOTIFY_CLIENT *client;

    assert(id != 0);

    pthread_mutex_lock(&image->notify_mutex);
    
    for (client = LIST_head(image->client_list); client != NULL;
         client = LIST_next(client)) {
        if (client->id == id) {
            fprintf(stderr, "Client %d already registered\n", id);
            goto finished;
        }
    }
        
    client = (IMAGE_NOTIFY_CLIENT *)malloc(sizeof(IMAGE_NOTIFY_CLIENT));
    assert(client != NULL);

    client->id = id;
    LIST_append(image->client_list, client);

 finished:
    
    pthread_mutex_unlock(&image->notify_mutex);
}

int
IMAGE_notify_event_exists(IMAGE *image)
{
    int status;
    
    pthread_mutex_lock(&image->notify_mutex);

    assert(image->event_list != NULL);
    status = LIST_count(image->event_list) > 0;

    pthread_mutex_unlock(&image->notify_mutex);

    return status;
}

int
IMAGE_notify_get_event(IMAGE *image, int id, IMAGE_NOTIFY_INFO *info)
{
    int i;
    IMAGE_NOTIFY_EVENT *event;

    pthread_mutex_lock(&image->notify_mutex);

    assert(image->event_list != NULL);
    
    if (LIST_count(image->event_list) == 0) {
        goto no_event;
    }

    event = LIST_head(image->event_list);
    do {
        for (i = 0; i < event->client_count; i++) {
            if (event->id[i] == id) {
                /* already reported this event to this client */
                break;
            }
        }

        if (i == event->client_count) {
            /* have not sent this event to this client */
            /* find an open slot (guaranteed) and fill it up */
            for (i = 0; i < event->client_count; i++) {
                if (event->id[i] == 0) {
                    event->id[i] = id;
                    break;
                }
            }
                
            break;
        }

        /* move on to the next event to see if this client has seen it yet */
        event = LIST_next(event);
        
    } while (event != NULL);

    if (event == NULL) {
        goto no_event;
    }
    
    /* check to see if all clients have seen this event */
    for (i = 0; i < event->client_count; i++) {
        if (event->id[i] == 0) {
            break;
        }
    }

    if (i == event->client_count) {
        LIST_delete(image->event_list, event);
    }
    
    pthread_mutex_unlock(&image->notify_mutex);

    info->type = event->type;
    info->viewport[0] = event->viewport[0];
    info->viewport[1] = event->viewport[1];
    info->viewport[2] = event->viewport[2];
    info->viewport[3] = event->viewport[3];
    info->image = image;

    return 0;

 no_event:

    pthread_mutex_unlock(&image->notify_mutex);
    return -1;
}

static IMAGE_NOTIFY_EVENT *
create_notify_event(IMAGE *image)
{
    int client_count;
    IMAGE_NOTIFY_EVENT *event;

    /* only add an event if someone is interested */
    assert(image->client_list != NULL);
    
    client_count = LIST_count(image->client_list);
    if (client_count == 0) {
        return NULL;
    }
    
    assert(image->event_list != NULL);

    event = (IMAGE_NOTIFY_EVENT *)malloc(sizeof(IMAGE_NOTIFY_EVENT));
    assert(event != NULL);

    event->id = (int *)malloc(client_count * sizeof(int));
    assert(event->id != NULL);
    memset(event->id, 0, client_count * sizeof(int));
    event->client_count = client_count;

    LIST_append(image->event_list, event);
    
    return event;
}

void
IMAGE_notify_start_processing(IMAGE *image)
{
    IMAGE_NOTIFY_EVENT *event;
    
    pthread_mutex_lock(&image->notify_mutex);

    event = create_notify_event(image);
    if (event != NULL) {
        event->type = IMAGE_NOTIFY_START;
    }

    pthread_mutex_unlock(&image->notify_mutex);
}
    
void
IMAGE_notify_stop_processing(IMAGE *image)
{
    IMAGE_NOTIFY_EVENT *event;
    
    pthread_mutex_lock(&image->notify_mutex);

    event = create_notify_event(image);
    if (event != NULL) {
        event->type = IMAGE_NOTIFY_STOP;
    }

    pthread_mutex_unlock(&image->notify_mutex);
}

void
IMAGE_notify_update_viewport(IMAGE *image, int viewport[4])
{
    IMAGE_NOTIFY_EVENT *event;
    
    pthread_mutex_lock(&image->notify_mutex);

    event = create_notify_event(image);
    if (event != NULL) {
        event->type = IMAGE_NOTIFY_VIEWPORT;
        
        if (viewport == NULL) {
            event->viewport[0] = image->viewport[0];
            event->viewport[1] = image->viewport[1];
            event->viewport[2] = image->viewport[2];
            event->viewport[3] = image->viewport[3];
        } else {
            event->viewport[0] = viewport[0];
            event->viewport[1] = viewport[1];
            event->viewport[2] = viewport[2];
            event->viewport[3] = viewport[3];
        }
    }

    pthread_mutex_unlock(&image->notify_mutex);
}

void
IMAGE_remove_notify_client(IMAGE *image, int id)
{
    IMAGE_NOTIFY_CLIENT *client;
    
    for (client = LIST_head(image->client_list);
         client != NULL;
         client = LIST_next(client)) {

        if (client->id == id) {
            LIST_delete(image->client_list, client);
            if (LIST_is_empty(image->client_list)) {
                while (LIST_pop(image->event_list) != NULL) {
                    /*EMPTY*/
                }
            }
            return;
        }
    }
}

