/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image_pixel.c,v 1.10 2001/12/20 21:25:32 wex Exp $
 *
 * $Log: image_pixel.c,v $
 * Revision 1.10  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.9  2001/12/19 00:59:08  wex
 * Win32 support (EXPORT), added image_rgba.c and VC++ project files.
 *
 * Revision 1.8  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 * Revision 1.7  2001/11/26 21:24:41  wex
 * A few extra checks in pixel routines, accumulate handles alpha.
 *
 * Revision 1.6  2001/11/16 16:39:25  wex
 * Safer (and a bit slower) get and set routines.
 *
 * Revision 1.5  2001/11/13 07:52:03  wex
 * Memory fixes
 *
 * Revision 1.4  2001/11/12 09:09:49  wex
 * Fixed up memory access errors.
 *
 * Revision 1.3  2001/11/10 05:57:12  wex
 * Allow viewport to be reset.
 *
 * Revision 1.2  2001/10/26 15:48:36  wex
 * First working version
 *
 * Revision 1.1  2001/07/18 04:58:51  wex
 * Added
 *
 *
 */

#include <assert.h>
#include <math.h>

#include "image_internal.h"

int
IMAGE_is_valid_pixel(IMAGE *image, int x, int y)
{
    return x >= image->viewport[0] && x <= image->viewport[2] &&
        y >= image->viewport[1] && y <= image->viewport[3];
}

void
IMAGE_get_viewport_wrapped_pixel(IMAGE *image, double sx, double sy,
    double *wx, double *wy)
{
    *wx = sx;
    *wy = sy;
    
    if (sx < image->viewport[0]) *wx = image->viewport[0];
    if (sy < image->viewport[1]) *wy = image->viewport[1];
    if (sx >= image->viewport[2]+1) *wx = image->viewport[2]-0.00001;
    if (sy >= image->viewport[3]+1) *wy = image->viewport[3]-0.00001;
}
    
int
IMAGE_set_color(IMAGE *image, int x, int y, double r, double g, double b)
{
    int i;
    IMAGE_TILE *tile;

    if (image->tile == NULL) {
        return -1;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    if (i < 0 || i >= image->x_tile_count * image->y_tile_count) {
        return -1;
    }
    
    assert(i >= 0 && i < image->x_tile_count * image->y_tile_count);
    tile = image->tile[i];
    if (tile == NULL) {
        tile = image_tile_create(image, i);
        if (tile == NULL) {
            return -1;
        }
    }

    if (tile->r == NULL) {
        IMAGE_errno = IMAGE_ERR_NO_CHANNEL;
        return -1;
    }
    
    i = IMAGE_PIXEL_INDEX(image, x, y);

    assert(i >= 0 && i < IMAGE_TILE_SIZE*IMAGE_TILE_SIZE);
    
    tile->r[i] = r;
    tile->g[i] = g;
    tile->b[i] = b;

    return 0;
}

int
IMAGE_set_alpha(IMAGE *image, int x, int y, double a)
{
    int i;
    IMAGE_TILE *tile;

    if (image->tile == NULL) {
        return -1;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    assert(i >= 0 && i < image->x_tile_count * image->y_tile_count);
    tile = image->tile[i];
    if (tile == NULL) {
        tile = image_tile_create(image, i);
        if (tile == NULL) {
            return -1;
        }
    }

    if (tile->a == NULL) {
        IMAGE_errno = IMAGE_ERR_NO_CHANNEL;
        return -1;
    }
    
    i = IMAGE_PIXEL_INDEX(image, x, y);

    assert(i >= 0 && i < IMAGE_TILE_SIZE*IMAGE_TILE_SIZE);
    
    tile->a[i] = a;

    return 0;
}
    
int
IMAGE_set_depth(IMAGE *image, int x, int y, double z)
{
    int i;
    IMAGE_TILE *tile;

    if (image->tile == NULL) {
        return -1;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    tile = image->tile[i];
    if (tile == NULL) {
        tile = image_tile_create(image, i);
        if (tile == NULL) {
            return -1;
        }
    }

    if (tile->z == NULL) {
        IMAGE_errno = IMAGE_ERR_NO_CHANNEL;
        return -1;
    }
    
    i = IMAGE_PIXEL_INDEX(image, x, y);

    tile->z[i] = z;

    return 0;
}

int
IMAGE_update_depth(IMAGE *image, int x, int y, double z)
{
    int i;
    IMAGE_TILE *tile;

    if (image->tile == NULL) {
        return -1;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    tile = image->tile[i];
    if (tile == NULL) {
        tile = image_tile_create(image, i);
        if (tile == NULL) {
            return -1;
        }
    }

    if (tile->z == NULL) {
        IMAGE_errno = IMAGE_ERR_NO_CHANNEL;
        return -1;
    }
    
    i = IMAGE_PIXEL_INDEX(image, x, y);

    assert(i >= 0 && i < IMAGE_TILE_SIZE*IMAGE_TILE_SIZE);
    
    if (z < tile->z[i]) {
        tile->z[i] = z;
        return 1;
    }

    return 0;
}

void
IMAGE_get_color(IMAGE *image, int x, int y, double *r, double *g, double *b)
{
    int i;
    IMAGE_TILE *tile;

    if (!image->is_color_defined || image->tile == NULL) {
        *r = *g = *b = 0;
        return;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    if (i < 0 || i >= image->x_tile_count * image->y_tile_count) {
        *r = *g = *b = 0;
        return;
    }

    tile = image->tile[i];
    if (tile == NULL) {
        *r = *g = *b = 0;
        return;
    }

    i = IMAGE_PIXEL_INDEX(image, x, y);
    *r = tile->r[i];
    *g = tile->g[i];
    *b = tile->b[i];
}

void
IMAGE_get_bilerp_color(IMAGE *image, double x, double y, double *r, double *g,
    double *b)
{
    int ix, iy;
    double fx, fy;
    double w0, w1, w2, w3;
    double tr, tg, tb;

    if (!image->is_color_defined || image->tile == NULL) {
        *r = *g = *b = 0;
        return;
    }

    ix = floor(x);
    iy = floor(y);
    fx = x - ix;
    fy = y - iy;
    
	w0 = (1.0 - fx) * (1.0 - fy);
	w1 = fx * (1.0 - fy);
	w2 = fx * fy;
	w3 = (1.0 - fx) * fy;
    
    IMAGE_get_color(image, ix, iy, &tr, &tg, &tb);
    *r = tr * w0;
    *g = tg * w0;
    *b = tb * w0;
    
    IMAGE_get_color(image, ix+1, iy, &tr, &tg, &tb);
    *r += tr * w1;
    *g += tg * w1;
    *b += tb * w1;
    
    IMAGE_get_color(image, ix+1, iy+1, &tr, &tg, &tb);
    *r += tr * w2;
    *g += tg * w2;
    *b += tb * w2;
    
    IMAGE_get_color(image, ix, iy+1, &tr, &tg, &tb);
    *r += tr * w3;
    *g += tg * w3;
    *b += tb * w3;
}

double
IMAGE_get_alpha(IMAGE *image, int x, int y)
{
    int i;
    IMAGE_TILE *tile;

    if (!image->is_alpha_defined || image->tile == NULL) {
        return 0;
    }

    i = IMAGE_TILE_INDEX(image, x, y);
    tile = image->tile[i];
    if (tile == NULL) {
        return 0;
    }

    i = IMAGE_PIXEL_INDEX(image, x, y);

    return tile->a[i];
}

double
IMAGE_get_depth(IMAGE *image, int x, int y)
{
    int i;
    IMAGE_TILE *tile;

    if (!image->is_depth_defined || image->tile == NULL) {
        return 0;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    tile = image->tile[i];
    if (tile == NULL) {
        return 0;
    }

    i = IMAGE_PIXEL_INDEX(image, x, y);
    return tile->a[i];
}

    
int
IMAGE_accumulate_color(IMAGE *image, int x, int y, double r, double g,
    double b, double a)
{
    int i;
    IMAGE_TILE *tile;

    if (image->tile == NULL) {
        return -1;
    }
    
    i = IMAGE_TILE_INDEX(image, x, y);
    if (i < 0 || i >= image->x_tile_count * image->y_tile_count) {
        return -1;
    }
    
    assert(i >= 0 && i < image->x_tile_count * image->y_tile_count);
    tile = image->tile[i];
    if (tile == NULL) {
        tile = image_tile_create(image, i);
        if (tile == NULL) {
            return -1;
        }
    }

    if (tile->r == NULL) {
        IMAGE_errno = IMAGE_ERR_NO_CHANNEL;
        return -1;
    }
    
    i = IMAGE_PIXEL_INDEX(image, x, y);

    if (i < 0 || i >= IMAGE_TILE_SIZE*IMAGE_TILE_SIZE) {
        /* clip to image to allow for culling */
        return -1;
    }
    
    tile->r[i] += r;
    tile->g[i] += g;
    tile->b[i] += b;

    if (tile->a != NULL) {
        tile->a[i] += a;
    }

    return 0;
}

