/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/image/image_ppm.c,v 1.6 2001/12/20 21:25:32 wex Exp $
 *
 * $Log: image_ppm.c,v $
 * Revision 1.6  2001/12/20 21:25:32  wex
 * Removed EXPORT in favor of .def file.
 *
 * Revision 1.5  2001/12/05 10:21:11  wex
 * Added notify and thread safe code, off-by-one bugs, bilerp
 *
 * Revision 1.4  2001/11/10 05:57:12  wex
 * Allow viewport to be reset.
 *
 * Revision 1.3  2001/11/06 06:55:38  wex
 * Set/get color, tiling for ppm
 *
 * Revision 1.2  2001/10/29 00:50:24  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:36  wex
 * First working version
 *
 *
 *
 */
#include <stdio.h>
#include <ctype.h>

#include "image_format.h"
#include "image_internal.h"

static int image_ppm_open(IMAGE_FORMAT_OPEN_INFO *info);
static int get_number(FILE *f);
static int read_header(FILE *f, int *type, int *width, int *height,
    int *max_value);
static int image_ppm_write(IMAGE_FORMAT_WRITE_INFO *info);
static int image_ppm_read_tile(IMAGE_FORMAT_READ_TILE_INFO *info);


void
IMAGE_support_ppm(void)
{
    static IMAGE_FORMAT *format = NULL;

    if (format != NULL) {
        return;
    }

    format = image_format_create("ppm");
    image_format_add_extension(format, "ppm");
    image_format_set_open_callback(format, image_ppm_open);
    image_format_set_read_tile_callback(format, image_ppm_read_tile);
    image_format_set_write_callback(format, image_ppm_write);
}

static int
image_ppm_open(IMAGE_FORMAT_OPEN_INFO *info)
{
    FILE *fd;
	int type;
	int width;
	int height;
    int max_value;
    int x, y;
    double r, g, b;
    IMAGE *image;

    fd = fopen(info->filename, "rb");
    if (fd == NULL) {
        IMAGE_errno = IMAGE_ERR_LIBC;
        return -1;
    }

	if (read_header(fd, &type, &width, &height, &max_value) == -1) {
        fclose(fd);
        return -1;
	}

    image = info->image;
    IMAGE_free(image);

    image->viewport[0] = 0;
    image->viewport[1] = 0;
    image->viewport[2] = width - 1;
    image->viewport[3] = height - 1;

    image_compute_tiling(image);
    
	image->is_alpha_defined = type == '2' || type == '5';
	image->is_color_defined = type == '3' || type == '6';
    image->is_depth_defined = 0;

    switch (type) {
    case '2':
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
                IMAGE_set_alpha(image, x, height-y-1,
                    get_number(fd)/(float)max_value);
			}
		}
        break;

    case '3':
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
                r = get_number(fd)/(float)max_value;
                g = get_number(fd)/(float)max_value;
                b = get_number(fd)/(float)max_value;
                
                IMAGE_set_color(image, x, height-y-1, r, g, b);
			}
		}
        break;
        
    case '6':
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
                r = fgetc(fd)/(float)max_value;
                g = fgetc(fd)/(float)max_value;
                b = fgetc(fd)/(float)max_value;
                IMAGE_set_color(image, x, height-y-1, r, g, b);
			}
		}
        break;

    default:
        abort();
    }

    fclose(fd);
    
    return 0;
}

static int
get_number(FILE *f)
{
	int number;
	int c;

	number = 0;

	c = fgetc(f);

	while (c != EOF) {
		if (isspace(c)) {
			c = fgetc(f);
		} else if (c == '#') {
			while (c != '\n' && c != EOF) {
				c = fgetc(f);
			}
		} else {
			break;
		}
	}

	if (isdigit(c)) {
		while (isdigit(c)) {
			number = number*10 + c - '0';
			c = fgetc(f);
		}

		ungetc(c, f); /* we'll get the space later */

		return number;
	}

	return -1;
}


static int
read_header(FILE *f, int *type, int *width, int *height, int *max_value)
{
	/* assumes the file is rewound.  leaves the file pointer at the
	   beginning of the image data */

	int c;

	c = fgetc(f);
	if (c != 'P') {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	*type = fgetc(f);
	if (*type < '1' || *type > '6') {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	if (*type == '1' || *type == '4') {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	*width = get_number(f);
	if (*width <= 0) {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	*height = get_number(f);
	if (*height <= 0) {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	*max_value = get_number(f);
	if (*max_value <= 0 || *max_value > 255) {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	c = fgetc(f);
	if (!isspace(c)) {
		IMAGE_errno = IMAGE_ERR_PPM;
		return -1;
	}

	return 0;
}

static int
image_ppm_read_tile(IMAGE_FORMAT_READ_TILE_INFO *info)
{
    /* all tiles read when open :( */
    return 0;
}

static int
image_ppm_write(IMAGE_FORMAT_WRITE_INFO *info)
{
    FILE *fd;
    int x, y;
    double r, g, b;
    char type;
    int width, height;
    IMAGE *image;
    
    image = info->image;

    fd = fopen(info->filename, "wb");
    if (fd == NULL) {
        IMAGE_errno = IMAGE_ERR_LIBC;
        return -1;
    }

    if (IMAGE_is_color_defined(image)) {
        type = '6';
    } else {
        type = '5';
    }

    width = IMAGE_get_width(image);
    height = IMAGE_get_height(image);

    fprintf(fd, "P%c %d %d 255 ", type, width, height);
    
    switch (type) {
    case '5':
        for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
                fputc((int)(IMAGE_get_alpha(image, x, height-y-1) * 255.9),fd);

			}
		}
        break;

    case '6':
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
                IMAGE_get_color(image, x, height - y - 1, &r, &g, &b);
                if (r < 0) r = 0;
                if (r > 1) r = 1;
                if (g < 0) g = 0;
                if (g > 1) g = 1;
                if (b < 0) b = 0;
                if (b > 1) b = 1;
				fputc((int)(r*255.9), fd);
				fputc((int)(g*255.9), fd);
				fputc((int)(b*255.9), fd);
			}
		}
    }

    fclose(fd);
    
    return 0;
}

