/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/ip/ip_font.c,v 1.2 2001/11/26 21:26:11 wex Exp $
 *
 * $Log: ip_font.c,v $
 * Revision 1.2  2001/11/26 21:26:11  wex
 * Basic image processing engine for point to point operations
 *
 *
 */
#include "ip.h"

#ifdef IP_FREETYPE_ENABLED


static FT_Library ft_library;       /* handle to library     */

static void ip_font_matte_buffer(IMAGE *image, int vp[4], unsigned char *a,
    double r, double g, double b);


int
IP_FONT_initialize(IP_FONT *font, char *fontname)
{
    FT_Face face;                   /* handle to face object */
    static int initialized = 0;
    int error;
    
    if (!initialized) {
        initialized = 1;
        if (FT_Init_FreeType(&ft_library)) {
            fprintf(stderr, "Cannot initialize FreeType library\n");
            return -1;
        }
    }

    error = FT_New_Face(ft_library, fontname, 0, &face);
    if (error == FT_Err_Unknown_File_Format) {
        fprintf(stderr, "Font \"%s\" could be opened and read, but it "
            "appears that its font format is unsupported\n", fontname);
        return -1;
    } else if (error) {
        fprintf(stderr, "Font \"%s\" could not be opened and read\n",fontname);
        return -1;
    }

    font->face = face;

    IP_FONT_set_size(font, 12);
    
    return 0;
}

void
IP_FONT_set_size(IP_FONT *font, double points)
{
    FT_Set_Char_Size(
        font->face,   /* handle to face object           */
        0,            /* char_width in 1/64th of points  */
        points*64,    /* char_height in 1/64th of points */
        72,           /* horizontal device resolution    */
        72);          /* vertical device resolution      */
}

int
IP_FONT_render_string(IP_FONT *font, IMAGE *image, int x, int y, char *s,
    double r, double g, double b)
{
    FT_GlyphSlot  slot = font->face->glyph;  // a small shortcut
    int           i, num_chars, error, vp[4];
    unsigned char *a;

    num_chars = strlen(s);

    for (i = 0; i < num_chars; i++ ) {
        /* load glyph image into the slot (erase previous one) */
        error = FT_Load_Char(font->face, s[i], FT_LOAD_RENDER);
        if (error) continue;  /* ignore errors */

        /* draw the bitmap with y as the bottom line of normal text */
        a = slot->bitmap.buffer;
        vp[0] = x + slot->bitmap_left;
        vp[1] = y - slot->bitmap.rows + slot->bitmap_top;
        vp[2] = x + slot->bitmap_left + slot->bitmap.width - 1;
        vp[3] = y + slot->bitmap_top - 1;

        ip_font_matte_buffer(image, vp, a, r, g, b);

        /* advance the location for the next character */
        x += slot->advance.x >> 6;
    }

    return 0;
}

static void
ip_font_matte_buffer(IMAGE *image, int vp[4], unsigned char *a, double r,
    double g, double b)
{
    int i, x, y;
    double t;
    double old_r, old_g, old_b;
    double new_r, new_g, new_b;
    
    for (y = vp[3], i = 0; y >= vp[1]; y--) {
        for (x = vp[0]; x <= vp[2]; x++, i++) {
            t = a[i] / 255.0;
            IMAGE_get_color(image, x, y, &old_r, &old_g, &old_b);
            new_r = t * r + (1 - t) * old_r;
            new_g = t * g + (1 - t) * old_g;
            new_b = t * b + (1 - t) * old_b;
            IMAGE_set_color(image, x, y, new_r, new_g, new_b);
        }
    }
}

    
void
IP_FONT_destroy(IP_FONT *font)
{
/*    font->face = NULL; */
    free(font);
}


#endif /* IP_FREETYPE_ENABLED */
