/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/ip/ip_glue.c,v 1.5 2002/01/07 22:26:14 wex Exp $
 *
 * $Log: ip_glue.c,v $
 * Revision 1.5  2002/01/07 22:26:14  wex
 * Port to Windows, destroy fp for cleaning up operations
 *
 * Revision 1.4  2001/12/08 23:28:04  wex
 * Updated FreeType defines, alpha_from_lum, referencing fixes.
 *
 * Revision 1.3  2001/12/05 10:47:42  wex
 * Global viewport
 *
 * Revision 1.2  2001/11/26 21:26:11  wex
 * Basic image processing engine for point to point operations
 *
 * Revision 1.1  2001/11/16 09:21:57  wex
 * Added
 *
 *
 */
#include <Python.h>
#include "ip.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}

static PyObject *ip_read_new(PyObject *self, PyObject *args);
static PyObject *ip_image_new(PyObject *self, PyObject *args);
static PyObject *ip_fill_new(PyObject *self, PyObject *args);
static PyObject *ip_gradient_new(PyObject *self, PyObject *args);
static PyObject *ip_matte_new(PyObject *self, PyObject *args);
static PyObject *ip_set_global_viewport(PyObject *self, PyObject *args);
static PyObject *ip_set_global_viewport_from_image(PyObject *self,
    PyObject *args);

#ifdef IP_FREETYPE_ENABLED
static PyObject *ip_font_new(PyObject *self, PyObject *args);
static void ip_font_dealloc(IP_FONT *self);
static int ip_font_print(IP_FONT *self, FILE *fp, int flags);
static PyObject *ip_font_render_string(PyObject *self, PyObject *args);
static PyObject *ip_font_set_size(PyObject *self, PyObject *args);
static PyObject *ip_font_getattr(IMAGE *self, char *name);

static struct PyMethodDef ip_font_methods[] = {
    {"draw_string",   (PyCFunction)ip_font_render_string,    1, NULL},
    {"set_size",      (PyCFunction)ip_font_set_size,         1, NULL},
    {NULL,          NULL}
};

static PyTypeObject IP_Font = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "ip_font",                                      /* tp_name               */
    sizeof(IP_FONT),                                /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)ip_font_dealloc,                    /* tp_dealloc refcount=0 */
    (printfunc)ip_font_print,                       /* tp_print   "print x"  */
    (getattrfunc)ip_font_getattr,                   /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

#endif  /* No FreeType */

static struct PyMethodDef iptype_methods[] = {
#ifdef IP_FREETYPE_ENABLED
    {"Font", ip_font_new, 1},
#endif    
    {"read", ip_read_new, 1},
    {"image", ip_image_new, 1},
    {"fill", ip_fill_new, 1},
    {"gradient", ip_gradient_new, 1},
    {"matte", ip_matte_new, 1},
    {"viewport", ip_set_global_viewport, 1},
    {"match_viewport", ip_set_global_viewport_from_image, 1},
    {NULL, NULL}
};


void
initipcmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32
    IP_Font.ob_type = &PyType_Type;
#endif    
    m = Py_InitModule("ipcmodule", iptype_methods);
    d = PyModule_GetDict(m);                        /* with 'Image' func */
    ErrorObject = Py_BuildValue("s", "ipcmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize ip C module");
    }
    IMAGE_support_all();
}

static PyObject *
ip_read_new(PyObject *self, PyObject *args)
{
    char *filename;
    
    if (!PyArg_ParseTuple(args, "s", &filename)) 
        onError("you must pass a filename string");

    return (PyObject *)IP_OP_read(filename);
}

static PyObject *
ip_image_new(PyObject *self, PyObject *args)
{
    IMAGE *im;
    
    if (!PyArg_ParseTuple(args, "O", &im)) 
        onError("you must pass an image");

    if (!IMAGE_check(im))
        onError("invalid image");
    
    return (PyObject *)IP_OP_image(im);
}

static PyObject *
ip_fill_new(PyObject *self, PyObject *args)
{
    double r, g, b, a, z;
    
    if (!PyArg_ParseTuple(args, "(ddddd)", &r, &g, &b, &a, &z)) 
        onError("you must pass a rgbaz tuple");

    return (PyObject *)IP_OP_fill(r, g, b, a, z);
}

static PyObject *
ip_gradient_new(PyObject *self, PyObject *args)
{
    double tl[5], tr[5], bl[5], br[5];
    
    if (!PyArg_ParseTuple(args, "(ddddd)(ddddd)(ddddd)(ddddd)",
            &tl[0], &tl[1], &tl[2], &tl[3], &tl[4], 
            &tr[0], &tr[1], &tr[2], &tr[3], &tr[4], 
            &bl[0], &bl[1], &bl[2], &bl[3], &bl[4], 
            &br[0], &br[1], &br[2], &br[3], &br[4])) 
        onError("you must pass four color tuples");

    return (PyObject *)IP_OP_gradient(tl, tr, bl, br);
}

static PyObject *
ip_matte_new(PyObject *self, PyObject *args)
{
    IP_OP *over, *under;
    
    if (!PyArg_ParseTuple(args, "OO", &over, &under)) 
        onError("you must pass an over and under operation");

    return (PyObject *)IP_OP_matte(over, under);
}

static PyObject *
ip_set_global_viewport(PyObject *self, PyObject *args)
{
    int vp[4];

    if (!PyArg_ParseTuple(args, "iiii", &vp[0], &vp[1], &vp[2], &vp[3]))
        onError("you must pass xmin,ymin,xmax,ymax to set global viewport");

    IP_OP_set_global_viewport(vp);
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
ip_set_global_viewport_from_image(PyObject *self, PyObject *args)
{
    int vp[4];
    IMAGE *image;

    if (!PyArg_ParseTuple(args, "O", &image))
        onError("you must pass an image to set the global viewport");

    IMAGE_get_viewport(image, &vp[0], &vp[1], &vp[2], &vp[3]);
    IP_OP_set_global_viewport(vp);
    
    Py_INCREF(Py_None);
    return Py_None;
}
    
#ifdef IP_FREETYPE_ENABLED

static PyObject *
ip_font_new(PyObject *self, PyObject *args)
{
    char *fontname;
    PyObject *font;
    
    if (!PyArg_ParseTuple(args, "s", &fontname)) 
        onError("font creation requires a valid font name argument");

    font = (PyObject *)PyObject_NEW(IP_FONT, &IP_Font);
    IP_FONT_initialize((IP_FONT *)font, fontname);
    return font;
}

static void
ip_font_dealloc(IP_FONT *self)
{
    PyMem_DEL(self);
}

static int
ip_font_print(IP_FONT *self, FILE *fp, int flags)
{
    fprintf(fp, "IP_FONT at %p\n", self);
    return 0;
}

static PyObject *
ip_font_render_string(PyObject *self, PyObject *args)
{
    char *text;
    int x, y;
    double r, g, b;
    IMAGE *image;

    if (!PyArg_ParseTuple(args, "Oiisddd", &image, &x, &y, &text, &r, &g, &b)) 
        onError("text rendering requires an image, x, y, \"text string\" "
            "and r,g,b");

    IP_FONT_render_string((IP_FONT *)self, image, x, y, text, r, g, b);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
ip_font_set_size(PyObject *self, PyObject *args)
{
    double points;

    if (!PyArg_ParseTuple(args, "d", &points))
        onError("text size requires a single double argument");

    IP_FONT_set_size((IP_FONT *)self, points);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
ip_font_getattr(IMAGE *self, char *name)
{
    return Py_FindMethod(ip_font_methods, (PyObject *)self, name);
}
#endif
