#include <assert.h>

#include "ip_op.h"


static void ip_op_dealloc(IP_OP *self);
static int ip_op_print(IP_OP *self, FILE *fp, int flags);
static PyObject *ip_op_getattr(IMAGE *self, char *name);
static PyObject *ip_op_get_image(PyObject *self, PyObject *args);
static PyObject *ip_op_alpha_from_lum_new(PyObject *self, PyObject *args);

static void ip_op_compute_pixel(IP_COMPUTE_INFO *info, IP_OP *op,
    double *output);
static void ip_op_compute_viewport(IP_OP *op, int viewport[4]);


static struct PyMethodDef ip_op_methods[] = {
    {"alpha_from_lum", (PyCFunction)ip_op_alpha_from_lum_new, 1, NULL},
    {"get_image",   (PyCFunction)ip_op_get_image,    1, NULL},
    {NULL,          NULL}
};


static int ip_op_viewport[4] = {0,0,719, 485};
static IP_QUALITY ip_op_quality = IP_QUALITY_HI;


static PyTypeObject IP_Op = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "ip_op",                                        /* tp_name               */
    sizeof(IP_OP),                                  /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)ip_op_dealloc,                      /* tp_dealloc refcount=0 */
    (printfunc)ip_op_print,                         /* tp_print   "print x"  */
    (getattrfunc)ip_op_getattr,                     /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};


IP_OP *
ip_op_create(char *name, int (*compute)(IP_COMPUTE_INFO *info),
    int (*viewport)(IP_VIEWPORT_INFO *info), void *user_data)
{
    IP_OP *op;

#ifdef WIN32    
    IP_Op.ob_type = &PyType_Type;
#endif
    
    op = PyObject_NEW(IP_OP, &IP_Op);
    
    op->name = strdup(name);
    op->input_count = 0;
    op->input = NULL;
    op->compute = compute;
    op->viewport = viewport;
    op->destroy = NULL;
    
    op->user_data = user_data;

    return op;
}

void
ip_op_add_input(IP_OP *op, char *name, IP_OP *input)
{
    int i;

    i = op->input_count;
    op->input_count++;
    assert(op->input_count < IP_MAX_INPUT);
    op->input = realloc(op->input, op->input_count * sizeof(IP_INPUT));
    assert(op->input != NULL);

    op->input[i].name = strdup(name);
    op->input[i].op = input;

    /* saving a reference */
    Py_INCREF(input);
}

static void
ip_op_dealloc(IP_OP *self)
{
    int i;

    if (self->destroy != NULL) {
        self->destroy(self);
    } else if (self->user_data != NULL) {
        /* assume it just needs to be free'd */
        free(self->user_data);
        self->user_data = NULL;
    }
    
    for (i = 0; i < self->input_count; i++) {
        Py_DECREF(self->input[i].op);
    }

    if (self->input_count > 0) {
        free(self->input);
        self->input = NULL;
    }
    
    PyMem_DEL(self);
}

static int
ip_op_print(IP_OP *self, FILE *fp, int flags)
{
    fprintf(fp, "IP_OP at %p\n", self);
    return 0;
}

static PyObject *
ip_op_getattr(IMAGE *self, char *name)
{
    return Py_FindMethod(ip_op_methods, (PyObject *)self, name);
}

static PyObject *
ip_op_get_image(PyObject *self, PyObject *args)
{
    return (PyObject *)IP_OP_compute_image((IP_OP *)self, NULL);
}

static PyObject *
ip_op_alpha_from_lum_new(PyObject *self, PyObject *args)
{
    return (PyObject *)IP_OP_alpha_from_lum((IP_OP *)self);
}

IMAGE *
IP_OP_compute_image(IP_OP *op, int viewport[4])
{
    IP_COMPUTE_INFO info;
    double c[5];
    IMAGE *output;

    assert(op != NULL);
    
    if (viewport == NULL) {
        info.viewport[0] = info.viewport[1] = INT_MAX;
        info.viewport[2] = info.viewport[3] = -INT_MAX;
        ip_op_compute_viewport(op, info.viewport);
    } else {
        info.viewport[0] = viewport[0];
        info.viewport[1] = viewport[1];
        info.viewport[2] = viewport[2];
        info.viewport[3] = viewport[3];
    }

    info.op = op;
    info.user_data = op->user_data;

    output = IMAGE_create();
    IMAGE_set_color_defined(output, 1);
    IMAGE_set_alpha_defined(output, 1);
    IMAGE_set_viewport(output, info.viewport[0], info.viewport[1],
        info.viewport[2], info.viewport[3]);
    
    for (info.y = info.viewport[1]; info.y <= info.viewport[3]; info.y++) {
        for (info.x = info.viewport[0]; info.x <= info.viewport[2]; info.x++) {
            ip_op_compute_pixel(&info, op, c);
            IMAGE_set_color(output, info.x, info.y, c[0], c[1], c[2]);
            IMAGE_set_alpha(output, info.x, info.y, c[3]);
        }
    }

    return output;
}

static void
ip_op_compute_viewport(IP_OP *op, int viewport[4])
{
    int i;
    IP_VIEWPORT_INFO info;

    /* compute the viewport of the input operations, recursing if necessary */
    for (i = 0; i < op->input_count; i++) {
        ip_op_compute_viewport(op->input[i].op, viewport);
    }

    /* add in the viewport of this operation */
    if (op->viewport != NULL) {
        info.op = op;
        info.user_data = op->user_data;
        info.viewport[0] = info.viewport[1] = INT_MAX;
        info.viewport[2] = info.viewport[3] = -INT_MAX;

        if (op->viewport(&info) < 0) return;
        
        if (info.viewport[0] < viewport[0]) viewport[0] = info.viewport[0];
        if (info.viewport[1] < viewport[1]) viewport[1] = info.viewport[1];
        if (info.viewport[2] > viewport[2]) viewport[2] = info.viewport[2];
        if (info.viewport[3] > viewport[3]) viewport[3] = info.viewport[3];
    }

    /* if this is an input node without a viewport, use the global one */
    if (op->viewport == NULL && op->input_count == 0) {
        if (ip_op_viewport[0] < viewport[0]) viewport[0] = ip_op_viewport[0];
        if (ip_op_viewport[1] < viewport[1]) viewport[1] = ip_op_viewport[1];
        if (ip_op_viewport[2] > viewport[2]) viewport[2] = ip_op_viewport[2];
        if (ip_op_viewport[3] > viewport[3]) viewport[3] = ip_op_viewport[3];
    }
}

static void
ip_op_compute_pixel(IP_COMPUTE_INFO *info, IP_OP *op, double *output)
{
    int i;

    /* compute the input pixels recursing if necessary */
    for (i = 0; i < op->input_count; i++) {
        ip_op_compute_pixel(info, op->input[i].op, info->input[i]);
    }

    /* compute this operation */
    info->op = op;
    info->output = output;
    info->user_data = op->user_data;
    
    op->compute(info);
}

void
IP_OP_set_global_viewport(int viewport[4])
{
    ip_op_viewport[0] = viewport[0];
    ip_op_viewport[1] = viewport[1];
    ip_op_viewport[2] = viewport[2];
    ip_op_viewport[3] = viewport[3];
}

IP_QUALITY
IP_OP_set_global_quality(IP_QUALITY quality)
{
    IP_QUALITY old;

    old = ip_op_quality;
    ip_op_quality = quality;
    return old;
}

void
ip_op_get_global_viewport(int viewport[4])
{
    viewport[0] = ip_op_viewport[0];
    viewport[1] = ip_op_viewport[1];
    viewport[2] = ip_op_viewport[2];
    viewport[3] = ip_op_viewport[3];
}

void
ip_op_set_destroy_fp(IP_OP *op, void (*destroy)(IP_OP *op))
{
    op->destroy = destroy;
}

