/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/list/list.c,v 1.6 2002/01/03 05:16:48 wex Exp $
 *
 * $Log: list.c,v $
 * Revision 1.6  2002/01/03 05:16:48  wex
 * Stub for distutils automatic define of initlist()
 *
 * Revision 1.5  2001/12/20 21:20:23  wex
 * Removed EXORTS in favor of .def file
 *
 * Revision 1.4  2001/12/19 00:56:32  wex
 * Added EXPORT keyword for Win32 DLL support.  Added VC++ project files.
 *
 * Revision 1.3  2001/11/13 11:45:18  wex
 * Added LIST_is_empty()
 *
 * Revision 1.2  2001/10/29 00:50:26  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:38  wex
 * First working version
 *
 *
 */

#include <assert.h>
#include <stdio.h>
#include <malloc.h>

#include "list_internal.h"

/* stub for linking on windows */
void initlist(void) {}


LIST *
LIST_create(void)
{
    LIST *list;

    list = (LIST *)malloc(sizeof(LIST));
    if (list == NULL) {
        LIST_errno = LIST_ERR_MEMORY;
        return NULL;
    }

    LIST_init(list);
    
    return list;
}

void
LIST_init(LIST *list)
{
    list->head = NULL;
    list->tail = NULL;
    list->count = 0;
}

void
LIST_delete(LIST *list, void *element)
{
    LIST_ELEMENT *e = element;
    
    if (list->head == e) {
        list->head = e->next;
    }

    if (list->tail == e) {
        list->tail = e->prev;
    }
    
    if (e->prev != NULL) {
        e->prev->next = e->next;
    }

    if (e->next != NULL) {
        e->next->prev = e->prev;
    }

    --list->count;
}

void
LIST_destroy(LIST *list)
{
    while (LIST_pop(list) != NULL) {
        /* empty */
    }

    free(list);
}

void
LIST_append(LIST *list, void *element)
{
    LIST_ELEMENT *e = element;
    
    if (list->head == NULL) {
        list->head = e;
    }
    
    if (list->tail != NULL) {
        assert(list->tail->next == NULL);
        list->tail->next = e;
    }

    e->prev = list->tail;
    e->next = NULL;

    list->tail = e;
    ++list->count;
}

void
LIST_prepend(LIST *list, void *element)
{
    LIST_ELEMENT *e = element;
    
    if (list->head != NULL) {
        assert(list->head->prev == NULL);
        list->head->prev = e;
    }

    if (list->tail == NULL) {
        list->tail = e;
    }

    e->prev = NULL;
    e->next = list->head;

    list->head = e;
    ++list->count;
}

void *
LIST_next(void *element)
{
    return ((LIST_ELEMENT *)element)->next;
}

void *
LIST_prev(void *element)
{
    return ((LIST_ELEMENT *)element)->prev;
}

void *
LIST_head(void *list)
{
    return ((LIST *)list)->head;
}

void *
LIST_tail(void *list)
{
    return ((LIST *)list)->tail;
}

void *
LIST_pop(LIST *list)
{
    LIST_ELEMENT *element;

    element = LIST_head(list);
    if (element == NULL) return NULL;
    
    LIST_delete(list, element);

    return element;
}

int
LIST_count(LIST *list)
{
    return list->count;
}

int
LIST_is_empty(LIST *list)
{
    return list->count == 0;
}

void
LIST_apply(LIST *list, void (*fp)(void *element, void *user_data),
    void *user_data)
{
    LIST_ELEMENT *cur;

    for (cur = list->head; cur != NULL; cur = LIST_next(cur)) {
        fp(cur, user_data);
    }
}

static void
split(LIST *list, LIST *new, void *new_head, int new_count)
{
	LIST_ELEMENT *tail, *cur;

	LIST_init(new);

	if (new_head == NULL) {
		return;
	}

	tail = ((LIST_ELEMENT *)new_head)->prev;

	if (tail == NULL) {
        /* old list will now be empty */
		*new = *list;
		LIST_init(list);
        return;
	} else {
        if (new_count == -1) {
            new_count = 0;
            for (cur = new_head; cur != NULL; cur = cur->next) {
                new_count++;
            }
        }

        new->head = new_head;
        new->tail = list->tail;
        new->count = new_count;

        list->tail = tail;
        list->count -= new_count;

        tail->next = NULL;
        ((LIST_ELEMENT *)new_head)->prev = NULL;
    }
}

void
LIST_sort(LIST *list,
    int (*compare)(void *element_a, void *element_b, void *user_data),
    void *user_data)
{
	int i;
	LIST new;
	LIST_ELEMENT *a, *b;

	if (list->count <= 1) {
		return;
	}

	a = list->head;
	for (i = 0; i < list->count/2; i++) {
		a = a->next;
	}

	split(list, &new, a, list->count - i);
	LIST_sort(list, compare, user_data);
	LIST_sort(&new, compare, user_data);

	a = list->head;

	while ((b = new.head) != NULL) {
		if (a == NULL) {
			LIST_delete(&new, b);
			LIST_append(list, b);
		} else if (compare(a, b, user_data) <= 0) {
			a = a->next;
		} else {
			LIST_delete(&new, b);
			LIST_insert_before(list, b, a);
		}
	}
}

void
LIST_insert_before(LIST *list, void *element, void *before_element)
{
    LIST_ELEMENT *e = element;
    LIST_ELEMENT *be = before_element;
    
	if (e == NULL)
		return;

	if (be == NULL) {
		e->next = NULL;
		e->prev = list->tail;

		if (list->tail == NULL)
			list->head = e;
		else
			list->tail->next = e;

		list->tail = e;
	} else {
		e->prev = be->prev;
		e->next = be;

		if (be->prev == NULL)
			list->head = e;
		else
			be->prev->next = e;

		be->prev = e;
	}

	list->count++;
}

