/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/matrix/matrix.h,v 1.14 2001/12/20 21:22:26 wex Exp $
 *
 * $Log: matrix.h,v $
 * Revision 1.14  2001/12/20 21:22:26  wex
 * Support for Win32
 *
 * Revision 1.13  2001/12/16 00:33:39  wex
 * Removed old code
 *
 * Revision 1.12  2001/12/08 23:23:58  wex
 * Return w when projecting, added matrix copy.
 *
 * Revision 1.11  2001/12/05 10:18:18  wex
 * Added a few more routines and macros
 *
 * Revision 1.10  2001/11/12 18:52:06  wex
 * Cleanup for Python1.5
 *
 * Revision 1.9  2001/11/12 09:10:19  wex
 * A few more macros.
 *
 * Revision 1.8  2001/11/10 05:58:15  wex
 * Added windowed projection
 *
 * Revision 1.7  2001/11/04 08:15:47  wex
 * MATRIX_check
 *
 * Revision 1.6  2001/11/02 19:54:56  wex
 * Added VECTOR_check
 *
 * Revision 1.5  2001/11/02 03:16:47  wex
 * Vector Python type and functions
 *
 * Revision 1.4  2001/11/01 04:26:56  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.3  2001/10/26 15:48:40  wex
 * First working version
 *
 * Revision 1.2  2001/07/18 04:58:04  wex
 * First round of bugs cleared, basically working.
 *
 * Revision 1.1  2001/07/17 09:37:26  wex
 * Added glue
 *
 *
 */

#ifndef _MATRIX_H_
#define _MATRIX_H_

#include <Python.h>

/* error return codes */
#define MATRIX_ERR_NONE         0
#define MATRIX_ERR_MEMORY       1
#define MATRIX_ERR_NOINVERSE    2

/* contains the error code if a function returns an error */
extern int MATRIX_errno;

/* returns a user readable string for error messages */
char *MATRIX_error_string(void);

typedef struct matrix MATRIX;

/* VECTOR type is public to allow for C stack allocation and macros */
typedef struct vector {
    PyObject_HEAD
    double v[3];
} VECTOR;


/** Matrix **/

MATRIX *MATRIX_create(void);
void MATRIX_destroy(MATRIX *m);

int MATRIX_check(MATRIX *m);

/* Sets the matrix to the specified affine transformation */
void MATRIX_set_identity(MATRIX *m);
void MATRIX_set_scale(MATRIX *m, double x, double y, double z);
void MATRIX_set_translate(MATRIX *m, double x, double y, double z);
void MATRIX_set_rotate(MATRIX *m, double x, double y, double z);/*ZXY*/
void MATRIX_set_lens_projection(MATRIX *m, double fov, double hither,
    double yon);
void MATRIX_set_window_projection(MATRIX *m, double window[4],
    double hither, double yon);

double MATRIX_determinant(MATRIX *m);
void MATRIX_copy(MATRIX *src, MATRIX *dst);

void MATRIX_transpose(MATRIX *src, MATRIX *dst);

/* All operations work in with dst=argument if necessary */
void MATRIX_multiply(MATRIX *a, MATRIX *b, MATRIX *dst);
void MATRIX_adjoint(MATRIX *m, MATRIX *dst);
int MATRIX_invert(MATRIX *m, MATRIX *dst);  /* returns -1 on error */
void MATRIX_lookat(MATRIX *m, double tx, double ty, double tz);

/* Post-multiplies the matrix by the specified affine transformation */
void MATRIX_scale(MATRIX *m, double x, double y, double z);
void MATRIX_translate(MATRIX *m, double x, double y, double z);
void MATRIX_rotate(MATRIX *m, double x, double y, double z); /* ZYX */

int MATRIX_print(MATRIX *self, FILE *fp);


/** Vector **/

VECTOR *VECTOR_create(double x, double y, double z);
void VECTOR_destroy(VECTOR *v);

/* returns true if this is a vector object */
int VECTOR_check(PyObject *pobj);

void VECTOR_set(VECTOR *v, double x, double y, double z);
double VECTOR_length(VECTOR *v);

/* These operations modify the vector */
void VECTOR_normalize(VECTOR *v);
void VECTOR_scale(VECTOR *a, double scale);
void VECTOR_add_scalar(VECTOR *a, double scale);

double VECTOR_dot(VECTOR *a, VECTOR *b);
void VECTOR_cross(VECTOR *a, VECTOR *b, VECTOR *dst);
void VECTOR_add(VECTOR *a, VECTOR *b, VECTOR *dst);

void VECTOR_print(VECTOR *v, FILE *fd);


/** Vector/Matrix Transformations **/

/* 4x4 matrix, w/o division */
void VECTOR_transform(VECTOR *v, MATRIX *m, VECTOR *dst);
/* 4x4 matrix, w/  division, returns w */
double VECTOR_project(VECTOR *v, MATRIX *m, VECTOR *dst);   
/* 3x3 w/o division */
void VECTOR_3x3_transform(VECTOR *v, MATRIX *m, VECTOR *dst);


#define V2_SUB(a, b, diff) ((diff)[0]=(a)[0]-(b)[0], (diff)[1]=(a)[1]-(b)[1])
#define V2_2X2_DET(a, b) ((a)[0]*(b)[1] - (a)[1]*(b)[0])

#define V3_SET(v, x, y, z) ((v)[0] = (x), (v)[1] = (y), (v)[2] = (z))
#define V3_ZERO(v) (V3_SET(v, 0,0,0))

#define V3_COPY(v, dst) ((dst)[0]=(v)[0], (dst)[1]=(v)[1], (dst)[2]=(v)[2])

#define V3_SUB(v0, v1, dst) \
	((dst)[0] = (v0)[0] - (v1)[0], \
	 (dst)[1] = (v0)[1] - (v1)[1], \
     (dst)[2] = (v0)[2] - (v1)[2])

#define V3_ADD(v0, v1, dst) \
	((dst)[0] = (v0)[0] + (v1)[0], \
	 (dst)[1] = (v0)[1] + (v1)[1], \
     (dst)[2] = (v0)[2] + (v1)[2])

#define V3_MUL(v0, v1, dst) \
	((dst)[0] = (v0)[0] * (v1)[0], \
	 (dst)[1] = (v0)[1] * (v1)[1], \
     (dst)[2] = (v0)[2] * (v1)[2])

#define V3_INC(a, dst) \
	((dst)[0] += (a)[0], \
	 (dst)[1] += (a)[1], \
	 (dst)[2] += (a)[2])

#define V3_CROSS(v0, v1, dst) \
	((dst)[0] = (v0)[1]*(v1)[2] - (v0)[2]*(v1)[1],	\
     (dst)[1] = (v0)[2]*(v1)[0] - (v0)[0]*(v1)[2],	\
     (dst)[2] = (v0)[0]*(v1)[1] - (v0)[1]*(v1)[0])

#define V3_SMUL(s, v, dst) \
	((dst)[0] = (s)*(v)[0],(dst)[1] = (s)*(v)[1], (dst)[2] = (s)*(v)[2])

#define V3_SDIV(s, v, dst) {double __s = 1.0 / (s);V3_SMUL(__s, v, dst);}

#define V3_DOT(v0, v1) ((v0)[0]*(v1)[0] + (v0)[1]*(v1)[1] + (v0)[2]*(v1)[2])

#define V3_LEN(v) (sqrt(V3_DOT(v, v)))

#define V3_NORMALIZE(v) \
	{double __len=V3_LEN(v);if(__len==0){V3_ZERO(v);}else{V3_SDIV(__len,v,v);}}

#define V3_LERP(v0, v1, t, dst) \
	(   (dst)[0] = (v0)[0] + t * ((v1)[0] - (v0)[0]), \
        (dst)[1] = (v0)[1] + t * ((v1)[1] - (v0)[1]), \
        (dst)[2] = (v0)[2] + t * ((v1)[2] - (v0)[2])  )


#endif /* _MATRIX_H_ */
