/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/matrix/matrix_glue.c,v 1.16 2002/01/03 05:16:09 wex Exp $
 *
 * $Log: matrix_glue.c,v $
 * Revision 1.16  2002/01/03 05:16:09  wex
 * Switched to matrixcmodule for package format
 *
 * Revision 1.15  2001/12/21 07:14:39  wex
 * Added POOL support (requires Python2.x)
 *
 * Revision 1.14  2001/12/20 21:22:26  wex
 * Support for Win32
 *
 * Revision 1.13  2001/12/08 23:23:58  wex
 * Return w when projecting, added matrix copy.
 *
 * Revision 1.12  2001/12/05 10:18:18  wex
 * Added a few more routines and macros
 *
 * Revision 1.11  2001/11/16 09:23:51  wex
 * Stats
 *
 * Revision 1.10  2001/11/13 07:53:01  wex
 * MATRIX objects are always python objects now.
 *
 * Revision 1.9  2001/11/06 06:56:53  wex
 * Basic transformations fixed up
 *
 * Revision 1.8  2001/11/04 08:15:48  wex
 * MATRIX_check
 *
 * Revision 1.7  2001/11/03 07:48:36  wex
 * Added vector_negative
 *
 * Revision 1.6  2001/11/02 19:54:56  wex
 * Added VECTOR_check
 *
 * Revision 1.5  2001/11/02 03:16:47  wex
 * Vector Python type and functions
 *
 * Revision 1.4  2001/11/01 04:26:56  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.3  2001/10/26 15:48:40  wex
 * First working version
 *
 * Revision 1.2  2001/07/18 04:58:04  wex
 * First round of bugs cleared, basically working.
 *
 * Revision 1.1  2001/07/17 09:37:26  wex
 * Added glue
 *
 *
 */

#include "matrix_internal.h"

#ifdef MATRIX_USE_POOL
#include <pool.h>

static POOL *matrix_pool = NULL;
static POOL *vector_pool = NULL;
#endif


#define DEGTORAD		.017453292519943295769


static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isMatrix(m) ((m)->ob_type == &Matrix)
#define isVector(m) ((m)->ob_type == &Vector)


static PyObject *matrix_set_identity(MATRIX *self, PyObject *args);
static PyObject *matrix_set_translate(MATRIX *self, PyObject *args);
static PyObject *matrix_set_scale(MATRIX *self, PyObject *args);
static PyObject *matrix_set_rotate(MATRIX *self, PyObject *args);
static PyObject *matrix_translate(MATRIX *self, PyObject *args);
static PyObject *matrix_scale(MATRIX *self, PyObject *args);
static PyObject *matrix_rotate(MATRIX *self, PyObject *args);
static PyObject *matrix_lookat(MATRIX *self, PyObject *args);
static PyObject *matrix_lens(MATRIX *self, PyObject *args);
static PyObject *matrix_copy(MATRIX *self, PyObject *args);
static PyObject *matrix_project(MATRIX *self, PyObject *args);
static PyObject *matrix_transform(MATRIX *self, PyObject *args);
static PyObject *matrix_3x3_transform(MATRIX *self, PyObject *args);
static PyObject *matrix_transpose(MATRIX *self, PyObject *args);
static PyObject *matrix_multiply(MATRIX *self, MATRIX *a);
static PyObject *matrix_invert(MATRIX *self);
static PyObject *matrix_new(PyObject *self, PyObject *args);
static void matrix_dealloc(MATRIX *self);
static int matrix_print(MATRIX *self, FILE *fp, int flags);
static PyObject *matrix_getattr(MATRIX *self, char *name);
static PyObject *matrix_item(MATRIX *self, int index);

static PyObject *vector_new(PyObject *self, PyObject *args);
static void vector_dealloc(VECTOR *self);
static int vector_print(VECTOR *self, FILE *fp, int flags);
static PyObject *vector_abs(VECTOR *self);
static PyObject *vector_getattr(VECTOR *self, char *name);
static PyObject *vector_item(VECTOR *self, int index);
static int vector_assign_item(VECTOR *self, int index, PyObject *args);
static int vector_nonzero(VECTOR *self);
static PyObject *vector_multiply(VECTOR *a, VECTOR *b);
static PyObject *vector_add(VECTOR *a, VECTOR *b);
static PyObject *vector_sub(VECTOR *a, VECTOR *b);
static PyObject *vector_dot(VECTOR *a, VECTOR *b);
static PyObject *vector_dot_method(VECTOR *self, PyObject *args);
static PyObject *vector_cross(VECTOR *a, VECTOR *b);
static PyObject *vector_cross_method(VECTOR *self, PyObject *args);
static PyObject *vector_length(VECTOR *self, PyObject *args);
static PyObject *vector_negative(VECTOR *self);
static int vector_coerce(PyObject **pv, PyObject **pw);

/** Type Descriptors **/

static struct PyMethodDef matrix_methods[] = {
    {"set_identity",  (PyCFunction)matrix_set_identity,  1, NULL},
    {"set_translate", (PyCFunction)matrix_set_translate, 1, NULL},
    {"set_scale",     (PyCFunction)matrix_set_scale,     1, NULL},
    {"set_rotate",    (PyCFunction)matrix_set_rotate,    1, NULL},
    {"translate",     (PyCFunction)matrix_translate,     1, NULL},
    {"scale",         (PyCFunction)matrix_scale,         1, NULL},
    {"rotate",        (PyCFunction)matrix_rotate,        1, NULL},
    {"lookat",        (PyCFunction)matrix_lookat,        1, NULL},
    {"lens",          (PyCFunction)matrix_lens,          1, NULL},
    {"copy",          (PyCFunction)matrix_copy,          1, NULL},
    {"project",       (PyCFunction)matrix_project,       1, NULL},
    {"transform",     (PyCFunction)matrix_transform,     1, NULL},
    {"transform_3x3", (PyCFunction)matrix_transform,     1, NULL},
    {"transpose",     (PyCFunction)matrix_transpose,     1, NULL},
    {NULL,            NULL}
};

static PySequenceMethods matrix_sequence = {
    (inquiry) NULL,                                 /* sq_length    "len(x)" */
    (binaryfunc) NULL,                              /* sq_concat    "x + y"  */
    (intargfunc) NULL,                              /* sq_repeat    "x * n"  */
    (intargfunc) matrix_item,                       /* sq_item      "x[i],in"*/
    (intintargfunc) NULL,                           /* sq_slice     "x[i:j]" */
    (intobjargproc) NULL,                           /* sq_ass_item  "x[i]=v" */
    (intintobjargproc) NULL,                        /* sq_ass_slice "x[i:j]=v*/
};

static PyNumberMethods matrix_number = {
    (binaryfunc) NULL,                              /* nb_add       "x + y"  */
    (binaryfunc) NULL,                              /* nb_sub       "x - y"  */
    (binaryfunc) matrix_multiply,                   /* nb_multiply  "x * y"  */
    (binaryfunc) NULL,                              /* nb_divide    "x / y"  */
    (binaryfunc) NULL,                              /* nb_remainder "x % y"  */
    (binaryfunc) NULL,                              /* nb_divmod    "divmod" */
    (ternaryfunc) NULL,                             /* nb_power     "pow(x)" */
    (unaryfunc) NULL,                               /* nb_negative  "-x"     */
    (unaryfunc) NULL,                               /* nb_positive  "+x"     */
    (unaryfunc) NULL,                               /* nb_absolute  "abs(x)" */
    (inquiry) NULL,                                 /* nb_nonzero   "if x"   */
    (unaryfunc) matrix_invert,                      /* nb_invert    "~x"     */
    (binaryfunc) NULL,                              /* nb_lshift    "x << y" */
    (binaryfunc) NULL,                              /* nb_rshift    "x >> y" */
    (binaryfunc) NULL,                              /* nb_and       "x & y"  */
    (binaryfunc) NULL,                              /* nb_xor       "x ^ y"  */
    (binaryfunc) NULL,                              /* nb_or        "x | y"  */
    (coercion) NULL,                                /* nb_coerce    (n/a)    */
    (unaryfunc) NULL,                               /* nb_int       "int(x)" */
    (unaryfunc) NULL,                               /* nb_long      "long(x)"*/
    (unaryfunc) NULL,                               /* nb_float     "float(x)*/
    (unaryfunc) NULL,                               /* nb_oct       "oct(x)" */
    (unaryfunc) NULL,                               /* nb_hex       "hex(x)" */
};

static PyTypeObject Matrix = {
    /* type header */
#ifndef WIN32    
    PyObject_HEAD_INIT(&PyType_Type)
#else
    PyObject_HEAD_INIT(NULL)
#endif    
    0,                                              /* ob_size               */
    "matrix",                                       /* tp_name               */
    sizeof(MATRIX),                                 /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)matrix_dealloc,                     /* tp_dealloc refcount=0 */
    (printfunc)matrix_print,                        /* tp_print   "print x"  */
    (getattrfunc)matrix_getattr,                    /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    &matrix_number,                                 /* tp_as_number          */
    &matrix_sequence,                               /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef vector_methods[] = {
    {"dot",  (PyCFunction)vector_dot_method,  1, NULL},
    {"cross", (PyCFunction)vector_cross_method, 1, NULL},
    {"length",     (PyCFunction)vector_length,     1, NULL},
    {NULL,            NULL}
};

static PySequenceMethods vector_sequence = {
    (inquiry) NULL,                                 /* sq_length    "len(x)" */
    (binaryfunc) NULL,                              /* sq_concat    "x + y"  */
    (intargfunc) NULL,                              /* sq_repeat    "x * n"  */
    (intargfunc) vector_item,                       /* sq_item      "x[i],in"*/
    (intintargfunc) NULL,                           /* sq_slice     "x[i:j]" */
    (intobjargproc) vector_assign_item,             /* sq_ass_item  "x[i]=v" */
    (intintobjargproc) NULL,                        /* sq_ass_slice "x[i:j]=v*/
};

static PyNumberMethods vector_number = {
    (binaryfunc) vector_add,                        /* nb_add       "x + y"  */
    (binaryfunc) vector_sub,                        /* nb_sub       "x - y"  */
    (binaryfunc) vector_multiply,                   /* nb_multiply  "x * y"  */
    (binaryfunc) NULL,                              /* nb_divide    "x / y"  */
    (binaryfunc) vector_cross,                      /* nb_remainder "x % y"  */
    (binaryfunc) NULL,                              /* nb_divmod    "divmod" */
    (ternaryfunc) NULL,                             /* nb_power     "pow(x)" */
    (unaryfunc) vector_negative,                    /* nb_negative  "-x"     */
    (unaryfunc) NULL,                               /* nb_positive  "+x"     */
    (unaryfunc) vector_abs,                         /* nb_absolute  "abs(x)" */
    (inquiry) vector_nonzero,                       /* nb_nonzero   "if x"   */
    (unaryfunc) NULL,                               /* nb_invert    "~x"     */
    (binaryfunc) NULL,                              /* nb_lshift    "x << y" */
    (binaryfunc) NULL,                              /* nb_rshift    "x >> y" */
    (binaryfunc) NULL,                              /* nb_and       "x & y"  */
    (binaryfunc) vector_dot,                        /* nb_xor       "x ^ y"  */
    (binaryfunc) NULL,                              /* nb_or        "x | y"  */
    (coercion) vector_coerce,                       /* nb_coerce    (n/a)    */
    (unaryfunc) NULL,                               /* nb_int       "int(x)" */
    (unaryfunc) NULL,                               /* nb_long      "long(x)"*/
    (unaryfunc) NULL,                               /* nb_float     "float(x)*/
    (unaryfunc) NULL,                               /* nb_oct       "oct(x)" */
    (unaryfunc) NULL,                               /* nb_hex       "hex(x)" */
};

static PyTypeObject Vector = {
    /* type header */
#ifndef WIN32    
    PyObject_HEAD_INIT(&PyType_Type)
#else
    PyObject_HEAD_INIT(NULL)
#endif    
    0,                                              /* ob_size               */
    "vector",                                       /* tp_name               */
    sizeof(VECTOR),                                 /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)vector_dealloc,                     /* tp_dealloc refcount=0 */
    (printfunc)vector_print,                        /* tp_print   "print x"  */
    (getattrfunc)vector_getattr,                    /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    &vector_number,                                 /* tp_as_number          */
    &vector_sequence,                               /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef matrixtype_methods[] = {
    {"Matrix", matrix_new, 1},
    {"Vector", vector_new, 1},
    {NULL, NULL}
};


int matrix_alloced = 0;
int matrix_destroyed = 0;
int matrix_max = 0;

int vector_alloced = 0;
int vector_destroyed = 0;
int vector_max = 0;



void
initmatrixcmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32    
    Matrix.ob_type = &PyType_Type;
    Vector.ob_type = &PyType_Type;
#endif

#ifdef MATRIX_USE_POOL    
    matrix_pool = POOL_create(sizeof(MATRIX), 1024);
    vector_pool = POOL_create(sizeof(VECTOR), 4096);
#endif
    
    m = Py_InitModule("matrixcmodule", matrixtype_methods);
    d = PyModule_GetDict(m);                        /* with 'Matrix' func */
    ErrorObject = Py_BuildValue("s", "matrixcmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize matrix C module");
    }
}


/** Instance Methods **/

static PyObject *
matrix_set_identity(MATRIX *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;   /* verify no arguments */
    MATRIX_set_identity(self);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_set_translate(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("translate requires three arguments - x,y,z");
    MATRIX_set_translate(self, x, y, z);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_set_scale(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("scale requires three arguments - x,y,z");
    MATRIX_set_scale(self, x, y, z);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_set_rotate(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("scale requires three arguments - x,y,z");
    MATRIX_set_rotate(self, x*DEGTORAD, y*DEGTORAD, z*DEGTORAD);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_translate(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("translate requires three arguments - x,y,z");
    MATRIX_translate(self, x, y, z);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_scale(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("scale requires three arguments - x,y,z");
    MATRIX_scale(self, x, y, z);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_rotate(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("rotate requires three arguments - x,y,z");
    MATRIX_rotate(self, x*DEGTORAD, y*DEGTORAD, z*DEGTORAD);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_lookat(MATRIX *self, PyObject *args)
{
    double x, y, z;
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) 
        onError("lookat requires three arguments - x,y,z");
    MATRIX_lookat(self, x, y, z);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_lens(MATRIX *self, PyObject *args)
{
    double fov, hither, yon;
    if (!PyArg_ParseTuple(args, "ddd", &fov, &hither, &yon)) 
        onError("lens requires three arguments - fov, hither, yon");
    MATRIX_set_lens_projection(self, fov*DEGTORAD, hither, yon);
    Py_INCREF(self);
    return (PyObject *)self;
}

static PyObject *
matrix_copy(MATRIX *self, PyObject *args)
{
    MATRIX *m;
    
    if (!PyArg_ParseTuple(args, ""))
        onError("copy cannot be passed any arguments");
    m = matrix_alloc();
    MATRIX_copy(self, m);
    return (PyObject *)m;
}

static PyObject *
matrix_project(MATRIX *self, PyObject *args)
{
    VECTOR *src, *dst;
    
    if (!PyArg_ParseTuple(args, "O", &src))
        onError("project requires a single argument");

    if (!VECTOR_check((PyObject *)src)) 
        onError("project requires a single vector argument");

    dst = VECTOR_create(0,0,0);
    VECTOR_project(src, self, dst);
    return (PyObject *)dst;
}

static PyObject *
matrix_transform(MATRIX *self, PyObject *args)
{
    VECTOR *src, *dst;
    
    if (!PyArg_ParseTuple(args, "O", &src))
        onError("transform requires a single argument");

    if (!VECTOR_check((PyObject *)src)) 
        onError("transform requires a single vector argument");

    dst = VECTOR_create(0,0,0);
    VECTOR_transform(src, self, dst);
    return (PyObject *)dst;
}
    
static PyObject *
matrix_3x3_transform(MATRIX *self, PyObject *args)
{
    VECTOR *src, *dst;
    
    if (!PyArg_ParseTuple(args, "O", &src))
        onError("transform requires a single argument");

    if (!VECTOR_check((PyObject *)src)) 
        onError("transform requires a single vector argument");

    dst = VECTOR_create(0,0,0);
    VECTOR_3x3_transform(src, self, dst);
    return (PyObject *)dst;
}

static PyObject *
matrix_transpose(MATRIX *self, PyObject *args)
{
    MATRIX *d;
    
    if (!PyArg_ParseTuple(args, ""))
        onError("transpose does not take any arguments");

    d = matrix_alloc();
    MATRIX_transpose(self, d);
    return (PyObject *)d;
}

static PyObject *
matrix_multiply(MATRIX *self, MATRIX *a)
{
    MATRIX *d = matrix_alloc();
    MATRIX_multiply(self, a, d);
    return (PyObject *)d;
}

static PyObject *
matrix_invert(MATRIX *self)
{
    MATRIX *d = matrix_alloc();
    MATRIX_invert(self, d);
    return (PyObject *)d;
}

/** Basic Type Methods **/

static PyObject *
matrix_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)matrix_alloc();
}

MATRIX *
matrix_alloc(void)
{
    MATRIX *m;
    matrix_alloced++;
    if (matrix_alloced - matrix_destroyed > matrix_max) {
        matrix_max = matrix_alloced - matrix_destroyed;
    }

#ifdef MATRIX_USE_POOL
    m = POOL_create_item(matrix_pool);
    PyObject_INIT(m, &Matrix);
#else
    m = PyObject_NEW(MATRIX, &Matrix);              /* malloc, init, incref */
#endif    

    matrix_initialize(m);
    return m;
}

static void
matrix_dealloc(MATRIX *self)
{
    matrix_destroyed++;
    matrix_delete(self);
#ifdef MATRIX_USE_POOL
    POOL_destroy_item(matrix_pool, self);
#else
    PyMem_DEL(self);
#endif    
}

static int
matrix_print(MATRIX *self, FILE *fp, int flags)
{
    MATRIX_print(self, fp);
    return 0;
}

static PyObject *
matrix_getattr(MATRIX *self, char *name)
{
    if (strcmp(name, "__members__") == 0) {
        return Py_BuildValue("[s]", "data");
    } else if (strcmp(name, "data") == 0) {
        return Py_BuildValue("((dddd)(dddd)(dddd)(dddd))",
            self->m[0][0], self->m[0][1], self->m[0][2], self->m[0][3],
            self->m[1][0], self->m[1][1], self->m[1][2], self->m[1][3],
            self->m[2][0], self->m[2][1], self->m[2][2], self->m[2][3],
            self->m[3][0], self->m[3][1], self->m[3][2], self->m[3][3]);
    }
#if 0    
    else if (strcmp(name, "parent") == 0) {
        return Py_BuildValue("O", self->parent);
    } else if (strcmp(name, "child_count") == 0) {
        return Py_BuildValue("i", self->child_count);
    } else if (strcmp(name, "child") == 0) {
        return Py_BuildValue("???");
    }
#endif

    return Py_FindMethod(matrix_methods, (PyObject *)self, name);
}


/** Sequence Type Operations **/

static PyObject *
matrix_item(MATRIX *self, int index)
{
    if (index < 0 || index > 3) {
        PyErr_SetString(PyExc_IndexError, "matricies only have 4 indicies");
        return NULL;
    }

    return Py_BuildValue("(dddd)", self->m[index][0], self->m[index][1],
        self->m[index][2], self->m[index][3]);
}

int
MATRIX_check(MATRIX *m)
{
    return isMatrix(m);
}


/**** Vector ****/

/** Instance Methods **/

static PyObject *
vector_dot(VECTOR *a, VECTOR *b)
{
    return (PyObject *)PyFloat_FromDouble(V3_DOT(a->v, b->v));
}

static PyObject *
vector_dot_method(VECTOR *self, PyObject *args)
{
    VECTOR *b;
    
    if (!PyArg_ParseTuple(args, "O", &b))
        onError("Dot product requires a vector argument");

    return PyFloat_FromDouble(VECTOR_dot(self, b));
}

static PyObject *
vector_cross(VECTOR *a, VECTOR *b)
{
    VECTOR *c;

    c = VECTOR_create(0,0,0);
    VECTOR_cross(a, b, c);
    return (PyObject *)c;
}

static PyObject *
vector_cross_method(VECTOR *self, PyObject *args)
{
    VECTOR *b, *c;

    if (!PyArg_ParseTuple(args, "O", &b)) 
        onError("Cross requires a vector argument");

    c = VECTOR_create(0,0,0);
    VECTOR_cross(self, b, c);

    return (PyObject *)c;
}

static PyObject *
vector_negative(VECTOR *self)
{
    VECTOR *v;

    v = VECTOR_create(-self->v[0], -self->v[1], -self->v[2]);
    return (PyObject *)v;
}

static PyObject *
vector_length(VECTOR *self, PyObject *args)
{
    return PyFloat_FromDouble(VECTOR_length(self));
}

/** Basic Type Methods **/

static PyObject *
vector_new(PyObject *self, PyObject *args)
{
    double x, y, z;
    
    if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) {
        fprintf(stderr, "Vector creation requires x, y and z\n");
        return NULL;
    }
        
    return (PyObject *)VECTOR_create(x, y, z);
}

VECTOR *
VECTOR_create(double x, double y, double z)
{
    VECTOR *v;

    vector_alloced++;
    if (vector_alloced - vector_destroyed > vector_max) {
        vector_max = vector_alloced - vector_destroyed;
    }

#ifdef MATRIX_USE_POOL
    v = POOL_create_item(vector_pool);
    PyObject_INIT(v, &Vector);
#else
    v = PyObject_NEW(VECTOR, &Vector);              /* malloc, init, incref */
#endif    
    V3_SET(v->v, x,y,z);
    return v;
}

static void
vector_dealloc(VECTOR *self)
{
    vector_destroyed++;
#ifdef MATRIX_USE_POOL
    POOL_destroy_item(vector_pool, self);
#else    
    PyMem_DEL(self);
#endif    
}

static int
vector_print(VECTOR *self, FILE *fp, int flags)
{
    VECTOR_print(self, fp);
    return 0;
}

static PyObject *
vector_abs(VECTOR *self)
{
    VECTOR *v;

    v = VECTOR_create(fabs(self->v[0]), fabs(self->v[1]), fabs(self->v[2]));
    return (PyObject *)v;
}

static PyObject *
vector_getattr(VECTOR *self, char *name)
{
    if (strcmp(name, "x") == 0 ||strcmp(name, "r") == 0) {
        return PyFloat_FromDouble(self->v[0]);
    }
    
    if (strcmp(name, "y") == 0 ||strcmp(name, "g") == 0) {
        return PyFloat_FromDouble(self->v[1]);
    }
    
    if (strcmp(name, "z") == 0 ||strcmp(name, "b") == 0) {
        return PyFloat_FromDouble(self->v[2]);
    }
    
    return Py_FindMethod(vector_methods, (PyObject *)self, name);
}


/** Sequence Type Operations **/

static PyObject *
vector_item(VECTOR *self, int index)
{
    if (index < 0 || index > 2) {
        PyErr_SetString(PyExc_IndexError, "vectors only have 3 indicies");
        return NULL;
    }

    return PyFloat_FromDouble(self->v[index]);
}

static int
vector_nonzero(VECTOR *self)
{
    return self->v[0] != 0 && self->v[1] != 0 && self->v[2] != 0;
}

static int
vector_assign_item(VECTOR *self, int index, PyObject *args)
{
    double v;
    
    if (index < 0 || index > 2) {
        PyErr_SetString(ErrorObject, "Vectors only have three indicies");
        return -1;
    }

    if (!PyArg_ParseTuple(args, "d", &v)) {
        PyErr_SetString(ErrorObject, "Can only assign double values to "
            "vector components");
        return -1;
    }

    self->v[index] = v;

    return 0;
}

int
VECTOR_check(PyObject *pobj)
{
    return isVector(pobj);
}

static int
vector_coerce(PyObject **pv, PyObject **pw)
{
    double a, x, y, z;
    PyObject *pf;

    if (PyInt_Check(*pw)) {
        a = PyInt_AsLong(*pw);
        *pw = (PyObject *)VECTOR_create(a, a, a);
        Py_INCREF(*pv);
        return 0;
    }

    if (PyFloat_Check(*pw)) {
        a = PyFloat_AsDouble(*pw);
        *pw = (PyObject *)VECTOR_create(a, a, a);
        Py_INCREF(*pv);
        return 0;
    }

    if (PyTuple_Check(*pw) && PyTuple_Size(*pw) == 3) {
        pf = PyTuple_GET_ITEM(*pw, 0);
        if (PyFloat_Check(pf)) {
            x = PyFloat_AsDouble(pf);
        } else if (PyInt_Check(pf)) {
            x = PyInt_AsLong(pf);
        } else {
            return 1;
        }
        
        pf = PyTuple_GET_ITEM(*pw, 1);
        if (PyFloat_Check(pf)) {
            y = PyFloat_AsDouble(pf);
        } else if (PyInt_Check(pf)) {
            y = PyInt_AsLong(pf);
        } else {
            return 1;
        }
        
        pf = PyTuple_GET_ITEM(*pw, 2);
        if (PyFloat_Check(pf)) {
            z = PyFloat_AsDouble(pf);
        } else if (PyInt_Check(pf)) {
            z = PyInt_AsLong(pf);
        } else {
            return 1;
        }

        *pw = (PyObject *)VECTOR_create(x, y, z);
        Py_INCREF(*pv);

        return 0;
    }
        
    return 1;
}

static PyObject *
vector_multiply(VECTOR *a, VECTOR *b)
{
    double p[3];
    
    V3_MUL(a->v, b->v, p);
    return (PyObject *)VECTOR_create(p[0], p[1], p[2]);
}

static PyObject *
vector_add(VECTOR *a, VECTOR *b)
{
    double p[3];
    
    V3_ADD(a->v, b->v, p);
    return (PyObject *)VECTOR_create(p[0], p[1], p[2]);
}

static PyObject *
vector_sub(VECTOR *a, VECTOR *b)
{
    double p[3];
    
    V3_SUB(a->v, b->v, p);
    return (PyObject *)VECTOR_create(p[0], p[1], p[2]);
}

