/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/shader/shader.h,v 1.9 2001/12/08 23:32:41 wex Exp $
 *
 * $Log: shader.h,v $
 * Revision 1.9  2001/12/08 23:32:41  wex
 * Geometry viewports and SHADER_LIGHT_destroy
 *
 * Revision 1.8  2001/12/05 10:26:42  wex
 * Error and type checking
 *
 * Revision 1.7  2001/11/16 00:52:26  wex
 * A few more memory leaks
 *
 * Revision 1.6  2001/11/12 09:11:38  wex
 * Fixed up some memory leakage.
 *
 * Revision 1.5  2001/11/06 06:53:25  wex
 * Added motion matrix for geometry
 *
 * Revision 1.4  2001/11/04 08:15:16  wex
 * Changes to support material bindings.
 *
 * Revision 1.3  2001/11/02 03:17:15  wex
 * Support for Python material and light shaders
 *
 * Revision 1.2  2001/11/01 04:26:57  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.1  2001/10/26 15:48:41  wex
 * First working version
 *
 *
 */

#ifndef _SHADER_H_
#define _SHADER_H_

#include <matrix.h>
#include <geometry.h>


/* error return codes */
#define SHADER_ERR_NONE         0
#define SHADER_ERR_MEMORY       1

/* contains the error code if a function returns an error */
extern int SHADER_errno;

/* returns a user readable string for error messages */
char *SHADER_error_string(void);

typedef enum {
    SHADER_TYPE_NONE,
    SHADER_TYPE_GEOMETRY,
    SHADER_TYPE_MATERIAL,
    SHADER_TYPE_LIGHT
} SHADER_TYPE;

typedef enum {
    SHADER_DATA_TYPE_INT,
    SHADER_DATA_TYPE_DOUBLE,
    SHADER_DATA_TYPE_VECTOR,
    SHADER_DATA_TYPE_STRING
} SHADER_DATA_TYPE;

typedef struct shader SHADER;
typedef struct shader_geometry SHADER_GEOMETRY;
typedef struct shader_material SHADER_MATERIAL;
typedef struct shader_light SHADER_LIGHT;

typedef int (*SHADER_DATA_CALLBACK)(SHADER *shader, void *user_data,
    char *name, SHADER_DATA_TYPE type, void *result);

SHADER_GEOMETRY *SHADER_GEOMETRY_create(PyObject *pobj);
SHADER_MATERIAL *SHADER_MATERIAL_create(PyObject *pobj);
SHADER_LIGHT *SHADER_LIGHT_create(PyObject *pobj);

void SHADER_GEOMETRY_destroy(SHADER_GEOMETRY *geometry);
void SHADER_LIGHT_destroy(SHADER_LIGHT *light);

/* all lights default to "on", use this to disable */
int SHADER_LIGHT_mask_material(SHADER_LIGHT *light, SHADER_MATERIAL *material,
    int on);

/* set the affine space of the shader for vector transformations */
void SHADER_LIGHT_set_space(SHADER_LIGHT *light, MATRIX *l2w);
void SHADER_MATERIAL_set_space(SHADER_MATERIAL *material, MATRIX *l2w);

PyObject *SHADER_MATERIAL_get_pobj(SHADER_MATERIAL *material);
void SHADER_MATERIAL_set_pobj(SHADER_MATERIAL *material, PyObject *pobj);
SHADER_MATERIAL *SHADER_MATERIAL_get_from_pobj(PyObject *pobj);

int SHADER_MATERIAL_check(PyObject *pobj);

/* initiate a shading thread on the specified material, call get
   data functions to retrieve final information */
SHADER *SHADER_MATERIAL_shade(PyObject *material,
    SHADER_DATA_CALLBACK get_data, SHADER_DATA_CALLBACK set_data,
    void *user_data);
void SHADER_destroy(SHADER *shader);

MATRIX *SHADER_GEOMETRY_get_l2w(SHADER_GEOMETRY *geometry);
MATRIX *SHADER_GEOMETRY_get_l2w_motion(SHADER_GEOMETRY *geometry);
void SHADER_GEOMETRY_get_bbox(SHADER_GEOMETRY *geometry,
    VECTOR *min, VECTOR *max);
int SHADER_GEOMETRY_generate(SHADER_GEOMETRY *geometry, MATRIX *w2s,
    int viewport[4], double area, GRID_CAPTURE grid_capture, void *user_data);

int SHADER_get_double(SHADER *shader, char *name, double *dst);
int SHADER_get_vector(SHADER *shader, char *name, VECTOR *dst);
int SHADER_get_string(SHADER *shader, char *name, char **dst);
int SHADER_get_int(SHADER *shader, char *name, int *dst);

int SHADER_set_double(SHADER *shader, char *name, double val);
int SHADER_set_vector(SHADER *shader, char *name, VECTOR *val);
int SHADER_set_string(SHADER *shader, char *name, char *val);
int SHADER_set_int(SHADER *shader, char *name, int val);

SHADER *SHADER_first_light(SHADER *shader);
SHADER *SHADER_next_light(SHADER *shader);
PyObject *SHADER_get_light_list(SHADER *shader);


#endif /* _SHADER_H_ */
