/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/shader/shader_data.c,v 1.5 2001/11/04 08:15:16 wex Exp $
 *
 * $Log: shader_data.c,v $
 * Revision 1.5  2001/11/04 08:15:16  wex
 * Changes to support material bindings.
 *
 * Revision 1.4  2001/11/02 19:57:13  wex
 * Check python object for shader data, handle tuples and vectors.
 *
 * Revision 1.3  2001/11/02 03:17:15  wex
 * Support for Python material and light shaders
 *
 * Revision 1.2  2001/10/29 00:50:28  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:41  wex
 * First working version
 *
 *
 */

#include "shader_internal.h"

static PyObject *shader_get_pobj(SHADER *shader);

/* only go to parent if data not in local shader or pobj */
static int
shader_get_data(SHADER *shader, char *name, SHADER_DATA_TYPE type, void *dst)
{
    SHADER_DATA *data;
    VECTOR *v;
    PyObject *pobj, *pattr, *pf;
    double x, y, z;
    
    if (shader->get_data_callback != NULL &&
        shader->get_data_callback(shader, shader->user_data,name,type,dst)==0){
        return 0;
    }

    /* first look in local shader? */
    for (data = LIST_head(shader->data_list); data != NULL;
         data = LIST_next(data)) {

        if (type == data->type && strcmp(name, data->name) == 0) {
            break;
        }
    }

    if (data != NULL) {
        switch (type) {
        case SHADER_DATA_TYPE_DOUBLE:
            *(double *)dst = data->data.d;
            break;
        case SHADER_DATA_TYPE_VECTOR:
            VECTOR_set((VECTOR *)dst, data->data.v[0], data->data.v[1],
                data->data.v[2]);
            break;
        case SHADER_DATA_TYPE_INT:
        case SHADER_DATA_TYPE_STRING:
            break;
        }

        return 0;
    }
        

    /* check the python object? */
    pobj = shader_get_pobj(shader);
    pattr = PyObject_GetAttrString(pobj, name);
    PyErr_Clear();
    if (pattr != NULL) {
        switch (type) {
        case SHADER_DATA_TYPE_DOUBLE:
            if (PyFloat_Check(pattr)) {
                *(double *)dst = PyFloat_AsDouble(pattr);
                Py_DECREF(pattr);
                return 0;
            }
            break;
        case SHADER_DATA_TYPE_VECTOR:
            if (VECTOR_check(pattr)) {
                v = (VECTOR *)pattr;
                VECTOR_set((VECTOR *)dst, v->v[0], v->v[1], v->v[2]);
                Py_DECREF(pattr);
                return 0;
            }
            if (PyTuple_Check(pattr) && PyTuple_Size(pattr) == 3) {
                pf = PyTuple_GET_ITEM(pattr, 0);
                if (!PyFloat_Check(pf) && !PyInt_Check(pf)) goto error;
                x = PyFloat_AsDouble(pf);
                pf = PyTuple_GET_ITEM(pattr, 1);
                if (!PyFloat_Check(pf) && !PyInt_Check(pf)) goto error;
                y = PyFloat_AsDouble(pf);
                pf = PyTuple_GET_ITEM(pattr, 2);
                if (!PyFloat_Check(pf) && !PyInt_Check(pf)) goto error;
                z = PyFloat_AsDouble(pf);
                VECTOR_set((VECTOR *)dst, x, y, z);
                Py_DECREF(pattr);
                return 0;
            }
            break;
        case SHADER_DATA_TYPE_INT:
        case SHADER_DATA_TYPE_STRING:
            break;
        }
        Py_DECREF(pattr);
    }

    if (shader->parent == NULL) {
        return -1;
    }
        
    return shader_get_data(shader->parent, name, type, dst);

 error:
    if (pattr != NULL) {
        Py_DECREF(pattr);
    }
    return -1;
}

static PyObject *
shader_get_pobj(SHADER *shader)
{
    switch (shader->type) {
    case SHADER_TYPE_MATERIAL: return shader->shader.material->pobj;
    case SHADER_TYPE_GEOMETRY: return shader->shader.geometry->pobj;
    case SHADER_TYPE_LIGHT: return shader->shader.light->pobj;
    case SHADER_TYPE_NONE:
    default:
        return NULL;
    }
}

    
int
SHADER_get_double(SHADER *shader, char *name, double *dst)
{
    return shader_get_data(shader, name, SHADER_DATA_TYPE_DOUBLE, dst);
}

int
SHADER_get_vector(SHADER *shader, char *name, VECTOR *dst)
{
    return shader_get_data(shader, name, SHADER_DATA_TYPE_VECTOR, dst);
}

int
SHADER_get_string(SHADER *shader, char *name, char **dst)
{
    return shader_get_data(shader, name, SHADER_DATA_TYPE_STRING, dst);
}

int
SHADER_get_int(SHADER *shader, char *name, int *dst)
{
    return shader_get_data(shader, name, SHADER_DATA_TYPE_INT, dst);
}



/* Ignore parents when setting data */
static int
shader_set_data(SHADER *shader, char *name, SHADER_DATA_TYPE type, void *val)
{
    SHADER_DATA *data;
    
    if (shader->set_data_callback != NULL &&
        shader->set_data_callback(shader, shader->user_data,name,type,val)==0){
        return 0;
    }

    /* store data in this shader for top level access */

    data = (SHADER_DATA *)malloc(sizeof(SHADER_DATA));
    if (data == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return -1;
    }

    data->name = strdup(name);
    data->type = type;
        
    switch (type) {
    case SHADER_DATA_TYPE_DOUBLE:
        data->data.d = *(double *)val;
        break;
    case SHADER_DATA_TYPE_VECTOR:
        data->data.v[0] = ((VECTOR *)val)->v[0];
        data->data.v[1] = ((VECTOR *)val)->v[1];
        data->data.v[2] = ((VECTOR *)val)->v[2];
        break;
    case SHADER_DATA_TYPE_INT:
    case SHADER_DATA_TYPE_STRING:
        break;
    }

    LIST_append(shader->data_list, data);

    return 0;
}

        
int
SHADER_set_double(SHADER *shader, char *name, double val)
{
    return shader_set_data(shader, name, SHADER_DATA_TYPE_DOUBLE, &val);
}

int
SHADER_set_vector(SHADER *shader, char *name, VECTOR *val)
{
    return shader_set_data(shader, name, SHADER_DATA_TYPE_VECTOR, val);
}

int
SHADER_set_string(SHADER *shader, char *name, char *val)
{
    return shader_set_data(shader, name, SHADER_DATA_TYPE_STRING, val);
}

int
SHADER_set_int(SHADER *shader, char *name, int val)
{
    return shader_set_data(shader, name, SHADER_DATA_TYPE_INT, &val);
}

