/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/shader/shader_geometry.c,v 1.10 2001/12/08 23:32:42 wex Exp $
 *
 * $Log: shader_geometry.c,v $
 * Revision 1.10  2001/12/08 23:32:42  wex
 * Geometry viewports and SHADER_LIGHT_destroy
 *
 * Revision 1.9  2001/11/16 00:52:26  wex
 * A few more memory leaks
 *
 * Revision 1.8  2001/11/13 07:49:12  wex
 * Python referencing cleanup, fixed leaks and bugs.
 *
 * Revision 1.7  2001/11/12 09:11:38  wex
 * Fixed up some memory leakage.
 *
 * Revision 1.6  2001/11/06 06:53:25  wex
 * Added motion matrix for geometry
 *
 * Revision 1.5  2001/11/04 08:15:16  wex
 * Changes to support material bindings.
 *
 * Revision 1.4  2001/11/03 07:42:51  wex
 * Updated setattr, error handling
 *
 * Revision 1.3  2001/11/02 03:17:15  wex
 * Support for Python material and light shaders
 *
 * Revision 1.2  2001/11/01 04:26:57  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.1  2001/10/26 15:48:41  wex
 * First working version
 *
 *
 */

#include <assert.h>

#include "shader_internal.h"

static int shader_geometry_default_bbox(SHADER *shader, VECTOR *min,
    VECTOR *max);

SHADER_GEOMETRY *
SHADER_GEOMETRY_create(PyObject *pobj)
{
    SHADER_GEOMETRY *geometry;

    geometry = (SHADER_GEOMETRY *)malloc(sizeof(SHADER_GEOMETRY));
    if (geometry == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return NULL;
    }

    Py_INCREF(pobj);
    
    geometry->pobj = pobj;
    geometry->bbox = shader_geometry_default_bbox;
    geometry->generate = PyObject_GetAttrString(pobj, "generate");

    return geometry;
}

void
SHADER_GEOMETRY_destroy(SHADER_GEOMETRY *geometry)
{
    assert(geometry->pobj != NULL);
    Py_DECREF(geometry->generate);
    Py_DECREF(geometry->pobj);
    geometry->pobj = NULL;
    free(geometry);
}

MATRIX *
SHADER_GEOMETRY_get_l2w(SHADER_GEOMETRY *geometry)
{
    PyObject *pattr;
    
    pattr = PyObject_GetAttrString(geometry->pobj, "l2w");
    if (pattr != NULL) {
        assert(MATRIX_check((MATRIX *)pattr));
        return (MATRIX *)pattr;
    }
    
    return MATRIX_create();
}

MATRIX *
SHADER_GEOMETRY_get_l2w_motion(SHADER_GEOMETRY *geometry)
{
    PyObject *pattr;

    if (PyObject_HasAttrString(geometry->pobj, "l2w_motion")) {
        pattr = PyObject_GetAttrString(geometry->pobj, "l2w_motion");
        if (pattr != NULL) {
            assert(MATRIX_check((MATRIX *)pattr));
            return (MATRIX *)pattr;
        }
    }

    return SHADER_GEOMETRY_get_l2w(geometry);
}

void
SHADER_GEOMETRY_get_bbox(SHADER_GEOMETRY *geometry, VECTOR *min, VECTOR *max)
{
    VECTOR_set(min, -1, -1, -1);
    VECTOR_set(max, 1, 1, 1);
}

int
SHADER_GEOMETRY_generate(SHADER_GEOMETRY *geometry, MATRIX *w2s,
    int viewport[4], double area, GRID_CAPTURE grid_capture, void *user_data)
{
    int status;
    SHADER *shader;
    PyObject *pargs, *pstatus, *pobj;
    MATRIX *l2w, *l2s;
    
    shader = shader_create(NULL, SHADER_TYPE_GEOMETRY, geometry);
    if (shader == NULL) {
        return -1;
    }
    
    shader->grid_capture = grid_capture;
    shader->grid_capture_data = user_data;
    
    assert(shader->type == SHADER_TYPE_GEOMETRY);

    /* extract and call the python generate function */
    pargs = Py_BuildValue("()");
    pstatus = PyEval_CallObject(geometry->generate, pargs);
    Py_DECREF(pargs);
    if (PyErr_Occurred()) PyErr_Print();
    if (pstatus != NULL) {
        PyArg_Parse(pstatus, "i", &status);
        Py_DECREF(pstatus);
    } else {
        status = 0;
    }

    l2w = SHADER_GEOMETRY_get_l2w(shader->shader.geometry);
    l2s = MATRIX_create();
    MATRIX_multiply(l2w, w2s, l2s);
    MATRIX_destroy(l2w);
    
    /* extract the geometry C type from the Python class */
    pobj = PyObject_GetAttrString(shader->shader.geometry->pobj, "geometry");
    GEOMETRY_generate_grids((GEOMETRY *)pobj, l2s, viewport, area,
        shader->grid_capture, shader->grid_capture_data);

    Py_DECREF(pobj);
    MATRIX_destroy(l2s);
    SHADER_destroy(shader);

    return status;
}

static int
shader_geometry_default_bbox(SHADER *shader, VECTOR *min, VECTOR *max)
{
    VECTOR_set(min, -3, -3, -3);
    VECTOR_set(max, 3, 3, 3);
    return 0;
}
