/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/shader/shader_glue.c,v 1.10 2002/01/03 05:39:46 wex Exp $
 *
 * $Log: shader_glue.c,v $
 * Revision 1.10  2002/01/03 05:39:46  wex
 * Changed to shadercmodule for package format
 *
 * Revision 1.9  2001/12/20 21:29:08  wex
 * Support for Win32.  Removed a DECREF and INCREF in set_pobj
 *
 * Revision 1.8  2001/12/05 10:26:42  wex
 * Error and type checking
 *
 * Revision 1.7  2001/11/16 09:24:45  wex
 * Removed an INCREF that was leaking vectors
 *
 * Revision 1.6  2001/11/13 07:49:12  wex
 * Python referencing cleanup, fixed leaks and bugs.
 *
 * Revision 1.5  2001/11/12 18:52:07  wex
 * Cleanup for Python1.5
 *
 * Revision 1.4  2001/11/04 08:15:16  wex
 * Changes to support material bindings.
 *
 * Revision 1.3  2001/11/03 07:42:51  wex
 * Updated setattr, error handling
 *
 * Revision 1.2  2001/11/02 19:57:13  wex
 * Check python object for shader data, handle tuples and vectors.
 *
 * Revision 1.1  2001/11/02 03:17:15  wex
 * Support for Python material and light shaders
 *
 *
 */

#include <assert.h>

#include "shader_internal.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isShader(m) ((m)->ob_type == &Shader)
#define isMaterial(m) ((m)->ob_type == &Material)


static PyObject *shader_new(PyObject *self, PyObject *args);
static void shader_dealloc(SHADER *shader);
static int shader_print(SHADER *shader, FILE *fp, int flags);
static PyObject *shader_getattr(SHADER *shader, char *name);
static int shader_setattr(SHADER *shader, char *name, PyObject *val);

static PyObject *material_new(PyObject *self, PyObject *args);
static SHADER_MATERIAL *material_create(void);
static void material_dealloc(SHADER_MATERIAL *material);
static int material_print(SHADER_MATERIAL *material, FILE *fp, int flags);
static int material_setattr(SHADER_MATERIAL *material, char *name, PyObject *val);
static int shader_material_check(PyObject *obj);
static char *shader_type(SHADER_TYPE type);




/** Type Descriptors **/

static PyTypeObject Shader = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else    
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "shader",                                       /* tp_name               */
    sizeof(SHADER),                                 /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)shader_dealloc,                     /* tp_dealloc refcount=0 */
    (printfunc)shader_print,                        /* tp_print   "print x"  */
    (getattrfunc)shader_getattr,                    /* tp_getattr "x.attr"   */
    (setattrfunc)shader_setattr,                    /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static PyTypeObject Material = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else    
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "material",                                     /* tp_name               */
    sizeof(SHADER),                                 /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)material_dealloc,                   /* tp_dealloc refcount=0 */
    (printfunc)material_print,                      /* tp_print   "print x"  */
    (getattrfunc)NULL,                              /* tp_getattr "x.attr"   */
    (setattrfunc)material_setattr,                  /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef shadertype_methods[] = {
    {"Shader", shader_new, 1},
    {"ShaderMaterial", material_new, 1},
    {NULL, NULL}
};


void
initshadercmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32    
    Shader.ob_type = &PyType_Type;
    Material.ob_type = &PyType_Type;
#endif    
    m = Py_InitModule("shadercmodule", shadertype_methods);
    d = PyModule_GetDict(m);                        /* with 'Shader' func */
    ErrorObject = Py_BuildValue("s", "shadercmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize shader C module");
    }
}

static PyObject *
shader_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)shader_create(NULL, SHADER_TYPE_NONE, NULL);
}

SHADER *
shader_create(SHADER *parent, SHADER_TYPE type, void *type_data)
{
    SHADER *shader;
    
    shader = PyObject_NEW(SHADER, &Shader);          /* malloc, init, incref */
    if (shader == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return NULL;
    }

    shader_initialize(shader, type, type_data);
    
    return shader;
}

static void
shader_dealloc(SHADER *shader)
{
    shader_delete(shader);
    PyMem_DEL(shader);
}

static char *
shader_type(SHADER_TYPE type)
{
    switch (type) {
    case SHADER_TYPE_NONE: return "none";
    case SHADER_TYPE_MATERIAL: return "material";
    case SHADER_TYPE_LIGHT: return "light";
    case SHADER_TYPE_GEOMETRY: return "geometry";
    }
    return "";
}
        
static int
shader_print(SHADER *shader, FILE *fp, int flags)
{
    fprintf(fp, "SHADER at %p: type=%s  parent=%p\n", shader,
        shader_type(shader->type), shader->parent);
    return 0;
}

static PyObject *
shader_getattr(SHADER *shader, char *name)
{
    double d;
    VECTOR v, *r;
    PyObject *map;
    SHADER *map_shader;

/*    fprintf(stderr, "Get shader type %s attribute %s\n", shader_type(shader->type), name); */
    
    if (strcmp(name, "light_list") == 0) {
        return SHADER_get_light_list(shader);
    }

    if (shader->type == SHADER_TYPE_MATERIAL) {
        SHADER_MATERIAL *shader_material;
        /* check for binding */
        shader_material = shader->shader.material;
        assert(isMaterial(shader_material));
        if (shader_material != NULL) {
            /* get the _material attribute */
            map = shader_material_get_binding(shader_material, name);
            if (map != NULL) {
                map_shader = shader_material_shade_child(map, shader);
                SHADER_get_vector(map_shader, "C", &v);
                SHADER_get_double(map_shader, "A", &d);
                Py_DECREF(map_shader);
                
/*                fprintf(stderr, "Binding eval (%g, %g, %g, %g)\n",
                  v.v[0], v.v[1], v.v[2], d); */
                return (PyObject*)VECTOR_create(v.v[0], v.v[1], v.v[2]);
            }
        }
    }

    if (SHADER_get_vector(shader, name, &v) == 0) {
        r = VECTOR_create(v.v[0], v.v[1], v.v[2]);
        return (PyObject *)r;
    }
        
    if (SHADER_get_double(shader, name, &d) == 0) {
        return PyFloat_FromDouble(d);
    }

    Py_INCREF(Py_None);
    return Py_None;
}

static int
shader_setattr(SHADER *shader, char *name, PyObject *val)
{
    double x, y, z;
    VECTOR v;
    PyObject *pf;
    char buf[1024];
    
    if (PyFloat_Check(val)) {
        SHADER_set_double(shader, name, PyFloat_AsDouble(val));
        return 0;
    }
    
    if (PyTuple_Check(val) && PyTuple_Size(val) == 3) {
        pf = PyTuple_GET_ITEM(val, 0);
        if (!PyFloat_Check(pf) && !PyInt_Check(pf)) return 0;
        x = PyFloat_AsDouble(pf);
        pf = PyTuple_GET_ITEM(val, 1);
        if (!PyFloat_Check(pf) && !PyInt_Check(pf)) return 0;
        y = PyFloat_AsDouble(pf);
        pf = PyTuple_GET_ITEM(val, 2);
        if (!PyFloat_Check(pf) && !PyInt_Check(pf)) return 0;
        z = PyFloat_AsDouble(pf);
        V3_SET(v.v, x, y, z);
        SHADER_set_vector(shader, name, &v);
        return 0;
    }

    if (VECTOR_check(val)) {
        SHADER_set_vector(shader, name, (VECTOR *)val);
        return 0;
    }

    sprintf(buf, "Attribute \"%s\" uses an unknown shader data type", name);
    
    PyErr_SetString(ErrorObject, buf);

    return -1;
}

static PyObject *
material_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)material_create();
}

static SHADER_MATERIAL *
material_create(void)
{
    SHADER_MATERIAL *material;

    material = (SHADER_MATERIAL *)PyObject_NEW(SHADER_MATERIAL, &Material);
    if (material == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return NULL;
    }

    shader_material_initialize(material);

    return material;
}

static void
material_dealloc(SHADER_MATERIAL *material)
{
    shader_material_delete(material);
    PyMem_DEL(material);
}

static int
material_print(SHADER_MATERIAL *material, FILE *fp, int flags)
{
    fprintf(fp, "MATERIAL at %p\n", material);
    return 0;
}

static int
material_setattr(SHADER_MATERIAL *material, char *name, PyObject *val)
{
    char buf[1024]; 

    if (shader_material_check(val)) {
        SHADER_MATERIAL *shader_material;
        /* get the material */
        shader_material = (SHADER_MATERIAL *)PyObject_GetAttrString(val,
            "_material");
        assert(isMaterial(shader_material));
        shader_material_binding_create(material, name, val);
        Py_DECREF(shader_material);

        return 0;
    }
    
    sprintf(buf, "Attribute \"%s\" uses an unknown shader data type", name);
    PyErr_SetString(ErrorObject, buf);

    return 1;
/*    return 0; */
}

int
SHADER_MATERIAL_check(PyObject *pobj)
{
    return isMaterial(pobj);
}

static int
shader_material_check(PyObject *obj)
{
    if (!PyInstance_Check(obj)) {
        return 0;
    }
    
/*    fprintf(stderr, "Check material class \"%s\"\n", PyString_AsString(((PyInstanceObject *)obj)->in_class->cl_name)); */

    return 1;
}
