/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/shader/shader_material.c,v 1.13 2001/12/20 21:29:08 wex Exp $
 *
 * $Log: shader_material.c,v $
 * Revision 1.13  2001/12/20 21:29:08  wex
 * Support for Win32.  Removed a DECREF and INCREF in set_pobj
 *
 * Revision 1.12  2001/12/05 10:26:42  wex
 * Error and type checking
 *
 * Revision 1.11  2001/11/16 00:52:27  wex
 * A few more memory leaks
 *
 * Revision 1.10  2001/11/13 18:30:30  wex
 * Added an assert to make sure we don't reset the function pointers
 *
 * Revision 1.9  2001/11/13 07:49:12  wex
 * Python referencing cleanup, fixed leaks and bugs.
 *
 * Revision 1.8  2001/11/12 18:52:07  wex
 * Cleanup for Python1.5
 *
 * Revision 1.7  2001/11/04 08:15:16  wex
 * Changes to support material bindings.
 *
 * Revision 1.6  2001/11/03 07:42:51  wex
 * Updated setattr, error handling
 *
 * Revision 1.5  2001/11/02 19:57:13  wex
 * Check python object for shader data, handle tuples and vectors.
 *
 * Revision 1.4  2001/11/02 03:17:15  wex
 * Support for Python material and light shaders
 *
 * Revision 1.3  2001/11/01 04:26:57  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:28  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:41  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <assert.h>

#include "shader_internal.h"

static int shader_material_shade(PyObject *material,
    SHADER_MATERIAL *shader_material, SHADER *shader);


#if 0
SHADER_MATERIAL *
SHADER_MATERIAL_create(PyObject *pobj)
{
    SHADER_MATERIAL *material;

    material = (SHADER_MATERIAL *)malloc(sizeof(SHADER_MATERIAL));
    if (material == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return NULL;
    }

    shader_material_initialize(pobj);
    
    return material;
}
#endif

void
shader_material_initialize(SHADER_MATERIAL *material)
{
    material->shade_pyobj_func = NULL;
    material->shade_initialized = 0;
    material->binding_list = NULL;
    material->pobj = NULL;
    material->error = 0;
}

void
shader_material_delete(SHADER_MATERIAL *material)
{
    SHADER_BINDING *binding;
    
    /* 911: Doing the DECREF causes a crash 
    Py_DECREF(material->shade_pyobj_func);  */

    if (material->pobj != NULL) {
        Py_DECREF(material->pobj);
        material->pobj = NULL;
    }
    
    material->shade_pyobj_func = NULL;
    if (material->binding_list != NULL) {
        while ((binding = LIST_pop(material->binding_list)) != NULL) {
            Py_DECREF(binding->map_class);
            free(binding);
        }
        LIST_destroy(material->binding_list);
        material->binding_list = NULL;   /* paranoid */
    }
}

SHADER *
SHADER_MATERIAL_shade(PyObject *material,
    SHADER_DATA_CALLBACK get_data, SHADER_DATA_CALLBACK set_data,
    void *user_data)
{
    SHADER *shader;
    SHADER_MATERIAL *shader_material;

    shader_material = SHADER_MATERIAL_get_from_pobj(material);
    assert(shader_material != NULL);

    if (shader_material->error) {
        return NULL;
    }
    
    shader = shader_create(NULL, SHADER_TYPE_MATERIAL, shader_material);
    if (shader == NULL) {
        return NULL;
    }

    shader->get_data_callback = get_data;
    shader->set_data_callback = set_data;
    shader->user_data = user_data;

    if (shader_material_shade(material, shader_material, shader) < 0) {
        SHADER_destroy(shader);
        return NULL;
    }

    return shader;
}

SHADER *
shader_material_shade_child(PyObject *material, SHADER *parent)
{
    SHADER *shader;
    SHADER_MATERIAL *shader_material;

    shader_material = SHADER_MATERIAL_get_from_pobj(material);
    assert(shader_material != NULL);

    SHADER_MATERIAL_set_pobj(shader_material, material);
    
    shader = shader_create(NULL, SHADER_TYPE_MATERIAL, shader_material);
    if (shader == NULL) {
        return NULL;
    }

    Py_INCREF(parent);
    shader->parent = parent;

    if (shader_material_shade(material, shader_material, shader) < 0) {
        SHADER_destroy(shader);
        return NULL;
    }
    
    return shader;
}
    
static int
shader_material_shade(PyObject *material, SHADER_MATERIAL *shader_material,
    SHADER *shader)
{
    PyObject *pargs, *pstatus;
    double a;
    VECTOR rgb;

    assert(shader_material->shade_pyobj_func != NULL);
    
    pargs = Py_BuildValue("(O)", shader);
    pstatus = PyEval_CallObject(shader_material->shade_pyobj_func, pargs);
    
    Py_DECREF(pargs);
    
    if (PyErr_Occurred()) PyErr_Print();
    if (pstatus != NULL) {
        PyArg_Parse(pstatus, "(dddd)", &rgb.v[0], &rgb.v[1], &rgb.v[2], &a);
        Py_DECREF(pstatus);
    } else {
        shader_material->error = 1;
        fprintf(stderr, "NULL material shade status\n");
        rgb.v[0] = rgb.v[1] = rgb.v[2] = 0;
        a = 1;
    }

/*    fprintf(stderr, "Material shader returned color (%g, %g, %g, %g)\n",
      rgb.v[0], rgb.v[1], rgb.v[2], a);   */

    SHADER_set_vector(shader, "C", &rgb);
    SHADER_set_double(shader, "A", a);

    return 0;
}

int
shader_material_binding_create(SHADER_MATERIAL *material, char *name,
    PyObject *map)
{
    SHADER_BINDING *binding;

    binding = (SHADER_BINDING *)malloc(sizeof(SHADER_BINDING));
    if (binding == NULL) {
        SHADER_errno = SHADER_ERR_MEMORY;
        return -1;
    }

    binding->attribute_name = strdup(name);
    binding->map_class = map;
    Py_INCREF(map); 

    if (material->binding_list == NULL) {
        material->binding_list = LIST_create();
    }
    
    LIST_append(material->binding_list, binding);

    return 0;
}

PyObject *
shader_material_get_binding(SHADER_MATERIAL *material, char *name)
{
    SHADER_BINDING *binding;

    if (material->binding_list == NULL) {
        return NULL;
    }

    for (binding = LIST_head(material->binding_list); binding != NULL;
         binding = LIST_next(binding)) {

        if (strcmp(binding->attribute_name, name) == 0) {
            return binding->map_class;
        }
    }

    return NULL;
}

SHADER_MATERIAL *
SHADER_MATERIAL_get_from_pobj(PyObject *pobj)
{
    SHADER_MATERIAL *shader_material;

    /* extract the SHADER_MATERIAL from the Material class */
    shader_material = (SHADER_MATERIAL *)PyObject_GetAttrString(pobj,
        "_material");

    /* decrement reference count now, since we only want this func to borrow */
    if (shader_material != NULL) {
        Py_DECREF(shader_material);
    } 

    return shader_material;
}

void
SHADER_MATERIAL_set_pobj(SHADER_MATERIAL *material, PyObject *pobj)
{
    if (material->shade_initialized) return;

#if 0
    /* don't reference count here? */
    if (material->pobj != NULL) {
        assert(material->pobj == NULL);
        Py_DECREF(material->pobj);
    }
    Py_INCREF(pobj);
#endif
    
    material->pobj = pobj;
        
    /* find the shade function */
    material->shade_pyobj_func = PyObject_GetAttrString(pobj, "shade");
    if (material->shade_pyobj_func == NULL ||
        !PyCallable_Check(material->shade_pyobj_func)) {
        fprintf(stderr, "Cannot find material shade function\n");
        material->shade_pyobj_func = NULL;
        return;
    }

    material->shade_initialized = 1;
}


PyObject *
SHADER_MATERIAL_get_pobj(SHADER_MATERIAL *material)
{
    return material->pobj;
}

