/* Copyright (c) 2000, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui.c,v 1.16 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui.c,v $
 * Revision 1.16  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.15  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.14  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.13  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.12  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.11  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.10  2000/10/05 10:29:41  wex
 * Added labels, titles, and textports.  Fixed parsing bug and radio look.
 *
 * Revision 1.9  2000/10/02 05:56:00  wex
 * Added radio buttons, fixed minor bugs.
 *
 * Revision 1.8  2000/10/01 18:27:37  wex
 * Cleanups for Windows.
 *
 * Revision 1.7  2000/09/30 18:00:24  wex
 * Handle more visual types.  Menubars resize correctly.
 *
 * Revision 1.6  2000/09/30 15:59:14  wex
 * Saving and loading generated header files.
 *
 * Revision 1.5  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 * Revision 1.4  2000/09/26 16:31:38  wex
 * Events working under Windows.
 *
 * Revision 1.3  2000/09/25 02:35:02  wex
 * Creating windows and widgets in win32, events still broken.
 *
 * Revision 1.2  2000/09/24 03:13:08  wex
 * Initial changes to support WIN32
 *
 * Revision 1.1.1.1  2000/09/16 23:38:08  wex
 * Initial revision
 *
 */

#include <stdlib.h>

#include "tui_internal.h"


int tui_major_version = 0;
int tui_minor_version = 1;


static char *HELV_14_NAME =
        "-adobe-helvetica-medium-r-normal--14-140-75-75-p-*-iso8859-1";
static char *HELV_BOLD_14_NAME =
        "-adobe-helvetica-bold-r-normal--14-140-75-75-p-*-iso8859-1";
static char *HELV_BOLD_ITALIC_14_NAME =
        "-adobe-helvetica-bold-o-normal--14-140-75-75-p-*-iso8859-1";


#ifdef TUI_XWIN
static int initialize_display(TUI_APPLICATION *app);
static void install_color_map(TUI_APPLICATION *app);
#endif


TUI_APPLICATION *
TUI_APPLICATION_create(
#ifdef TUI_WIN32
    HINSTANCE hInstance
#endif
    )
{
    TUI_APPLICATION *app;
#ifdef TUI_WIN32
	WNDCLASS wc;
    static int wc_initialized = 0;
#endif

    app = (TUI_APPLICATION *)malloc(sizeof(TUI_APPLICATION));
    assert(app != NULL);

    app->client_list = LIST_create();
    app->break_out_of_mainloop = 0;
    
#ifdef TUI_XWIN
    if (initialize_display(app) < 0) {
        free(app);
        fprintf(stderr, "Cannot open X display\n");
        abort();
        return NULL;
    }
    install_color_map(app);
#endif

#ifdef TUI_WIN32
    app->hInstance = hInstance;

    if (!wc_initialized) {
        wc_initialized = 1;
        wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
        wc.lpfnWndProc = tui_win32_wndproc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = 0;
        wc.hInstance = hInstance;
        wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
        wc.hCursor = LoadCursor(NULL, IDC_ARROW);
        wc.hbrBackground = (HBRUSH)GetStockObject(LTGRAY_BRUSH);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = TUI_WINDOW_CLASS_NAME;
        if (RegisterClass(&wc) == 0) {
            free(app);
            return NULL;
        }
    }
#endif    

	TUI_errno = -1;
    return app;
}


TUI *
TUI_create(TUI_APPLICATION *app)
{
	TUI *tui;

	tui = (TUI *)malloc(sizeof(TUI));
	if (tui == NULL) {
		TUI_errno = TUI_ERR_MEMORY;
		return NULL;
	}

	tui->root = tui_widget_allocate();
    tui->app = app;
	
	return tui;
}


void
TUI_destroy(TUI *tui)
{
    TUI_CLIENT *client;

	assert(tui != NULL);
	assert(tui->root != NULL);
	assert(tui->root->next == NULL);

    for (client = LIST_head(tui->app->client_list); client != NULL;
         client = LIST_next(client)) {
        if (client->callback != NULL) {
            TUI_CLIENT_INFO info;
            info.user_data = client->user_data;
            info.message = TUI_CLIENT_MESSAGE_DESTROY;
            client->callback(&info);
        }
    }
    
	TUI_WIDGET_destroy(tui->root);

#ifdef TUI_XWIN	
	assert(tui->app->display != NULL);
	XCloseDisplay(tui->app->display);
#endif
	
	free(tui);
}


#ifdef TUI_XWIN
Display *
TUI_get_display(TUI *tui)
{
	return tui->app->display;
}


static unsigned long
alloc_color(Display *display, int r, int g, int b)
{
    Colormap cmap;
    XColor xcolor;

    cmap = DefaultColormap(display, DefaultScreen(display));

    xcolor.flags = DoRed | DoGreen | DoBlue;
    xcolor.red = r << 8;
    xcolor.green = g << 8;
    xcolor.blue = b << 8;

    XAllocColor(display, cmap, &xcolor);

    return xcolor.pixel;
}

static int
initialize_display(TUI_APPLICATION *app)
{
    int screen, depth, visual_count;
    XVisualInfo visual_template, *visual_list;
	
	app->display = XOpenDisplay(NULL);
	if (app->display == NULL) {
		TUI_errno = TUI_ERR_DISPLAY;
		return -1;
	}

    screen = DefaultScreen(app->display);
    depth = DefaultDepth(app->display, screen);
    
    if (XMatchVisualInfo(app->display, screen, depth, PseudoColor,
        &app->visual_info)==0) {
        visual_template.visualid = XVisualIDFromVisual(
            DefaultVisual(app->display, screen));
        visual_list = XGetVisualInfo(app->display, VisualIDMask,
            &visual_template, &visual_count);

        if (visual_list == NULL || visual_count == 0) {
            TUI_errno = TUI_ERR_VISUAL;
            return -1;
        }

        app->visual_info = visual_list[0];
        XFree(visual_list);
    }

    return 0;
}

static void
install_color_map(TUI_APPLICATION *app)
{
    XGCValues values;
	Window root_window;

#define ALLOC_COLOR_GC(R, G, B) \
    (values.foreground = alloc_color(app->display, R, G, B), \
    XCreateGC(app->display, root_window, \
        GCForeground | GCGraphicsExposures, &values))

	root_window = DefaultRootWindow(app->display);
    values.graphics_exposures = 0;

    app->gc[TUI_GC_BLACK]           = ALLOC_COLOR_GC(0x00, 0x00, 0x00);
    app->gc[TUI_GC_VERY_DARK_GRAY]  = ALLOC_COLOR_GC(0x22, 0x22, 0x22);
    app->gc[TUI_GC_DARK_GRAY]       = ALLOC_COLOR_GC(0x55, 0x55, 0x55);
    app->gc[TUI_GC_MEDIUM_GRAY]     = ALLOC_COLOR_GC(0x88, 0x88, 0x88);
    app->gc[TUI_GC_LIGHT_GRAY]      = ALLOC_COLOR_GC(0xaa, 0xaa, 0xaa);
    app->gc[TUI_GC_MENU_GRAY]       = ALLOC_COLOR_GC(0xc5, 0xc5, 0xc5);
    app->gc[TUI_GC_VERY_LIGHT_GRAY] = ALLOC_COLOR_GC(0xdd, 0xdd, 0xdd);
    app->gc[TUI_GC_WHITE]           = ALLOC_COLOR_GC(0xff, 0xff, 0xff);

    app->gc[TUI_GC_LIGHT_BLUE]      = ALLOC_COLOR_GC(0x77, 0x77, 0xcc);
    app->gc[TUI_GC_BLUE]            = ALLOC_COLOR_GC(0x00, 0x00, 0xff);
    
    app->gc[TUI_GC_TERRACOTTA]      = ALLOC_COLOR_GC(0xbb, 0x88, 0x88);
    
    app->gc[TUI_GC_RED]             = ALLOC_COLOR_GC(0xff, 0x00, 0x00);
    app->gc[TUI_GC_LIGHT_RED]       = ALLOC_COLOR_GC(0xcc, 0x77, 0x77);
    app->gc[TUI_GC_YELLOW]          = ALLOC_COLOR_GC(0xff, 0xff, 0x00);

    app->gc[TUI_GC_CHECKMARK]       = ALLOC_COLOR_GC(0xdc, 0x20, 0x20);
    app->gc[TUI_GC_SHADOW]          = ALLOC_COLOR_GC(0x40, 0x40, 0x40);

    values.foreground = alloc_color(app->display, 0x00, 0x00, 0x00);
    app->text_font = XLoadQueryFont(app->display, HELV_14_NAME);
    values.font = app->text_font->fid;
    app->gc[TUI_GC_TEXT] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    values.foreground = alloc_color(app->display, 0x00, 0x00, 0x00);
    app->title_font = XLoadQueryFont(app->display, HELV_BOLD_14_NAME);
    values.font = app->title_font->fid;
    app->gc[TUI_GC_TEXT_BOLD] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    values.foreground = alloc_color(app->display, 0x55, 0x55, 0x55);
    values.font = app->text_font->fid;
    app->gc[TUI_GC_TEXT_GRAY] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    values.foreground = alloc_color(app->display, 0x00, 0x00, 0x00);
    app->menu_font = XLoadQueryFont(app->display,
        HELV_BOLD_ITALIC_14_NAME);
    values.font = app->menu_font->fid;
    app->gc[TUI_GC_TEXT_MENUBAR] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    values.foreground = alloc_color(app->display, 0x55, 0x55, 0x55);
    values.font = app->menu_font->fid;
    app->gc[TUI_GC_TEXT_MENUBAR_GRAY] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    values.foreground = alloc_color(app->display, 0x88, 0x88, 0x88);
    values.font = app->menu_font->fid;
    app->gc[TUI_GC_TEXT_MENUITEM_GRAY] = XCreateGC(app->display, root_window,
        GCForeground | GCGraphicsExposures | GCFont, &values);

    app->color[TUI_COLOR_LIGHT_GRAY] = alloc_color(app->display,
		0xdd, 0xdd, 0xdd);
    app->color[TUI_COLOR_MENU_GRAY] = alloc_color(app->display,
		0xd4, 0xd4, 0xd4);
}
#endif

void
TUI_add_client(TUI_APPLICATION *app, void (*callback)(TUI_CLIENT_INFO *info),
    void *user_data)
{
    TUI_CLIENT *client;

    assert(callback != NULL);
    
    client = (TUI_CLIENT *)malloc(sizeof(TUI_CLIENT));
    assert(client != NULL);

    client->callback = callback;
    client->user_data = user_data;

    LIST_append(app->client_list, client);
}

void
TUI_remove_client(TUI_APPLICATION *app,
    void (*callback)(TUI_CLIENT_INFO *info), void *user_data)
{
    TUI_CLIENT *client;

    for (client = LIST_head(app->client_list); client != NULL;
         client = LIST_next(client)) {

        if (client->callback == callback && client->user_data == user_data) {
            LIST_delete(app->client_list, client);
            return;
        }
    }
}

void
TUI_break(TUI *tui)
{
    tui->app->break_out_of_mainloop = 1;
}
