/* Copyright (c) 2000, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_button.c,v 1.14 2002/01/03 06:50:13 wex Exp $
 *
 * $Log: tui_button.c,v $
 * Revision 1.14  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.13  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.12  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.11  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.10  2000/10/06 03:48:46  wex
 * Slightly better events, widget windows capture other events.
 *
 * Revision 1.9  2000/10/05 10:29:41  wex
 * Added labels, titles, and textports.  Fixed parsing bug and radio look.
 *
 * Revision 1.8  2000/10/02 09:28:34  wex
 * Minor bug fixes.
 *
 * Revision 1.7  2000/10/02 05:56:00  wex
 * Added radio buttons, fixed minor bugs.
 *
 * Revision 1.6  2000/10/01 18:27:37  wex
 * Cleanups for Windows.
 *
 * Revision 1.5  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 * Revision 1.4  2000/09/26 16:31:38  wex
 * Events working under Windows.
 *
 * Revision 1.2  2000/09/24 03:13:08  wex
 * Initial changes to support WIN32
 *
 * Revision 1.1.1.1  2000/09/16 23:38:08  wex
 * Initial revision
 *
 */


#include "tui_internal.h"

#include <string.h>


#define TUI_BUTTON_HEIGHT			28
#define TUI_BUTTON_MIN_WIDTH        75

#ifdef TUI_XWIN

#define TUI_BUTTON_MARGIN			10
#define TUI_BUTTON_BOTTOM_MARGIN	9

#define TUI_RADIO_SIZE              24
#define TUI_RADIO_SPACING           6

static void tui_button_draw(TUI_WIDGET *widget);
static void tui_radio_button_draw(TUI_WIDGET *widget);
static void tui_button_expose(TUI_WIDGET *widget, XExposeEvent *event);
static void tui_button_cross(TUI_WIDGET *widget, XCrossingEvent *event);
static void tui_button_button(TUI_WIDGET *widget, XButtonEvent *event);

#endif /* TUI_XWIN */

static TUI_WIDGET *
create_button(TUI_WIDGET *window, char *text, int x, int y,
    TUI_WIDGET_TYPE type)
{
	TUI_WIDGET *widget;
    int width;

#ifdef TUI_XWIN	
	Window button_window;
	
    width = XTextWidth(window->tui->app->text_font, text, strlen(text)) +
        2 * TUI_BUTTON_MARGIN;

    if (width < TUI_BUTTON_MIN_WIDTH) {
        width = TUI_BUTTON_MIN_WIDTH;
    }
    
    if (type == TUI_RADIO_BUTTON) {
        width += TUI_RADIO_SPACING + TUI_RADIO_SIZE;
    }
    
	button_window = tui_create_xwin_window(window->tui, window->window, x, y,
		width, TUI_BUTTON_HEIGHT);
	if (button_window == 0) {
		return NULL;
	}
	
    XSelectInput(window->tui->app->display, button_window,
		EnterWindowMask | LeaveWindowMask | ExposureMask | PointerMotionMask |
		ButtonPressMask | ButtonReleaseMask | StructureNotifyMask);
#endif

#ifdef TUI_WIN32
	HWND hwnd;
	unsigned long win32_type;
	static int cur_id = 101;

	assert(window->hwnd != NULL);

    width = 100;
    if (type == TUI_BUTTON) {
		win32_type = WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON;
	} else if (type == TUI_RADIO_BUTTON) {
		win32_type = WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_RADIOBUTTON;
	}

	hwnd = CreateWindow("button", text,
		win32_type, x, y, width, TUI_BUTTON_HEIGHT,
		window->hwnd, (HMENU)cur_id, window->tui->app->hInstance, NULL);
	if (hwnd == NULL) {
		return NULL;
	}
#endif

#ifdef TUI_OSX
	ControlRef control;
    Rect rect;
    unsigned char buf[255];
    
    rect.top = y;
    rect.left = x;
    rect.right = x + 100;
    rect.bottom = y + TUI_BUTTON_HEIGHT;

    sprintf(buf, " %s", text);
    buf[0] = strlen(text);
    
    control = NewControl(window->window, &rect, buf, 1, 0, 0, 1,
        kControlPushButtonProc, 101);
    if (control == NULL) {
        fprintf(stderr, "Cannot create button\n");
        return NULL;
    }
#endif
    
	assert(window->type == TUI_WINDOW);
	assert(text != NULL);

	widget = tui_widget_allocate();
	if (widget == NULL) {
		return NULL;
	}

	widget->tui = window->tui;
	widget->type = type;

	widget->parent = window;
    tui_widget_append_to_list(&window->child, widget);

#ifdef TUI_XWIN
	widget->window = button_window;
    if (type == TUI_RADIO_BUTTON) {
        widget->draw_fp = tui_radio_button_draw;
    } else {
        widget->draw_fp = tui_button_draw;
    }
	widget->expose_fp = tui_button_expose;
	widget->cross_fp = tui_button_cross;
	widget->button_fp = tui_button_button;
#endif

#ifdef TUI_WIN32
	widget->hwnd = hwnd;
	widget->control_id = cur_id;
	cur_id++;
#endif

#ifdef TUI_OSX
    widget->control = control;
#endif    
	widget->x = x;
	widget->y = y;
	widget->w = width;
	widget->h = TUI_BUTTON_HEIGHT;
	widget->text = strdup(text);

	TUI_WIDGET_set_enabled(widget, 1);
	TUI_WIDGET_set_visible(widget, 1);
	
	return widget;
}


TUI_WIDGET *
TUI_window_create_button(TUI_WIDGET *window, char *text, int x, int y)
{
    return create_button(window, text, x, y, TUI_BUTTON);
}

    
TUI_WIDGET *
TUI_window_create_radio_button(TUI_WIDGET *window, char *text, int x, int y)
{
    return create_button(window, text, x, y, TUI_RADIO_BUTTON);
}

#ifdef TUI_XWIN

static void
draw_diagonal_frame(TUI *tui, Window window, int x, int y, int w, int h,
    TUI_GC *gc)
{
    int i, n;
    XPoint pt[8];

    n = 0;
    while (gc[n] != TUI_GC_NULL) {
        n++;
    }

    /* draw the inner region first with padding, since it doesn't align well */
    if (n % 2 == 1) {
        pt[0].x = x;
        pt[0].y = y + h/2 - 2;
        pt[1].x = x + w/2 - 2;
        pt[1].y = y;
        pt[2].x = x + w/2 + 1;
        pt[2].y = y;
        pt[3].x = x + w - 1;
        pt[3].y = y + h/2 - 2;
        pt[4].x = x + w - 1;
        pt[4].y = y + h/2 + 1;
        pt[5].x = x + w/2 + 1;
        pt[5].y = y + h - 1;
        pt[6].x = x + w/2 - 2;
        pt[6].y = y + h - 1;
        pt[7].x = x;
        pt[7].y = y + h/2 + 1;

        XFillPolygon(tui->app->display, window, tui->app->gc[gc[n/2]], pt, 8,
            Convex, CoordModeOrigin);
    }

    for (i = 0; i < n/2; i++) {
        XDrawLine(tui->app->display, window, tui->app->gc[gc[i]],
            x + i, y + h/2 - 2, x + w/2 - 2, y + i);
        XDrawLine(tui->app->display, window, tui->app->gc[gc[i]],
            x + w/2 + 1, y + i, x + w - 1 - i, y + h/2 - 2);

        XDrawLine(tui->app->display, window, tui->app->gc[gc[n - 1 - i]],
            x + i, y + h/2 + 1, x + w/2 - 2, y + h - 1 - i);
        XDrawLine(tui->app->display, window, tui->app->gc[gc[n - 1 - i]],
            x + w/2 + 1, y + h - 1 - i, x + w - 1 - i, y + h/2 + 1);

        if (i == 1) {
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[1]], 1, h/2);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[1]],w-2,h/2);
        } else if (i > 1) {
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[i]],
                x + w/2 - 1, y+i-1);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[i]],
                x + w/2, y+i-1);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[i]],
                x + i - 1, y + h/2 - 1);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[i]],
                x + w - i, y + h/2 - 1);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1-i]],
                x + w/2 - 1, y + h - i);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1-i]],
                x + w/2, y + h - i);
            
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1-i]],
                x + i - 1, y + h/2);
            XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1-i]],
                x + w - i, y + h/2);
        }
    }
    
    XDrawPoint(tui->app->display, window, tui->app->gc[gc[0]],
        x, y + h/2 - 1);
    XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1]],
        x, y + h/2);
    XDrawPoint(tui->app->display, window, tui->app->gc[gc[0]],
        x + w - 1, y + h/2 - 1);
    XDrawPoint(tui->app->display, window, tui->app->gc[gc[n-1]],
        x + w - 1, y + h/2);
    
    XDrawLine(tui->app->display, window, tui->app->gc[gc[0]],
        x + w/2 - 1, y, x + w/2, y);
    XDrawLine(tui->app->display, window, tui->app->gc[gc[n - 1]],
        x + w/2 - 1, y + h - 1, x + w/2, y + h - 1);
}

static void
tui_radio_button_draw(TUI_WIDGET *widget)
{
    TUI_GC gc_neutral[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_WHITE,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_located[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_loc_sel[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_BLACK,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_NULL
    };
    TUI_GC gc_selected[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_WHITE,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_disabled[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_tri_normal[] = {
        TUI_GC_LIGHT_BLUE,
        TUI_GC_BLUE,
        TUI_GC_BLACK,
        TUI_GC_DARK_GRAY
    };
    TUI_GC gc_tri_disable[] = {
        TUI_GC_LIGHT_BLUE,
        TUI_GC_LIGHT_BLUE,
        TUI_GC_LIGHT_BLUE,
        TUI_GC_BLUE
    };
    TUI_APPLICATION *app;
    TUI_GC *gc, gc_text, *gc_tri;
    int x, y, i;

    if (!widget->enabled) {
        gc = gc_disabled;
        gc_text = TUI_GC_TEXT_GRAY;
        gc_tri = gc_tri_disable;
    } else if (widget->located && widget->selected) {
        gc = gc_loc_sel;
        gc_text = TUI_GC_TEXT;
        gc_tri = gc_tri_normal;
    } else if (widget->located) {
        gc = gc_located;
        gc_text = TUI_GC_TEXT;
        gc_tri = gc_tri_normal;
    } else if (widget->selected) {
        gc = gc_selected;
        gc_text = TUI_GC_TEXT;
        gc_tri = gc_tri_normal;
    } else {
        gc = gc_neutral;
        gc_text = TUI_GC_TEXT;
        gc_tri = gc_tri_normal;
    }
    
    draw_diagonal_frame(widget->tui, widget->window, 0, 0,
        TUI_RADIO_SIZE, TUI_RADIO_SIZE, gc);

    /* 911: Won't work correctly if widget is resized! */
    if (widget->value) {
        app = widget->tui->app;
        y = TUI_RADIO_SIZE/2 - 7;
        x = TUI_RADIO_SIZE/2 + 5;
        for (i = 0; i < 3; i++, x--, y++) {
            XDrawLine(app->display, widget->window, app->gc[gc_tri[0]],
                TUI_RADIO_SIZE/2, y, x, TUI_RADIO_SIZE/2 - 2);
        }
        XDrawLine(app->display, widget->window, app->gc[gc_tri[1]],
            TUI_RADIO_SIZE/2, y, x, TUI_RADIO_SIZE/2 - 2);
        XDrawLine(app->display, widget->window, app->gc[gc_tri[1]],
            x, TUI_RADIO_SIZE/2 - 1, TUI_RADIO_SIZE/2, TUI_RADIO_SIZE/2 + 1);
        XDrawLine(app->display, widget->window, app->gc[gc_tri[1]],
            TUI_RADIO_SIZE/2, TUI_RADIO_SIZE/2, TUI_RADIO_SIZE/2, y);
        XDrawLine(app->display, widget->window, app->gc[gc_tri[1]],
            TUI_RADIO_SIZE/2 + 1, y + 2, TUI_RADIO_SIZE/2 + 1, y + 3);

        x += 1;
        y = TUI_RADIO_SIZE/2 + 2;
        XDrawLine(app->display, widget->window, app->gc[gc_tri[2]],
            x++, TUI_RADIO_SIZE/2 - 1, TUI_RADIO_SIZE/2, y++);
        XDrawLine(app->display, widget->window, app->gc[gc_tri[2]],
            x++, TUI_RADIO_SIZE/2 - 1, TUI_RADIO_SIZE/2, y);
        XDrawLine(app->display, widget->window, app->gc[gc_tri[2]],
            x++, TUI_RADIO_SIZE/2 - 1, TUI_RADIO_SIZE/2 + 1, y++);

        XDrawLine(app->display, widget->window, app->gc[gc_tri[3]],
            x, TUI_RADIO_SIZE/2 - 1, TUI_RADIO_SIZE/2 + 1, y);
    }

    x = TUI_RADIO_SIZE + TUI_RADIO_SPACING;
    y = widget->h - TUI_BUTTON_BOTTOM_MARGIN;
    
    XDrawString(widget->tui->app->display, widget->window,
        widget->tui->app->gc[gc_text], x,y, widget->text,strlen(widget->text));
}

static void
tui_button_draw(TUI_WIDGET *widget)
{
    TUI_GC gc_neutral[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_located[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_loc_sel[] = {
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_BLACK,
        TUI_GC_WHITE,
        TUI_GC_WHITE,
        TUI_GC_NULL
    };
    TUI_GC gc_selected[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC gc_disabled[] = {
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_DARK_GRAY,
        TUI_GC_NULL
    };
    TUI_GC *gc, gc_text;
    int x, y;

    if (!widget->enabled) {
        gc = gc_disabled;
        gc_text = TUI_GC_TEXT_GRAY;
    } else if (widget->located && widget->selected) {
        gc = gc_loc_sel;
        gc_text = TUI_GC_TEXT;
    } else if (widget->located) {
        gc = gc_located;
        gc_text = TUI_GC_TEXT;
    } else if (widget->selected) {
        gc = gc_selected;
        gc_text = TUI_GC_TEXT;
    } else {
        gc = gc_neutral;
        gc_text = TUI_GC_TEXT;
    }

    tui_draw_frame(widget->tui,widget->window, 0,0,widget->w,widget->h,gc);
    x = widget->w/2 - XTextWidth(widget->tui->app->text_font, widget->text,
        strlen(widget->text))/2;
    y = widget->h - TUI_BUTTON_BOTTOM_MARGIN;
    
    XDrawString(widget->tui->app->display, widget->window,
        widget->tui->app->gc[gc_text], x,y, widget->text,strlen(widget->text));
}

static void
tui_button_expose(TUI_WIDGET *widget, XExposeEvent *event)
{
    widget->draw_fp(widget);
}

static void
tui_button_cross(TUI_WIDGET *widget, XCrossingEvent *event)
{
    widget->located = event->type == EnterNotify;
    widget->draw_fp(widget);
}


static void
tui_button_button(TUI_WIDGET *widget, XButtonEvent *event)
{
	TUI_WIDGET_INFO info;
    
    widget->selected = event->type == ButtonPress;
    widget->draw_fp(widget);

    if (event->type == ButtonRelease && widget->located &&
        widget->callback != NULL && widget->enabled && widget->visible) {

        if (widget->type == TUI_RADIO_BUTTON) {
            tui_window_disable_matching_radios(widget->parent, widget);
            widget->value = widget->key;
            widget->draw_fp(widget);
        }

		info.widget = widget;
		info.user_data = widget->user_data;
		
        widget->callback(&info);
    }
}

#endif /* TUI_XWIN */


void
TUI_radio_button_set_selected_value(TUI_WIDGET *radio, char value)
{
    assert(radio->type == TUI_RADIO_BUTTON);
    assert(value != 0);
    radio->key = value;
}

