/* Copyright (c) 2000, Flarg Inc, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_construct.c,v 1.11 2002/01/03 06:50:13 wex Exp $
 *
 * $Log: tui_construct.c,v $
 * Revision 1.11  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.10  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.9  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.8  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.7  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.6  2000/10/06 03:48:46  wex
 * Slightly better events, widget windows capture other events.
 *
 * Revision 1.5  2000/10/05 10:29:41  wex
 * Added labels, titles, and textports.  Fixed parsing bug and radio look.
 *
 * Revision 1.4  2000/10/02 09:28:34  wex
 * Minor bug fixes.
 *
 * Revision 1.3  2000/10/02 07:16:48  wex
 * Keyboard equivalents.
 *
 * Revision 1.2  2000/10/02 06:04:08  wex
 * Fixed parsing of invisible and disabled.
 *
 * Revision 1.1  2000/10/02 05:56:00  wex
 * Added radio buttons, fixed minor bugs.
 *
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#if defined(TUI_XWIN)
#include <dlfcn.h>
#endif

#ifdef TUI_OSX
#include <stdio.h>
#include <stdlib.h>
#include <mach-o/dyld.h>
#endif

#include "tui_internal.h"


static const char *
next_child_block(const char *cur, char *value)
{
    int i, end_token, new_block, brace_count;

    if (cur == NULL) {
        return NULL;
    }
    
    brace_count = 0;

    /* search forward to find the next open brace at this level */
    for (i = 0; cur[i] != '\0'; i++) {
        if (cur[i] == '}') return NULL;
        if (cur[i] == '{') break;
    }

    new_block = i + 1;

    /* now back up to get the token for this block */
    /*   first get rid of any whitespace */
    for (--i; isspace(cur[i]) || cur[i] == '\n'; i--);
    end_token = i;
    /*   next find the beginning of the token */
    for (--i; isalpha(cur[i]) || cur[i] == '_'; i--);
    i++;

    assert(end_token-i > 0 && end_token-i < 50);

    strncpy(value, &cur[i], end_token-i+1);
    value[end_token-i+1] = '\0';

    return &cur[new_block];
}


static const char *
skip_block(const char *cur)
{
    int i, brace_count;

    brace_count = 0;
    for (i = 0; cur[i] != '\0'; i++) {
        if (cur[i] == '{') {
            brace_count++;
            continue;
        }
        if (cur[i] == '}') {
            if (--brace_count < 0) {
                return &cur[i+1];
            }
        }
    }

    return NULL;
}


/* search ahead and return first matching token at this block level */
static char *
get_attribute_value(const char *cur, char *token)
{
    static char buffer[512];
    int i, token_start, token_end, len, brace_count, started_string;

    len = strlen(token);
    
    i = 0;
    brace_count = 0;
    token_start = i;
    token_end = i;
    
    for (i = 0; cur[i] != '\0'; i++) {
        if (cur[i] == '{') {                            /* begin block */
            brace_count++;
            continue;
        }

        if (cur[i] == '}') {                            /* end block */
            if (--brace_count < 0) {
                return NULL;
            }
        }

        if (cur[i] == '\'') {                           /* string */
            while (cur[++i] != '\'');
        }

        if (cur[i] == '=' ||
            isspace(cur[i])) {                          /* new token */

            if (brace_count == 0 && token_end-token_start == len) {
                if (strncmp(token,&cur[token_start],token_end-token_start)==0){
                    /* found it */
                    break;
                }
            }
            
            token_end = i+1;
            token_start = i+1;
        }

        if (isalpha(cur[i]) || cur[i] == '_') {
            token_end++;
        }
    }

    assert(cur[i] != '\0');

    /* find the value of this token */
    token_start = token_end+1;
    started_string = 0;
    for (i = token_end+1; cur[i] != ';'; i++) {
        if (cur[i] == '\'') {
            if (started_string) {
                break;
            }
            started_string = 1;
            token_start = i+1;
            token_end = token_start;
        } else if ((!started_string && isspace(cur[i])) || cur[i] == '=') {
            if (started_string && cur[i] == '=') {
                fprintf(stderr, "Found equals inside of string\n");
            }
            token_start = i+1;
            token_end = token_start;
        } else {
            token_end++;
        }
    }

    assert(token_end - token_start > 0);

    strncpy(buffer, &cur[token_start], token_end-token_start);
    buffer[token_end-token_start] = '\0';

    return buffer;
}


static int
tui_find_widget_name_structure_index(TUI_WIDGET *widget, char *name,int *index)
{
    if (widget->name != NULL) {
        if (strcmp(name, widget->name) == 0) {
            return 0;
        }

        (*index)++;
    }

    if (widget->child != NULL) {
        if (tui_find_widget_name_structure_index(widget->child,name,index)==0){
            return 0;
        }
    }

    if (widget->next != NULL) {
        if (tui_find_widget_name_structure_index(widget->next,name,index)==0){
            return 0;
        }
    }

    return -1;
}


static void
set_callback(TUI_WIDGET *widget, char *text)
{
	void *fp = NULL;

#ifdef TUI_XWIN
    void *dl_handle;
    char *python_path;
    char *directory;
    char buffer[4096];

    /* check for module DSO in each PYTHONPATH directory */
    python_path = getenv("PYTHONPATH");
    assert(python_path != NULL);
    python_path = strdup(python_path);
    dl_handle = NULL;
    for (directory = strtok(python_path, ":");
         directory != NULL;
         directory = strtok(NULL, ":")) {
        
        sprintf(buffer, "%s/parc/tuicmodule.so", directory);
        dl_handle = dlopen(buffer, RTLD_NOW);
        if (dl_handle != NULL) break;
    }

    free(python_path);
    
    if (dl_handle != NULL) {
        fp = dlsym(dl_handle, text);
        dlclose(dl_handle);
    }

    if (fp == NULL) {
        dl_handle = dlopen(NULL, RTLD_NOW);
        assert(dl_handle != NULL);
        fp = dlsym(dl_handle, text);
        dlclose(dl_handle);
    }
#endif

#ifdef TUI_WIN32
	HMODULE hModule;

	hModule = GetModuleHandle(NULL);
	if (hModule == NULL) {
		fprintf(stderr, "Cannot get module handle\n");
		return;
	}

	fp = GetProcAddress(hModule, text);
#endif

#ifdef TUI_OSX
    char *underscore_symbol_name;
    NSSymbol symbol;

    underscore_symbol_name = (char *)malloc(strlen(text) + 2);
    sprintf(underscore_symbol_name, "_%s", text);
    
    symbol = NSLookupAndBindSymbol(underscore_symbol_name);
    if (symbol == NULL) {
        fprintf(stderr, "Cannot find symbol \"%s\"\n", symbol);
        return;
    }

    fp = NSAddressOfSymbol(symbol);
#endif /* TUI_OSX */
    
    if (fp == NULL) {
        fprintf(stderr, "Cannot find function \"%s\" in application.\n", text);
        return;
    }

    TUI_WIDGET_set_callback(widget, (void(*)(TUI_WIDGET_INFO *))fp, NULL);
}

    
static void
get_standard_widget_attributes(const char *cur, void *ui, TUI_WIDGET *widget)
{
    int index;
    char *text;
    char key;
    
    text = get_attribute_value(cur, "name");
    if (text != NULL) {
        TUI_WIDGET_set_name(widget, text);

        index = 0;
        if (tui_find_widget_name_structure_index(widget->tui->root, text,
            &index) == 0) {

            /* save in structure */
            *((TUI_WIDGET **)ui + index + 1) = widget;
        }
    }

    text = get_attribute_value(cur, "callback");
    if (text != NULL) {
        TUI_WIDGET_set_callback_name(widget, text);
        set_callback(widget, text);
    }

    text = get_attribute_value(cur, "key");
    if (text != NULL) {
        assert(widget->type == TUI_MENUITEM);
        TUI_menuitem_set_key(widget, text[0]);
    }

    text = get_attribute_value(cur, "radio_value");
    if (text != NULL) {
        assert(widget->type == TUI_RADIO_BUTTON);
        key = (char)atoi((const char *)text);
        TUI_radio_button_set_selected_value(widget, key);
    }

    text = get_attribute_value(cur, "value");
    if (text != NULL) {
        TUI_WIDGET_set_value(widget, atoi(text));
    }

    text = get_attribute_value(cur, "disabled");
    TUI_WIDGET_set_enabled(widget, text == NULL);

    text = get_attribute_value(cur, "invisible");
    TUI_WIDGET_set_visible(widget, text == NULL);
}


static void
parse_pulldown_block(const char *cur, void *ui, TUI_WIDGET *menubar)
{
    char value[512], *text;
    TUI_WIDGET *pulldown, *menuitem;

    text = get_attribute_value(cur, "text");
    pulldown = TUI_menubar_create_pulldown(menubar, text);
    get_standard_widget_attributes(cur, ui, pulldown);

    for (cur = next_child_block(cur, value);
         cur != NULL;
         cur = skip_block(cur), cur = next_child_block(cur, value)) {

        if (strcmp(value, "menuitem") == 0) {
            text = get_attribute_value(cur, "text");
            menuitem = TUI_pulldown_create_menuitem(pulldown, text);
            get_standard_widget_attributes(cur, ui, menuitem);
        } else if (strcmp(value, "checkmenuitem") == 0) {
            text = get_attribute_value(cur, "text");
            menuitem = TUI_pulldown_create_check_menuitem(pulldown, text);
            get_standard_widget_attributes(cur, ui, menuitem);
        } else if (strcmp(value, "menuseparator") == 0) {
            TUI_pulldown_create_separator(pulldown);
        } else {
            abort();
        }
    }
}

        
        
static TUI_WIDGET *
parse_child_block(const char *cur, char *type, void *ui, TUI_WIDGET *parent)
{
    TUI_WIDGET *widget;
    char value[512], *text;
    int x, y, w, h;
    
    if (strcmp(type, "button") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        text = get_attribute_value(cur, "text");
        widget = TUI_window_create_button(parent, text, x, y);
        get_standard_widget_attributes(cur, ui, widget);

        return widget;
    }

    if (strcmp(type, "radio_button") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        text = get_attribute_value(cur, "text");
        widget = TUI_window_create_radio_button(parent, text, x, y);
        get_standard_widget_attributes(cur, ui, widget);
        
        return widget;
    }

    if (strcmp(type, "label") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        text = get_attribute_value(cur, "text");
        widget = TUI_window_create_label(parent, text, x, y);
        get_standard_widget_attributes(cur, ui, widget);
        
        return widget;
    }

    if (strcmp(type, "title") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        text = get_attribute_value(cur, "text");
        widget = TUI_window_create_title(parent, text, x, y);
        get_standard_widget_attributes(cur, ui, widget);
        
        return widget;
    }

    if (strcmp(type, "textport") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        w = atoi(get_attribute_value(cur, "w"));
        text = get_attribute_value(cur, "text");
        widget = TUI_window_create_textport(parent, text, x, y, w);
        get_standard_widget_attributes(cur, ui, widget);
        
        return widget;
    }

    if (strcmp(type, "oglwindow") == 0) {
        assert(parent->type == TUI_WINDOW);
        x = atoi(get_attribute_value(cur, "x"));
        y = atoi(get_attribute_value(cur, "y"));
        w = atoi(get_attribute_value(cur, "w"));
        h = atoi(get_attribute_value(cur, "h"));
        widget = TUI_window_create_oglwindow(parent, x, y, w, h);
        get_standard_widget_attributes(cur, ui, widget);
        
        return widget;
    }

    if (strcmp(type, "menubar") == 0) {
        widget = TUI_window_create_menubar(parent);

        for (cur = next_child_block(cur, value);
             cur != NULL;
             cur = skip_block(cur), cur = next_child_block(cur, value)) {

            assert(strcmp(value, "pulldown") == 0);
            parse_pulldown_block(cur, ui, widget);
        }

        return widget;
    }

    abort();
    
    return NULL;
}

        
static TUI_WIDGET *
parse_window_block(const char *cur, TUI *tui, void *ui)
{
    TUI_WIDGET *window;
    char value[512];
    char *text;
    int x, y, w, h;
    
    x = atoi(get_attribute_value(cur, "x"));
    y = atoi(get_attribute_value(cur, "y"));
    w = atoi(get_attribute_value(cur, "w"));
    h = atoi(get_attribute_value(cur, "h"));
    text = get_attribute_value(cur, "text");
    window = TUI_create_window(tui, text, x, y, w, h);
    get_standard_widget_attributes(cur, ui, window);
    
    for (cur = next_child_block(cur, value);
         cur != NULL;
         cur = skip_block(cur), cur = next_child_block(cur, value)) {

        parse_child_block(cur, value, ui, window);
    }

	return window;
}


static TUI **
get_app_structure_tui(void *app_structure)
{
    return app_structure;
}

    
int
TUI_construct(TUI_APPLICATION *app, void *app_structure, const char *layout)
{
    TUI **tui;
    const char *cur;
    char value[512], *text;

    cur = layout;

    cur = next_child_block(layout, value);
    assert(strcmp(value, "ui") == 0);

    text = get_attribute_value(cur, "major_version");
    if (atoi(text) < tui_major_version) {
        TUI_errno = TUI_ERR_VERSION;
        return -1;
    }
        
    text = get_attribute_value(cur, "minor_version");
    if (atoi(text) < tui_minor_version) {
        TUI_errno = TUI_ERR_VERSION;
        return -1;
    }

    tui = get_app_structure_tui(app_structure);
    *tui = TUI_create(app);

    for (cur = next_child_block(cur, value); cur != NULL;
         cur = skip_block(cur), cur = next_child_block(cur, value)) {
        parse_window_block(cur, *tui, app_structure);
    }

    assert(next_child_block(cur, value) == NULL);

    return 0;
}

