/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_glue.c,v 1.5 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui_glue.c,v $
 * Revision 1.5  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.4  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.3  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.2  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.1  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 *
 */

#ifdef TUI_PYTHON

#include <Python.h>
#include <pthread.h>

#ifdef TUI_XWIN
#include <unistd.h>
#endif

#include <image.h>
#include <zb.h>

#include "tui_internal.h"
#include "tui_imv_ui.h"
#include "tui_imv.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isTuiImageViewer(m) ((m)->ob_type == &TuiImageViewer)

#define RED_SHIFT		0
#define GRN_SHIFT		8
#define BLU_SHIFT		16
#define ALP_SHIFT		24

#define RGBA(r,g,b,a)  \
   ((a) << ALP_SHIFT | (r) << RED_SHIFT | (g) << GRN_SHIFT | (b) << BLU_SHIFT)


typedef struct tui_imv_wrapper {
    PyObject_HEAD
    TUI_APPLICATION *app;
    TUI_IMV *imv;
    TUI_IMV_UI *ui;
    IMAGE *image;
    unsigned long *rgba;
    int viewport[4];
    int frame;
    pthread_mutex_t mutex;
} TUI_IMV_WRAPPER;
    
    
static PyObject *tui_imv_new(PyObject *self, PyObject *args);
static void tui_imv_dealloc(TUI_IMV_WRAPPER *wrapper);
static int tui_imv_print(TUI_IMV_WRAPPER *self, FILE *fp, int flags);
static PyObject *tui_imv_getattr(TUI_IMV_WRAPPER *self, char *name);
static PyObject *tui_imv_set_frame(TUI_IMV_WRAPPER *self, PyObject *args);
static PyObject *tui_imv_set_previous(TUI_IMV_WRAPPER *self, PyObject *args);
static PyObject *tui_imv_wait_for_exit(TUI_IMV_WRAPPER *self, PyObject *args);
static void setup_image(TUI_IMV_WRAPPER *wrapper);

static int imv_wait_for_exit = 0;

/** Type Descriptors **/

static struct PyMethodDef tui_imv_methods[] = {
    {"set_frame", (PyCFunction)tui_imv_set_frame, 1, NULL},
    {"set_previous", (PyCFunction)tui_imv_set_previous, 1, NULL},
    {"wait_for_exit", (PyCFunction)tui_imv_wait_for_exit, 1, NULL},
};

static PyTypeObject TuiImageViewer = {
    /* type header */
#ifndef WIN32    
    PyObject_HEAD_INIT(&PyType_Type)
#else
    PyObject_HEAD_INIT(NULL)
#endif
    0,                                              /* ob_size               */
    "tui_imv",                                      /* tp_name               */
    sizeof(TUI_IMV_WRAPPER),                        /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)tui_imv_dealloc,                    /* tp_dealloc refcount=0 */
    (printfunc)tui_imv_print,                       /* tp_print   "print x"  */
    (getattrfunc)tui_imv_getattr,                   /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef tuitype_methods[] = {
    {"ImageViewer", tui_imv_new, 1},
    {NULL, NULL}
};


void
inittuicmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32
    TuiImageViewer.ob_type = &PyType_Type;
#endif    
    m = Py_InitModule("tuicmodule", tuitype_methods);/* make the module */
    d = PyModule_GetDict(m);                        /* with 'Tui' func */
    ErrorObject = Py_BuildValue("s", "tuicmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize tui C module");
    }
}

static void
tui_imv_open(TUI_WIDGET_INFO *info)
{}

static void
tui_imv_close(TUI_WIDGET_INFO *info)
{
    fprintf(stderr, "Closing image...\n");
}

static void
tui_imv_stop_processing(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    if (wrapper->image != NULL) {
        IMAGE_notify_stop_processing(wrapper->image);
    }
}

static void
tui_imv_save_as(TUI_WIDGET_INFO *info)
{}

static void
set_play_status(TUI_IMV_WRAPPER *wrapper, TUI_WIDGET *widget)
{
    TUI_WIDGET_set_value(wrapper->ui->play, widget == wrapper->ui->play);
    TUI_WIDGET_set_value(wrapper->ui->reverse, widget == wrapper->ui->reverse);
    TUI_WIDGET_set_value(wrapper->ui->stop, widget == wrapper->ui->stop);
}

static void
tui_imv_forward(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    set_play_status(wrapper, wrapper->ui->play);
    TUI_IMV_stop(wrapper->imv);
    TUI_IMV_play(wrapper->imv);
}

static void
tui_imv_frame_forward(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;
    
    set_play_status(wrapper, wrapper->ui->stop);
    TUI_IMV_stop(wrapper->imv);
    TUI_IMV_step_frame(wrapper->imv, 1);
}

static void
tui_imv_frame_reverse(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    set_play_status(wrapper, wrapper->ui->stop);
    TUI_IMV_stop(wrapper->imv);
    TUI_IMV_step_frame(wrapper->imv, -1);
}

static void
tui_imv_reverse(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    set_play_status(wrapper, wrapper->ui->reverse);
    TUI_IMV_stop(wrapper->imv);
    TUI_IMV_reverse(wrapper->imv);
}

static void
tui_imv_stop(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    set_play_status(wrapper, wrapper->ui->stop);
    TUI_IMV_stop(wrapper->imv);
}

static void
tui_imv_loop(TUI_WIDGET_INFO *info)
{
    int state;
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    state = !TUI_WIDGET_get_value(wrapper->ui->loop);
    TUI_WIDGET_set_value(wrapper->ui->loop, state);
    TUI_WIDGET_set_value(wrapper->ui->autoreverse, 0);
    if (state) {
        TUI_IMV_set_looping(wrapper->imv, TUI_IMV_LOOP_FORWARD);
    } else {
        TUI_IMV_set_looping(wrapper->imv, TUI_IMV_LOOP_STOP);
    }
}

static void
tui_imv_autoreverse(TUI_WIDGET_INFO *info)
{
    int state;
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    state = !TUI_WIDGET_get_value(wrapper->ui->autoreverse);
    TUI_WIDGET_set_value(wrapper->ui->loop, 0);
    TUI_WIDGET_set_value(wrapper->ui->autoreverse, state);
    if (state) {
        TUI_IMV_set_looping(wrapper->imv, TUI_IMV_LOOP_REVERSE);
    } else {
        TUI_IMV_set_looping(wrapper->imv, TUI_IMV_LOOP_STOP);
    }
}

static void
tui_imv_fps_30(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    TUI_WIDGET_set_value(wrapper->ui->fps_30, 1);
    TUI_WIDGET_set_value(wrapper->ui->fps_24, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_12, 0);
    TUI_WIDGET_set_value(wrapper->ui->every_frame, 0);
    TUI_IMV_set_rate(wrapper->imv, 30);
}

static void
tui_imv_fps_24(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    TUI_WIDGET_set_value(wrapper->ui->fps_30, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_24, 1);
    TUI_WIDGET_set_value(wrapper->ui->fps_12, 0);
    TUI_WIDGET_set_value(wrapper->ui->every_frame, 0);
    TUI_IMV_set_rate(wrapper->imv, 24);
}

static void
tui_imv_fps_12(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    TUI_WIDGET_set_value(wrapper->ui->fps_30, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_24, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_12, 1);
    TUI_WIDGET_set_value(wrapper->ui->every_frame, 0);
    TUI_IMV_set_rate(wrapper->imv, 12);
}

void
tui_imv_every_frame(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    TUI_WIDGET_set_value(wrapper->ui->fps_30, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_24, 0);
    TUI_WIDGET_set_value(wrapper->ui->fps_12, 0);
    TUI_WIDGET_set_value(wrapper->ui->every_frame, 1);
    TUI_IMV_set_rate(wrapper->imv, -1);
}

void
tui_imv_exit(TUI_WIDGET_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;

    if (wrapper->image != NULL) {
        IMAGE_notify_stop_processing(wrapper->image);
    }

    imv_wait_for_exit = 0;
    TUI_break(TUI_WIDGET_get_tui(info->widget)); 
}

static void
compute_8bit_color(IMAGE *image, int x, int y, unsigned long *rgba)
{
    double rf, gf, bf, af;
    unsigned char r, g, b, a;
    
    IMAGE_get_color(image, x, y, &rf, &gf, &bf);
    af = IMAGE_get_alpha(image, x, y);

    /* NOTE: apply appropriate gamma here */

    if (rf < 0) r = 0;
    else if (rf > 1) r = 255;
    else r = rf * 255.9;
    
    if (gf < 0) g = 0;
    else if (gf > 1) g = 255;
    else g = gf * 255.9;
    
    if (bf < 0) b = 0;
    else if (bf > 1) b = 255;
    else b = bf * 255.9;
    
    if (af < 0) a = 0;
    else if (af > 1) a = 255;
    else a = bf * 255.9;

    *rgba = RGBA(r, g, b, a);
}

static void
update_image_viewport(TUI_IMV_WRAPPER *wrapper, int x0, int y0, int x1, int y1)
{
    int i, x, y, w;

    w = wrapper->viewport[2] - wrapper->viewport[0] + 1;

    if (x0 < wrapper->viewport[0]) x0 = wrapper->viewport[0];
    if (x1 > wrapper->viewport[2]) x1 = wrapper->viewport[2];
    if (y0 < wrapper->viewport[1]) y0 = wrapper->viewport[1];
    if (y1 > wrapper->viewport[3]) y1 = wrapper->viewport[3];
    if (x1 < x0 || y1 < y0) return;  /* should never happen */
    
    for (y = y0; y <= y1; y++) {
        if (y < wrapper->viewport[1] || y > wrapper->viewport[3]) {
            abort();
        }
        
        i = y * w + x0;
        
        for (x = x0; x <= x1; x++, i++) {
            if (x < wrapper->viewport[0] || x > wrapper->viewport[2]) {
                abort();
            }
            
            compute_8bit_color(wrapper->image, x, y, &wrapper->rgba[i]);
        }
    }
}

static void
create_image_buffers(TUI_IMV_WRAPPER *wrapper)
{
    int x0, y0, x1, y1, n;

    if (wrapper->image == NULL) {
        return;
    }
    
    IMAGE_get_viewport(wrapper->image, &x0, &y0, &x1, &y1);
    n = (x1 - x0 + 1) * (y1 - y0 + 1);

    if (wrapper->rgba == NULL ||
        x0 != wrapper->viewport[0] ||
        y0 != wrapper->viewport[1] ||
        x1 != wrapper->viewport[2] ||
        y1 != wrapper->viewport[3]) {

        if (wrapper->rgba != NULL) {
            free(wrapper->rgba); 
        }
        wrapper->rgba = (unsigned long *)malloc(n * sizeof(unsigned long));
        assert(wrapper->rgba != NULL);
        memset(wrapper->rgba, 0, n * sizeof(unsigned long)); 

        wrapper->viewport[0] = x0;
        wrapper->viewport[1] = y0;
        wrapper->viewport[2] = x1;
        wrapper->viewport[3] = y1; 
    }
    
    update_image_viewport(wrapper, x0, y0, x1, y1); 
}

static void
tui_client_callback(TUI_CLIENT_INFO *info)
{
    TUI_IMV_WRAPPER *wrapper = info->user_data;
    IMAGE_NOTIFY_INFO event;
    int x0, y0, x1, y1;

    if (info->message == TUI_CLIENT_MESSAGE_DESTROY) return;
    
    if (pthread_mutex_trylock(&wrapper->mutex) != 0) {
        return;
    }

#if 0    
    /* handle only one of the expose events */
    if (IMAGE_notify_get_event(wrapper->image, (int)wrapper, &event) == 0) {
        update_image_viewport(wrapper, event.viewport[0]-2, event.viewport[1]-2,
            event.viewport[2]+2, event.viewport[3]+2);
        TUI_IMV_update_rect(wrapper->imv, event.viewport[0] - 2,
            event.viewport[1] - 2, event.viewport[2] + 2, event.viewport[3]+2);
    }

    return;
#endif

    if (wrapper->image == NULL) {
        pthread_mutex_unlock(&wrapper->mutex);
        return;
    }

    /* check to see if the image has changed size since we created buffers */
    IMAGE_get_viewport(wrapper->image, &x0, &y0, &x1, &y1);
    if (wrapper->rgba == NULL ||
        x0 != wrapper->viewport[0] || y0 != wrapper->viewport[1] ||
        x1 != wrapper->viewport[2] || y1 != wrapper->viewport[3]) {
        setup_image(wrapper);
    }

    /* check and handle all exposed image viewports */
    while (IMAGE_notify_get_event(wrapper->image, (int)wrapper, &event) == 0) {
        if (event.type == IMAGE_NOTIFY_START) {
            TUI_WIDGET_set_enabled(wrapper->ui->play_menu, 0);
            TUI_WIDGET_set_enabled(wrapper->ui->close, 0);
            TUI_WIDGET_set_enabled(wrapper->ui->stop_processing, 1);
            continue;
        } else if (event.type == IMAGE_NOTIFY_STOP) {
            TUI_WIDGET_set_enabled(wrapper->ui->play_menu, 1);
            TUI_WIDGET_set_enabled(wrapper->ui->close, 1);
            TUI_WIDGET_set_enabled(wrapper->ui->stop_processing, 0);
            continue;
        } else if (event.type != IMAGE_NOTIFY_VIEWPORT) {
            abort();
        }


/*        fprintf(stderr, "client event notify wrapper=%p  "
            "viewport=[%d, %d, %d, %d]\n", wrapper, event.viewport[0],
            event.viewport[1], event.viewport[2], event.viewport[3]); */
        update_image_viewport(wrapper, event.viewport[0]-2, event.viewport[1]-2,
            event.viewport[2]+2, event.viewport[3]+2);
        TUI_IMV_update_rect(wrapper->imv, event.viewport[0] - 2,
            event.viewport[1] - 2, event.viewport[2] + 2, event.viewport[3]+2);
    }

    pthread_mutex_unlock(&wrapper->mutex);
}

static void *
start_viewer(void *data)
{
    TUI_IMV_WRAPPER *wrapper = data;

    wrapper->app = TUI_APPLICATION_create(
#ifdef TUI_WIN32
            tui_hInstance
#endif        
        );

    wrapper->ui = (TUI_IMV_UI *)malloc(sizeof(TUI_IMV_UI));
    if (TUI_construct(wrapper->app, wrapper->ui, tui_imv_ui_layout) < 0) {
        fprintf(stderr, "Unable to construct UI\n");
        exit(EXIT_FAILURE);
    }

    wrapper->imv = TUI_IMV_create(wrapper->ui->oglwindow);
    assert(wrapper->imv != NULL);

    if (wrapper->image != NULL) {
        setup_image(wrapper);
        IMAGE_add_notify_client(wrapper->image, (int)wrapper);
    }

    TUI_WIDGET_set_callback(wrapper->ui->open, tui_imv_open, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->stop_processing,
        tui_imv_stop_processing, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->close, tui_imv_close, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->save_as, tui_imv_save_as, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->quit, tui_imv_exit, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->play, tui_imv_forward, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->reverse, tui_imv_reverse, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->stop, tui_imv_stop, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->frame_forward, tui_imv_frame_forward,
        wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->frame_reverse, tui_imv_frame_reverse,
        wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->loop, tui_imv_loop, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->autoreverse, tui_imv_autoreverse,
        wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->fps_30, tui_imv_fps_30, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->fps_24, tui_imv_fps_24, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->fps_12, tui_imv_fps_12, wrapper);
    TUI_WIDGET_set_callback(wrapper->ui->every_frame, tui_imv_every_frame,
        wrapper);
    
    TUI_add_client(wrapper->app, tui_client_callback, wrapper);

    pthread_mutex_unlock(&wrapper->mutex);
    
    TUI_mainloop(wrapper->ui->tui);

    /* deallocate and destroy here instead of in tui_imv_dealloc()
       to avoid destroying memory that may be used inside of the
       client_callback */
    pthread_mutex_lock(&wrapper->mutex);

    TUI_remove_client(wrapper->app, tui_client_callback, wrapper);

    if (wrapper->image != NULL) {
        IMAGE_destroy(wrapper->image);
        wrapper->image = NULL;
    }

    if (wrapper->rgba != NULL) {
        free(wrapper->rgba);
        wrapper->rgba = NULL;
    }

    TUI_IMV_destroy(wrapper->imv);
    wrapper->imv = NULL;
    TUI_destroy(wrapper->ui->tui);
    wrapper->ui->tui = NULL;
    wrapper->ui = NULL;
    
    pthread_mutex_unlock(&wrapper->mutex);
    pthread_mutex_destroy(&wrapper->mutex);
    
    /* Someone might have hit exit - destroy the UI, but keep the Python
       object around because the Python script still has a handle */
    if (wrapper->ob_refcnt == 0) {
        PyMem_DEL(wrapper);
    }

    return NULL;
}

static void
setup_image(TUI_IMV_WRAPPER *wrapper)
{
    create_image_buffers(wrapper);

    TUI_IMV_set_image(wrapper->imv, wrapper->frame,
        wrapper->viewport[0], wrapper->viewport[1],
        wrapper->viewport[2], wrapper->viewport[3],
        wrapper->rgba);

    TUI_WIDGET_set_size(wrapper->ui->main_window,
        wrapper->viewport[2] - wrapper->viewport[0] + 1,
        wrapper->viewport[3] - wrapper->viewport[1] + 27 /*menubar*/);

#ifdef TUI_WIN32
    TUI_WIDGET_set_location(wrapper->ui->oglwindow, 0, 0);
#endif

    TUI_IMV_set_window_viewport(wrapper->imv, 
        wrapper->viewport[0], wrapper->viewport[1],
        wrapper->viewport[2], wrapper->viewport[3]);
}

static PyObject *
tui_imv_new(PyObject *self, PyObject *args)
{
    pthread_t thread;
    pthread_attr_t attr;
    IMAGE *image;
    int frame, background;
    TUI_IMV_WRAPPER *wrapper;
    
    image = NULL;
    frame = 0;
    background = 1;
    if (!PyArg_ParseTuple(args, "|Oii", &image, &frame, &background)) 
        onError("You must pass in a valid image when creating a viewer");

    wrapper = PyObject_NEW(TUI_IMV_WRAPPER, &TuiImageViewer);

    if (image != NULL) {
        assert(IMAGE_check(image));
        IMAGE_incref(image);
        wrapper->image = image;
    } else {
        wrapper->image = NULL;
    }

    wrapper->app = NULL;
    wrapper->imv = NULL;
    wrapper->ui = NULL;
    wrapper->rgba = NULL;
    wrapper->frame = frame;
    wrapper->viewport[0] = wrapper->viewport[1] = 0;
    wrapper->viewport[2] = wrapper->viewport[3] = -1;
    pthread_mutex_init(&wrapper->mutex, NULL);

    pthread_mutex_lock(&wrapper->mutex);
    
    if (background) {
        /* start viewer in separate thread in background */
        pthread_attr_init(&attr);
        if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED) != 0) {
            fprintf(stderr, "Cannot set detached state\n");
        }

        if (pthread_create(&thread, &attr, start_viewer, wrapper) != 0) {
            onError("Cannot start a new thread for the image viewer\n");
        }
    } else {
        /* run in the foreground (blocks until exit) */
        start_viewer(wrapper);
    }

    return (PyObject *)wrapper;
}

static void
tui_imv_dealloc(TUI_IMV_WRAPPER *wrapper)
{
    /* Trouble if we are destroying the image at the same time as we
       are in the callback.  This happens because this dealloc function
       is called by the main thread while the update is happening in
       a separate thread.  */
    /* Don't de-allocate anything here, just make sure we are
       broken out of the mainloop, and destroy things after mainloop()
       returns. */
    if (wrapper != NULL && wrapper->ui != NULL && wrapper->ui->tui != NULL) {
        TUI_break(wrapper->ui->tui);
    }
}

static int
tui_imv_print(TUI_IMV_WRAPPER *self, FILE *fp, int flags)
{
    int i;

    if (self->imv->image != NULL) {
        i = self->frame - self->imv->min_frame;
        fprintf(fp, "IMV at %p: image viewport [%d, %d, %d, %d]\n", self->imv,
            self->imv->image[i]->x0,
            self->imv->image[i]->y0,
            self->imv->image[i]->x1,
            self->imv->image[i]->y1);
    } else {
        fprintf(fp, "IMV at %p\n", self->imv);
    }
    return 0;
}

static PyObject *
tui_imv_getattr(TUI_IMV_WRAPPER *self, char *name)
{
    return Py_FindMethod(tui_imv_methods, (PyObject *)self, name);
}

static PyObject *
tui_imv_set_frame(TUI_IMV_WRAPPER *wrapper, PyObject *args)
{
    IMAGE *image;
    int frame, empty;
    int x0, y0, x1, y1;

    image = NULL;
    if (!PyArg_ParseTuple(args, "Oi", &image, &frame))
        onError("set_frame requires an image argument");

    pthread_mutex_lock(&wrapper->mutex);
    
    assert(image != NULL);
    assert(IMAGE_check(image));

    IMAGE_get_viewport(image, &x0, &y0, &x1, &y1);
    empty = IMAGE_empty(image);
            
    if (image != wrapper->image) {
        if (wrapper->image != NULL) {
            /* clean up old image */
            IMAGE_remove_notify_client(wrapper->image, (int)wrapper);
            IMAGE_destroy(wrapper->image);
        }

        /* set up new image */
        wrapper->image = image;
        IMAGE_incref(image);
        IMAGE_add_notify_client(image, (int)wrapper);
        wrapper->rgba = NULL;
    }

    wrapper->frame = frame;

    if (!empty) {
        IMAGE_notify_update_viewport(image, NULL);
        TUI_oglwindow_set_redraw(wrapper->imv->oglwindow, 0);
        setup_image(wrapper);
        TUI_oglwindow_set_redraw(wrapper->imv->oglwindow, 1);
    }
  
    pthread_mutex_unlock(&wrapper->mutex);

    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *
tui_imv_set_previous(TUI_IMV_WRAPPER *self, PyObject *args)
{
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
tui_imv_wait_for_exit(TUI_IMV_WRAPPER *self, PyObject *args)
{
    imv_wait_for_exit = 1;

    if (self->ui == NULL || self->imv == NULL) {
        goto finished;
    }
    
    set_play_status(self, self->ui->play);
    TUI_IMV_play(self->imv);
    
    while (imv_wait_for_exit) {
        sleep(1);
    }


 finished:
    
    Py_INCREF(Py_None);
    return Py_None;
}


#endif /* TUI_PYTHON */
