/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_imv.c,v 1.5 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui_imv.c,v $
 * Revision 1.5  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.4  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.3  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.2  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.1  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include <image.h>

#include "tui_imv.h"
#include "tui_internal.h"

#ifdef TUI_XWIN
#include <sys/time.h>
#endif

#define RED_SHIFT		0
#define GRN_SHIFT		8
#define BLU_SHIFT		16
#define ALP_SHIFT		24
#define RGBA(r,g,b,a)  \
   ((a) << ALP_SHIFT | (r) << RED_SHIFT | (g) << GRN_SHIFT | (b) << BLU_SHIFT)

static void tui_imv_widget_callback(TUI_WIDGET_INFO *info);
static TUI_IMV_IMAGE *tui_imv_image_create(int x0, int y0, int x1, int y1,
    unsigned long *rgba);


TUI_IMV *
TUI_IMV_create(TUI_WIDGET *oglwindow)
{
    TUI_IMV *imv;

    imv = (TUI_IMV *)malloc(sizeof(TUI_IMV));
    if (imv == NULL) {
        TUI_errno = TUI_ERR_MEMORY;
        return NULL;
    }

    imv->oglwindow = oglwindow;
    imv->image = NULL;
    imv->min_frame = INT_MAX;
    imv->max_frame = -INT_MAX;
    imv->cur_frame = 0;
    imv->step = 0;
    imv->previous = NULL;
    imv->last_seconds = 0;
    imv->rate = 24;
    imv->looping = TUI_IMV_LOOP_FORWARD;
    
    TUI_WIDGET_set_callback(oglwindow, tui_imv_widget_callback, imv);

    return imv;
}

void
TUI_IMV_destroy(TUI_IMV *imv)
{
    int i, n;

    n = imv->max_frame - imv->min_frame + 1;
    if (n > 0) {
        /* allocated and destroyed by caller */
        for (i = 0; i < n; i++) {
            if (imv->image[i] != NULL) {
                imv->image[i]->image = NULL;
                free(imv->image[i]); 
                imv->image[i] = NULL;
            }
        }
        free(imv->image);
    }
}

void
TUI_IMV_set_image(TUI_IMV *imv, int frame, int x0, int y0, int x1, int y1,
    unsigned long *rgba)
{
    int n, i, j, max, min;
    TUI_IMV_IMAGE **image;

    if (frame < imv->min_frame) min = frame;
    else min = imv->min_frame;
    if (frame > imv->max_frame) max = frame;
    else max = imv->max_frame;
    
    n = max - min + 1;
    
    image = (TUI_IMV_IMAGE **)malloc(n * sizeof(TUI_IMV_IMAGE *));
    for (i = min; i <= max; i++) {
        j = i - min;
        if (i < imv->min_frame || i > imv->max_frame) {
            image[j] = NULL;
        } else {
            image[j] = imv->image[i - imv->min_frame];
        }
    }

    if (imv->image != NULL) {
        free(imv->image);
    }
    imv->image = image;
    if (min < imv->min_frame) imv->min_frame = min;
    if (max > imv->max_frame) imv->max_frame = max;

    i = frame - min;
    if (imv->image[i] != NULL) {
        free(imv->image[i]); 
    }
    imv->image[i] = tui_imv_image_create(x0, y0, x1, y1, rgba);
    imv->cur_frame = frame;
}

static TUI_IMV_IMAGE *
tui_imv_image_create(int x0, int y0, int x1, int y1, unsigned long *rgba)
{
/*    int w, h, i; */
    TUI_IMV_IMAGE *im;

    im = (TUI_IMV_IMAGE *)malloc(sizeof(TUI_IMV_IMAGE));
    if (im == NULL) {
        TUI_errno = TUI_ERR_MEMORY;
        return NULL;
    }

#if 0    
    w = x1 - x0 + 1;
    h = y1 - y0 + 1;
    
    im->image = (unsigned long *)malloc(w * h * sizeof(unsigned long));
    if (im->image == NULL) {
        TUI_errno = TUI_ERR_MEMORY;
        free(im);
        return NULL;
    }

    for (i = 0; i < w * h; i++) {
        im->image[i] = RGBA(red[i], grn[i], blu[i], alp[i]);
    }

#else
    im->image = rgba;
#endif
    
    im->x0 = x0;
    im->y0 = y0;
    im->x1 = x1;
    im->y1 = y1;
    
    return im;
}

static void
tui_imv_widget_callback(TUI_WIDGET_INFO *info)
{
    int w, h, i, im_w, im_h, x, y;
    TUI_IMV *imv = info->user_data;

    w = TUI_WIDGET_get_width(info->widget);
    h = TUI_WIDGET_get_height(info->widget);

	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
    glOrtho(0, w, 0, h, -1, 1); 
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    glDisable(GL_DEPTH_TEST);

	glClearColor(0.5, 0.5, 0.5, 1);
	glClear(GL_COLOR_BUFFER_BIT);


    i = imv->cur_frame - imv->min_frame;
    if (imv->image != NULL && imv->image[i] != NULL) {

        im_w = imv->image[i]->x1 - imv->image[i]->x0 + 1;
        im_h = imv->image[i]->y1 - imv->image[i]->y0 + 1;

        if (w > im_w) {
            x = (w - im_w) / 2;
        } else {
            x = 0;
        }
        if (h > im_h) {
            y = (h - im_h) / 2;
        } else {
            y = 0;
        }
        glRasterPos2i(x, y);
    
        glDrawPixels(im_w, im_h, GL_RGBA, GL_UNSIGNED_BYTE,
            imv->image[i]->image);
    }
}

static int
advance_frame_count(TUI_IMV *imv)
{
#ifdef TUI_XWIN
    struct timeval curtime;
    struct timezone timezone;
#endif
    double seconds;
    int frame_count;
    
#ifdef TUI_XWIN    
    gettimeofday(&curtime, &timezone);
    seconds = curtime.tv_sec + 0.000001 * curtime.tv_usec;
#endif
#ifdef TUI_WIN32
    seconds = GetTickCount() * 0.001;
#endif    

    if (imv->last_seconds == 0) {
        /* start time */
        frame_count = imv->step;
    } else if (imv->rate > 0) {
        frame_count = imv->step * (seconds - imv->last_seconds) * imv->rate;
    } else {
        /* every frame at 100 fps */
        frame_count = imv->step * (seconds - imv->last_seconds) * 100;
        if (frame_count > 0) frame_count = 1;
        else if (frame_count < 0) frame_count = -1;
    }

    if (frame_count != 0) {
        imv->last_seconds = seconds;
    }
    
    return frame_count;
}

void
TUI_IMV_update_rect(TUI_IMV *imv, int x0, int y0, int x1, int y1)
{
    int i;

    i = imv->cur_frame - imv->min_frame;
    assert(i >= 0 && i < imv->max_frame - imv->min_frame + 1);

    /* make sure all of this is valid */
    if (imv->image[i] == NULL ||
        imv->image[i]->x1 < imv->image[i]->x0 ||
        imv->image[i]->y1 < imv->image[i]->y0) {
        return;
    }

    tui_oglwindow_force_redraw(imv->oglwindow);
}

void
TUI_IMV_set_window_viewport(TUI_IMV *imv, int x0, int y0, int x1, int y1)
{
/*    TUI_WIDGET_set_size(imv->oglwindow, x1-x0+1, y1-y0+1); */
}

static void
step_frame(TUI_IMV *imv, int step)
{
    imv->cur_frame += step;

    if (imv->cur_frame > imv->max_frame) {
        switch (imv->looping) {
        case TUI_IMV_LOOP_STOP:
            imv->cur_frame = imv->max_frame;
            break;
        case TUI_IMV_LOOP_FORWARD:
            imv->cur_frame = imv->min_frame;
            break;
        case TUI_IMV_LOOP_REVERSE:
            imv->step = -imv->step;
            imv->cur_frame = imv->max_frame;
            break;
        }
    } else if (imv->cur_frame < imv->min_frame) {
        switch (imv->looping) {
        case TUI_IMV_LOOP_STOP:
            imv->cur_frame = imv->min_frame;
            break;
        case TUI_IMV_LOOP_FORWARD:
            imv->cur_frame = imv->max_frame;
            break;
        case TUI_IMV_LOOP_REVERSE:
            imv->step = -imv->step;
            imv->cur_frame = imv->min_frame;
            break;
        }
    }
}

static void
tui_imv_play_callback(TUI_CLIENT_INFO *info)
{
    TUI_IMV *imv = info->user_data;
    int frame_count;
    
    if (info->message == TUI_CLIENT_MESSAGE_DESTROY) {
        return;
    }

    frame_count = advance_frame_count(imv);
    step_frame(imv, frame_count);
    tui_oglwindow_force_redraw(imv->oglwindow);
}

void
TUI_IMV_step_frame(TUI_IMV *imv, int step)
{
    step_frame(imv, step * imv->step);
    tui_oglwindow_force_redraw(imv->oglwindow);
}
    
void
TUI_IMV_play(TUI_IMV *imv)
{
    imv->step = 1;
    TUI_add_client(imv->oglwindow->tui->app, tui_imv_play_callback, imv);
}

void
TUI_IMV_reverse(TUI_IMV *imv)
{
    imv->step = -1;
    TUI_add_client(imv->oglwindow->tui->app, tui_imv_play_callback, imv);
}

void
TUI_IMV_stop(TUI_IMV *imv)
{
    TUI_remove_client(imv->oglwindow->tui->app, tui_imv_play_callback, imv);
}

void
TUI_IMV_set_rate(TUI_IMV *imv, float frames_per_second)
{
    imv->rate = frames_per_second;
}

void
TUI_IMV_set_looping(TUI_IMV *imv, TUI_IMV_LOOPING looping)
{
    imv->looping = looping;
}

