/* Copyright (c) 2000, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_mainloop.c,v 1.14 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui_mainloop.c,v $
 * Revision 1.14  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.13  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.12  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.11  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.10  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.9  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.8  2000/10/06 03:48:46  wex
 * Slightly better events, widget windows capture other events.
 *
 * Revision 1.7  2000/10/02 09:28:34  wex
 * Minor bug fixes.
 *
 * Revision 1.6  2000/09/30 18:00:24  wex
 * Handle more visual types.  Menubars resize correctly.
 *
 * Revision 1.5  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 * Revision 1.4  2000/09/26 16:31:38  wex
 * Events working under Windows.
 *
 * Revision 1.3  2000/09/25 02:35:02  wex
 * Creating windows and widgets in win32, events still broken.
 *
 * Revision 1.2  2000/09/24 03:13:08  wex
 * Initial changes to support WIN32
 *
 * Revision 1.1.1.1  2000/09/16 23:38:08  wex
 * Initial revision
 *
 */

#include <stdio.h>

#include "tui_internal.h"

#ifdef TUI_XWIN
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define XK_LATIN1
#define XK_MISCELLANY
#include <X11/keysymdef.h>

static TUI_WIDGET *
tui_window_find_widget_by_location(TUI_WIDGET *window, int x, int y)
{
    TUI_WIDGET *widget;

    for (widget = window->child; widget != NULL; widget = widget->next) {
        if (x >= widget->x && x <= widget->x + widget->w &&
            y >= widget->y && y <= widget->y + widget->h) {
            return widget;
        }
    }

    return NULL;
}

static int
tui_widget_handle_event(TUI_WIDGET *widget, XEvent *event)
{
    TUI_EVENT_INFO info;
    TUI_WIDGET *window;
    char buffer[128];
    XComposeStatus status;
    KeySym keysym;
    Window child_return;
    XEvent tmp_event;
    
    if (widget->type != TUI_WINDOW) {
        window = widget->parent;
        if (window->type != TUI_WINDOW) {
            assert(window->type == TUI_MENUBAR);
            window = window->parent;
        }
        assert(window->type == TUI_WINDOW);
    } else {
        window = widget;
    }
    
    if (window->intercept != NULL) {
        info.window = window;
        info.widget = widget;
        info.user_data = window->intercept_user_data;
        info.x = 0;
        info.y = 0;
        info.key = 0;
        info.state = 0;
        info.type = TUI_EVENT_NONE;
        switch (event->type) {
        case ButtonPress:
        case ButtonRelease:
            info.type = TUI_EVENT_BUTTON;
            info.state = event->type == ButtonPress;
            if (widget == window) {
                info.x = event->xbutton.x;
                info.y = event->xbutton.y;
                info.widget = tui_window_find_widget_by_location(window,
                    info.x, info.y);
            } else {
                XTranslateCoordinates(window->tui->app->display,
                    widget->window, window->window, event->xbutton.x,
                    event->xbutton.y, &info.x, &info.y, &child_return);
            }
            
            break;
        case KeyPress:
        case KeyRelease:
            info.type = TUI_EVENT_KEY;
            info.state = event->type == KeyPress;
            XLookupString(&event->xkey, buffer, 128, &keysym, &status);
            info.key = ((int) keysym - XK_0 + '0');
            break;
        case MotionNotify:
            info.type = TUI_EVENT_MOTION;
            if (widget == window) {
                info.x = event->xmotion.x;
                info.y = event->xmotion.y;
                info.widget = tui_window_find_widget_by_location(window,
                    info.x, info.y);
            } else {
                XTranslateCoordinates(window->tui->app->display,
                    widget->window, window->window, event->xmotion.x,
                    event->xmotion.y, &info.x, &info.y, &child_return);
            }
                
            break;
        }

        if (window->intercept(&info) != 0) {
            return 0;
        }
    }


    /* handle event normally */
	switch (event->type) {

	case Expose:
		if (widget->expose_fp != NULL) {
			widget->expose_fp(widget, &event->xexpose);
		}
		break;

	case EnterNotify:
	case LeaveNotify:
		if (widget->cross_fp != NULL) {
			widget->cross_fp(widget, &event->xcrossing);
		}
		break;

	case ButtonPress:
	case ButtonRelease:
		if (widget->button_fp != NULL) {
			widget->button_fp(widget, &event->xbutton);
		}
		break;

	case KeyPress:
	case KeyRelease:
		if (widget->key_fp != NULL) {
			widget->key_fp(widget, &event->xkey);
		}
		break;

	case MotionNotify:
		if (widget->motion_fp != NULL) {
			widget->motion_fp(widget, &event->xmotion);
		}
		break;

	case ConfigureNotify:
#if 1
        if (XCheckMaskEvent(widget->tui->app->display,
                StructureNotifyMask, &tmp_event)) {
            if (tmp_event.type == ConfigureNotify) {
                /* replace this event */
                *event = tmp_event;
            } else {
                tui_widget_handle_event(widget, &tmp_event);
            }
            
            while (XCheckMaskEvent(widget->tui->app->display,
                StructureNotifyMask, &tmp_event)) {
                if (tmp_event.type == ConfigureNotify) {
                    /* replace this event */
                    *event = tmp_event;
                } else {
                    tui_widget_handle_event(widget, &tmp_event);
                }
            }
        }
#endif        
        tui_widget_resize(widget->child, event->xconfigure.width,
            event->xconfigure.height);

        break;
        
	case ReparentNotify:
	case MapNotify:
	case UnmapNotify:
	case DestroyNotify:
		/*EMPTY*/
		break;

	default:
		TUI_errno = TUI_ERR_EVENT;
		return -1;
	}

	return 0;
}


static TUI_WIDGET *
find_widget_by_window(TUI_WIDGET *widget, Window window)
{
	TUI_WIDGET *sibling, *child;

	for (sibling = widget->child; sibling != NULL; sibling = sibling->next) {
		if (sibling->window == window) {
			return sibling;
		}

		child = find_widget_by_window(sibling, window);
		if (child != NULL) {
			return child;
		}
	}

	TUI_errno = TUI_ERR_NO_WINDOW;
	
	return NULL;
}

int
TUI_mainloop(TUI *tui)
{
	XEvent event;
	TUI_WIDGET *widget;
    TUI_CLIENT *client;
	int pending_count;
    int select_count;
    Display *display;
	struct timeval timeout;
	fd_set read_fds;
	fd_set write_fds;

    display = tui->app->display;

	while (!tui->app->break_out_of_mainloop) {

        /* handle all existing X events */
        pending_count = XPending(display);
        while (pending_count > 0 && XPending(display)) {
            XNextEvent(display, &event);
            widget = find_widget_by_window(tui->root, event.xany.window);
            if (widget != NULL) {
                if (tui_widget_handle_event(widget, &event) < 0) {
                    fprintf(stderr, "Widget handler returned error\n");
                }
            } else {
                fprintf(stderr, "Cannot find widget window\n");
            }
            pending_count--;
        }

        /* call any client routines */
        for (client = LIST_head(tui->app->client_list); client != NULL;
             client = LIST_next(client)) {
            if (client->callback != NULL) {
                TUI_CLIENT_INFO info;
                info.user_data = client->user_data;
                info.message = TUI_CLIENT_MESSAGE_UPDATE;
                client->callback(&info);
            }
        }
        
        /* now block in select for 1/10th of a second, but break
           out if any X events come up -- later add file descriptors
           for the clients */
        timeout.tv_sec = 0;
        timeout.tv_usec = 20000;   /* 1/50th of a second */
        FD_ZERO(&read_fds);
        FD_ZERO(&write_fds);
        XFlush(display);
        FD_SET(ConnectionNumber(display), &read_fds);

        select_count = select(FD_SETSIZE, &read_fds, &write_fds,NULL,&timeout);
	}

	return 0;
}

#endif /* TUI_XWIN */


#ifdef TUI_WIN32

HINSTANCE tui_hInstance = NULL;

static TUI_WIDGET *
find_widget_by_id(TUI_WIDGET *widget, int id)
{
	TUI_WIDGET *child;
	
	if (widget->control_id == id) {
		return widget;
	}

	if (widget->child != NULL) {
		child = find_widget_by_id(widget->child, id);
		if (child != NULL) {
			return child;
		}
	}

	if (widget->next != NULL) {
		return find_widget_by_id(widget->next, id);
	}

	return NULL;
}

static TUI_WIDGET *
find_widget_by_hwnd(TUI_WIDGET *widget, HWND hwnd)
{
	TUI_WIDGET *child;
	
	if (widget->hwnd == hwnd) {
		return widget;
	}

	if (widget->child != NULL) {
		child = find_widget_by_hwnd(widget->child, hwnd);
		if (child != NULL) {
			return child;
		}
	}

	if (widget->next != NULL) {
		return find_widget_by_hwnd(widget->next, hwnd);
	}

	return NULL;
}


static void
tui_win32_command(HWND hwnd, int id, HWND hwndctl, UINT codenotify)
{
    TUI *tui;
	TUI_WIDGET *widget;
	TUI_WIDGET_INFO info;

    tui = GetWindowLong(hwnd, GWL_USERDATA);
    if (tui == NULL) {
        return;
    }
    
    widget = find_widget_by_id(tui->root, id);

    if (widget != NULL && widget->type == TUI_RADIO_BUTTON) {
        TUI_WIDGET_set_value(widget, widget->key);
    }

    if (widget != NULL && widget->callback != NULL) {
        info.widget = widget;
        info.user_data = widget->user_data;
        widget->callback(&info);
        return;
    }

	FORWARD_WM_COMMAND(hwnd, id, hwndctl, codenotify, DefWindowProc);
}


static void
tui_win32_destroy(HWND hwnd)
{
	PostQuitMessage(0);
}


LRESULT CALLBACK
tui_win32_wndproc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    TUI *tui;
    TUI_CLIENT *client;
    TUI_WIDGET *widget;
    PAINTSTRUCT ps;

    tui = GetWindowLong(hwnd, GWL_USERDATA);

	switch (message) {
    case WM_CREATE:
        return 0;
    case WM_CLOSE:
        PostQuitMessage(0);
        return 0;

    case WM_KEYDOWN:
        switch (wparam) {
        case VK_ESCAPE:
            PostQuitMessage(0);
        }
        return 0;

	case WM_COMMAND:
		return HANDLE_WM_COMMAND(hwnd, wparam, lparam, tui_win32_command);
	case WM_DESTROY:
		return HANDLE_WM_DESTROY(hwnd, wparam, lparam, tui_win32_destroy);
	case WM_SETFOCUS:
		SetFocus(hwnd);
		return 0;

    case WM_TIMER:
        /* call any client routines */
        if (tui == NULL) {
            fprintf(stderr, "NULL tui in win32_wndproc (timer)\n");
        } else {
            for (client = LIST_head(tui->app->client_list); client != NULL;
                 client = LIST_next(client)) {
                if (client->callback != NULL) {
                    TUI_CLIENT_INFO info;
                    info.user_data = client->user_data;
                    info.message = TUI_CLIENT_MESSAGE_UPDATE;
                    client->callback(&info);
                }
            }
        }
        
	case WM_SIZE:
        fprintf(stderr, "Got resize event\n");
        if (tui == NULL) {
            fprintf(stderr, "NULL tui in win32_wndproc (size)\n");
        } else {
            widget = find_widget_by_hwnd(tui->root, hwnd);
            if (widget != NULL) {
                fprintf(stderr, "Found widget to resize\n");
                tui_widget_resize(widget->child, LOWORD(lparam),
                    HIWORD(lparam));
            }
        }
        return 0;

    case WM_PAINT:
        if (tui == NULL) {
            fprintf(stderr, "NULL tui in win32_wndproc (paint)\n");
        } else {
            widget = find_widget_by_hwnd(tui->root, hwnd);
            if (widget != NULL) {
                if (widget->type == TUI_OGLWINDOW) {
                    /* call begin/end to stop paint events */
                    BeginPaint(hwnd, &ps);
                    tui_oglwindow_force_redraw(widget);
                    EndPaint(hwnd, &ps);
                    return 0;
                }
                if (widget->type == TUI_WINDOW) {
                    /* forward this message on to the subwindows */
                    BeginPaint(hwnd, &ps);
                    tui_redraw_all_oglwindows(widget->child);
                    EndPaint(hwnd, &ps);
                    return 0;
                }
            }
        }
        /* otherwise handle as usual */
		return DefWindowProc(hwnd, message, wparam, lparam);
                
	default:
		return DefWindowProc(hwnd, message, wparam, lparam);
	}
}

HWND
find_first_hwnd(TUI_WIDGET *widget)
{
    if (widget->type == TUI_WINDOW) return widget->hwnd;
    if (widget->child != NULL) {
        HWND hwnd;
        hwnd = find_first_hwnd(widget->child);
        if (hwnd != NULL) {
            return hwnd;
        }
    }

    if (widget->next != NULL) return find_first_hwnd(widget->next);

    return NULL;
}

int
TUI_mainloop(TUI *tui)
{
	MSG msg;
	HWND hwnd;
    
    /* find a top level window and start a timer */
    hwnd = find_first_hwnd(tui->root);
    if (hwnd != NULL && !LIST_is_empty(tui->app->client_list)) {
        SetTimer(hwnd, 101, 25, (TIMERPROC)NULL);
    }
    
	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return 0;
}

#endif /* TUI_WIN32 */



#ifdef TUI_OSX



static void
redraw_window(WindowPtr window)
{
/*    EraseRect(GetWindowPort(window)); */
}

static void
activate_window(WindowPtr window, int state)
{
    ControlHandle control;

    GetRootControl(window,&control);

    if (state) {
        ActivateControl(control);
        ShowFloatingWindows();
    } else {
        DeactivateControl(control);
        HideFloatingWindows();
    }

    redraw_window(window);
}


static void
handle_os_event(EventRecord *eventStrucPtr)
{
    int state;

    state = (eventStrucPtr->message & resumeFlag) == 0;
    
    switch((eventStrucPtr->message >> 24) & 0x000000FF) {
    case suspendResumeMessage:
        activate_window(FrontWindow(), state);
        break;
    case mouseMovedMessage:
        break;
    }
}

static TUI_WIDGET *
find_widget_from_control(TUI_WIDGET *widget, ControlHandle control)
{
    TUI_WIDGET *found;
    
    if (widget->control == control) {
        return widget;
    }

    if (widget->child) {
        found = find_widget_from_control(widget->child, control);
        if (found != NULL) return found;
    }

    if (widget->next) {
        return find_widget_from_control(widget->next, control);
    }

    return NULL;
}

    
static void
handle_in_content(EventRecord *event, WindowPtr window)
{
    ControlHandle control;
    SInt16 part, value;
    TUI_WIDGET *widget, *window_widget;
    TUI_WIDGET_INFO info;
    
    SetPortWindowPort(window);
    GlobalToLocal(&event->where);

    window_widget = (TUI_WIDGET *)GetWRefCon(window);
    assert(window_widget != NULL);
    
    control = FindControlUnderMouse(event->where, window, &part);
    if (control != NULL) {
        HandleControlClick(control, event->where, event->modifiers, NULL);
        widget = find_widget_from_control(window_widget->tui->root, control);
        if (widget != NULL && widget->callback != NULL) {
            info.widget = widget;
            info.user_data = widget->user_data;
            widget->callback(&info);
        }
    }
}


static void
handle_mouse_down(EventRecord *event)
{
    SInt16     part;
    WindowPtr  window;
    Rect       grow_rect;
    SInt32     new_size;

    part = FindWindow(event->where,&window);

    switch(part) {
    case inMenuBar:
/*        doMenuChoice(MenuSelect(event->where)); */
        break;

    case inContent:
        if (window != FrontWindow()) {
            SelectWindow(window);
        } else {
            handle_in_content(event,window);
        }
        break;

    case inDrag:
        DragWindow(window, event->where, NULL);
        break;

    case inGoAway:
        if (TrackGoAway(window, event->where) == true) {
            /* all done */
            exit(0);
        }
        break;

    case inGrow:
        grow_rect.top = 0;
        grow_rect.left = 0;
        grow_rect.right = 1024;
        grow_rect.bottom = 1024;
        new_size = GrowWindow(window, event->where,&grow_rect);
        if(new_size != 0) {
            SizeWindow(window, LoWord(new_size), HiWord(new_size), true);
/*            doAdjustScrollBars(window); */
            redraw_window(window);
        }
        break;

    case inZoomIn:
    case inZoomOut:
        if (TrackBox(window, event->where, part)) {
            SetPortWindowPort(window);
/*           EraseRect(GetWindowPort(window)); */
            ZoomWindow(window, part, false);
/*            doAdjustScrollBars(window); */
        }
        break;
    }
}


int
TUI_mainloop(TUI *tui)
{
    EventRecord event;
    SInt16 part;
    WindowPtr window;
    RgnHandle region;

    while (!tui->app->break_out_of_mainloop) {
        if (WaitNextEvent(everyEvent, &event, 60L, NULL)) {

            if (tui->app->break_out_of_mainloop) break;
            
            switch (event.what) {
            case mouseDown:
                handle_mouse_down(&event);
                break;

            case mouseUp:
                
            case keyDown:
            case autoKey:
                break;

            case activateEvt:
                window = (WindowPtr)event.message;
                activate_window(window,
                    ((event.modifiers & activeFlag) == activeFlag));
                break;

            case kHighLevelEvent:
                AEProcessAppleEvent(event);
                break;

            case updateEvt:
                window = (WindowPtr)event.message;
                BeginUpdate(window);
                GetPortVisibleRegion((CGrafPtr)window, region);
                UpdateControls(window, region);
                SetPortWindowPort(window);
                redraw_window(window);
                EndUpdate(window);
                break;

            case osEvt:
                handle_os_event(&event);
                break;
                
            default:
                fprintf(stderr, "Default event %d\n", event.what);
                break;
            }
            
        } /* WaitNextEvent */
    }

    return 0;
}

#endif
