/* Copyright (c) 2001, Flarg Inc, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_oglwindow.c,v 1.4 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui_oglwindow.c,v $
 * Revision 1.4  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.3  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.2  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.1  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "tui_internal.h"

#ifdef TUI_XWIN
#include <GL/glx.h>
#define XK_LATIN1
#define XK_MISCELLANY
#include <X11/keysymdef.h>

static Window tui_ogl_create_xwindow(TUI_WIDGET *window, int x, int y,
    int w, int h, GLXContext *glx_context);
static void tui_ogl_draw(TUI_WIDGET *widget);
static void tui_ogl_expose(TUI_WIDGET *widget, XExposeEvent *event);
static void tui_ogl_button(TUI_WIDGET *widget, XButtonEvent *event);
static void tui_ogl_resize(TUI_WIDGET *widget, int width, int height);
static void tui_ogl_key(TUI_WIDGET *widget, XKeyEvent *event);
#endif

#ifdef TUI_WIN32
static HWND tui_ogl_create_win32_window(TUI_WIDGET *window, int x, int y,
    int w, int h, HGLRC *hrc, HDC *hdc);
#endif

static int tui_ogl_make_current(TUI_WIDGET *widget);


TUI_WIDGET *
TUI_window_create_oglwindow(TUI_WIDGET *window, int x, int y, int w, int h)
{
    TUI_WIDGET *widget;

#ifdef TUI_XWIN
    Window ogl_window;
    GLXContext glx_context;
    
    ogl_window = tui_ogl_create_xwindow(window, x, y, w, h, &glx_context);
    if (ogl_window == 0) {
        return NULL;
    }

    XSelectInput(window->tui->app->display, ogl_window,
        StructureNotifyMask | ExposureMask | KeyPressMask | ButtonPressMask |
        PointerMotionMask | ButtonReleaseMask);
#endif    

#ifdef TUI_WIN32
    HWND hwnd;
    HGLRC hrc;
    HDC hdc;
    
    hwnd = tui_ogl_create_win32_window(window, x, y, w, h, &hrc, &hdc);
    if (hwnd == NULL) {
        return NULL;
    }
    SetWindowLong(hwnd, GWL_USERDATA, (long)window->tui);
#endif
    
    widget = tui_widget_allocate();
    if (widget == NULL) {
        return NULL;
    }

    widget->tui = window->tui;
    widget->type = TUI_OGLWINDOW;
    widget->parent = window;
    tui_widget_append_to_list(&window->child, widget);
    
#ifdef TUI_XWIN
	widget->window = ogl_window;
    widget->glx_context = glx_context;
    widget->draw_fp = tui_ogl_draw;
    widget->expose_fp = tui_ogl_expose;
    widget->key_fp = tui_ogl_key;
    widget->button_fp = tui_ogl_button;
    widget->resize_fp = tui_ogl_resize;
#endif

#ifdef TUI_WIN32
    widget->hwnd = hwnd;
    widget->hrc = hrc;
    widget->hdc = hdc;
#endif
    
	widget->x = x;
	widget->y = y;
	widget->w = w;
	widget->h = h;
	widget->text = NULL;

	TUI_WIDGET_set_visible(widget, 1);

	return widget;
}


#ifdef TUI_XWIN

static Window
tui_ogl_create_xwindow(TUI_WIDGET *window, int x, int y, int w, int h,
    GLXContext *glx_context)
{
    int n;
	int error_base;
	int event_base;
	int screen;
    Window ogl_window;
	int glx_attrib_list[16];
	XSetWindowAttributes win_attr;
    Display *display;
	Colormap colormap;
	XVisualInfo *visual_info;
    
    display = window->tui->app->display;
    screen = DefaultScreen(display);
    
    if (!glXQueryExtension(display, &error_base, &event_base)) {
		fprintf(stderr, "(TUI) GLX extension not supported.\n");
		exit(EXIT_FAILURE);
	}

    n = 0;
    glx_attrib_list[n++] = GLX_RGBA;
    glx_attrib_list[n++] = GLX_DOUBLEBUFFER;
    glx_attrib_list[n++] = GLX_RED_SIZE;
    glx_attrib_list[n++] = 1;
    glx_attrib_list[n++] = GLX_GREEN_SIZE;
    glx_attrib_list[n++] = 1;
    glx_attrib_list[n++] = GLX_BLUE_SIZE;
    glx_attrib_list[n++] = 1;
    glx_attrib_list[n++] = GLX_DEPTH_SIZE;
    glx_attrib_list[n++] = 1;
	glx_attrib_list[n++] = None;

	visual_info = glXChooseVisual(display, screen, glx_attrib_list);
	if (visual_info == NULL) {
		fprintf(stderr, "(TUI) Cannot choose GLX visual.\n");
		exit(EXIT_FAILURE);
	}


	*glx_context = glXCreateContext(display, visual_info, NULL, True);
	if (*glx_context == NULL) {
		fprintf(stderr, "(TUI) Cannot create GLX context.\n");
		exit(EXIT_FAILURE);
	}

	colormap = XCreateColormap(display, RootWindow(display, screen),
        visual_info->visual, AllocNone);
	
    win_attr.border_pixel = 0;
	win_attr.colormap = colormap;
	
	ogl_window = XCreateWindow(display, window->window,
		x, y, w, h, 0,
		visual_info->depth, InputOutput, visual_info->visual,
		CWBorderPixel | CWColormap, &win_attr);

    return ogl_window;
}

static void
tui_ogl_draw(TUI_WIDGET *widget)
{
    TUI_GC gc[] = {
        TUI_GC_LIGHT_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,

        TUI_GC_TERRACOTTA,

        TUI_GC_MEDIUM_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_LIGHT_GRAY,
        TUI_GC_VERY_LIGHT_GRAY,
        TUI_GC_NULL
    };

    if (widget->callback == NULL) {
        tui_draw_frame(widget->tui, widget->window,0,0,widget->w,widget->h,gc);
        return;
    }

    if (tui_ogl_make_current(widget) == 0) {
        glDrawBuffer(GL_BACK);
        
        if (widget->callback != NULL) {
            TUI_WIDGET_INFO info;
            info.widget = widget;
            info.x = 0;
            info.y = 0;
            info.state = 1;
            info.user_data = widget->user_data;
            widget->callback(&info);
        }

        glXSwapBuffers(widget->tui->app->display, widget->window);
    }
}

static void
tui_ogl_expose(TUI_WIDGET *widget, XExposeEvent *event)
{
    widget->draw_fp(widget);
}

static void
tui_ogl_button(TUI_WIDGET *widget, XButtonEvent *event)
{
}


static void
tui_ogl_key(TUI_WIDGET *widget, XKeyEvent *event)
{
    widget->parent->key_fp(widget->parent, event);

#if 0    
    char buffer[128], key;
    KeySym keysym;
    XComposeStatus status;
    TUI_WIDGET_INFO info;

    if (event->type != KeyPress || (event->state & ControlMask)) {
        return;
    }
    
    XLookupString(event, buffer, 128, &keysym, &status);
        widget->parent->key_fp(widget->parent, event);

    if (keysym == XK_Tab || !widget->selected) {
        /* forward back up to the window */
        widget->parent->key_fp(widget->parent, event);
        return;
    } else if (keysym == XK_BackSpace || keysym == XK_Delete) {
    } else if (keysym == XK_Return) {
        if (widget->callback != NULL) {
            info.widget = widget;
            info.user_data = widget->user_data;
            widget->callback(&info);
        }
    } else if (keysym == XK_Left) {
    } else if (keysym == XK_Right) {
    } else if (keysym < XK_space || keysym > XK_asciitilde) {
        return;
    } else {
        key = (char) ((int) keysym - XK_0 + '0');
    }
#endif    
}

static void
tui_ogl_resize(TUI_WIDGET *widget, int width, int height)
{
    
    height -= TUI_MENUBAR_SPACING;
    TUI_WIDGET_set_size(widget, width, height);
}

#endif /* TUI_XWIN */


#ifdef TUI_WIN32

static HWND
tui_ogl_create_win32_window(TUI_WIDGET *window, int x, int y,int w, int h,
    HGLRC *hrc, HDC *hdc)
{
    HWND hwnd;
    PIXELFORMATDESCRIPTOR pfd;
    int format;
    
    /* Create the opengl window */
    hwnd = CreateWindow(TUI_WINDOW_CLASS_NAME, "OpenGL Win32",
        WS_CHILD | WS_VISIBLE, x, y, w, h, window->hwnd, NULL,
        window->tui->app->hInstance, NULL );

    if (hwnd == NULL) {
        return NULL;
    }
    
    *hdc = GetDC(hwnd);
    
    /* set the pixel format for the DC */
    ZeroMemory(&pfd, sizeof(pfd));
    pfd.nSize = sizeof(pfd);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 24;
    pfd.cDepthBits = 16;
    pfd.iLayerType = PFD_MAIN_PLANE;
    format = ChoosePixelFormat(*hdc, &pfd);
    SetPixelFormat(*hdc, format, &pfd);
 
    /* create and enable the render context (RC) */
    *hrc = wglCreateContext(*hdc);

    return hwnd;
}

static void
tui_ogl_release(TUI_WIDGET *widget)
{
    SwapBuffers(widget->hdc);
    wglMakeCurrent(NULL, NULL);
    ReleaseDC(widget->hwnd, widget->hdc);
}

#endif

void
tui_oglwindow_force_redraw(TUI_WIDGET *widget)
{
#ifdef TUI_XWIN
    widget->draw_fp(widget);
#endif

#ifdef TUI_WIN32    
    if (tui_ogl_make_current(widget) == 0) {
        if (widget->callback != NULL) {
            TUI_WIDGET_INFO info;
            info.widget = widget;
            info.x = 0;
            info.y = 0;
            info.state = 1;
            info.user_data = widget->user_data;
            widget->callback(&info);
        }

        tui_ogl_release(widget);
    }
    
#endif    
}

void
tui_redraw_all_oglwindows(TUI_WIDGET *widget)
{
    if (widget == NULL) return;
    if (widget->type == TUI_OGLWINDOW) tui_oglwindow_force_redraw(widget);
    if (widget->child != NULL) tui_redraw_all_oglwindows(widget->child);
    if (widget->next != NULL) tui_redraw_all_oglwindows(widget->next);
}

static int
tui_ogl_make_current(TUI_WIDGET *widget) 
{
    if (widget->selected) return -1;
    
#ifdef TUI_XWIN
	if (!glXMakeCurrent(widget->tui->app->display, widget->window,
            widget->glx_context)) {
		fprintf(stderr, "(TUI) Cannot make GLX context current.\n");
        return -1;
		exit(EXIT_FAILURE);
	}
#endif

#ifdef TUI_WIN32
    widget->hdc = GetDC(widget->hwnd);
    wglMakeCurrent(widget->hdc, widget->hrc);
#endif
    
    return 0;
}

int
TUI_oglwindow_set_redraw(TUI_WIDGET *oglwindow, int state)
{
    int oldstate;
    oldstate = !oglwindow->selected;
    oglwindow->selected = !state;
    return oldstate;
}

