/* Copyright (c) 2000, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_pulldown.c,v 1.13 2002/01/03 06:50:13 wex Exp $
 *
 * $Log: tui_pulldown.c,v $
 * Revision 1.13  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.12  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.11  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.10  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.9  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.8  2000/10/05 10:29:41  wex
 * Added labels, titles, and textports.  Fixed parsing bug and radio look.
 *
 * Revision 1.7  2000/10/02 09:28:34  wex
 * Minor bug fixes.
 *
 * Revision 1.6  2000/10/02 07:16:48  wex
 * Keyboard equivalents.
 *
 * Revision 1.5  2000/09/30 15:59:14  wex
 * Saving and loading generated header files.
 *
 * Revision 1.4  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 * Revision 1.3  2000/09/25 02:35:02  wex
 * Creating windows and widgets in win32, events still broken.
 *
 * Revision 1.2  2000/09/24 03:13:08  wex
 * Initial changes to support WIN32
 *
 * Revision 1.1.1.1  2000/09/16 23:38:08  wex
 * Initial revision
 *
 */

#include <ctype.h>
#include <string.h>

#include "tui_internal.h"

#ifdef TUI_XWIN

#define TUI_PULLDOWN_MARGIN				10
#define TUI_PULLDOWN_KEY_MARGIN         58
#define TUI_PULLDOWN_KEY_SPACE          15
#define TUI_PULLDOWN_TOP				26
#define TUI_PULLDOWN_ITEM_HEIGHT		22
#define TUI_PULLDOWN_SEPARATOR_HEIGHT	5
#define TUI_PULLDOWN_TOP_MARGIN			0
#define TUI_PULLDOWN_BOTTOM_MARGIN		10
#define TUI_PULLDOWN_OFFSET				13
#define TUI_PULLDOWN_CHECK_WIDTH        12

#define TUI_PULLDOWN_HIGHLIGHTED_LEFT 5
#define TUI_PULLDOWN_HIGHLIGHTED_RIGHT 5
#define TUI_PULLDOWN_HIGHLIGHTED_TOP 15
#define TUI_PULLDOWN_HIGHLIGHTED_BOTTOM 5
#define TUI_PULLDOWN_HIGHLIGHTED_HEIGHT 20

static void tui_pulldown_draw(TUI_WIDGET *pulldown);
static void tui_pulldown_expose(TUI_WIDGET *pulldown, XExposeEvent *event);
static void tui_pulldown_motion(TUI_WIDGET *pulldown, XMotionEvent *event);

#endif /* TUI_XWIN */


TUI_WIDGET *
TUI_menubar_create_pulldown(TUI_WIDGET *menubar, char *text)
{
	TUI_WIDGET *widget;
#ifdef TUI_OSX
    unsigned char buf[255];
#endif
    
	assert(menubar->type == TUI_MENUBAR);
	assert(text != NULL);
	
	widget = tui_widget_allocate();
	if (widget == NULL) {
		return NULL;
	}

	widget->tui = menubar->tui;
	widget->type = TUI_PULLDOWN;

	widget->parent = menubar;
	widget->text = strdup(text);

#ifdef TUI_XWIN	
	widget->draw_fp = tui_pulldown_draw;
	widget->expose_fp = tui_pulldown_expose;
	widget->motion_fp = tui_pulldown_motion;
#endif

#ifdef TUI_WIN32
	assert(menubar->hmenu != NULL);
	widget->hmenu = CreatePopupMenu();
	if (!AppendMenu(menubar->hmenu, MF_POPUP | MF_STRING,
		(UINT)widget->hmenu, text)) {
		free(widget->text);
		free(widget);
		return NULL;
	}
	assert(menubar->parent->type == TUI_WINDOW);
	assert(menubar->parent->hwnd != NULL);
	DrawMenuBar(menubar->parent->hwnd);
#endif

#ifdef TUI_OSX
    sprintf(buf, " %s", text);
    buf[0] = strlen(text);
    widget->menu = NewMenu(hierMenu, buf);
    DrawMenuBar();
#endif
    
	tui_widget_append_to_list(&menubar->child, widget);

	TUI_WIDGET_set_enabled(widget, 1);
	TUI_WIDGET_set_visible(widget, 1); 

#ifdef TUI_XWIN
    /* repaint the menubar */
    menubar->draw_fp(menubar);
#endif

	return widget;
}


#ifdef TUI_XWIN

static int
item_height(TUI_WIDGET *item)
{
	if (item->text[0] == '\0') {
		return TUI_PULLDOWN_SEPARATOR_HEIGHT;
	}

	return TUI_PULLDOWN_ITEM_HEIGHT;
}


/* 911: Don't store width and height in pulldown since that is used
   to store the width of the pulldown title for menubar selection */

static void
compute_pulldown_dimensions(TUI_WIDGET *pulldown, int *w, int *h)
{
	int item_width;
	TUI_WIDGET *menuitem;

	*w = 50;
	*h = TUI_PULLDOWN_TOP_MARGIN + TUI_PULLDOWN_BOTTOM_MARGIN;

	for (menuitem=pulldown->child; menuitem != NULL; menuitem=menuitem->next) {
		*h += item_height(menuitem);


		item_width = XTextWidth(pulldown->tui->app->menu_font, menuitem->text,
			strlen(menuitem->text)) + TUI_PULLDOWN_MARGIN * 2;

        if (menuitem->type == TUI_CHECK_MENUITEM) {
            item_width += TUI_PULLDOWN_CHECK_WIDTH;
        }
        
        if (menuitem->key != 0) {
            item_width += TUI_PULLDOWN_KEY_MARGIN + TUI_PULLDOWN_KEY_SPACE;
        }
        
		if (item_width > *w) *w = item_width;
	}
}

		
void
tui_pulldown_open(TUI_WIDGET *pulldown)
{
	int x, y, w, h;
	Window root_window, child_window;
    XSetWindowAttributes attr;

	pulldown->selected = 1;
	
	compute_pulldown_dimensions(pulldown, &w, &h);

	root_window = DefaultRootWindow(pulldown->tui->app->display);
	
	XTranslateCoordinates(pulldown->tui->app->display,
        pulldown->parent->window, root_window, pulldown->x -
        TUI_MENUBAR_SPACING/2, TUI_PULLDOWN_TOP, &x, &y, &child_window);
	
	/* must use OverrideRedirect to avoid WM adjusting positioning */
    attr.override_redirect = True;
	pulldown->window = XCreateWindow(pulldown->tui->app->display, root_window,
        x, y, w, h, 0, CopyFromParent, InputOutput,
		CopyFromParent, CWOverrideRedirect, &attr);
	XSelectInput(pulldown->tui->app->display, pulldown->window,
		ExposureMask | PointerMotionMask | ButtonReleaseMask);
	XMapWindow(pulldown->tui->app->display, pulldown->window);
}


void
tui_pulldown_close(TUI_WIDGET *pulldown)
{
	pulldown->selected = 0;
	XDestroyWindow(pulldown->tui->app->display, pulldown->window);
	pulldown->window = 0;
}


int
tui_pulldown_all_disabled(TUI_WIDGET *pulldown)
{
    TUI_WIDGET *child;
    
    for (child = pulldown->child; child != NULL; child = child->next) {
        if (child->enabled) {
            return 0;
        }
    }
    
    return 1;
}


static void
draw_check(Display* display, Window window, GC color, int x, int y)
{
	XPoint points[7];
	
	x += 2;
	y += 1;
	
	points[0].x = x + 1;    points[0].y = y + 3;
	points[1].x = x + 4;    points[1].y = y + 6;
	points[2].x = x + 8;    points[2].y = y - 3;
	points[3].x = x + 9;    points[3].y = y - 4;
	points[4].x = x + 9;    points[4].y = y - 2;
	points[5].x = x + 4;    points[5].y = y + 7;
	points[6].x = x + 0;    points[6].y = y + 3;
	
	XDrawLines(display, window, color, points, 7, CoordModeOrigin);
}

static int
get_text_x(TUI_WIDGET *pulldown)
{
	TUI_WIDGET *menuitem;

	for (menuitem=pulldown->child; menuitem != NULL; menuitem=menuitem->next) {
        if (menuitem->type == TUI_CHECK_MENUITEM) {
            return TUI_PULLDOWN_CHECK_WIDTH + TUI_PULLDOWN_MARGIN;
        }
    }

    return TUI_PULLDOWN_MARGIN;
}


static void
tui_pulldown_draw(TUI_WIDGET *pulldown)
{
	int y, w, h, all_disabled;
    GC gc_text, gc_check;
    char key_text[64];
	TUI_WIDGET *child;
    TUI_GC gc_normal[] = {
        TUI_GC_DARK_GRAY,
        TUI_GC_WHITE,
        TUI_GC_VERY_LIGHT_GRAY,
        /* TUI_GC_VERY_LIGHT_GRAY, */
        TUI_GC_MENU_GRAY,
        /* TUI_GC_MEDIUM_GRAY, */
        TUI_GC_LIGHT_GRAY,
        TUI_GC_MEDIUM_GRAY,
        TUI_GC_VERY_DARK_GRAY,
        TUI_GC_NULL
    };

	if (pulldown->window == 0) {
		return;
	}

	compute_pulldown_dimensions(pulldown, &w, &h);

    all_disabled = tui_pulldown_all_disabled(pulldown);
    
	tui_draw_frame(pulldown->tui, pulldown->window, 0, 0, w, h, gc_normal);

	y = TUI_PULLDOWN_TOP_MARGIN;
	
	for (child = pulldown->child; child != NULL; child = child->next) {

		y += item_height(child);
		
		if (child->text[0] == '\0') {	/* separator */
			XDrawLine(pulldown->tui->app->display, pulldown->window,
				pulldown->tui->app->gc[TUI_GC_DARK_GRAY], 3, y+3, w - 5, y+3);
			XDrawLine(pulldown->tui->app->display, pulldown->window,
				pulldown->tui->app->gc[TUI_GC_WHITE], 3, y+4, w - 5, y+4);
			continue;
		}
			
		if (child->enabled && child->selected) {
			/* draw highlight box */
			XFillRectangle(pulldown->tui->app->display, pulldown->window,
				pulldown->tui->app->gc[TUI_GC_VERY_LIGHT_GRAY],
				TUI_PULLDOWN_HIGHLIGHTED_LEFT,
				y - TUI_PULLDOWN_HIGHLIGHTED_TOP, w -
				TUI_PULLDOWN_HIGHLIGHTED_RIGHT -
				TUI_PULLDOWN_HIGHLIGHTED_LEFT - 1,
				TUI_PULLDOWN_HIGHLIGHTED_HEIGHT);
			XDrawLine(pulldown->tui->app->display, pulldown->window,
				pulldown->tui->app->gc[TUI_GC_VERY_DARK_GRAY],
				TUI_PULLDOWN_HIGHLIGHTED_LEFT,
				y + TUI_PULLDOWN_HIGHLIGHTED_BOTTOM,
				w - TUI_PULLDOWN_HIGHLIGHTED_RIGHT - 1,
				y + TUI_PULLDOWN_HIGHLIGHTED_BOTTOM);
			XDrawLine(pulldown->tui->app->display, pulldown->window,
				pulldown->tui->app->gc[TUI_GC_VERY_DARK_GRAY],
				w - TUI_PULLDOWN_HIGHLIGHTED_RIGHT -
				1, y - TUI_PULLDOWN_HIGHLIGHTED_TOP,
				w - TUI_PULLDOWN_HIGHLIGHTED_RIGHT -
				1, y + TUI_PULLDOWN_HIGHLIGHTED_BOTTOM);
		}

		if (child->enabled && child->parent->enabled && !all_disabled) {
			gc_text = pulldown->tui->app->gc[TUI_GC_TEXT_MENUBAR];
		} else {
			gc_text = pulldown->tui->app->gc[TUI_GC_TEXT_MENUITEM_GRAY];
		}

        if (child->type == TUI_CHECK_MENUITEM) {
            if (child->value != 0) {
                if (pulldown->enabled && child->enabled) {
                    gc_check = pulldown->tui->app->gc[TUI_GC_BLACK];
                } else {
                    gc_check = pulldown->tui->app->gc[TUI_GC_MEDIUM_GRAY];
                }
                draw_check(pulldown->tui->app->display, pulldown->window,
                    gc_check, 6, y - 9);
            }
        }
        
		XDrawString(pulldown->tui->app->display, pulldown->window, gc_text,
			get_text_x(pulldown), y, child->text, strlen(child->text));

        if (child->key != 0) {
            sprintf(key_text, "Ctrl-%c", toupper(child->key));
            XDrawString(pulldown->tui->app->display, pulldown->window, gc_text,
                w - TUI_PULLDOWN_KEY_MARGIN, y, key_text, strlen(key_text));
        }
	}
}


static void
tui_pulldown_expose(TUI_WIDGET *pulldown, XExposeEvent *event)
{
	tui_pulldown_draw(pulldown);
}


static void
tui_pulldown_motion(TUI_WIDGET *pulldown, XMotionEvent *event)
{
	int w, h, x, y, iy;
	int changed;
	TUI_WIDGET *child;
	Window child_window;

	/* convert location into proper coordinates */
	XTranslateCoordinates(pulldown->tui->app->display,
        pulldown->parent->window, pulldown->window, event->x, event->y,
        &x, &y, &child_window);
	
	compute_pulldown_dimensions(pulldown, &w, &h);

	iy = TUI_PULLDOWN_TOP_MARGIN;
	changed = 0;
	
	/* run through all children each time to select or deselect */
	for (child = pulldown->child; child != NULL; child = child->next, iy+=h) {
		h = item_height(child);

		if (!child->parent->enabled || x < 0 || x > w) {
			if (child->selected) changed = 1;
			child->selected = 0;
			continue;
		}

		if (y > iy && y < iy+h) {
			if (!child->selected) changed = 1;
			child->selected = 1;
		} else {
			if (child->selected) changed = 1;
			child->selected = 0;
		}
	}

	if (changed) {
		tui_pulldown_draw(pulldown);
	}
}


#endif /* TUI_XWIN */
