/* Copyright (c) 2000, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_widget.c,v 1.18 2002/01/07 22:31:59 wex Exp $
 *
 * $Log: tui_widget.c,v $
 * Revision 1.18  2002/01/07 22:31:59  wex
 * Many fixes for both Windows and Linux, looped playback working
 *
 * Revision 1.17  2002/01/03 06:50:13  wex
 * Removed EXPORT, updated WIN32, handle some thread memory bugs.
 *
 * Revision 1.16  2001/12/19 01:01:44  wex
 * Support for Win32 including OpenGL and compiled as DLL.
 *
 * Revision 1.15  2001/12/16 00:32:58  wex
 * Updates for viewer, added checkmenuitem in Xlib.
 *
 * Revision 1.14  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.13  2000/11/04 18:00:08  wex
 * Initial support for Mac OSX.
 *
 * Revision 1.12  2000/10/06 03:48:46  wex
 * Slightly better events, widget windows capture other events.
 *
 * Revision 1.11  2000/10/05 10:29:41  wex
 * Added labels, titles, and textports.  Fixed parsing bug and radio look.
 *
 * Revision 1.10  2000/10/02 09:28:34  wex
 * Minor bug fixes.
 *
 * Revision 1.9  2000/10/02 05:56:00  wex
 * Added radio buttons, fixed minor bugs.
 *
 * Revision 1.8  2000/10/01 18:27:37  wex
 * Cleanups for Windows.
 *
 * Revision 1.7  2000/09/30 18:00:24  wex
 * Handle more visual types.  Menubars resize correctly.
 *
 * Revision 1.6  2000/09/30 15:59:14  wex
 * Saving and loading generated header files.
 *
 * Revision 1.5  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 * Revision 1.4  2000/09/26 16:31:38  wex
 * Events working under Windows.
 *
 * Revision 1.3  2000/09/25 02:35:02  wex
 * Creating windows and widgets in win32, events still broken.
 *
 * Revision 1.2  2000/09/24 03:13:08  wex
 * Initial changes to support WIN32
 *
 * Revision 1.1.1.1  2000/09/16 23:38:08  wex
 * Initial revision
 *
 */

#include <stdlib.h>
#include <string.h>

#include "tui_internal.h"


void
TUI_WIDGET_destroy(TUI_WIDGET *widget)
{
	TUI_WIDGET *prev;

    if (widget->parent != NULL) {
        /* update parent and sibling lists */
        prev = widget->parent->child;
        if (prev == widget) {
            /* head of parent list */
            widget->parent->child = widget->next;
        } else {
            while (prev->next != widget) {
                prev = prev->next;
                assert(prev != NULL);
            }

            prev->next = widget->next;
        }
    }

	/* destroy all child widgets */
	while (widget->child != NULL) {
		TUI_WIDGET_destroy(widget->child);
	}

#ifdef TUI_XWIN
    if (widget->window != 0) {
        XDestroyWindow(widget->tui->app->display, widget->window);
    }
#endif

#ifdef TUI_WIN32
    if (widget->type == TUI_WINDOW) {
        DestroyWindow(widget->hwnd);
    }
#endif    
        
	free(widget);
}


TUI_WIDGET *
tui_widget_allocate(void)
{
	TUI_WIDGET *widget;

	widget = (TUI_WIDGET *)malloc(sizeof(TUI_WIDGET));
	if (widget == NULL) {
		TUI_errno = TUI_ERR_MEMORY;
		return NULL;
	}

	widget->tui = NULL;
	widget->type = TUI_NONE;
	widget->parent = NULL;
	widget->child = NULL;
	widget->next = NULL;

#ifdef TUI_XWIN	
	widget->window = 0;
    widget->glx_context = 0;
	widget->draw_fp = NULL;
	widget->expose_fp = NULL;
	widget->cross_fp = NULL;
	widget->button_fp = NULL;
	widget->key_fp = NULL;
	widget->motion_fp = NULL;
	widget->resize_fp = NULL;
#endif
#ifdef TUI_WIN32
	widget->hwnd = NULL;
	widget->hmenu = NULL;
#endif
#ifdef TUI_OSX
    widget->window = NULL;
    widget->control = NULL;
    widget->menu = NULL;
#endif
    
	widget->value = 0;
	widget->x = -1;
	widget->y = -1;
	widget->w = -1;
	widget->h = -1;
	widget->text = NULL;
	widget->key = 0;
	
	widget->visible = 0;
	widget->enabled = 0;
	widget->located = 0;
	widget->selected = 0;
	
	widget->callback = NULL;
	widget->user_data = NULL;

    widget->intercept = NULL;
    widget->intercept_user_data = NULL;
    
    widget->name = NULL;
    widget->callback_name = NULL;

	return widget;
}


int
TUI_WIDGET_set_enabled(TUI_WIDGET *widget, int state)
{
	int old_value;

	old_value = widget->enabled;
	widget->enabled = state;

#ifdef TUI_XWIN	
	if (widget->draw_fp) {
		widget->draw_fp(widget);
	}
    if (widget->type == TUI_PULLDOWN) {
        assert(widget->parent->type == TUI_MENUBAR);
        assert(widget->parent->draw_fp != NULL);
        widget->parent->draw_fp(widget->parent);
    }
#endif

#ifdef TUI_WIN32
    if (widget->type == TUI_MENUITEM || widget->type == TUI_CHECK_MENUITEM) {
        UINT mf_state;

        if (state == 1) {
            mf_state = MF_ENABLED;
        } else {
            mf_state = MF_DISABLED;
        }

        EnableMenuItem(widget->parent->hmenu, widget->control_id, mf_state);
    }
#endif
    
	return old_value;
}


int
TUI_WIDGET_set_visible(TUI_WIDGET *widget, int state)
{
	int old_value;

	assert(widget != NULL);

	old_value = widget->visible;
	widget->visible = state;

#ifdef TUI_XWIN	
	if (widget->window != 0) {
		if (widget->visible) {
			XMapWindow(widget->tui->app->display, widget->window);
		} else {
			XUnmapWindow(widget->tui->app->display, widget->window);
		}
	}
#endif

#ifdef TUI_WIN32
	if (widget->hwnd != NULL) {
		ShowWindow(widget->hwnd, SW_SHOW);
		UpdateWindow(widget->hwnd);
	}
#endif

#ifdef TUI_OSX
    if (widget->window != NULL) {
        TransitionWindow(widget->window, kWindowZoomTransitionEffect,
            kWindowShowTransitionAction, NULL);
        BringToFront(widget->window);
        SelectWindow(widget->window);
    }
#endif
    
	return old_value;
}

void
TUI_WIDGET_set_callback(TUI_WIDGET *widget,
	void (*callback)(TUI_WIDGET_INFO *info), void *user_data)
{
	widget->callback = callback;
	widget->user_data = user_data;
}


void *
TUI_WIDGET_set_user_data(TUI_WIDGET *widget, void *user_data)
{
    void *old_data;

    old_data = widget->user_data;
    
    widget->user_data = user_data;

    return old_data;
}

    
char *
tui_widget_type_to_string(TUI_WIDGET_TYPE widget_type)
{
    switch (widget_type) {
    case TUI_NONE: return "none";
    case TUI_WINDOW: return "window";
    case TUI_BUTTON: return "button";
    case TUI_RADIO_BUTTON: return "radio_button";
    case TUI_TEXTPORT: return "textport";
    case TUI_MENUBAR: return "menubar";
    case TUI_PULLDOWN: return "pulldown";
    case TUI_MENUITEM: return "menuitem";
    case TUI_CHECK_MENUITEM: return "checkmenuitem";
    case TUI_OGLWINDOW: return "oglwindow";
    default:
        return NULL;
    }
}

void
TUI_WIDGET_set_name(TUI_WIDGET *widget, char *name)
{
    if (widget->name != NULL) {
        free(widget->name);
    }

    if (name == NULL) {
        widget->name = NULL;
    } else {
        widget->name = strdup(name);
    }
}

void
TUI_WIDGET_set_text(TUI_WIDGET *widget, char *text)
{
    if (widget->text != NULL) {
        free(widget->text);
    }

    if (text == NULL) {
        widget->text = NULL;
    } else {
        widget->text = strdup(text);
    }
}

void
TUI_WIDGET_set_callback_name(TUI_WIDGET *widget, char *name)
{
    if (widget->callback_name != NULL) {
        free(widget->callback_name);
    }

    if (name == NULL) {
        widget->callback_name = NULL;
    } else {
        widget->callback_name = strdup(name);
    }
}


void
TUI_WIDGET_set_value(TUI_WIDGET *widget, int value)
{
    widget->value = value;

    if (widget->type == TUI_RADIO_BUTTON) {
        if (widget->value) {
            widget->value = widget->key;
        }
        tui_window_disable_matching_radios(widget->parent, widget);
#ifdef TUI_WIN32
        CheckDlgButton(widget->parent->hwnd, widget->control_id, 1);
#endif                
    }

#ifdef TUI_WIN32
    if (widget->type == TUI_CHECK_MENUITEM) {
        UINT state;

        if (value == 1) {
            state = MF_CHECKED;
        } else {
            state = MF_UNCHECKED;
        }
        
        CheckMenuItem(widget->parent->hmenu, widget->control_id, state);
    }
#endif    

#ifdef TUI_XWIN
    if (widget->draw_fp) {
        widget->draw_fp(widget);
    }
#endif

#if 0    
    if (widget->callback != NULL) {
        info.widget = widget;
        info.user_data = widget->user_data;
        widget->callback(&info);
    }
#endif    
}


int
TUI_WIDGET_get_value(TUI_WIDGET *widget)
{
    return widget->value;
}


void
tui_widget_append_to_list(TUI_WIDGET **head, TUI_WIDGET *widget)
{
    TUI_WIDGET *prev;
    
	if (*head == NULL) {
		*head = widget;
	} else {
		for (prev = *head; prev->next != NULL; prev = prev->next) {
			/*EMPTY*/
		}
		prev->next = widget;
	}
}


void
tui_widget_deselect_all_textports(TUI_WIDGET *widget)
{
    if (widget->type == TUI_TEXTPORT) {
        if (widget->selected) {
            widget->selected = 0;
#ifdef TUI_XWIN
            widget->draw_fp(widget);
#endif
        }
    }

    if (widget->child != NULL) {
        tui_widget_deselect_all_textports(widget->child);
    }

    if (widget->next != NULL) {
        tui_widget_deselect_all_textports(widget->next);
    }
}

    
void
TUI_WIDGET_set_location(TUI_WIDGET *widget, int x, int y)
{
#ifdef TUI_XWIN
    XMoveWindow(widget->tui->app->display,widget->window, x, y);
    if (widget->draw_fp != NULL) {
        widget->draw_fp(widget);
    }
#endif

#ifdef TUI_WIN32
    SetWindowPos(widget->hwnd, HWND_TOP, x, y, widget->w, widget->h, 0);
#endif
    
    widget->x = x;
    widget->y = y;
}

void
tui_widget_resize(TUI_WIDGET *widget, int w, int h)
{
    if (widget == NULL) {
        /* no widgets in window */
        return;
    }

#ifdef TUI_XWIN    
    if (widget->resize_fp != NULL) {
        widget->resize_fp(widget, w, h);
    }
#endif

#ifdef TUI_WIN32
    if (widget->type == TUI_WINDOW || widget->type ==  TUI_OGLWINODW) {
        fprintf(stderr, "Setting window resize to %d x %d\n", width, height);
        TUI_WIDGET_set_size(widget, w, h);
    }
#endif


    if (widget->child != NULL) {
        tui_widget_resize(widget->child, w, h);
    }
    
    if (widget->next != NULL) {
        tui_widget_resize(widget->next, w, h);
    }
}

void
TUI_WIDGET_set_size(TUI_WIDGET *widget, int w, int h)
{
#ifdef TUI_XWIN
    XResizeWindow(widget->tui->app->display, widget->window, w, h);
    if (widget->draw_fp != NULL) {
        widget->draw_fp(widget);
    }
#endif

#ifdef TUI_WIN32
    SetWindowPos(widget->hwnd, HWND_TOP, widget->x, widget->y, w, h, 0);
#endif
    
    widget->w = w;
    widget->h = h;
}

void
TUI_WIDGET_translate_coordinates(TUI_WIDGET *src_widget,
    TUI_WIDGET *dst_widget, int src_x, int src_y, int *dst_x, int *dst_y)
{
#ifdef TUI_XWIN
    Window child_return;
    
    XTranslateCoordinates(src_widget->tui->app->display,
        src_widget->window, dst_widget->window, src_x, src_y,
        dst_x, dst_y, &child_return);
#endif

#ifdef TUI_WIN32
	assert(src_widget->parent->type == TUI_WINDOW);
	assert(dst_widget->type == TUI_WINDOW);

	*dst_x = src_x + src_widget->x;
	*dst_y = src_y + src_widget->y;
#endif
}

int
TUI_WIDGET_get_width(TUI_WIDGET *widget)
{
    return widget->w;
}

int
TUI_WIDGET_get_height(TUI_WIDGET *widget)
{
    return widget->h;
}

TUI *
TUI_WIDGET_get_tui(TUI_WIDGET *widget)
{
    return widget->tui;
}

