/* Copyright (c) 2000, Flarg Inc, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/tui/tui_write.c,v 1.7 2001/12/05 10:47:01 wex Exp $
 *
 * $Log: tui_write.c,v $
 * Revision 1.7  2001/12/05 10:47:01  wex
 * Changes to support image viewer in Python, compilation without warnings
 *
 * Revision 1.6  2000/10/06 03:48:46  wex
 * Slightly better events, widget windows capture other events.
 *
 * Revision 1.5  2000/10/02 07:16:48  wex
 * Keyboard equivalents.
 *
 * Revision 1.4  2000/10/02 05:56:00  wex
 * Added radio buttons, fixed minor bugs.
 *
 * Revision 1.3  2000/10/01 18:27:37  wex
 * Cleanups for Windows.
 *
 * Revision 1.2  2000/09/30 15:59:14  wex
 * Saving and loading generated header files.
 *
 * Revision 1.1  2000/09/28 16:06:27  wex
 * Added TUI_APPLICATION, beginning layout and code output.
 *
 *
 */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "tui_internal.h"

#define INDENT_COUNT 2



static char *
get_lower_root_name(char *filename)
{
    int i, len;
    char *root_name;
    static char buf[1024];

    len = strlen(filename);
    assert(len < 1024);
    
    strncpy(buf, filename, len);

    root_name = strrchr(buf, '/');
    if (root_name == NULL) {
        root_name = buf;
    } else {
        root_name++;
    }

    len = strlen(root_name);
    
    root_name[len-2] = '\0';

    for (i = 0; i < len-2; i++) {
        if (isupper(root_name[i])) {
            root_name[i] = tolower(root_name[i]);
        }
    }
    
    return root_name;
}

static char *
get_upper_root_name(char *filename)
{
    int i, len;
    char *root_name;
    static char buf[1024];

    len = strlen(filename);
    assert(len < 1024);
    
    strncpy(buf, filename, len);

    root_name = strrchr(buf, '/');
    if (root_name == NULL) {
        root_name = buf;
    } else {
        root_name++;
    }

    len = strlen(root_name);
    
    root_name[len-2] = '\0';

    for (i = 0; i < len-2; i++) {
        if (islower(root_name[i])) {
            root_name[i] = toupper(root_name[i]);
        }
    }
    
    return root_name;
}

static char *
get_pad(int spaces)
{
    static char buf[1024];
    static int initialized = 0;
    int len;

    if (!initialized) {
        memset(buf, ' ', 1024);
        buf[0] = '"';
        buf[1] = '\0';
        initialized = 1;
    }

    assert(spaces > 1);
    assert(spaces < 1024);

    len = strlen(buf);
    assert(len >= 0 && len < 1024);
    
    buf[len] = ' ';
    buf[spaces+1] = '\0';

    return buf;
}


static void
write_header(FILE *stream, char *upper_root, char *lower_root)
{
    fprintf(stream, "/* Do not edit by hand.  This file was machine "
        "generated. */\n\n");

    fprintf(stream, "#ifndef _%s_H_\n", upper_root);
    fprintf(stream, "#define _%s_H_\n", upper_root);
    fprintf(stream, "\n");

    fprintf(stream, "/* Use the structure and string defined in this file "
        "like this:\n\n");
    fprintf(stream, "       TUI_APPLICATION *app;\n");
    fprintf(stream, "       %s %s;\n\n", upper_root, lower_root);
    fprintf(stream, "       app = TUI_APPLICATION_create();\n");
    fprintf(stream, "       TUI_construct(app, &%s, %s_layout);\n",
        lower_root, lower_root);
    fprintf(stream, "       TUI_WIDGET_set_callback_data(%s.foo, ptr);\n",
        lower_root);
    fprintf(stream, "*/\n");
}


static void
write_footer(FILE *stream, char *upper_root)
{
    fprintf(stream, "\n");
    fprintf(stream, "#endif /* _%s_H_ */\n", upper_root);
}


static void
write_widget_layout(TUI_WIDGET *widget, FILE *stream, int level)
{
    char *pad = NULL;

    assert(widget != NULL);

    if (widget->type == TUI_MENUITEM && widget->text[0] == '\0') {
        fprintf(stream, "%smenuseparator {}\"\n",get_pad(level*INDENT_COUNT));
        goto next_widget;
    }
    
    fprintf(stream, "%s%s {\"\n", get_pad(level*INDENT_COUNT),
        tui_widget_type_to_string(widget->type));

    pad = strdup(get_pad((level+1)*INDENT_COUNT));

    if (widget->text != NULL &&
        widget->type != TUI_MENUBAR) {
        fprintf(stream, "%stext = '%s';\"\n", pad, widget->text);
    }

    if (widget->type == TUI_WINDOW ||
        widget->type == TUI_BUTTON ||
        widget->type == TUI_RADIO_BUTTON ||
        widget->type == TUI_TEXTPORT ||
        widget->type == TUI_OGLWINDOW) {

        fprintf(stream, "%sx = %d;\"\n", pad, widget->x);
        fprintf(stream, "%sy = %d;\"\n", pad, widget->y);
    }

    if (widget->type == TUI_WINDOW || widget->type == TUI_OGLWINDOW) {
        fprintf(stream, "%sw = %d;\"\n", pad, widget->w);
        fprintf(stream, "%sh = %d;\"\n", pad, widget->h);
    }

    if (widget->type == TUI_RADIO_BUTTON) {
        fprintf(stream, "%sradio_value = %d;\"\n", pad, (int)widget->key);
    } else if (widget->key != 0) {
        fprintf(stream, "%skey = '%c';\"\n", pad, widget->key);
    }

    if (widget->visible == 0) {
        fprintf(stream, "%sinvisible;\"\n", pad);
    }

    if (widget->enabled == 0) {
        fprintf(stream, "%sdisabled;\"\n", pad);
    }

    if (widget->name != NULL) {
        fprintf(stream, "%sname = '%s';\"\n", pad, widget->name);
    }

    if (widget->callback_name != NULL) {
        fprintf(stream, "%scallback = '%s';\"\n", pad, widget->callback_name);
    }
    
    if (widget->child) {
        write_widget_layout(widget->child, stream, level+1);
    }

    fprintf(stream, "%s}\"\n", get_pad(level*INDENT_COUNT));

 next_widget:
    
    if (widget->next) {
        write_widget_layout(widget->next, stream, level);
    }

    if (pad != NULL) free(pad);
}


static void
write_layout(TUI *tui, FILE *stream, char *lower_root)
{
    char *pad;
    
    assert(tui->root->type == TUI_NONE);
    assert(tui->root->next == NULL);
    assert(tui->root->child != NULL);

    fprintf(stream, "\n");
    fprintf(stream, "/* This string should be passed to TUI_construct() by "
        "your application to\n");
    fprintf(stream, "   build your user interface. */\n\n");
    
    fprintf(stream, "static const char *%s_layout = \n", lower_root);
    fprintf(stream, "%s%s {\"\n", get_pad(INDENT_COUNT), lower_root);

    pad = get_pad(INDENT_COUNT*2);
    
    fprintf(stream, "%smajor_version = %d;\"\n", pad, tui_major_version);
    fprintf(stream, "%sminor_version = %d;\"\n", pad, tui_minor_version);
    
    write_widget_layout(tui->root->child, stream, 2);

    fprintf(stream, "%s}\"\n", get_pad(INDENT_COUNT));
    fprintf(stream, ";\n\n");
}


#if 0
static void
write_widget_code(TUI_WIDGET *widget, FILE *stream, char *lower_root)
{
    char *pad, namebuf[512];

    pad = get_pad(INDENT_COUNT);
    
    if (widget->name != NULL) {
        sprintf(namebuf, "%s->%s", lower_root, widget->name);
    } else if (widget->key != 0 || widget->enabled == 0 ||
        widget->visible == 0) {
        sprintf(namebuf, "widget");
    } else {
        namebuf[0] = '\0';
    }
            
    switch (widget->type) {

    case TUI_MENUBAR:
        fprintf(stream, "%smenubar = TUI_window_create_menubar(window);\n",
            pad);
        break;

    case TUI_PULLDOWN:
        fprintf(stream, "\n%spulldown = TUI_menubar_create_pulldown("
            "menubar, \"%s\");\n", pad, widget->text);
        break;
                
    case TUI_MENUITEM:
        assert(widget->text != NULL);
        if (widget->text[0] == '\0') {
            fprintf(stream,"%sTUI_pulldown_create_separator(pulldown);\n",pad);
            goto next_widget;
        } else if (namebuf[0] != '\0') {
            fprintf(stream, "%s%s = TUI_pulldown_create_menuitem("
                "pulldown, \"%s\");\n", pad, namebuf, widget->text);
        } else {
            fprintf(stream, "%sTUI_pulldown_create_menuitem("
                "pulldown, \"%s\");\n", pad, widget->text);
        }
            
        break;
                
    case TUI_BUTTON:
        if (namebuf[0] != '\0') {
            fprintf(stream, "%s%s = TUI_window_create_button(window, "
                "\"%s\", %d, %d, %d, %d);\n", pad, namebuf,
                widget->text, widget->x, widget->y, widget->w,widget->h);
        } else {
            fprintf(stream, "%sTUI_window_create_button(window, "
                "\"%s\", %d, %d, %d, %d);\n", pad, widget->text, widget->x,
                widget->y, widget->w,widget->h);
        }
            
        break;
    }

    if (widget->key != 0) {
        assert(namebuf[0] != '\0');
        fprintf(stream, "%sTUI_WIDGET_set_key(%s, \"%c\");\n",
            pad, namebuf, widget->key);
    }

    if (widget->enabled == 0) {
        assert(namebuf[0] != '\0');
        fprintf(stream, "%sTUI_WIDGET_set_enabled(%s, 0);\n", pad, namebuf);
    }

    if (widget->visible == 0) {
        assert(namebuf[0] != '\0');
        fprintf(stream, "%sTUI_WIDGET_set_visible(%s, 0);\n", pad, namebuf);
    }

    if (widget->child != NULL) {
        write_widget_code(widget->child, stream, lower_root);
    }

 next_widget:
    
    if (widget->next != NULL) {
        write_widget_code(widget->next, stream, lower_root);
    }
}


static int
menubar_exists(TUI_WIDGET *widget)
{
    int found;
    
    if (widget->type == TUI_MENUBAR) {
        return 1;
    }

    if (widget->child != NULL) {
        found = menubar_exists(widget->child);
        if (found) {
            return 1;
        }
    }

    if (widget->next != NULL) {
        return menubar_exists(widget->next);
    }

    return 0;
}


static void
write_code(TUI *tui, FILE *stream, char *)
{
    TUI_WIDGET *window;
    char *lower_root, *upper_root, *pad;

    fprintf(stream, "\n#include \"%s.h\"\n\n", lower_root);

    upper_root = get_upper_root_name(filename);
    pad = get_pad(INDENT_COUNT);
    
    fprintf(stream, "%s *\n", upper_root);
    fprintf(stream, "%s_create(TUI_APPLICATION *app)\n", upper_root);
    fprintf(stream, "{\n%s%s *%s;\n", pad, upper_root, lower_root);
    fprintf(stream, "%sTUI_WIDGET *window;\n", pad);
    
    if (menubar_exists(tui->root)) {
        fprintf(stream, "%sTUI_WIDGET *menubar, *pulldown;\n", pad);
    }

    fprintf(stream, "\n");
    
    fprintf(stream, "%s%s->tui = TUI_create(app);\n", pad, lower_root);

    for (window = tui->root->child; window != NULL; window = window->next) {
        fprintf(stream, "%swindow = TUI_create_window(%s->tui, %d, %d, "
            "%d, %d);\n\n", pad, lower_root, window->x, window->y, window->w,
            window->h);
        write_widget_code(window->child, stream, lower_root);
    }
    
    fprintf(stream, "\n%sreturn %s;\n}\n", pad, lower_root);
}
#endif


static void
write_widget_structure(TUI_WIDGET *widget, FILE *stream)
{
    if (widget->name != NULL) {
        fprintf(stream, "    TUI_WIDGET *%s;\n", widget->name);
    }

    if (widget->child != NULL) {
        write_widget_structure(widget->child, stream);
    }
    
    if (widget->next != NULL) {
        write_widget_structure(widget->next, stream);
    }
}


static void
write_structure(TUI *tui, FILE *stream, char *upper_root, char *lower_root)
{
    TUI_WIDGET *window;
    char *pad;
    
    pad = strdup(get_pad(INDENT_COUNT));

    fprintf(stream, "\n");
    fprintf(stream, "/* This structure can be used in your application to set "
        "specific values and\n");
    fprintf(stream, "   set callback data for your widgets.  Only widgets "
        "that have been given\n");
    fprintf(stream, "   a name are stored in this structure. */\n\n");
    
    fprintf(stream, "typedef struct %s {\n", lower_root);
    fprintf(stream, "    TUI *tui;\n");

    for (window = tui->root->child; window != NULL; window = window->next) {
        write_widget_structure(window, stream);
    }
    
    fprintf(stream, "} %s;\n\n", upper_root);

    free(pad);
}


int
TUI_write(TUI *tui, char *filename)
{
    int len;
    char *upper_root, *lower_root;
    FILE *stream;

    assert(tui != NULL);
    assert(filename != NULL);

    /* make sure the file is a C file */
    len = strlen(filename);
    if (len < 3) {
        TUI_errno = TUI_ERR_FILENAME;
        return -1;
    }
    
    if (filename[len-1] != 'h' || filename[len-2] != '.') {
        TUI_errno = TUI_ERR_FILENAME;
        return -1;
    }
        
    stream = fopen(filename, "w");
    if (stream == NULL) {
        TUI_errno = TUI_ERR_OPEN;
        return -1;
    }

    upper_root = get_upper_root_name(filename);
    lower_root = get_lower_root_name(filename);
    
    write_header(stream, upper_root, lower_root);
    write_structure(tui, stream, upper_root, lower_root);
    write_layout(tui, stream, lower_root);
    write_footer(stream, upper_root);
    
    fclose(stream);

    return 0;
}

