/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb.c,v 1.10 2001/12/21 07:20:12 wex Exp $
 *
 * $Log: zb.c,v $
 * Revision 1.10  2001/12/21 07:20:12  wex
 * Removed memory leaks
 *
 * Revision 1.9  2001/12/20 22:00:45  wex
 * Updates for Win32, fixed some matrix memory leaks
 *
 * Revision 1.8  2001/12/16 00:28:34  wex
 * Fixes to frustum culling
 *
 * Revision 1.7  2001/11/13 18:32:34  wex
 * Memory management and reference counting fixes.
 *
 * Revision 1.6  2001/11/13 11:46:27  wex
 * Removed ZB_GRID_DATA, ZB_GEOMETRY_DATA, and tile grid and geometry lists
 *
 * Revision 1.5  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.4  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.3  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:29  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "zb_internal.h"
#include "zb_camera.h"
#include "zb_geometry.h"
#include "zb_grid.h"
#include "zb_pixel.h"

struct zb_light {
    LIST_ELEMENT list_element;
    SHADER_LIGHT *light;
};

static void zb_light_destroy(ZB_LIGHT *light);


ZB *
ZB_create(void)
{
    ZB *zb;
    
    zb = (ZB *)malloc(sizeof(ZB));
    if (zb == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    zb_initialize(zb);
    
    return zb;
}

void
zb_initialize(ZB *zb)
{
    zb->rgba = IMAGE_create();
    IMAGE_add_notify_client(zb->rgba, (int)zb);
    zb->geometry_list = LIST_create();
    zb->tile_list = LIST_create();
    zb->grid_list = LIST_create();
    zb->light_list = LIST_create();
    zb->camera = zb_camera_create();
    zb->filter = zb_filter_create(ZB_FILTER_GAUSSIAN, 2, 1);
    zb->dof = NULL;
    
    IMAGE_set_color_defined(zb->rgba, 1);
    IMAGE_set_alpha_defined(zb->rgba, 1);
    
    zb->nonempty_tiles = 0;
    zb->captured_grids = 0;
    zb->scanned_grids = 0;
}

void
ZB_destroy(ZB *zb)
{
    zb_delete(zb);
    free(zb);
}

void
ZB_print_statistics(ZB *zb)
{
    extern int zb_shaded_points, zb_grid_points;
    extern int zb_grids_created, zb_grids_destroyed, zb_max_grids;
    extern float zb_grid_mb;
    extern int zb_geometry_alloced, zb_geometry_destroyed, zb_geometry_max;
    extern int geometry_alloced, geometry_destroyed, geometry_max;
    extern int vertex_alloced, vertex_destroyed, vertex_max;
    extern int vector_alloced, vector_destroyed, vector_max;
    extern int matrix_alloced, matrix_destroyed, matrix_max;
    extern int geometry_outside_vp_count;
    
    fprintf(stderr, "%d grid points (%d%% shaded)\n", zb_grid_points,
        (int)(100.0 * zb_shaded_points / zb_grid_points));
    fprintf(stderr, "geometry: %d max (%3.2f MB) %d created, %d destroyed\n",
        geometry_max, geometry_max * 512 / (1024.0*1024),
        geometry_alloced, geometry_destroyed);
    fprintf(stderr, "zb_geometry: %d max (%3.2f MB) %d created, %d destroyed, "
        "%d remaining, %d outside vp\n", zb_geometry_max,
        zb_geometry_max*512/(1024.0*1024),
        zb_geometry_alloced, zb_geometry_destroyed,
        zb->geometry_list != NULL ? LIST_count(zb->geometry_list): 0,
        geometry_outside_vp_count);
    fprintf(stderr, "grids: %d max (%3.2f MB) %d created, %d destroyed, "
        " %d remaining\n", zb_max_grids, zb_grid_mb, zb_grids_created,
        zb_grids_destroyed, zb->grid_list != NULL ? LIST_count(zb->grid_list) : 0);
    fprintf(stderr, "vertex: %d max (%3.2f MB) %d created %d destroyed\n",
        vertex_max, vertex_max * sizeof(VERTEX) / (1024.0*1024),
        vertex_alloced, vertex_destroyed);
    fprintf(stderr, "vector: %d max (%3.2f MB) %d created %d destroyed\n",
        vector_max, vector_max * sizeof(VECTOR) / (1024.0*1024),
        vector_alloced, vector_destroyed);
    fprintf(stderr, "matrix: %d max (%3.2f MB) %d created %d destroyed\n",
        matrix_max, matrix_max * (16*sizeof(double)+32) / (1024.0*1024),
        matrix_alloced, matrix_destroyed);
    fprintf(stderr, "%d captured grids\n", zb->captured_grids);
    fprintf(stderr, "%d non-empty tiles\n", zb->nonempty_tiles);
    fprintf(stderr, "%d scanned grids\n", zb->scanned_grids);
}

void
zb_delete(ZB *zb)
{
    ZB_GEOMETRY *geometry;
    ZB_GRID *grid;
    ZB_LIGHT *light;
    
    /* delete any geometry and grids left after the last row */
    /*   (or we might clip the bbox to the top of the window? and remove
         objects outside the original window) */
    assert(zb->geometry_list != NULL);
    while ((geometry = LIST_pop(zb->geometry_list)) != NULL) {
        zb_geometry_destroy(geometry);
    }
    LIST_destroy(zb->geometry_list);
    zb->geometry_list = NULL;

    assert(zb->grid_list != NULL);
    while ((grid = LIST_pop(zb->grid_list)) != NULL) {
        zb_grid_destroy(grid);
    }
    LIST_destroy(zb->grid_list);
    zb->grid_list = NULL;

    while ((light = LIST_pop(zb->light_list)) != NULL) {
        zb_light_destroy(light);
    }
    
    if (zb->rgba != NULL) {
        IMAGE_destroy(zb->rgba);
        zb->rgba = NULL;
    }

    zb_camera_destroy(zb->camera);
    zb->camera = NULL;
}

ZB_LOCATION
zb_compare_viewports(int a0, int a1, int a2, int a3, int b0, int b1, int b2,
    int b3)
{
    if (a0 > b2) return ZB_BBOX_RIGHT;
    if (a1 > b3) return ZB_BBOX_ABOVE;
    if (a2 < b0) return ZB_BBOX_LEFT;
    if (a3 < b1) return ZB_BBOX_BELOW;

    return ZB_BBOX_INTERSECT;
}

int
zb_viewport_contained(int a0, int a1, int a2, int a3, int b0, int b1,
    int b2, int b3)
{
    return a0 > b0 && a1 > b1 && a2 < b2 && a3 < b3;
}

ZB_TILE *
zb_get_tile(ZB *zb)
{
    return LIST_head(zb->tile_list);
}

int
ZB_set_filename(ZB *zb, char *filename)
{
    if (filename == NULL || filename[0] == '\0') {
        ZB_errno = ZB_ERR_FILENAME;
        return -1;
    }
    
    zb->filename = strdup(filename);

    return 0;
}

IMAGE *
ZB_get_image(ZB *zb)
{
    return zb->rgba;
}

int
ZB_add_light(ZB *zb, PyObject *light)
{
    SHADER_LIGHT *shader_light;
    ZB_LIGHT *zb_light;
    
    shader_light = SHADER_LIGHT_create(light);
    if (shader_light == NULL) {
        return -1;
    }
    
    zb_light = (ZB_LIGHT *)malloc(sizeof(ZB_LIGHT));
    assert(zb_light != NULL);
    
    zb_light->light = shader_light;

    LIST_append(zb->light_list, zb_light);

    return 0;
}

void
zb_light_destroy(ZB_LIGHT *light)
{
    SHADER_LIGHT_destroy(light->light);
    free(light);
}
