/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_camera.c,v 1.11 2001/12/20 22:00:46 wex Exp $
 *
 * $Log: zb_camera.c,v $
 * Revision 1.11  2001/12/20 22:00:46  wex
 * Updates for Win32, fixed some matrix memory leaks
 *
 * Revision 1.10  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.9  2001/11/16 09:25:19  wex
 * Added get_viewport
 *
 * Revision 1.8  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.7  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.6  2001/11/06 06:52:37  wex
 * Motion blur, intersection cleanup
 *
 * Revision 1.5  2001/11/02 03:17:34  wex
 * Support for Python material and light shaders, updated VECTOR type
 *
 * Revision 1.4  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.3  2001/10/29 00:17:08  wex
 * Reduce default viewport size.
 *
 * Revision 1.2  2001/10/28 20:03:11  wex
 * Added
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <assert.h>

#include "zb_camera.h"
#include "zb_pixel.h"

#define ZB_CAMERA_DEFAULT_WIDTH     1024
#define ZB_CAMERA_DEFAULT_HEIGHT    1024


struct zb_camera {
    MATRIX *w2c;
    MATRIX *w2c_motion;
    MATRIX *c2n;
    MATRIX *c2s;
    MATRIX *w2s;
    MATRIX *w2s_motion;
    double hither, yon;
    int supersample_rate;
    int viewport[4];         /* xmin, ymin, xmax, ymax */
    double window[4];        /* xmin, ymin, xmax, ymax */
    double range[2];         /* hither, yon */
};

typedef struct zb_camera_format {
    char *name;
    double window[4];
    int viewport[4];
} ZB_CAMERA_FORMAT;

#define ZB_CAMERA_FORMAT_COUNT 5

ZB_CAMERA_FORMAT zb_camera_format[ZB_CAMERA_FORMAT_COUNT] = {
    {"d1",      {-1, -0.75,      1, 0.75},      {0, 0, 719,  485}},
    {"ntsc",    {-1, -0.75,      1, 0.75},      {0, 0, 511,  485}},
    {"1.85",    {-1, -0.541575,  1, 0.541575},  {0, 0, 2047, 1106}},
    {"academy", {-1, -0.728665,  1, 0.728665},  {0, 0, 2047, 1240}}
};

static void update_camera_matrices(ZB_CAMERA *camera);


ZB_CAMERA *
zb_camera_create(void)
{
    ZB_CAMERA *camera;

    camera = (ZB_CAMERA *)malloc(sizeof(ZB_CAMERA));
    if (camera == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    camera->w2c = MATRIX_create();
    camera->c2n = MATRIX_create();
    camera->c2s = MATRIX_create();
    camera->w2s = MATRIX_create();
    camera->w2c_motion = MATRIX_create();
    camera->w2s_motion = MATRIX_create();
    camera->hither = 0.01;
    camera->yon = 100000;

    camera->supersample_rate = 1;
    camera->viewport[0] = 0;
    camera->viewport[1] = 0;
    camera->viewport[2] = ZB_CAMERA_DEFAULT_WIDTH - 1;
    camera->viewport[3] = ZB_CAMERA_DEFAULT_HEIGHT - 1;

    camera->window[0] = -1;
    camera->window[1] = -1;
    camera->window[2] = 1;
    camera->window[3] = 1;

    camera->range[0] = 0;
    camera->range[1] = 10000;

    return camera;
}

void
zb_camera_destroy(ZB_CAMERA *camera)
{
    MATRIX_destroy(camera->w2c);
    MATRIX_destroy(camera->w2c_motion);
    MATRIX_destroy(camera->c2n);
    MATRIX_destroy(camera->c2s);
    MATRIX_destroy(camera->w2s);
    MATRIX_destroy(camera->w2s_motion);
    free(camera);
}

int
zb_camera_get_width(ZB_CAMERA *camera)
{
    return camera->viewport[2] - camera->viewport[0] + 1;
}

int
zb_camera_get_height(ZB_CAMERA *camera)
{
    return camera->viewport[3] - camera->viewport[1] + 1;
}

void
zb_camera_get_viewport(ZB_CAMERA *camera, int viewport[4])
{
    viewport[0] = camera->viewport[0];
    viewport[1] = camera->viewport[1];
    viewport[2] = camera->viewport[2];
    viewport[3] = camera->viewport[3];
}

void
ZB_set_camera_transforms(ZB *zb, MATRIX *w2c, MATRIX *w2c_motion,
    MATRIX *c2n)
{
    if (w2c != NULL) {
        MATRIX_copy(w2c, zb->camera->w2c);
    }
    
    if (w2c_motion != NULL) {
        MATRIX_copy(w2c_motion, zb->camera->w2c_motion);
    }
    
    if (c2n != NULL) {
        MATRIX_copy(c2n, zb->camera->c2n);
    }

    update_camera_matrices(zb->camera);
}

void
ZB_set_camera_w2c(ZB *zb, MATRIX *w2c, MATRIX *w2c_motion)
{
    ZB_set_camera_transforms(zb, w2c, w2c_motion, NULL);
}

void
ZB_set_camera_viewport(ZB *zb, int xmin, int ymin, int xmax, int ymax)
{
    zb->camera->viewport[0] = xmin * zb->camera->supersample_rate;
    zb->camera->viewport[1] = ymin * zb->camera->supersample_rate;
    zb->camera->viewport[2] = xmax * zb->camera->supersample_rate;
    zb->camera->viewport[3] = ymax * zb->camera->supersample_rate;

    update_camera_matrices(zb->camera);

    /* destroy and recreate output image */
    IMAGE_set_viewport(zb->rgba, xmin, ymin, xmax, ymax);
}

void
ZB_set_camera_window(ZB *zb, double x0, double y0, double x1, double y1)
{
    zb->camera->window[0] = x0;
    zb->camera->window[1] = y0;
    zb->camera->window[2] = x1;
    zb->camera->window[3] = y1;

    update_camera_matrices(zb->camera);
}

static void
update_camera_matrices(ZB_CAMERA *camera)
{
    int w, h;
    
    MATRIX_multiply(camera->w2c, camera->c2n, camera->w2s);
    MATRIX_multiply(camera->w2c_motion, camera->c2n, camera->w2s_motion);
    MATRIX_copy(camera->c2n, camera->c2s);
    
    /* translate the window to the origin */
    MATRIX_translate(camera->c2s, 0.5, 0.5, 0);
    MATRIX_translate(camera->w2s, 0.5, 0.5, 0);
    MATRIX_translate(camera->w2s_motion, 0.5, 0.5, 0);

    /* scale by the viewport */
    w = camera->viewport[2] - camera->viewport[0] + 1;
    h = camera->viewport[3] - camera->viewport[1] + 1;
    
    MATRIX_scale(camera->c2s, w, h, 1);
    MATRIX_scale(camera->w2s, w, h, 1);
    MATRIX_scale(camera->w2s_motion, w, h, 1);
}

    
MATRIX *
ZB_get_camera_w2s(ZB *zb)
{
    return zb->camera->w2s;
}

MATRIX *
zb_camera_get_w2c(ZB_CAMERA *camera)
{
    return camera->w2c;
}

MATRIX *
zb_camera_get_w2c_motion(ZB_CAMERA *camera)
{
    return camera->w2c_motion;
}

MATRIX *
zb_camera_get_c2s(ZB_CAMERA *camera)
{
    return camera->c2s;
}

MATRIX *
zb_camera_get_w2s(ZB_CAMERA *camera)
{
    return camera->w2s;
}

MATRIX *
zb_camera_get_w2s_motion(ZB_CAMERA *camera)
{
    return camera->w2s_motion;
}

MATRIX *
ZB_camera_get_c2s(ZB *zb)
{
    return zb->camera->c2s;
}

MATRIX *
ZB_camera_get_w2c(ZB *zb)
{
    return zb->camera->w2c;
}

int
ZB_set_camera_format(ZB *zb, char *format)
{
    int i;

    for (i = 0; i < ZB_CAMERA_FORMAT_COUNT; i++) {
        if (strcmp(format, zb_camera_format[i].name) == 0) {
            MATRIX_set_window_projection(zb->camera->c2n,
                zb_camera_format[i].window,
                zb->camera->hither, zb->camera->yon);
            ZB_set_camera_window(zb,
                zb_camera_format[i].window[0], zb_camera_format[i].window[1],
                zb_camera_format[i].window[2], zb_camera_format[i].window[3]);
            ZB_set_camera_viewport(zb,
                zb_camera_format[i].viewport[0],
                zb_camera_format[i].viewport[1],
                zb_camera_format[i].viewport[2],
                zb_camera_format[i].viewport[3]);
            return 0;
        }
    }

    return -1;
}

void
ZB_scale_camera_viewport(ZB *zb, float scale)
{
    assert(scale > 0 && scale < 10);

    ZB_set_camera_viewport(zb,
        zb->camera->viewport[0] / scale,
        zb->camera->viewport[1] / scale,
        zb->camera->viewport[2] / scale,
        zb->camera->viewport[3] / scale);
}


void
ZB_set_camera_samples(ZB *zb, int supersample_rate)
{
    int old_rate, vp[4];

    assert(supersample_rate > 0 && supersample_rate < 64);

    old_rate = zb->camera->supersample_rate;
    vp[0] = zb->camera->viewport[0] / old_rate;
    vp[1] = zb->camera->viewport[1] / old_rate;
    vp[2] = zb->camera->viewport[2] / old_rate;
    vp[3] = zb->camera->viewport[3] / old_rate;
    
    zb->camera->supersample_rate = supersample_rate;
    zb_filter_set_supersample_rate(zb->filter, supersample_rate);
    ZB_set_camera_viewport(zb, vp[0], vp[1], vp[2], vp[3]);
}

int
zb_camera_get_supersample_rate(ZB_CAMERA *camera)
{
    return camera->supersample_rate;
}

void
zb_camera_get_image_viewport(ZB_CAMERA *camera, int src[4], int dst[4])
{
    dst[0] = src[0] / camera->supersample_rate;
    dst[1] = src[1] / camera->supersample_rate;
    dst[2] = src[2] / camera->supersample_rate;
    dst[3] = src[3] / camera->supersample_rate;
}

    
