/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_dof.c,v 1.2 2001/11/12 18:52:08 wex Exp $
 *
 * $Log: zb_dof.c,v $
 * Revision 1.2  2001/11/12 18:52:08  wex
 * Cleanup for Python1.5
 *
 *
 */

#include <assert.h>
#include <math.h>

#include "zb_dof.h"
#include "zb_camera.h"

struct zb_dof {
    double scale_factor;
    double focal_length;
    double focus_image_dist;
    double confusion_factor;
    double width;
};


static ZB_DOF *zb_dof_create(double focus_dist, double scale_factor,
    double focal_length, double format_size, int width, double aperture);


/*
  focus_dist - the distance in world-space units that you want to be
      in focus.  This corresponds to the focus ring on a camera lens.

  scale_factor - specified by the user for converting world-space
      units to meters.

  focal_length - focal length of lens, specified in millimeters.
      This is typically taken from the ADB camera definition.

  aperture - the aperture f-stop, typically between 2.0 and 22.0.
      This is probably the most significant parameter for determining
	  depth-of-field.  A smaller number will result in more blurring.

  format_size - film format size in millimeters.  We always use 35.
      This is significant because we express lens view angles as a focal
	  length measure instead of degrees.
*/

void
ZB_set_dof(ZB *zb, double focus_dist, double scale_factor,
    double focal_length, double aperture)
{
    int width;

    width = zb_camera_get_width(zb->camera);
    
    zb->dof = zb_dof_create(focus_dist, scale_factor, focal_length, 50,
        width, aperture);
}

static ZB_DOF *
zb_dof_create(double focus_dist, double scale_factor, double focal_length,
	double format_size, int width, double aperture)
{
	ZB_DOF *dof;

	dof = (ZB_DOF *)malloc(sizeof(ZB_DOF));
	assert(dof != NULL);

	dof->scale_factor = scale_factor;
	
    /* convert from world-space units to meters */
    focus_dist *= scale_factor;

    /* convert distances to meters */
    focal_length *= 0.001;
    format_size *= 0.001;

	dof->focal_length = focal_length;
	dof->width = width;
	
    /* compute constant terms for subsequent calculations */
    dof->focus_image_dist = focal_length*focus_dist/(focus_dist-focal_length);
    dof->confusion_factor = (focal_length * width*0.5)/(aperture*format_size);

	return dof;
}


/*
 *  Compute circle of confusion radius (in pixels) given
 *    an object distance IN METERS.
 *
 *  This has been optimized using precomputed terms.
 */
double
zb_dof_confusion_radius(ZB_DOF *dof, double z)
{
    double image_dist, radius;

    if (dof == NULL) return 0;
    
    if (z < 0) z = -z;
    
	z *= dof->scale_factor;
	
    /*
     *  Prevent things from blowing up too close (division by zero).
     *  In the limit, radius is confusion factor, but we just clamp
     *  at some reasonable close distance to avoid numerical problems.
     */
    if (z < dof->focal_length * 1.05) {
        z = dof->focal_length * 1.05;
    }

    image_dist = (z * dof->focal_length) / (z - dof->focal_length);

    radius = fabs(image_dist - dof->focus_image_dist) * dof->confusion_factor
        / image_dist;
    
    {
        static double max = 0;
        if (radius > max + 4) {
            max = radius;
            fprintf(stderr, "%g = (%g - %g) * %g / %g\n",
                radius, image_dist, dof->focus_image_dist,
                dof->confusion_factor, image_dist);
        }
    }
	
    return radius;
}
