/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_geometry.c,v 1.15 2001/12/21 07:20:12 wex Exp $
 *
 * $Log: zb_geometry.c,v $
 * Revision 1.15  2001/12/21 07:20:12  wex
 * Removed memory leaks
 *
 * Revision 1.14  2001/12/16 00:28:34  wex
 * Fixes to frustum culling
 *
 * Revision 1.13  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.12  2001/11/16 09:26:36  wex
 * Statistics
 *
 * Revision 1.11  2001/11/16 01:04:18  wex
 * Memory leak cleanup
 *
 * Revision 1.10  2001/11/13 18:32:34  wex
 * Memory management and reference counting fixes.
 *
 * Revision 1.9  2001/11/13 11:46:27  wex
 * Removed ZB_GRID_DATA, ZB_GEOMETRY_DATA, and tile grid and geometry lists
 *
 * Revision 1.8  2001/11/13 07:47:18  wex
 * Python references and minor fixes
 *
 * Revision 1.7  2001/11/13 07:16:07  wex
 * Minor memory and code cleanup.
 *
 * Revision 1.6  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.5  2001/11/06 06:52:37  wex
 * Motion blur, intersection cleanup
 *
 * Revision 1.4  2001/11/04 08:14:59  wex
 * Changes to support material bindings.
 *
 * Revision 1.3  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:29  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <float.h>
#include <math.h>
#include <assert.h>

#include "zb_geometry.h"
#include "zb_camera.h"
#include "zb_grid.h"
#include "zb_dof.h"

struct zb_geometry {
    LIST_ELEMENT list_element;  /* only in one list at a time */
    ZB *zb;
    SHADER_GEOMETRY *geometry;
    SHADER_MATERIAL *material;
    SHADER_MATERIAL *displacement;
    int computed_grids;
    double camera_depth;         /* for sorting in tile */
    int computed_scr_bbox;
    double shading_rate;
    VECTOR *lcl_min, *lcl_max;  /* bbox in local space */
    VECTOR *scr_min, *scr_max;  /* bbox in camera space */
    MATRIX *l2w, *l2w_motion;
};

static ZB_GEOMETRY *zb_geometry_create(ZB *zb, SHADER_GEOMETRY *geometry,
    SHADER_MATERIAL *material, SHADER_MATERIAL *displacement,
    double shading_rate);
static void zb_geometry_compute_scr_bbox(ZB_GEOMETRY *geometry);
static void get_screen_space_bbox(ZB_DOF *dof, MATRIX *w2s,
    MATRIX *w2s_motion, MATRIX *l2w, MATRIX *l2w_motion,
    VECTOR *vlmin, VECTOR *vlmax, VECTOR *smin, VECTOR *smax);
static void transform_and_update(ZB_DOF *dof, MATRIX *m,
    double x, double y, double z, VECTOR *min, VECTOR *max);
static void zb_geometry_compute_depth(ZB_GEOMETRY *geometry);


int zb_geometry_alloced = 0;
int zb_geometry_destroyed = 0;
int zb_geometry_max = 0;



int
ZB_add_geometry(ZB *zb, PyObject *geometry, PyObject *material,
    PyObject *displacement, double shading_rate)
{
    ZB_GEOMETRY *zbg;
    SHADER_GEOMETRY *shader_geometry;
    SHADER_MATERIAL *shader_material;
    SHADER_MATERIAL *shader_displacement;

    /* we're saving a reference to this material (and SHADER_MATERIAL) */

    shader_material = SHADER_MATERIAL_get_from_pobj(material);
    Py_INCREF(material);

    if (displacement != NULL) {
        shader_displacement = SHADER_MATERIAL_get_from_pobj(displacement);
        Py_INCREF(displacement);
    } else {
        shader_displacement = NULL;
    }

    /* This saves a reference automatically */
    shader_geometry = SHADER_GEOMETRY_create(geometry);
    if (shader_geometry == NULL) {
        goto error;
    }
    
    zbg = zb_geometry_create(zb, shader_geometry, shader_material,
        shader_displacement, shading_rate);
    if (zbg == NULL) {
        goto error;
    }

    LIST_append(zb->geometry_list, zbg);

    SHADER_MATERIAL_set_pobj(shader_material, material);
    if (displacement != NULL) {
        SHADER_MATERIAL_set_pobj(shader_displacement, displacement);
    } 
    
    return 0;

 error:
    /* cleanup */
    Py_DECREF(material);
    if (displacement != NULL) {
        Py_DECREF(displacement);
    }
    if (shader_geometry != NULL) {
        SHADER_GEOMETRY_destroy(shader_geometry);
    }

    return -1;
}

static ZB_GEOMETRY *
zb_geometry_create(ZB *zb, SHADER_GEOMETRY *geometry,
    SHADER_MATERIAL *material, SHADER_MATERIAL *displacement,
    double shading_rate)
{
    ZB_GEOMETRY *zbg;

    zbg = (ZB_GEOMETRY *)malloc(sizeof(ZB_GEOMETRY));
    if (zbg == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    zbg->zb = zb;
    zbg->geometry = geometry;
    zbg->material = material;
    zbg->displacement = displacement;
    zbg->computed_grids = 0;
    zbg->camera_depth = -1;
    zbg->computed_scr_bbox = 0;
    zbg->shading_rate = shading_rate;
    zbg->l2w = NULL;
    zbg->l2w_motion = NULL;
    zbg->lcl_min = VECTOR_create( FLT_MAX,  FLT_MAX,  FLT_MAX);
    zbg->lcl_max = VECTOR_create(-FLT_MAX, -FLT_MAX, -FLT_MAX);
    zbg->scr_min = VECTOR_create( FLT_MAX,  FLT_MAX,  FLT_MAX);
    zbg->scr_max = VECTOR_create(-FLT_MAX, -FLT_MAX, -FLT_MAX);
    
    zb_geometry_compute_scr_bbox(zbg);

    if (zbg->scr_max->v[2] < -1) {
        /* behind camera */
        VECTOR_destroy(zbg->lcl_min);
        VECTOR_destroy(zbg->lcl_max);
        VECTOR_destroy(zbg->scr_min);
        VECTOR_destroy(zbg->scr_max);
        MATRIX_destroy(zbg->l2w);
        MATRIX_destroy(zbg->l2w_motion);
        free(zbg);
        return NULL;
    }
    
    zb_geometry_alloced++;
    if (zb_geometry_alloced - zb_geometry_destroyed > zb_geometry_max) {
        zb_geometry_max = zb_geometry_alloced - zb_geometry_destroyed;
    }
    
    return zbg;
}

void
zb_geometry_destroy(ZB_GEOMETRY *geometry)
{
    PyObject *pobj;

    zb_geometry_destroyed++;
    
    VECTOR_destroy(geometry->lcl_min);
    VECTOR_destroy(geometry->lcl_max);
    VECTOR_destroy(geometry->scr_min);
    VECTOR_destroy(geometry->scr_max);
    SHADER_GEOMETRY_destroy(geometry->geometry);
    pobj = SHADER_MATERIAL_get_pobj(geometry->material);
    assert(pobj != NULL);
    Py_DECREF(pobj); 
    MATRIX_destroy(geometry->l2w);
    MATRIX_destroy(geometry->l2w_motion);
    free(geometry);
}

ZB_LOCATION
zb_geometry_compare_viewports(ZB_GEOMETRY *geometry, int viewport[4])
{
    VECTOR *min, *max;
    min = geometry->scr_min;
    max = geometry->scr_max;
    return zb_compare_viewports(min->v[0], min->v[1], max->v[0], max->v[1],
        viewport[0], viewport[1], viewport[2], viewport[3]);
}

int
zb_geometry_viewport_contained(ZB_GEOMETRY *geometry, int viewport[4])
{
    VECTOR *min, *max;
    min = geometry->scr_min;
    max = geometry->scr_max;
    return zb_viewport_contained(min->v[0], min->v[1], max->v[0], max->v[1],
        viewport[0], viewport[1], viewport[2], viewport[3]);
}

int
zb_geometry_below_viewport_top(ZB_GEOMETRY *geometry, int top)
{
    return geometry->scr_max->v[1] < top;
}

int
zb_geometry_compare(ZB_GEOMETRY *a, ZB_GEOMETRY *b, void *data)
{
    if (a->camera_depth < 0) zb_geometry_compute_depth(a);
    if (b->camera_depth < 0) zb_geometry_compute_depth(b);

    return a->camera_depth < b->camera_depth;
}

static void
zb_geometry_compute_depth(ZB_GEOMETRY *geometry)
{
    VECTOR avg;

    assert(geometry->camera_depth < 0);

    /* compute depth as distance from COP to middle of bbox */
    VECTOR_add(geometry->scr_min, geometry->scr_max, &avg);
    VECTOR_scale(&avg, 0.5);

    geometry->camera_depth = VECTOR_length(&avg);
}

static void
zb_geometry_compute_scr_bbox(ZB_GEOMETRY *geometry)
{
    assert(geometry->computed_scr_bbox == 0);

    SHADER_GEOMETRY_get_bbox(geometry->geometry, geometry->lcl_min,
        geometry->lcl_max);
    geometry->l2w = SHADER_GEOMETRY_get_l2w(geometry->geometry);
    geometry->l2w_motion = SHADER_GEOMETRY_get_l2w_motion(geometry->geometry);

    get_screen_space_bbox(geometry->zb->dof,
        zb_camera_get_w2s(geometry->zb->camera),
        zb_camera_get_w2s_motion(geometry->zb->camera),
        geometry->l2w, geometry->l2w_motion,
        geometry->lcl_min, geometry->lcl_max,
        geometry->scr_min, geometry->scr_max);

    geometry->computed_scr_bbox = 1;
}

static void
get_screen_space_bbox(ZB_DOF *dof, MATRIX *w2s, MATRIX *w2s_motion,
    MATRIX *l2w, MATRIX *l2w_motion, VECTOR *vlmin, VECTOR *vlmax,
    VECTOR *smin, VECTOR *smax)
{
    MATRIX *l2s;
    double *lmin, *lmax;
    
    VECTOR_set(smin, FLT_MAX, FLT_MAX, FLT_MAX);
    VECTOR_set(smax, -FLT_MAX, -FLT_MAX, -FLT_MAX);

    l2s = MATRIX_create();
    MATRIX_multiply(l2w, w2s, l2s);

    lmin = vlmin->v;
    lmax = vlmax->v;

    transform_and_update(dof, l2s, lmin[0], lmin[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmin[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmax[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmin[0], lmax[1], lmin[2], smin, smax);
    
    transform_and_update(dof, l2s, lmin[0], lmin[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmin[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmax[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmin[0], lmax[1], lmax[2], smin, smax);

    MATRIX_multiply(l2w_motion, w2s_motion, l2s);
    
    transform_and_update(dof, l2s, lmin[0], lmin[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmin[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmax[1], lmin[2], smin, smax);
    transform_and_update(dof, l2s, lmin[0], lmax[1], lmin[2], smin, smax);
    
    transform_and_update(dof, l2s, lmin[0], lmin[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmin[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmax[0], lmax[1], lmax[2], smin, smax);
    transform_and_update(dof, l2s, lmin[0], lmax[1], lmax[2], smin, smax);

    MATRIX_destroy(l2s);
}

static void
transform_and_update(ZB_DOF *dof, MATRIX *m, double x, double y, double z,
    VECTOR *min, VECTOR *max)
{
    double r;
    VECTOR p, q;

    r = zb_dof_confusion_radius(dof, z);
    
    V3_SET(p.v, x, y, z);
    if (VECTOR_project(&p, m, &q) < 0) {
        q.v[2] = -fabs(q.v[2]);
    }

    if (q.v[0]-r < min->v[0]) min->v[0] = q.v[0]-r;
    if (q.v[1]-r < min->v[1]) min->v[1] = q.v[1]-r;
    if (q.v[2] < min->v[2]) min->v[2] = q.v[2];

    if (q.v[0]+r > max->v[0]) max->v[0] = q.v[0]+r;
    if (q.v[1]+r > max->v[1]) max->v[1] = q.v[1]+r;
    if (q.v[2] > max->v[2]) max->v[2] = q.v[2];
}


void
zb_geometry_create_grids(ZB_GEOMETRY *geometry, void *data)
{
    int ss;
    int viewport[4];
    double area;
    
    if (geometry->computed_grids) return;

    ss = zb_camera_get_supersample_rate(geometry->zb->camera);

    /* shading rate is in pixels, convert to subpixels */
    area = geometry->shading_rate * ss * ss;

    zb_camera_get_viewport(geometry->zb->camera, viewport);
    
    SHADER_GEOMETRY_generate(geometry->geometry,
        zb_camera_get_w2s(geometry->zb->camera), viewport, area,
        (GRID_CAPTURE)zb_grid_capture, data);

    geometry->computed_grids = 1;
}

ZB *
zb_geometry_get_zb(ZB_GEOMETRY *geometry)
{
    return geometry->zb;
}

SHADER_MATERIAL *
zb_geometry_get_material(ZB_GEOMETRY *geometry)
{
    return geometry->material;
}

SHADER_MATERIAL *
zb_geometry_get_displacement(ZB_GEOMETRY *geometry)
{
    return geometry->displacement;
}

MATRIX *
zb_geometry_get_l2w(ZB_GEOMETRY *geometry)
{
    return geometry->l2w;
}

MATRIX *
zb_geometry_get_l2w_motion(ZB_GEOMETRY *geometry)
{
    return geometry->l2w_motion;
}
