/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_glue.c,v 1.11 2002/01/03 05:43:26 wex Exp $
 *
 * $Log: zb_glue.c,v $
 * Revision 1.11  2002/01/03 05:43:26  wex
 * micropolygon interpolation controls
 *
 * Revision 1.10  2001/12/20 22:00:46  wex
 * Updates for Win32, fixed some matrix memory leaks
 *
 * Revision 1.9  2001/12/16 00:28:34  wex
 * Fixes to frustum culling
 *
 * Revision 1.8  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.7  2001/11/16 01:04:18  wex
 * Memory leak cleanup
 *
 * Revision 1.6  2001/11/13 18:32:34  wex
 * Memory management and reference counting fixes.
 *
 * Revision 1.5  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.4  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.3  2001/11/04 08:14:59  wex
 * Changes to support material bindings.
 *
 * Revision 1.2  2001/11/02 03:17:34  wex
 * Support for Python material and light shaders, updated VECTOR type
 *
 * Revision 1.1  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 *
 */

#include "zb_internal.h"
#include "zb_camera.h"

static PyObject *ErrorObject;                       /* local exceptions */

#define onError(message) {PyErr_SetString(ErrorObject, message); return NULL;}
#define isZb(m) ((m)->ob_type == &Zb)


static PyObject *zb_new(PyObject *self, PyObject *args);
static ZB *zb_alloc(void);
static void zb_dealloc(ZB *zb);
static int zb_print(ZB *zb, FILE *fp, int flags);
static PyObject *zb_getattr(ZB *zb, char *name);
static PyObject *zb_add_light(ZB *zb, PyObject *args);
static PyObject *zb_add_geometry(ZB *zb, PyObject *args);
static PyObject *zb_render(ZB *zb, PyObject *args);
static PyObject *zb_get_image(ZB *zb, PyObject *args);
static PyObject *zb_set_format(ZB *zb, PyObject *args);
static PyObject *zb_scale_viewport(ZB *zb, PyObject *args);
static PyObject *zb_set_samples(ZB *zb, PyObject *args);
static PyObject *zb_set_filter(ZB *zb, PyObject *args);
static PyObject *zb_set_dof(ZB *zb, PyObject *args);
static PyObject *zb_checkpoint(ZB *zb, PyObject *args);
static PyObject *zb_material_checkpoint(ZB *zb, PyObject *args);
static PyObject *zb_interpolate_mp(ZB *zb, PyObject *args);
static PyObject *zb_print_stats(ZB *zb, PyObject *args);


/** Type Descriptors **/

static struct PyMethodDef zbuffer_methods[] = {
    {"add_light",   (PyCFunction)zb_add_light,           1, NULL},
    {"add_geometry",(PyCFunction)zb_add_geometry,        1, NULL},
    {"render",      (PyCFunction)zb_render,              1, NULL},
    {"get_image",   (PyCFunction)zb_get_image,           1, NULL},
    {"set_format",  (PyCFunction)zb_set_format,          1, NULL},
    {"scale_viewport",(PyCFunction)zb_scale_viewport,    1, NULL},
    {"set_samples", (PyCFunction)zb_set_samples,         1, NULL},
    {"set_filter",  (PyCFunction)zb_set_filter,          1, NULL},
    {"set_dof",     (PyCFunction)zb_set_dof,             1, NULL},
    {"checkpoint",  (PyCFunction)zb_checkpoint,          1, NULL},
    {"material_checkpoint",  (PyCFunction)zb_material_checkpoint,          1, NULL},
    {"interpolate_mp", (PyCFunction)zb_interpolate_mp,   1, NULL},
    {"statistics",  (PyCFunction)zb_print_stats,         1, NULL},
    {NULL,          NULL}
};

static PyTypeObject Zb = {
    /* type header */
#ifdef WIN32
    PyObject_HEAD_INIT(NULL)
#else    
    PyObject_HEAD_INIT(&PyType_Type)
#endif    
    0,                                              /* ob_size               */
    "zbuffer",                                      /* tp_name               */
    sizeof(ZB),                                     /* tp_basicsize          */
    0,                                              /* tp_itemsize           */

    /* standard methods */
    (destructor)zb_dealloc,                         /* tp_dealloc refcount=0 */
    (printfunc)zb_print,                            /* tp_print   "print x"  */
    (getattrfunc)zb_getattr,                        /* tp_getattr "x.attr"   */
    0,                                              /* tp_setattr "x.attr=y" */
    (cmpfunc)NULL,                                  /* tp_compare "x > y"    */
    (reprfunc)NULL,                                 /* tp_repr    `x`,print x*/

    /* type categories */
    NULL,                                           /* tp_as_number          */
    NULL,                                           /* tp_as_sequence        */
    NULL,                                           /* tp_as_mapping         */
};

static struct PyMethodDef zbuffertype_methods[] = {
    {"ZBuffer", zb_new, 1},
    {NULL, NULL}
};


void
initzbcmodule(void)
{
    PyObject *m, *d;

#ifdef WIN32    
    Zb.ob_type = &PyType_Type;
#endif
    
    m = Py_InitModule("zbcmodule", zbuffertype_methods);/* make the module */
    d = PyModule_GetDict(m);                        /* with 'Zb' func */
    ErrorObject = Py_BuildValue("s", "zbcmodule.error");
    PyDict_SetItemString(d, "error", ErrorObject);  /* export exception */
    if (PyErr_Occurred()) {
        Py_FatalError("cannot initialize zb C module");
    }
}

static PyObject *
zb_new(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, "")) return NULL;
    return (PyObject *)zb_alloc();
}

static ZB *
zb_alloc(void)
{
    ZB *zb;
    MATRIX *w2c;
    
    zb = PyObject_NEW(ZB, &Zb);              /* malloc, init, incref */
    zb_initialize(zb);
    w2c = MATRIX_create();
    MATRIX_set_translate(w2c, 0, 0, -4);
    MATRIX_lookat(w2c, 0, 0, 0);
    ZB_set_camera_w2c(zb, w2c, w2c);
    ZB_set_camera_format(zb, "d1");
    ZB_set_camera_samples(zb, 1);
    ZB_set_filter_kernel(zb, ZB_FILTER_GAUSSIAN, 2);

#if 0    
    c2s = MATRIX_create();
#define DEGTORAD		.017453292519943295769
    MATRIX_set_lens_projection(c2s, DEGTORAD * 90, 0.01, 10000);
    ZB_set_camera_transforms(zb, w2c, w2c, c2s);
#endif

    MATRIX_destroy(w2c);
    
    return zb;
}

static void
zb_dealloc(ZB *zb)
{
    zb_delete(zb);
    PyMem_DEL(zb);
}

static int
zb_print(ZB *zb, FILE *fp, int flags)
{
    MATRIX *m;
    
    fprintf(fp, "ZB at %p: %d nonempty tiles, %d captured grids, %d scanned "
        "grids\n", zb, zb->nonempty_tiles, zb->captured_grids,
        zb->scanned_grids);
    m = zb_camera_get_w2c(zb->camera);
    fprintf(fp, "w2c:\n");
    MATRIX_print(m, fp);
    m = zb_camera_get_c2s(zb->camera);
    fprintf(fp, "c2s:\n");
    MATRIX_print(m, fp);
    
    return 0;
}

static PyObject *
zb_getattr(ZB *zb, char *name)
{
     return Py_FindMethod(zbuffer_methods, (PyObject *)zb, name);
}

static PyObject *
zb_add_light(ZB *zb, PyObject *args)
{
    PyObject *light;
    
    IMAGE_support_all();
    if (!PyArg_ParseTuple(args, "O", &light))
        onError("add_light requires a single Light argument");

    ZB_add_light(zb, light);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_add_geometry(ZB *zb, PyObject *args)
{
    PyObject *geometry, *material, *displacement;
    SHADER_MATERIAL *shader_material;
    double shading_rate;

    displacement = NULL;
    if (!PyArg_ParseTuple(args, "OOd|O", &geometry, &material, &shading_rate,
            &displacement))
        onError("add_geometry requires a geometry object and a material "
            "and a shading rate");

/*    Py_INCREF(geometry); */
/*    Py_INCREF(material); */

    if (displacement != NULL) {
        shader_material = SHADER_MATERIAL_get_from_pobj(displacement);
    } else {
        shader_material = NULL;
    }
    
    if (shader_material == NULL ||
        !SHADER_MATERIAL_check((PyObject *)shader_material)) {
        displacement = NULL;
    } 

    ZB_add_geometry(zb, geometry, material, displacement, shading_rate);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_render(ZB *zb, PyObject *args)
{
    int status;
    
    if (!PyArg_ParseTuple(args, ""))
        onError("render does not take arguments");

    status = ZB_render(zb);

    return Py_BuildValue("i", status);
}

static PyObject *
zb_get_image(ZB *zb, PyObject *args)
{
    IMAGE *image;
    
    if (!PyArg_ParseTuple(args, ""))
        onError("get_image does not take arguments");

    image = ZB_get_image(zb);
    IMAGE_incref(image);

    return Py_BuildValue("O", ZB_get_image(zb));
}

static PyObject *
zb_set_format(ZB *zb, PyObject *args)
{
    char *s;
    
    if (!PyArg_ParseTuple(args, "s", &s))
        onError("set format requires a single string argument");

    ZB_set_camera_format(zb, s);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_scale_viewport(ZB *zb, PyObject *args)
{
    double scale;
    
    if (!PyArg_ParseTuple(args, "d", &scale))
        onError("scale viewport requires a single double argument");

    ZB_scale_camera_viewport(zb, scale);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_set_samples(ZB *zb, PyObject *args)
{
    int s;
    
    if (!PyArg_ParseTuple(args, "i", &s))
        onError("set samples requires a single integer argument");

    ZB_set_camera_samples(zb, s);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_set_filter(ZB *zb, PyObject *args)
{
    char *s;
    int r;
    ZB_FILTER_TYPE type;
    
    if (!PyArg_ParseTuple(args, "si", &s, &r))
        onError("set filter requires a string name and integer radius "
            "argument");

    if (strcmp(s, "gaussian") == 0) {
        type = ZB_FILTER_GAUSSIAN;
    } else if (strcmp(s, "triangle") == 0) {
        type = ZB_FILTER_TRIANGLE;
    } else if (strcmp(s, "box") == 0) {
        type = ZB_FILTER_BOX;
    } else {
        onError("set filter received an unknown filter type");
    }

    if (r < 0 || r > 64) {
        onError("set filter received an invalid radius");
    }
    
    ZB_set_filter_kernel(zb, type, r);
    

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_set_dof(ZB *zb, PyObject *args)
{
    double focus_dist, scale_factor, focal_length, aperture;
    
    if (!PyArg_ParseTuple(args, "dddd", &focus_dist, &scale_factor,
            &focal_length, &aperture)) {
        onError("set filter requires a string name and integer radius "
            "argument");
    }

    ZB_set_dof(zb, focus_dist, scale_factor, focal_length, aperture);
    
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_checkpoint(ZB *zb, PyObject *args)
{
    fprintf(stderr, "checkpoint: zb=%p  ref=%d\n", zb, zb->ob_refcnt);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_material_checkpoint(ZB *zb, PyObject *args)
{
    PyObject *material;

    if (!PyArg_ParseTuple(args, "O", &material)) abort();
    
    fprintf(stderr, "checkpoint: zb=%p  ref=%d  material ref=%d\n", zb,
        zb->ob_refcnt, material->ob_refcnt);
    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
zb_interpolate_mp(ZB *zb, PyObject *args)
{
    int state;

    if (!PyArg_ParseTuple(args, "i", &state))
        onError("Pass a state when setting the micropolygon interpolator");

    ZB_set_interpolate_mp(zb, state);
    Py_INCREF(Py_None);
    return Py_None;
}

    
static PyObject *
zb_print_stats(ZB *zb, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
        onError("statistics does not take arguments");

    ZB_print_statistics(zb);
    Py_INCREF(Py_None);
    return Py_None;
}

