/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_grid.c,v 1.18 2002/01/03 05:43:26 wex Exp $
 *
 * $Log: zb_grid.c,v $
 * Revision 1.18  2002/01/03 05:43:26  wex
 * micropolygon interpolation controls
 *
 * Revision 1.17  2001/12/20 22:00:46  wex
 * Updates for Win32, fixed some matrix memory leaks
 *
 * Revision 1.16  2001/12/16 00:28:34  wex
 * Fixes to frustum culling
 *
 * Revision 1.15  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.14  2001/11/16 09:26:37  wex
 * Statistics
 *
 * Revision 1.13  2001/11/13 11:46:27  wex
 * Removed ZB_GRID_DATA, ZB_GEOMETRY_DATA, and tile grid and geometry lists
 *
 * Revision 1.12  2001/11/13 07:47:18  wex
 * Python references and minor fixes
 *
 * Revision 1.11  2001/11/13 07:16:07  wex
 * Minor memory and code cleanup.
 *
 * Revision 1.10  2001/11/12 18:52:08  wex
 * Cleanup for Python1.5
 *
 * Revision 1.9  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.8  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.7  2001/11/06 06:52:37  wex
 * Motion blur, intersection cleanup
 *
 * Revision 1.6  2001/11/04 08:14:47  wex
 * Switched from grid to micropolygon scan conversion for 14x speedup!
 *
 * Revision 1.5  2001/11/02 03:17:34  wex
 * Support for Python material and light shaders, updated VECTOR type
 *
 * Revision 1.4  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.3  2001/10/29 00:50:29  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.2  2001/10/29 00:16:48  wex
 * Interpolate all vertices together, rather than for each micropolygon
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <assert.h>

#include <geometry.h>

#include "zb_grid.h"
#include "zb_tile.h"
#include "zb_geometry.h"
#include "zb_pixel.h"
#include "zb_camera.h"
#include "zb_dof.h"

typedef struct zb_hit {
    double u, v;
    int tri;
} ZB_HIT;

typedef struct zb_mp {
    double *p0;
    double *p1;
    double *p2;
    double *p3;
} ZB_MP;

struct zb_grid {
    LIST_ELEMENT list_element;  /* only in one list at a time! */
    int ns, nt;
    double (*p)[3];             /* camera space position at frame open */
    double (*s)[3];             /* screen space position at frame open */
    double (*t)[3];             /* screen space position at frame close */
    double (*f)[3];             /* screen space interpolated position (tmp) */
    double (*c)[4];             /* shaded grid color (NULL if not shaded) */
    double (*r)[2];             /* depth of field circle of confusion radius */
    double (*uv)[2];            /* uv value */
    int bbox[4];                /* screen space viewport */
    SHADER_MATERIAL *material;
    MATRIX *l2c, *c2l;          /* to transform data values */
};

typedef struct zb_grid_shade_data {
    ZB_GRID *grid;
    int u, v;
} ZB_GRID_SHADE_DATA;

static double *create_color_grid(int n);
static int zb_grid_get_data(SHADER *material, ZB_GRID_SHADE_DATA *data,
    char *name, SHADER_DATA_TYPE type, void *result);
static int zb_grid_set_data(SHADER *material, ZB_GRID_SHADE_DATA *data,
    char *name, SHADER_DATA_TYPE type, void *result);
static int zb_grid_hit_interpolated_mp(ZB_GRID *grid, int u, int v, double x,
    double y, double t, ZB_HIT **hit);
static void zb_grid_render_mp(ZB_GRID *grid, ZB_TILE *tile, int u, int v,
    int w, int h, int tile_viewport[4]);
static int zb_grid_compute_mp_bbox(ZB_GRID *grid, int i0, int i1,int i2,int i3,
    int w, int tile_viewport[4], int *xmin, int *ymin, int *xmax, int *ymax);
static int mp_hit_triangles(ZB_MP *mp, double sx, double sy, ZB_HIT **hit);
static int mp_hit_edges(ZB_MP *mp, double sx, double sy, ZB_HIT **hit);
static ZB_MP *zb_grid_get_interpolated_mp(ZB_GRID *grid, int u,int v);
static void get_interpolated_color_and_depth(ZB_GRID *grid, ZB_HIT *hit,
    int u, int v, double color[4], double *depth);
static void get_color_and_depth(ZB_GRID *grid, ZB_HIT *hit,
    int u, int v, double color[4], double *depth);
static void linearly_interpolate(double *t0, double *t1, double t, double *d);
static int hit_triangle(double P[2], double V0[2], double V1[2], double V2[2],
    double *u, double *v);
static double *create_transformed_grid(int count, double (*p)[3], MATRIX *l2c);
static double *create_projected_grid(int count, double (*p)[3], MATRIX *l2s);
static double *create_displacement_grid(int , ZB_GRID *grid,
    SHADER_MATERIAL *displacement);
static void copy_displacement_grid_and_convert_to_offsets(int n,
    double (*d)[3], double (*p)[3]);
static void apply_displacement_offsets(int n, double (*p)[3], double (*d)[3]);
static void zb_grid_shade_point(ZB_GRID *grid, SHADER_MATERIAL *material,
    int u, int v, double c[4]);


static int (*zb_mp_hit)(ZB_MP *mp, double sx, double sy, ZB_HIT **hit) =
	mp_hit_edges;
static void (*zb_grid_get_color_and_depth)(ZB_GRID *grid, ZB_HIT *hit,
    int u, int v, double color[4], double *depth) = get_color_and_depth;


int zb_grid_points = 0;
int zb_shaded_points = 0;
int zb_grids_created = 0;
int zb_grids_destroyed = 0;
int zb_max_grids = 0;
float zb_grid_mb = 0;


void
ZB_set_interpolate_mp(ZB *zb, int state)
{
    if (state) {
        zb_mp_hit = mp_hit_triangles;
        zb_grid_get_color_and_depth = get_interpolated_color_and_depth;
    } else {
        zb_mp_hit = mp_hit_edges;
        zb_grid_get_color_and_depth = get_color_and_depth;
    }
}

ZB_GRID *
zb_grid_create(ZB *zb, GRID *ggrid, ZB_GEOMETRY *geometry)
{
    ZB_GRID *grid;
    MATRIX *l2w, *c2s, *w2c, *l2c_motion;
    SHADER_MATERIAL *displacement;
    int i, n;
    double r;
    double (*q)[3];
    double (*d)[3];
    
    zb_grids_created++;

    if (zb_grids_created - zb_grids_destroyed > zb_max_grids) {
        zb_max_grids = zb_grids_created - zb_grids_destroyed;
    }
    
    grid = (ZB_GRID *)malloc(sizeof(ZB_GRID));
    if (grid == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    zb_grid_mb += sizeof(ZB_GRID) / (1024.0 * 1024.0);
    
    n = ggrid->ns * ggrid->nt;
    grid->ns = ggrid->ns;
    grid->nt = ggrid->nt;
    grid->material = zb_geometry_get_material(geometry);

    zb_grid_points += n;
    
    grid->p = NULL;
    grid->s = NULL;
    grid->t = NULL;
    grid->f = NULL;
    grid->c = NULL;
    grid->r = NULL;
    grid->uv = NULL;
    grid->l2c = NULL;
    grid->c2l = NULL;
    d = NULL;
    q = NULL;
    
    grid->uv = (double (*)[2])malloc(n * 2 * sizeof(double));
    if (grid->uv == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }
    memcpy(grid->uv, ggrid->uv, n * 2 * sizeof(double));
    
    /* cached interpolated values computed for each pixel on-the-fly */
    grid->f = (double (*)[3])malloc(n * 3 * sizeof(double));
    if (grid->f == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }
    zb_grid_mb += (n * 3 * sizeof(double)) / (1024.0 * 1024.0);
    
    /* create camera space transformed points for shading */
    l2w = zb_geometry_get_l2w(geometry);
    w2c = zb_camera_get_w2c(zb->camera);
    grid->l2c = MATRIX_create();
    MATRIX_multiply(l2w, w2c, grid->l2c);

    grid->p = (double (*)[3])create_transformed_grid(n, ggrid->p, grid->l2c);
    zb_grid_mb += (n * 3 * sizeof(double)) / (1024.0 * 1024.0);
    grid->c2l = MATRIX_create();
    MATRIX_invert(grid->l2c, grid->c2l);

    /* compute the displacement grid */
    displacement = zb_geometry_get_displacement(geometry);
    if (displacement != NULL) {
        /* compute and apply the displacement map to the position
           before shading, and retain for motion positions */
        d = (double (*)[3])create_displacement_grid(n, grid, displacement);
        copy_displacement_grid_and_convert_to_offsets(n, d, grid->p);
    }
        
    /* shade the grid updating any displaced vertices */
    grid->c = (double (*)[4])create_color_grid(n);

    /* project the camera space points into screen space */
    c2s = zb_camera_get_c2s(zb->camera);
    grid->s = (double (*)[3])create_projected_grid(n, grid->p, c2s);
    zb_grid_mb += (n * 3 * sizeof(double)) / (1024.0 * 1024.0);

    l2w = zb_geometry_get_l2w_motion(geometry);

    /* always compute q for dof and mb */

    w2c = zb_camera_get_w2c_motion(zb->camera);
    l2c_motion = MATRIX_create();
    MATRIX_multiply(l2w, w2c, l2c_motion);

    /* apply displacement offsets to the motion position */
    q = (double (*)[3])create_transformed_grid(n, ggrid->q, l2c_motion);
    MATRIX_destroy(l2c_motion);

    if (displacement != NULL) {
        apply_displacement_offsets(n, q, d);
        free(d);
        d = NULL;
    }

    /* project motion into screen space */
    grid->t = (double (*)[3])create_projected_grid(n, q, c2s);
    zb_grid_mb += (n * 3 * sizeof(double)) / (1024.0 * 1024.0);

    /* compute screen space bbox */
    grid->bbox[0] = grid->bbox[1] = INT_MAX;
    grid->bbox[2] = grid->bbox[3] = -INT_MAX;

    grid->r = (double (*)[2])malloc(2 * n * sizeof(double));
    zb_grid_mb += (2 * n * sizeof(double)) / (1024.0 * 1024.0);
    
    for (i = 0; i < n; i++) {
        /* compute dof at both points, and add to bbox computation */
        grid->r[i][0] = r = zb_dof_confusion_radius(zb->dof, grid->p[i][2]);
        if (grid->s[i][0] - r < grid->bbox[0]) grid->bbox[0] = grid->s[i][0]-r;
        if (grid->s[i][1] - r < grid->bbox[1]) grid->bbox[1] = grid->s[i][1]-r;
        if (grid->s[i][0] + r > grid->bbox[2]) grid->bbox[2] = grid->s[i][0]+r;
        if (grid->s[i][1] + r > grid->bbox[3]) grid->bbox[3] = grid->s[i][1]+r;

        grid->r[i][1] = r = zb_dof_confusion_radius(zb->dof, q[i][2]);
        if (grid->t[i][0] - r < grid->bbox[0]) grid->bbox[0] = grid->t[i][0]-r;
        if (grid->t[i][1] - r < grid->bbox[1]) grid->bbox[1] = grid->t[i][1]-r;
        if (grid->t[i][0] + r > grid->bbox[2]) grid->bbox[2] = grid->t[i][0]+r;
        if (grid->t[i][1] + r > grid->bbox[3]) grid->bbox[3] = grid->t[i][1]+r;
    }

    free(q);

    /* add this grid to the remaining grid list */
    LIST_append(zb->grid_list, grid);
    
    return grid;
}

static double *
create_transformed_grid(int count, double (*p)[3], MATRIX *l2c)
{
    int i;
    double (*q)[3];
    VECTOR v, s;
    
    q = (double (*)[3])malloc(count * 3 * sizeof(double));
    if (q == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    for (i = 0; i < count; i++) {
        V3_SET(v.v, p[i][0], p[i][1], p[i][2]);
        VECTOR_transform(&v, l2c, &s);
        q[i][0] = s.v[0];
        q[i][1] = s.v[1];
        q[i][2] = s.v[2];
    }

    return (double *)q;
}

static double *
create_projected_grid(int count, double (*p)[3], MATRIX *l2s)
{
    int i;
    double (*q)[3];
    double w;
    VECTOR v, s;
    
    q = (double (*)[3])malloc(count * 3 * sizeof(double));
    if (q == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    for (i = 0; i < count; i++) {
        V3_SET(v.v, p[i][0], p[i][1], p[i][2]);
        w = VECTOR_project(&v, l2s, &s);
        if (w < 0) {
            q[i][2] = -fabs(s.v[2]);
        } else {
            q[i][2] = s.v[2];
        }
        q[i][0] = s.v[0];
        q[i][1] = s.v[1];
    }
    
    return (double *)q;
}

static double *
create_displacement_grid(int n, ZB_GRID *grid, SHADER_MATERIAL *displacement)
{
    int i, u, v;
    double (*d)[3];
    double p[4];

    d = (double (*)[3])malloc(n * 3 * sizeof(double));
    if (d == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    for (v = 0, i = 0; v < grid->nt; v++) {
        for (u = 0; u < grid->ns; u++, i++) {
            /* compute and store the displaced position */
            p[3] = -1;
            zb_grid_shade_point(grid, displacement, u, v, p);
            d[i][0] = p[0];
            d[i][1] = p[1];
            d[i][2] = p[2];
        }
    }

    assert(i == n);
    
    return (double *)d;
}

static void
copy_displacement_grid_and_convert_to_offsets(int n, double (*d)[3],
    double (*p)[3])
{
    int i;
    double off[3];
    
    for (i = 0; i < n; i++) {
        V3_SUB(d[i], p[i], off);
        V3_COPY(d[i], p[i]);
        V3_COPY(off, d[i]);
    }
}

static void
apply_displacement_offsets(int n, double (*p)[3], double (*d)[3])
{
    int i;

    for (i = 0; i < n; i++) {
        V3_INC(d[i], p[i]);
    }
}

void
zb_grid_destroy(ZB_GRID *grid)
{
    int n;
    
    zb_grids_destroyed++;

    assert(grid->p != NULL);
    assert(grid->s != NULL);
    assert(grid->t != NULL);
    assert(grid->f != NULL);
    assert(grid->c != NULL);
    assert(grid->r != NULL);
    assert(grid->uv != NULL);
    free(grid->p);
    free(grid->s);
    free(grid->t);
    free(grid->f);
    free(grid->c);
    free(grid->r);
    free(grid->uv);

    assert(grid->l2c != NULL);
    assert(grid->c2l != NULL);
    MATRIX_destroy(grid->l2c);
    MATRIX_destroy(grid->c2l);
    
    n = grid->ns * grid->nt;
    zb_grid_mb -= n * (3 /*f*/ + 3 /*p*/ + 3 /*s*/ + 3 /*t*/ + 2 /*r*/) *
        sizeof(double) / (1024.0 * 1024);
    
    /* paranoia */
    grid->p = NULL;
    grid->s = NULL;
    grid->t = NULL;
    grid->f = NULL;
    grid->c = NULL;
    grid->l2c = NULL;
    grid->c2l = NULL;
    
    free(grid);
    zb_grid_mb -= sizeof(ZB_GRID) / (1024.0 * 1024);
}

void
zb_grid_render(ZB_GRID *grid, ZB_TILE *tile, int tile_viewport[4])
{
    ZB *zb;
    int u, v, w, h;

    zb = zb_tile_get_zb(tile);
    ++zb->scanned_grids;

    zb_tile_create_pixels(tile);

    w = tile_viewport[2] - tile_viewport[0] + 1;
    h = tile_viewport[3] - tile_viewport[1] + 1;

    for (v = 0; v < grid->nt - 1; v++) {
        for (u = 0; u < grid->ns - 1; u++) {
            zb_grid_render_mp(grid, tile, u, v, w, h, tile_viewport);
        }
    }
}

static void
zb_grid_render_mp(ZB_GRID *grid, ZB_TILE *tile, int u, int v, int w, int h,
    int tile_viewport[4])
{
    int i0, i1, i2, i3;
    int x, y;
    int contained;
    int xmin, xmax, ymin, ymax;
    ZB_PIXEL *p;
    double xs, ys, t, lu, lv, r;
    double color[4], depth;
    ZB_HIT *hit;
    ZB *zb;
    
    i0 = v * grid->ns + u;
    i1 = i0 + 1;
    i2 = i1 + grid->ns;
    i3 = i2 - 1;


    contained = zb_grid_compute_mp_bbox(grid, i0, i1, i2, i3, w, tile_viewport,
        &xmin, &ymin, &xmax, &ymax);

    if (contained < 0) {
        /* bbox cull */
        return;
    }
    
    /* now loop over mp bbox */
    for (y = ymin; y <= ymax; y++) {
        for (x = xmin; x <= xmax; x++) {
                    
            p = zb_tile_get_pixel(tile, y * w + x);
            zb_pixel_get_offset_position(p, &xs, &ys);
            t = zb_pixel_get_time(p);
                    
            /* interpolate this mp's vertices (t is different for each pixel)*/
            linearly_interpolate(grid->s[i0], grid->t[i0], t, grid->f[i0]);
            linearly_interpolate(grid->s[i1], grid->t[i1], t, grid->f[i1]);
            linearly_interpolate(grid->s[i2], grid->t[i2], t, grid->f[i2]);
            linearly_interpolate(grid->s[i3], grid->t[i3], t, grid->f[i3]);

            /* add in a linearly interpolated lens offset if dof is nonzero */
            zb = zb_tile_get_zb(tile);
            if (zb->dof != NULL) {
                zb_pixel_get_lens_offset(p, &lu, &lv);
                r = grid->r[i0][0] + t * (grid->r[i0][1] - grid->r[i0][0]);
                grid->f[i0][0] += lu * r;
                grid->f[i0][1] += lv * r;

                r = grid->r[i1][0] + t * (grid->r[i1][1] - grid->r[i1][0]);
                grid->f[i1][0] += lu * r;
                grid->f[i1][1] += lv * r;

                r = grid->r[i2][0] + t * (grid->r[i2][1] - grid->r[i2][0]);
                grid->f[i2][0] += lu * r;
                grid->f[i2][1] += lv * r;

                r = grid->r[i3][0] + t * (grid->r[i3][1] - grid->r[i3][0]);
                grid->f[i3][0] += lu * r;
                grid->f[i3][1] += lv * r;
            }
            
            if (zb_grid_hit_interpolated_mp(grid, u, v, xs, ys, t, &hit)) {
                zb_grid_get_color_and_depth(grid, hit, u, v, color, &depth);
                if (depth > -1) zb_pixel_insert_surface(p, color, depth);
            }
        }
    }
}

static int
zb_grid_compute_mp_bbox(ZB_GRID *grid, int i0, int i1, int i2, int i3,
    int w, int tile_viewport[4], int *xmin, int *ymin, int *xmax, int *ymax)
{
    int h, contained;
    double r;
    
    h = tile_viewport[3] - tile_viewport[1] + 1;
    
    /* compute the bbox for this mp */
    r = grid->r[i0][0];
    *xmin = grid->s[i0][0] - r;
    *xmax = grid->s[i0][0] + r;
    *ymin = grid->s[i0][1] - r;
    *ymax = grid->s[i0][1] + r;
    
    r = grid->r[i1][0];
    if (grid->s[i1][0] - r < *xmin) *xmin = grid->s[i1][0] - r;
    if (grid->s[i1][0] + r > *xmax) *xmax = grid->s[i1][0] + r;
    if (grid->s[i1][1] - r < *ymin) *ymin = grid->s[i1][1] - r;
    if (grid->s[i1][1] + r > *ymax) *ymax = grid->s[i1][1] + r;

    r = grid->r[i2][0];
    if (grid->s[i2][0] - r < *xmin) *xmin = grid->s[i2][0] - r;
    if (grid->s[i2][0] + r > *xmax) *xmax = grid->s[i2][0] + r;
    if (grid->s[i2][1] - r < *ymin) *ymin = grid->s[i2][1] - r;
    if (grid->s[i2][1] + r > *ymax) *ymax = grid->s[i2][1] + r;

    r = grid->r[i3][0];
    if (grid->s[i3][0] - r < *xmin) *xmin = grid->s[i3][0] - r;
    if (grid->s[i3][0] + r > *xmax) *xmax = grid->s[i3][0] + r;
    if (grid->s[i3][1] - r < *ymin) *ymin = grid->s[i3][1] - r;
    if (grid->s[i3][1] + r > *ymax) *ymax = grid->s[i3][1] + r;

    /* motion position */

    r = grid->r[i0][1];
    if (grid->t[i0][0] - r < *xmin) *xmin = grid->t[i0][0] - r;
    if (grid->t[i0][0] + r > *xmax) *xmax = grid->t[i0][0] + r;
    if (grid->t[i0][1] - r < *ymin) *ymin = grid->t[i0][1] - r;
    if (grid->t[i0][1] + r > *ymax) *ymax = grid->t[i0][1] + r;

    r = grid->r[i1][1];
    if (grid->t[i1][0] - r < *xmin) *xmin = grid->t[i1][0] - r;
    if (grid->t[i1][0] + r > *xmax) *xmax = grid->t[i1][0] + r;
    if (grid->t[i1][1] - r < *ymin) *ymin = grid->t[i1][1] - r;
    if (grid->t[i1][1] + r > *ymax) *ymax = grid->t[i1][1] + r;

    r = grid->r[i2][1];
    if (grid->t[i2][0] - r < *xmin) *xmin = grid->t[i2][0] - r;
    if (grid->t[i2][0] + r > *xmax) *xmax = grid->t[i2][0] + r;
    if (grid->t[i2][1] - r < *ymin) *ymin = grid->t[i2][1] - r;
    if (grid->t[i2][1] + r > *ymax) *ymax = grid->t[i2][1] + r;

    r = grid->r[i3][1];
    if (grid->t[i3][0] - r < *xmin) *xmin = grid->t[i3][0] - r;
    if (grid->t[i3][0] + r > *xmax) *xmax = grid->t[i3][0] + r;
    if (grid->t[i3][1] - r < *ymin) *ymin = grid->t[i3][1] - r;
    if (grid->t[i3][1] + r > *ymax) *ymax = grid->t[i3][1] + r;

    /* convert viewport to tile offsets */
    contained = 1;
    
    if (*xmin > tile_viewport[0]) {
        *xmin -= tile_viewport[0];
        if (*xmin > w) return -1;
    } else {
        *xmin = 0;
        contained = 0;
    }
            
    if (*ymin > tile_viewport[1]) {
        *ymin -= tile_viewport[1];
        if (*ymin > h) return -1;
    } else {
        *ymin = 0;
        contained = 0;
    }
            
    if (*xmax < tile_viewport[2]) {
        *xmax -= tile_viewport[0];
        if (*xmax < 0) return -1;
    } else {
        *xmax = w - 1;
        contained = 0;
    }

    if (*ymax < tile_viewport[3]) {
        *ymax -= tile_viewport[1];
        if (*ymax < 0) return -1;
    } else {
        *ymax = h - 1;
        contained = 0;
    }

    return contained;
}

static double *
create_color_grid(int n)
{
    int i;
    double (*c)[4];

    c = (double (*)[4])malloc(n * 4 * sizeof(double));
    if (c == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }
    
    zb_grid_mb += (n * 4 * sizeof(double)) / (1024.0 * 1024.0);
    
    for (i = 0; i < n; i++) {
        c[i][3] = -1;   /* invalid color */
    }

    return (double *)c;
}

static void
zb_grid_shade_point(ZB_GRID *grid, SHADER_MATERIAL *material,
    int u, int v, double c[4])
{
    ZB_GRID_SHADE_DATA data;
    SHADER *shader;
    VECTOR C;
    double A;

    if (c[3] != -1) {
        /* already shaded this point */
        return;
    }

    zb_shaded_points++;
    
    data.u = u;
    data.v = v;
    data.grid = grid;
    V3_ZERO(C.v);
    A = 0;

    shader = SHADER_MATERIAL_shade(SHADER_MATERIAL_get_pobj(material),
        (SHADER_DATA_CALLBACK)zb_grid_get_data,
        (SHADER_DATA_CALLBACK)zb_grid_set_data, &data);

    if (shader == NULL) {
        V3_ZERO(c);
        c[3] = 0;
        return;
    }
    
    if (SHADER_get_vector(shader, "C", &C) < 0) {
        V3_ZERO(c);
    } else {
        V3_COPY(C.v, c);
    }

    if (SHADER_get_double(shader, "A", &A) < 0) {
        c[3] = 0;
    } else {
        c[3] = A;
    }

    SHADER_destroy(shader);
}

static int
zb_grid_get_data(SHADER *material, ZB_GRID_SHADE_DATA *data, char *name,
    SHADER_DATA_TYPE type, void *result)
{
    int i;
    double t0[3], t1[3];
    double avg[3];
    double (*p)[3];
    double *fr;
    VECTOR v;

    i = data->v * data->grid->ns + data->u;
    p = data->grid->p;

    switch (type) {
    case SHADER_DATA_TYPE_INT:
    case SHADER_DATA_TYPE_STRING:
        break;

    case SHADER_DATA_TYPE_DOUBLE:
        fr = (double *)result;
        
        if (strcmp(name, "du") == 0) {
            *fr = 0.1;
            return 0;
        }
        if (strcmp(name, "dv") == 0) {
            *fr = 0.1;
            return 0;
        }
        if (strcmp(name, "u") == 0) {
            *fr = data->grid->uv[i][0];  
/*            *fr = data->u / (float)(data->grid->ns+1);*/
            return 0;
        }
        if (strcmp(name, "v") == 0) {
            *fr = data->grid->uv[i][1]; 
/*            *fr = data->v / (float)(data->grid->nt+1); */
            return 0;
        }
            
        break;

    case SHADER_DATA_TYPE_VECTOR:
        if (strcmp(name, "P") == 0) {
            V3_SET(((VECTOR *)result)->v, p[i][0], p[i][1], p[i][2]);
            return 0;
        }
        if (strcmp(name, "N") == 0) {
            /* compute the average of the four face normals around point */
            V3_ZERO(avg);
            
            if (data->u < data->grid->ns - 1 && data->v < data->grid->nt - 1) {
                /* +/+ face */
                V3_SUB(p[i+1], p[i], t0);
                V3_SUB(p[i+data->grid->ns], p[i], t1);
                V3_CROSS(t0, t1, v.v);
                V3_NORMALIZE(v.v);
                VECTOR_3x3_transform(&v, data->grid->c2l, &v);
                V3_NORMALIZE(v.v);
                *(VECTOR *)result = v;
                return 0;
            }

            if (data->u < data->grid->ns - 1 && data->v > 0) {
                assert(data->v == data->grid->nt - 1);
                /* +/- face */
                V3_SUB(p[i+1], p[i], t0);
                V3_SUB(p[i-data->grid->ns], p[i], t1);
                V3_CROSS(t1, t0, v.v);
                V3_NORMALIZE(v.v);
                VECTOR_3x3_transform(&v, data->grid->c2l, &v);
                V3_NORMALIZE(v.v);
                *(VECTOR *)result = v;
                return 0;
            }

            if (data->v < data->grid->nt - 1) {
                assert(data->u == data->grid->ns - 1);
                /* -/+ face */
                V3_SUB(p[i-1], p[i], t0);
                V3_SUB(p[i+data->grid->ns], p[i], t1);
                V3_CROSS(t1, t0, v.v);
                VECTOR_3x3_transform(&v, data->grid->c2l, &v);
                V3_NORMALIZE(v.v);
                *(VECTOR *)result = v;
                return 0;
            }

            /* -/- face guarenteed to exist */
            assert(data->u == data->grid->ns - 1);
            assert(data->v == data->grid->nt - 1);
            V3_SUB(p[i-1], p[i], t0);
            V3_SUB(p[i-data->grid->ns], p[i], t1);
            V3_CROSS(t0, t1, v.v);
            V3_NORMALIZE(v.v);
            VECTOR_3x3_transform(&v, data->grid->c2l, &v);
            V3_NORMALIZE(v.v);
            *(VECTOR *)result = v;
            return 0;
        }
        if (strcmp(name, "dPdu") == 0) {
            return 0;
        }
        if (strcmp(name, "dPdv") == 0) {
            return 0;
        }
        break;
    }

    return -1;
}

static int
zb_grid_set_data(SHADER *material, ZB_GRID_SHADE_DATA *data, char *name,
    SHADER_DATA_TYPE type, void *result)
{
    int i;
    

    if (type == SHADER_DATA_TYPE_VECTOR && strcmp(name, "P") == 0) {
        i = data->v * data->grid->ns + data->u;
        /*
        grid->p[i][0] = fr[0];
        grid->p[i][1] = fr[1];
        grid->p[i][2] = fr[2]; */
        return 0;
    }
    
    return -1;
}

static int
mp_hit_edges(ZB_MP *mp, double sx, double sy, ZB_HIT **hit)
{
    static ZB_HIT hit_data = {0, 0, 0};

    *hit = NULL;
    
    /* backfacing test using screen space area */
	if ((mp->p1[0] - mp->p2[0]) * (mp->p0[1] - mp->p2[1]) -
		(mp->p0[0] - mp->p2[0]) * (mp->p1[1] - mp->p2[1]) >= 0) {
        return 0;
    }

    /* test point against line equations for micropolygon edges */
#define LINE_EQ(px, py, x0, y0, x1, y1) \
    (((py) - (y0)) * ((x1) - (x0)) - ((px) - (x0)) * ((y1) - (y0)))
    
    if (LINE_EQ(sx,sy, mp->p0[0],mp->p0[1], mp->p1[0],mp->p1[1]) < 0) return 0;
    if (LINE_EQ(sx,sy, mp->p1[0],mp->p1[1], mp->p2[0],mp->p2[1]) < 0) return 0;
    if (LINE_EQ(sx,sy, mp->p2[0],mp->p2[1], mp->p3[0],mp->p3[1]) < 0) return 0;
    if (LINE_EQ(sx,sy, mp->p3[0],mp->p3[1], mp->p0[0],mp->p0[1]) < 0) return 0;

    *hit = &hit_data;
    
    return 1;
}

static int
zb_grid_hit_interpolated_mp(ZB_GRID *grid, int u, int v, double x, double y,
    double t, ZB_HIT **hit)
{
    ZB_MP *mp;
    
    mp = zb_grid_get_interpolated_mp(grid, u, v);
    if (mp == NULL) {
        return 0;
    }

    return zb_mp_hit(mp, x, y, hit);
}

static int
mp_hit_triangles(ZB_MP *mp, double sx, double sy, ZB_HIT **hit)
{
    static ZB_HIT hit_data;
    double P[2];

    /* 911: switch to 2D intersect micropolygon quad */

    /* this is wrong since we have u&v triangle barycentric used for
       bilinear quad interpolation to get color and depth! */
    P[0] = sx;
    P[1] = sy;

    /* backfacing test using screen space area */
	if ((mp->p1[0] - mp->p2[0]) * (mp->p0[1] - mp->p2[1]) -
		(mp->p0[0] - mp->p2[0]) * (mp->p1[1] - mp->p2[1]) < 0) {
        /* backfacing test using screen space area */
        if (hit_triangle(P, mp->p2, mp->p1, mp->p0, &hit_data.u, &hit_data.v)){
            hit_data.tri = 0;
            *hit = &hit_data;
            return 1;
        }
    }
    
	if ((mp->p2[0] - mp->p3[0]) * (mp->p0[1] - mp->p3[1]) -
		(mp->p0[0] - mp->p3[0]) * (mp->p2[1] - mp->p3[1]) < 0) {
        if (hit_triangle(P, mp->p3, mp->p2, mp->p0, &hit_data.u, &hit_data.v)){
            hit_data.tri = 1;
            *hit = &hit_data;
            return 1;
        }
    }

    return 0;
}

static void
get_color_and_depth(ZB_GRID *grid, ZB_HIT *hit, int u, int v,
    double color[4], double *depth)
{
    int i;
    
    i = v * grid->ns + u;
    zb_grid_shade_point(grid, grid->material, u,   v,   grid->c[i]);
    color[0] = grid->c[i][0];
    color[1] = grid->c[i][1];
    color[2] = grid->c[i][2];
    color[3] = grid->c[i][3];
    *depth = grid->s[i][2];
}

static void
get_interpolated_color_and_depth(ZB_GRID *grid, ZB_HIT *hit, int u, int v,
    double color[4], double *depth)
{
    int i0, i1, i2, i3;
    double (*c)[4];
    double g;

    i0 = v * grid->ns + u;
    i1 = i0 + 1;
    i2 = i1 + grid->ns;
    i3 = i2 - 1;

    zb_grid_shade_point(grid, grid->material, u,   v,   grid->c[i0]);
    zb_grid_shade_point(grid, grid->material, u+1, v,   grid->c[i1]);
    zb_grid_shade_point(grid, grid->material, u+1, v+1, grid->c[i2]);
    zb_grid_shade_point(grid, grid->material, u,   v+1, grid->c[i3]);
    
    g = 1 - (hit->u + hit->v);
    
    c = grid->c;
    if (hit->tri == 0) {
        /* interpolate barycentric from p0->p1->p2 */
        color[0] = hit->u * c[i1][0] + hit->v * c[i0][0] + g * c[i2][0];
        color[1] = hit->u * c[i1][1] + hit->v * c[i0][1] + g * c[i2][1];
        color[2] = hit->u * c[i1][2] + hit->v * c[i0][2] + g * c[i2][2];
        color[3] = hit->u * c[i1][3] + hit->v * c[i0][3] + g * c[i2][3];
        *depth = hit->u * grid->s[i1][2] +
            hit->v * grid->s[i0][2] + g * grid->s[i2][2];
    } else {
        /* interpolate p0->p2->p3 */
        color[0] = hit->u * c[i2][0] + hit->v * c[i0][0] + g * c[i3][0];
        color[1] = hit->u * c[i2][1] + hit->v * c[i0][1] + g * c[i3][1];
        color[2] = hit->u * c[i2][2] + hit->v * c[i0][2] + g * c[i3][2];
        color[3] = hit->u * c[i2][3] + hit->v * c[i0][3] + g * c[i3][3];
        *depth = hit->u * grid->s[i2][2] +
            hit->v * grid->s[i0][2] + g * grid->s[i3][2];
    }
}

/* Tests if the point P is inside the triangle and returns the
   barycentric coordinates (u, v). */
static int
hit_triangle(double P[2], double V0[2], double V1[2], double V2[2],
    double *u, double *v)
{
	double d;
	double E1[2], E2[2], T[2];

	V2_SUB(V1, V0, E1);
	V2_SUB(V2, V0, E2);

	V2_SUB(P, V0, T);

	d = V2_2X2_DET(E1, E2);

	/* snapping vertices in TESS avoids degenerate triangles */
	if (d == 0) {
		/* triangle is degenerate */
		return 0;
	}
    
	/* invert to avoid extra division */
	d = 1 / d;

	*u = d * V2_2X2_DET(T, E2);

	if (*u < 0 || *u > 1) {
		/* outside triangle */
		return 0;
	}

	*v = d * V2_2X2_DET(E1, T);

	if (*v < 0 || *u + *v > 1) {
		/* outside triangle */
		return 0;
	}
	
	return 1;
}

static ZB_MP *
zb_grid_get_interpolated_mp(ZB_GRID *grid, int u, int v)
{
    int i;
    static ZB_MP mp;

    /* 911: this routine is redundant and should be removed */
    i = v * grid->ns + u;
    mp.p0 = grid->f[i];
    ++i;
    mp.p1 = grid->f[i];
    i += grid->ns;
    mp.p2 = grid->f[i];
    i -= 1;
    mp.p3 = grid->f[i];

    /* check to see if part of mp crosses behind camera? */
    if (mp.p0[2] < -1 || mp.p1[2] < -1 || mp.p2[2] < -1 || mp.p3[2] < -1) {
        return NULL;
    }

    return &mp;
}

static void
linearly_interpolate(double *t0, double *t1, double t, double *d)
{
    d[0] = t0[0] + t * (t1[0] - t0[0]);
    d[1] = t0[1] + t * (t1[1] - t0[1]);
    d[2] = t0[2] + t * (t1[2] - t0[2]);
}

ZB_LOCATION
zb_grid_compare_viewports(ZB_GRID *grid, int viewport[4])
{
    return zb_compare_viewports( grid->bbox[0], grid->bbox[1], grid->bbox[2],
        grid->bbox[3], viewport[0], viewport[1], viewport[2], viewport[3]);
}

int
zb_grid_viewport_contained(ZB_GRID *grid, int viewport[4])
{
    return zb_viewport_contained( grid->bbox[0], grid->bbox[1], grid->bbox[2],
        grid->bbox[3], viewport[0], viewport[1], viewport[2], viewport[3]);
}

int
zb_grid_below_viewport_top(ZB_GRID *grid, int top)
{
    return grid->bbox[3] < top;
}
