/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_pixel.c,v 1.13 2002/01/03 05:43:26 wex Exp $
 *
 * $Log: zb_pixel.c,v $
 * Revision 1.13  2002/01/03 05:43:26  wex
 * micropolygon interpolation controls
 *
 * Revision 1.12  2001/12/21 07:18:44  wex
 * Switched to tables from lists to avoid memory fragmentation
 *
 * Revision 1.11  2001/12/16 00:28:34  wex
 * Fixes to frustum culling
 *
 * Revision 1.10  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.9  2001/11/26 21:27:57  wex
 * Place alpha into output image
 *
 * Revision 1.8  2001/11/13 11:46:27  wex
 * Removed ZB_GRID_DATA, ZB_GEOMETRY_DATA, and tile grid and geometry lists
 *
 * Revision 1.7  2001/11/12 18:52:08  wex
 * Cleanup for Python1.5
 *
 * Revision 1.6  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.5  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.4  2001/11/06 06:52:38  wex
 * Motion blur, intersection cleanup
 *
 * Revision 1.3  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:29  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "zb_pixel.h"

#define ZB_SURFACE_TABLE_BLOCK_SIZE       3

#define ZB_SURFACE_IS_INVALID(surface) ((surface)->color[3] < 0)
#define ZB_SURFACE_INVALIDATE(surface) ((surface)->color[3] = -1)

typedef struct zb_surface {
    float depth;
    float color[4];
} ZB_SURFACE;

struct zb_pixel {
    ZB_SURFACE *surface_table;
    unsigned short table_count; /* number of surfaces in table */
    unsigned short table_size;  /* total size of table */
    double t;                   /* sample time */
    double x, y;                /* sample position offset within pixel */
    double u, v;                /* sample location on lens */
};

struct zb_filter {
    ZB_FILTER_TYPE type;
    double *w;
    int r;
    int ss;
};

static void zb_surface_initialize(ZB_SURFACE *surface, double color[4],
    double depth);
static void zb_pixel_filter(ZB_PIXEL *pixel, IMAGE *output, ZB_FILTER *filter);
static int surface_depth_compare(ZB_SURFACE *a, ZB_SURFACE *b);
static void filter_color(IMAGE *output, int x, int y, double r, double g,
    double b, double a, ZB_FILTER *filter);
static double kernel(ZB_FILTER_TYPE type, double x, double y, int r);
static double *zb_filter_create_weights(ZB_FILTER_TYPE type, int r, int ss);
static double zb_rand(void);



ZB_PIXEL *
zb_pixel_create_array(int count, int viewport[4])
{
    int i, x, y;
    ZB_PIXEL *pixel;

    pixel = (ZB_PIXEL *)malloc(count * sizeof(ZB_PIXEL));
    if (pixel == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    for (y = viewport[1], i = 0; y <= viewport[3]; y++) {
        for (x = viewport[0]; x <= viewport[2]; x++, i++) {
            pixel[i].surface_table = NULL;
            pixel[i].table_count = 0;
            pixel[i].table_size = 0;
            pixel[i].t = zb_rand();
            pixel[i].x = x + zb_rand();
            pixel[i].y = y + zb_rand();
            do {
                pixel[i].u = zb_rand();
                pixel[i].v = zb_rand();
            } while (pixel[i].u * pixel[i].u + pixel[i].v * pixel[i].v > 1);
        }
    }
    
    return pixel;
}

void
zb_pixel_destroy_array(ZB_PIXEL *pixel, int count)
{
    int i;

    for (i = 0; i < count; i++) {
        if (pixel[i].surface_table != NULL) {
            free(pixel[i].surface_table);
            pixel[i].surface_table = NULL;
        }
        pixel[i].table_count = 0;
        pixel[i].table_size = 0;
    }
    
    free(pixel);
}

double
zb_pixel_get_time(ZB_PIXEL *pixel)
{
    return pixel->t;
}

void
zb_pixel_get_offset_position(ZB_PIXEL *pixel, double *x, double *y)
{
    *x = pixel->x;
    *y = pixel->y;
}

void
zb_pixel_get_lens_offset(ZB_PIXEL *pixel, double *u, double *v)
{
    *u = pixel->u;
    *v = pixel->v;
}

void
zb_pixel_filter_array(ZB_PIXEL *pixel, int count, IMAGE *output,
    ZB_FILTER *filter)
{
    int i;

    for (i = 0; i < count; i++) {
        zb_pixel_filter(&pixel[i], output, filter);
    }
}

static void
zb_pixel_filter(ZB_PIXEL *pixel, IMAGE *output, ZB_FILTER *filter)
{
    ZB_SURFACE *surface;
    double p[4], a;
    int i;
    
    /* go through surface list and composite */
    if (pixel->table_count > 1) {
        qsort(pixel->surface_table, pixel->table_count, sizeof(ZB_SURFACE),
            (int (*)(const void*,const void*))surface_depth_compare);
    }
    
    p[0] = p[1] = p[2] = p[3] = 0;                   /* accumulated color */

    for (i = 0; i < pixel->table_count; i++) {
        surface = &pixel->surface_table[i];

        if (ZB_SURFACE_IS_INVALID(surface)) {
            continue;
        }
        
        /* this is suspiciously wrong... */
        a = surface->color[3];
        p[0] += (1 - p[3]) * a * surface->color[0];
        p[1] += (1 - p[3]) * a * surface->color[1];
        p[2] += (1 - p[3]) * a * surface->color[2]; 
        
        p[3] = 1 - (1-p[3])*(1-surface->color[3]);  /* multiply transparency */
    }

/*    if (p[3] > 0 || p[0] != 0 || p[1] != 0 || p[2] != 0) { */
        if (p[0] < 0) p[0] = 0;
        if (p[1] < 0) p[1] = 0;
        if (p[2] < 0) p[2] = 0;
        if (p[0] > 1) p[0] = 1;
        if (p[1] > 1) p[1] = 1;
        if (p[2] > 1) p[2] = 1;

        filter_color(output, pixel->x, pixel->y, p[0], p[1], p[2],p[3],filter);
/*    }*/
}

static void
filter_color(IMAGE *output, int x, int y, double r, double g, double b,
    double a, ZB_FILTER *filter)
{
    int i, px, py, ss;
    double w;

    ss = filter->ss;
    for (py = y - filter->r, i = 0; py <= y + filter->r; py++) {
        for (px = x - filter->r; px <= x + filter->r; px++, i++) {
            w = filter->w[i];
            if (w == 0) continue;
            IMAGE_accumulate_color(output, px/ss, py/ss, r*w, g*w, b*w, a*w);
        }
    }
}

ZB_FILTER *
zb_filter_create(ZB_FILTER_TYPE type, int r, int ss)
{
    ZB_FILTER *filter;

    assert(r > 0 && r < 10);
    assert(ss > 0 && ss < 65);
    
    filter = (ZB_FILTER *)malloc(sizeof(ZB_FILTER));
    if (filter == NULL) {
        return NULL;
    }

    filter->w = zb_filter_create_weights(type, r, ss);
    filter->ss = ss;
    filter->r = r * ss;     /* scale by supersample rate */
    filter->type = type;
    
    return filter;
}

static double *
zb_filter_create_weights(ZB_FILTER_TYPE type, int r, int ss)
{
    int i, n, x, y;
    double *w, total;

    r *= ss;
    
    n = (2 * r + 1) * (2 * r + 1);
    
    w = (double *)malloc(n * sizeof(double));
    if (w == NULL) {
        return NULL;
    }
    
    total = 0;
    for (y = -r, i = 0; y <= r; y++) {
        for (x = -r; x <= r; x++, i++) {
            w[i] = kernel(type, x + 0.5, y + 0.5, r);
            assert(w[i] >= 0);
            total += w[i];
            assert(i < n);
        }
    }

    assert(total > 0);
    total *= ss * ss; 
    for (i = 0; i < n; i++) {
        w[i] /= total;
        assert(w[i] >= 0 && w[i] < 1);
    }

    return w;
}

void
zb_filter_set_supersample_rate(ZB_FILTER *filter, int ss)
{
    int old_r;
    assert(ss > 0 && ss < 64);
    old_r = filter->r / filter->ss;
    filter->ss = ss;
    free(filter->w);
    filter->w = zb_filter_create_weights(filter->type, old_r, ss);
}

void
zb_filter_set_radius(ZB_FILTER *filter, int r)
{
    free(filter->w);
    filter->w = zb_filter_create_weights(filter->type, r, filter->ss);
    filter->r = r * filter->ss;
}

void
zb_filter_set_kernel(ZB_FILTER *filter, ZB_FILTER_TYPE type, int r)
{
    free(filter->w);
    filter->w = zb_filter_create_weights(type, r, filter->ss);
    filter->type = type;
    filter->r = r * filter->ss;
}

void
ZB_set_filter_kernel(ZB *zb, ZB_FILTER_TYPE type, int r)
{
    zb_filter_set_kernel(zb->filter, type, r);
}

static double
kernel(ZB_FILTER_TYPE type, double x, double y, int r)
{
    double d;
    
    switch (type) {
    case ZB_FILTER_BOX:
    default:
        if (fabs(x) > r || fabs(y) > r)
            return 0;
        return 1;
    case ZB_FILTER_TRIANGLE:
        x = fabs(x);
        y = fabs(y);
        d = (x > y) ? x : y;
        if (d >= r) return 0;
        return 1.0 - d / r;
    case ZB_FILTER_GAUSSIAN:
        d = sqrt(x * x + y * y);
        if (d > r) return 0;
        d /= r;
        return exp(-4.05 * d * d);          /* 1.8 * true x^2 */
    }

    return 1;
}

static int
surface_depth_compare(ZB_SURFACE *a, ZB_SURFACE *b)
{
    return a->depth > b->depth;
}

void
zb_pixel_insert_surface(ZB_PIXEL *pixel, double color[4], double depth)
{
    int i, inserted, invalid_idx;
    ZB_SURFACE *surface;

    assert(depth > -1);

    inserted = 0;
    invalid_idx = -1;

    /* go from end to beginning so we can trim the list if possible */
    for (i = pixel->table_count - 1; i >= 0; i--) {
        surface = &pixel->surface_table[i];

        if (ZB_SURFACE_IS_INVALID(surface)) {
            assert(i < pixel->table_count - 1);   /* must be in middle */
            invalid_idx = i;
            continue;
        }
        
        if (surface->depth < depth) {
            if (surface->color[3] > 0.99) {
                /* occluded by opaque object */
                return;
            }
            /* move on to next surface */
            continue;
        }

        /* new surface is in front of surface in list */
        if (color[3] > 0.99) {
            /* new surface occludes this surface */
            if (!inserted) {
                inserted = 1;
                zb_surface_initialize(surface, color, depth);
            } else {
                if (i == pixel->table_count - 1) {
                    /* trim the size of the table whenever possible */
                    --pixel->table_count;
                }
                ZB_SURFACE_INVALIDATE(surface);
            }
        } 
    }

    if (!inserted) {
        if (invalid_idx >= 0) {
            /* use an open slot in the table */
            surface = &pixel->surface_table[invalid_idx];
        } else if (pixel->table_count >= pixel->table_size) {
            /* allocate some more space in the surface table */
            if (pixel->table_size == 0) {
                pixel->table_size = 2;
            } else if (pixel->table_size < 32) {
                pixel->table_size *= 2;
            } else {
                pixel->table_size += 8;
            }
            pixel->surface_table = realloc(pixel->surface_table,
                pixel->table_size * sizeof(ZB_SURFACE));
            assert(pixel->surface_table != NULL);
            surface = &pixel->surface_table[pixel->table_count];
            ++pixel->table_count;
        } else {
            /* already some extra space in the table */
            surface = &pixel->surface_table[pixel->table_count];
            ++pixel->table_count;
        }

        zb_surface_initialize(surface, color, depth);
    }
}

static void
zb_surface_initialize(ZB_SURFACE *surface, double color[4], double depth)
{
    assert(depth > -1);
    
    surface->depth = depth;
    surface->color[0] = color[0];
    surface->color[1] = color[1];
    surface->color[2] = color[2];
    surface->color[3] = color[3];
}

ZB_PIXEL *
zb_pixel_array_get(ZB_PIXEL *pixel, int i)
{
    return &pixel[i];
}

 
static double
zb_rand(void)
{
    return (double)rand() / RAND_MAX;
}
