/* Copyright (c) 2001, Daniel Wexler, All Rights Reserved */
/*
 * $Header: /usr/local/cvsroot/flarg/lib/zb/zb_tile.c,v 1.10 2001/12/05 10:40:44 wex Exp $
 *
 * $Log: zb_tile.c,v $
 * Revision 1.10  2001/12/05 10:40:44  wex
 * Displacement, transparency, viewport scaling
 *
 * Revision 1.9  2001/11/13 18:32:34  wex
 * Memory management and reference counting fixes.
 *
 * Revision 1.8  2001/11/13 11:46:27  wex
 * Removed ZB_GRID_DATA, ZB_GEOMETRY_DATA, and tile grid and geometry lists
 *
 * Revision 1.7  2001/11/13 07:16:07  wex
 * Minor memory and code cleanup.
 *
 * Revision 1.6  2001/11/12 09:14:12  wex
 * depth of field, shading rate, bug fixes and optimizations
 *
 * Revision 1.5  2001/11/10 06:00:39  wex
 * Support for camera window, supersampling and filtering.
 *
 * Revision 1.4  2001/11/04 08:14:59  wex
 * Changes to support material bindings.
 *
 * Revision 1.3  2001/11/01 04:26:59  wex
 * Working with Python scene description and geometry shaders
 *
 * Revision 1.2  2001/10/29 00:50:29  wex
 * Fixed -Wall warnings.
 *
 * Revision 1.1  2001/10/26 15:48:42  wex
 * First working version
 *
 *
 */

#include <stdio.h>
#include <assert.h>

#include "zb_tile.h"
#include "zb_camera.h"
#include "zb_grid.h"
#include "zb_geometry.h"
#include "zb_pixel.h"

/* defined in subpixels, so typically 128/4=32 pixels */
#define ZB_TILE_SIZE 64


struct zb_tile {
    LIST_ELEMENT list_element;  /* tiles are stored in a global list in ZB */
    ZB *zb;                     /* backpointer */
    int viewport[4];            /* [xmin, ymin, xmax, ymax] */
    ZB_PIXEL *pixel;            /* array of pixels */
};

typedef struct zb_tile_capture_data {
    ZB_TILE *tile;
    ZB_GEOMETRY *geometry;
    LIST *active_list;
} ZB_TILE_CAPTURE_DATA;

static ZB_TILE *zb_tile_create(ZB *zb, int xmin, int ymin, int xmax, int ymax);
static void zb_tile_handle_grid(ZB_TILE *tile, ZB_GRID *grid,
    LIST *active_list);


int
zb_tile_initialize(ZB *zb)
{
    int w, h;
    int x, y;
    int nx, ny;
    int xmin, ymin, xmax, ymax;
    ZB_TILE *tile;

    w = zb_camera_get_width(zb->camera);
    nx = w / ZB_TILE_SIZE + ((w % ZB_TILE_SIZE) != 0);
    h = zb_camera_get_height(zb->camera);
    ny = h / ZB_TILE_SIZE + ((h % ZB_TILE_SIZE) != 0);

    for (y = 0; y < ny; y++) {
        ymin = y * ZB_TILE_SIZE;
        assert(ymin < h);
        ymax = ymin + ZB_TILE_SIZE - 1;
        if (ymax >= h) {
            ymax = h - 1;
        }
        
        for (x = 0; x < nx; x++) {
            /* compute the viewport without overlap */
            xmin = x * ZB_TILE_SIZE;
            xmax = xmin + ZB_TILE_SIZE - 1;
            if (xmax >= w) {
                xmax = w - 1;
            }
        
            tile = zb_tile_create(zb, xmin, ymin, xmax, ymax);
            if (tile == NULL) {
                return -1;
            }
        }
    }

    return 0;
}

static ZB_TILE *
zb_tile_create(ZB *zb, int xmin, int ymin, int xmax, int ymax)
{
    ZB_TILE *tile;

    tile = (ZB_TILE *)malloc(sizeof(ZB_TILE));
    if (tile == NULL) {
        ZB_errno = ZB_ERR_MEMORY;
        return NULL;
    }

    tile->zb = zb;
    tile->viewport[0] = xmin;
    tile->viewport[1] = ymin;
    tile->viewport[2] = xmax;
    tile->viewport[3] = ymax;
    tile->pixel = NULL;

    LIST_append(zb->tile_list, tile);
        
    return tile;
}

int
zb_tile_render(ZB_TILE *tile)
{
    ZB_GEOMETRY *geometry, *next_geometry;
    ZB_GRID *grid, *next_grid;
    LIST *active_geometry_list;
    LIST *active_grid_list;
    int viewport[4];
    IMAGE_NOTIFY_INFO event;
    
    /* clip all remaining geometry against this tile,
         anything above and/or to the left is completely finished, destroy
         anything else outside gets moved to list for next tile
         anything touching this tile is placed in "active" list
    */
    active_geometry_list = LIST_create();
    geometry = LIST_head(tile->zb->geometry_list);
    while (geometry != NULL) {

        next_geometry = LIST_next(geometry);  /* before we delete */
        
        switch (zb_geometry_compare_viewports(geometry, tile->viewport)) {
            
        case ZB_BBOX_ABOVE:
            /* ignore this object for this tile */
            break;
            
        case ZB_BBOX_RIGHT:
            /* ignore this object for this tile */
            break;

        case ZB_BBOX_LEFT:
            /* only remove if the top of the geometry object is
               below the top of this tile's viewport otherwise it
               may be required for the next row of tiles */
            if (zb_geometry_below_viewport_top(geometry, tile->viewport[3])) {
                /* remove from remaining list */
                LIST_delete(tile->zb->geometry_list, geometry);
                zb_geometry_destroy(geometry);
            }
            /* ignored left */
            break;

        case ZB_BBOX_BELOW:
            /* remove from remaining list */
            LIST_delete(tile->zb->geometry_list, geometry);
            zb_geometry_destroy(geometry);
            break;
            
        case ZB_BBOX_INTERSECT:
            /* remove from remaining and add to the active list */
            LIST_delete(tile->zb->geometry_list, geometry);
            LIST_append(active_geometry_list, geometry);
            break;
        }

        /* advance to next item accounting for possible deletion */
        geometry = next_geometry;
    }

    /* go through remaining grids from previous tiles and compare
       them to this tile's viewport, if they are above and/or to the
       left, destroy them, if they intersect this tile, add them to
       the active list, and if they are below and to the right, then
       leave them on the list. */
    active_grid_list = LIST_create();
    grid = LIST_head(tile->zb->grid_list);
    while (grid != NULL) {
        next_grid = LIST_next(grid);
        zb_tile_handle_grid(tile, grid, active_grid_list);
        grid = next_grid;
    }

    /* setup to handle any active geometry */
    geometry = LIST_pop(active_geometry_list);
    
    do {
/*        fprintf(stderr, "Tile [%d, %d, %d, %d]  geom = %d (act)  %d (rem) "
            "grid = %d (act)  %d (rem)\n", tile->viewport[0],
            tile->viewport[1], tile->viewport[2], tile->viewport[3],
            LIST_count(active_geometry_list),
            LIST_count(tile->zb->geometry_list),
            LIST_count(active_grid_list), LIST_count(tile->zb->grid_list));
*/      
        /* render any grids on the active list (previous tiles or geometry) */
        while ((grid = LIST_pop(active_grid_list)) != NULL) {
            zb_grid_render(grid, tile, tile->viewport);
            if (zb_grid_viewport_contained(grid, tile->viewport)) {
                zb_grid_destroy(grid);
            } else {
                /* place this back on the remaining list for other tiles */
                LIST_append(tile->zb->grid_list, grid);
            }
        }

        /* render all active geometry objects by converting them to
           grids (if they haven't already been converted) and then
           rendering the grids as they are generated.  Grids that are
           above and to the left of the current tile are destroyed,
           as are grids that lie completely within the tile, while
           grids that hit other tiles to the left and below are placed
           into the remaining grid list in the global view */
        
        if (geometry != NULL) {     /* (there may not be any active objects) */
            ZB_TILE_CAPTURE_DATA data;

            next_geometry = LIST_pop(active_geometry_list);
            data.tile = tile;
            data.active_list = active_grid_list;
            data.geometry = geometry;
            zb_geometry_create_grids(geometry, &data);
            if (zb_geometry_viewport_contained(geometry, tile->viewport)) {
                zb_geometry_destroy(geometry);
            } else {
                /* place this back on the remaining list for other tiles */
                LIST_append(tile->zb->geometry_list, geometry);
            }

            /* advance to next geometry object */
            geometry = next_geometry;
        }

    } while (geometry != NULL || !LIST_is_empty(active_grid_list));

    assert(LIST_is_empty(active_geometry_list));
    assert(LIST_is_empty(active_grid_list));
    LIST_destroy(active_geometry_list);
    LIST_destroy(active_grid_list);

    if (tile->pixel != NULL) {
        zb_pixel_filter_array(tile->pixel, zb_tile_get_pixel_count(tile),
            tile->zb->rgba, tile->zb->filter);
    }

    zb_camera_get_image_viewport(tile->zb->camera, tile->viewport, viewport);
    IMAGE_notify_update_viewport(tile->zb->rgba, viewport);
    
    while (IMAGE_notify_get_event(tile->zb->rgba, (int)tile->zb, &event) == 0){
        if (event.type == IMAGE_NOTIFY_STOP) {
            return -1;
        }
    }
    
    return 0;
}

int
zb_grid_capture(GRID_INFO *info)
{
    ZB_GRID *grid;
    ZB_TILE_CAPTURE_DATA *data = info->user_data;
    
    ++data->tile->zb->captured_grids;
    grid = zb_grid_create(data->tile->zb, info->grid, data->geometry);
    zb_tile_handle_grid(data->tile, grid, data->active_list);
    
    return 0;
}

static void
zb_tile_handle_grid(ZB_TILE *tile, ZB_GRID *grid, LIST *active_list)
{
    switch (zb_grid_compare_viewports(grid, tile->viewport)) {
    case ZB_BBOX_ABOVE:
    case ZB_BBOX_RIGHT:
        /* ignore this object for this tile but leave for later */
        return;

    case ZB_BBOX_LEFT:
        if (zb_grid_below_viewport_top(grid, tile->viewport[3])) {
            LIST_delete(tile->zb->grid_list, grid);
            zb_grid_destroy(grid);
            return;
        }
        /* ignore to left but hits next row */
        return;
        
    case ZB_BBOX_BELOW:
        /* remove from remaining list */
        LIST_delete(tile->zb->grid_list, grid);
        zb_grid_destroy(grid);
        return;
        
    case ZB_BBOX_INTERSECT:
        /* remove from remaining and add to the active list */
        LIST_delete(tile->zb->grid_list, grid);
        LIST_append(active_list, grid);
        return;
    }
}

void
zb_tile_destroy(ZB_TILE *tile)
{
    tile->zb = NULL;
    
    if (tile->pixel != NULL) {
        zb_pixel_destroy_array(tile->pixel, zb_tile_get_pixel_count(tile));
        tile->pixel = NULL;
    }
    
    tile->viewport[0] = 0;
    tile->viewport[1] = 0;
    tile->viewport[2] = 0;
    tile->viewport[3] = 0;
    
    free(tile);
}

int
zb_tile_get_pixel_count(ZB_TILE *tile)
{
    return (tile->viewport[2] - tile->viewport[0] + 1) *
        (tile->viewport[3] - tile->viewport[1] + 1);
}

ZB_PIXEL *
zb_tile_get_pixel(ZB_TILE *tile, int i)
{
    assert(i >= 0 && i < zb_tile_get_pixel_count(tile));
    return zb_pixel_array_get(tile->pixel, i);
}

int
zb_tile_create_pixels(ZB_TILE *tile)
{
    if (tile->pixel != NULL) return 0;
    
    tile->pixel = zb_pixel_create_array(zb_tile_get_pixel_count(tile),
        tile->viewport);
    if (tile->pixel == NULL) {
        return -1;
    }

    return 0;
}

ZB *
zb_tile_get_zb(ZB_TILE *tile)
{
    return tile->zb;
}
