#!/usr/bin/env python

import os, sys
from distutils.core import setup, Extension
from distutils import util

# Notes:
#
#   On Linux, must make symlinks from the build directory's shared
#   library for each of our modules into the current directory before
#   the build begins.  I don't know how to do this in a platform
#   independent manner yet.
#
#   On Windows, must copy the .lib files from the build temp directory
#   to the current directory after each library is built.  We also
#   need to have dummy initmodule functions since the \EXPORT initfoo
#   is always added to the command line.
#
#   Earlier versions of Python (<2.0?) do not provide a PyObject_INIT
#   function.  This is used in the matrix c module to reduce memory
#   fragmentation using the POOL library.  You can compile for these
#   earlier versions of Python by removing the MATRIX_USE_POOL define.
#
#   On Windows, must add Pthread stuff, and a define WIN32
#   Must put pthreadVC.dll in with other pyd files in parc.

def relink(dir, lib):
    try:
        os.remove(lib)
    except:
        print "Cannot remove file"
        
    os.symlink(dir+lib, lib)

    
if util.get_platform() == "win32":
    lib_suffix = ".lib"
    pthread_lib = "lib/PTHREADS/Pre-built/lib/pthreadVC.lib"
    pthread_include = "lib/PTHREADS/Pre-built/include"
    matrix_define_macros = [('WIN32', None), ('MATRIX_USE_POOL', None)]
    image_extra_objects = ["list" + lib_suffix, pthread_lib]
    image_include_dirs = ["lib/list", pthread_include]
    ip_extra_objects = ["list" + lib_suffix,
                        "imagecmodule" + lib_suffix,
                        pthread_lib, "lib/FreeType/freetype204MT.lib"]
    ip_define_macros = [('WIN32', None), ('IP_FREETYPE_ENABLED', None)]
    ip_libraries = []
    ip_include_dirs=["lib/list", "lib/image", "lib/FreeType/include"]
    geometry_define_macros=[('WIN32', None)]
    shader_define_macros=[('WIN32', None)]
    zb_define_macros=[('WIN32', None)]
    tui_libraries = ["GLU32", "OpenGL32", "odbccp32", "user32", "gdi32"]
    tui_include_dirs = ["lib/list", "lib/image", "lib/zb",
                        "lib/shader", "lib/geometry", "lib/matrix",
                        pthread_include]
    tui_extra_objects = ["imagecmodule" + lib_suffix,
                         "list" + lib_suffix,
                         pthread_lib]
    tui_define_macros=[('TUI_PYTHON', None),
                       ('TUI_WIN32', None),
                       ('WIN32', None)]
    
else :
    lib_suffix = ".so"
    matrix_define_macros = [('MATRIX_USE_POOL', None)]
    image_extra_objects = ["list" + lib_suffix]
    image_include_dirs = ["lib/list"]
    ip_extra_objects = ["list" + lib_suffix,
                        "imagecmodule" + lib_suffix]
    ip_libraries = ["freetype"]
    ip_define_macros = [('IP_FREETYPE_ENABLED', None)]
    ip_include_dirs=["lib/list", "lib/image", "/usr/include/freetype2"]
    geometry_define_macros=[]
    shader_define_macros=[]
    zb_define_macros=[]
    tui_libraries = ["GLU", "GL", "pthread", "X11", "Xi", "Xext", "Xmu", "m"]
    tui_include_dirs = ["lib/list", "lib/image", "lib/zb",
                        "lib/shader", "lib/geometry", "lib/matrix"]
    tui_extra_objects = ["imagecmodule" + lib_suffix, "list" + lib_suffix]
    tui_define_macros=[('TUI_PYTHON', None), ('TUI_XWIN', None)]

    # Setup links to the .so files
    build_dir = os.path.join('build', 'lib.%s-%s/parc/' %
                             (util.get_platform(), sys.version[0:3]))
    relink(build_dir, "pool.so")
    relink(build_dir, "list.so")
    relink(build_dir, "matrixcmodule.so")
    relink(build_dir, "geometrycmodule.so")
    relink(build_dir, "shadercmodule.so")
    relink(build_dir, "imagecmodule.so")
    relink(build_dir, "ipcmodule.so")


setup(name="Parc",
      version="0.5",
      description="Tools for creating photorealistic animation effects",
      author="Daniel Wexler",
      author_email="wex@flarg.com",
      url="http://www.flarg.com/parc",
      package_dir = {'parc': 'lib'},
      packages=['parc',
                'parc.matrix',
                'parc.image',
                'parc.ip',
                'parc.geometry',
                'parc.shader',
                'parc.zb',
                'parc.tui'],
      ext_modules=[Extension("parc.list",
                             ["lib/list/list.c",
                              "lib/list/list_error.c"],
                             export_symbols=["LIST_error_string",
                                             "LIST_create",
                                             "LIST_init",
                                             "LIST_destroy",
                                             "LIST_append",
                                             "LIST_prepend",
                                             "LIST_delete",
                                             "LIST_insert_before",
                                             "LIST_next",
                                             "LIST_prev",
                                             "LIST_head",
                                             "LIST_tail",
                                             "LIST_pop",
                                             "LIST_count",
                                             "LIST_is_empty",
                                             "LIST_apply",
                                             "LIST_sort"]),
                   Extension("parc.pool",
                             ["lib/pool/pool.c",
                              "lib/pool/pool_error.c"],
                             include_dirs=["lib/list"], 
                             extra_objects=["list" + lib_suffix],
                             export_symbols=["POOL_create",
                                             "POOL_destroy",
                                             "POOL_create_item",
                                             "POOL_destroy_item"]),
                   Extension("parc.matrixcmodule",
                             ["lib/matrix/matrix.c",
                              "lib/matrix/matrix_xform.c",
                              "lib/matrix/matrix_invert.c",
                              "lib/matrix/matrix_error.c",
                              "lib/matrix/matrix_glue.c",
                              "lib/matrix/vector.c",
                              "lib/matrix/vector_transform.c"],
                             define_macros=matrix_define_macros,
                             include_dirs=["lib/list", "lib/pool"],
                             extra_objects=["list" + lib_suffix,
                                            "pool" + lib_suffix],
                             export_symbols=["MATRIX_create",
                                             "MATRIX_destroy",
                                             "MATRIX_check",
                                             "MATRIX_set_identity",
                                             "MATRIX_set_scale",
                                             "MATRIX_set_translate",
                                             "MATRIX_set_rotate",
                                             "MATRIX_set_lens_projection",
                                             "MATRIX_set_window_projection",
                                             "MATRIX_determinant",
                                             "MATRIX_copy",
                                             "MATRIX_multiply",
                                             "MATRIX_adjoint",
                                             "MATRIX_invert",
                                             "MATRIX_lookat",
                                             "MATRIX_scale",
                                             "MATRIX_translate",
                                             "MATRIX_rotate",
                                             "MATRIX_print",
                                             "VECTOR_create",
                                             "VECTOR_destroy",
                                             "VECTOR_check",
                                             "VECTOR_set",
                                             "VECTOR_length",
                                             "VECTOR_normalize",
                                             "VECTOR_scale",
                                             "VECTOR_add_scalar",
                                             "VECTOR_dot",
                                             "VECTOR_cross",
                                             "VECTOR_add",
                                             "VECTOR_print",
                                             "VECTOR_transform",
                                             "VECTOR_project",
                                             "VECTOR_3x3_transform",
                                             "vector_alloced",
                                             "vector_destroyed",
                                             "vector_max",
                                             "matrix_alloced",
                                             "matrix_destroyed",
                                             "matrix_max"]),
                   Extension("parc.imagecmodule",
                             ["lib/image/image.c",
                              "lib/image/image_error.c",
                              "lib/image/image_format.c",
                              "lib/image/image_glue.c",
                              "lib/image/image_notify.c",
                              "lib/image/image_pixel.c",
                              "lib/image/image_ppm.c",
                              "lib/image/image_read.c",
                              "lib/image/image_rgba.c",
                              "lib/image/image_support.c",
                              "lib/image/image_tile.c",
                              "lib/image/image_write.c"],
                             include_dirs=image_include_dirs,
                             extra_objects=image_extra_objects,
                             export_symbols=["IMAGE_support_all",
                                             "IMAGE_create",
                                             "IMAGE_incref",
                                             "IMAGE_destroy",
                                             "IMAGE_free",
                                             "IMAGE_read",
                                             "IMAGE_write",
                                             "IMAGE_check",
                                             "IMAGE_empty",
                                             "IMAGE_set_viewport",
                                             "IMAGE_set_color_defined",
                                             "IMAGE_set_alpha_defined",
                                             "IMAGE_set_depth_defined",
                                             "IMAGE_get_width",
                                             "IMAGE_get_height",
                                             "IMAGE_get_viewport",
                                             "IMAGE_is_valid_pixel",
                                             "IMAGE_get_viewport_wrapped_pixel",
                                             "IMAGE_set_color",
                                             "IMAGE_set_alpha",
                                             "IMAGE_set_depth",
                                             "IMAGE_get_color",
                                             "IMAGE_get_alpha",
                                             "IMAGE_get_depth",
                                             "IMAGE_get_bilerp_color",
                                             "IMAGE_accumulate_color",
                                             "IMAGE_update_depth",
                                             "IMAGE_create_rgba",
                                             "IMAGE_rgba_update_viewport",
                                             "IMAGE_add_notify_client",
                                             "IMAGE_remove_notify_client",
                                             "IMAGE_notify_update_viewport",
                                             "IMAGE_notify_get_event",
                                             "IMAGE_allocated_kilobytes",
                                             "IMAGE_error_string",
                                             "IMAGE_errno"]),
                   Extension("parc.ipcmodule",
                             ["lib/ip/ip_font.c",
                              "lib/ip/ip_op.c",
                              "lib/ip/ip_glue.c",
                              "lib/ip/ip_error.c",
                              "lib/ip/ip_fill.c",
                              "lib/ip/ip_gradient.c",
                              "lib/ip/ip_image.c",
                              "lib/ip/ip_alpha_from_lum.c",
                              "lib/ip/ip_map.c",
                              "lib/ip/ip_matte.c",
                              "lib/ip/ip_misc.c",
                              "lib/ip/ip_read.c"],
                             include_dirs=ip_include_dirs,
                             define_macros=ip_define_macros,
                             libraries=ip_libraries,
                             extra_objects=ip_extra_objects),
                   Extension("parc.geometrycmodule",
                             ["lib/geometry/geometry.c",
                              "lib/geometry/geometry_error.c",
                              "lib/geometry/geometry_glue.c",
                              "lib/geometry/grid.c",
                              "lib/geometry/subd.c",
                              "lib/geometry/vertex.c"],
                             include_dirs=["lib/matrix",
                                           "lib/list",
                                           "lib/pool"],
                             define_macros=geometry_define_macros,
                             extra_objects=["matrixcmodule" + lib_suffix,
                                            "pool" + lib_suffix,
                                            "list" + lib_suffix],
                             export_symbols= ["GEOMETRY_generate_grids",
                                              "geometry_alloced",
                                              "geometry_destroyed",
                                              "geometry_max",
                                              "vertex_alloced",
                                              "vertex_destroyed",
                                              "vertex_max"]),
                   Extension("parc.shadercmodule",
                             ["lib/shader/shader.c",
                              "lib/shader/shader_data.c",
                              "lib/shader/shader_error.c",
                              "lib/shader/shader_geometry.c",
                              "lib/shader/shader_glue.c",
                              "lib/shader/shader_light.c",
                              "lib/shader/shader_material.c"],
                             include_dirs=["lib/list", "lib/pool",
                                           "lib/matrix", "lib/geometry"],
                             extra_objects=["geometrycmodule" + lib_suffix,
                                            "matrixcmodule" + lib_suffix,
                                            "pool" + lib_suffix,
                                            "list" + lib_suffix],
                             define_macros=shader_define_macros,
                             export_symbols=["SHADER_GEOMETRY_create",
                                             "SHADER_LIGHT_create",
                                             "SHADER_GEOMETRY_destroy",
                                             "SHADER_LIGHT_destroy",
                                             "SHADER_MATERIAL_get_pobj",
                                             "SHADER_MATERIAL_set_pobj",
                                             "SHADER_MATERIAL_get_from_pobj",
                                             "SHADER_MATERIAL_check",
                                             "SHADER_MATERIAL_shade",
                                             "SHADER_destroy",
                                             "SHADER_GEOMETRY_get_l2w",
                                             "SHADER_GEOMETRY_get_l2w_motion",
                                             "SHADER_GEOMETRY_get_bbox",
                                             "SHADER_GEOMETRY_generate",
                                             "SHADER_get_double",
                                             "SHADER_get_vector",
                                             "SHADER_get_string",
                                             "SHADER_get_int",
                                             "SHADER_set_double",
                                             "SHADER_set_vector",
                                             "SHADER_set_string",
                                             "SHADER_set_int",
                                             "SHADER_get_light_list"]),
                   Extension("parc.zbcmodule",
                             ["lib/zb/zb.c",
                              "lib/zb/zb_camera.c",
                              "lib/zb/zb_dof.c",
                              "lib/zb/zb_error.c",
                              "lib/zb/zb_geometry.c",
                              "lib/zb/zb_glue.c",
                              "lib/zb/zb_grid.c",
                              "lib/zb/zb_pixel.c",
                              "lib/zb/zb_point.c",
                              "lib/zb/zb_render.c",
                              "lib/zb/zb_tile.c"],
                             define_macros=zb_define_macros,
                             include_dirs=["lib/list", "lib/pool",
                                           "lib/matrix", "lib/image",
                                           "lib/geometry", "lib/shader"],
                             extra_objects=["shadercmodule" + lib_suffix,
                                            "geometrycmodule" + lib_suffix,
                                            "imagecmodule" + lib_suffix,
                                            "matrixcmodule" + lib_suffix,
                                            "pool" + lib_suffix,
                                            "list" + lib_suffix]),
                   Extension("parc.tuicmodule",
                             ["lib/tui/tui.c",
                              "lib/tui/tui_button.c",
                              "lib/tui/tui_construct.c",
                              "lib/tui/tui_error.c",
                              "lib/tui/tui_filebrowser.c",
                              "lib/tui/tui_glue.c",
                              "lib/tui/tui_imv.c",
                              "lib/tui/tui_label.c",
                              "lib/tui/tui_mainloop.c",
                              "lib/tui/tui_menubar.c",
                              "lib/tui/tui_menuitem.c",
                              "lib/tui/tui_misc.c",
                              "lib/tui/tui_oglwindow.c",
                              "lib/tui/tui_pulldown.c",
                              "lib/tui/tui_textport.c",
                              "lib/tui/tui_widget.c",
                              "lib/tui/tui_window.c",
                              "lib/tui/tui_write.c"],
                             include_dirs=tui_include_dirs,
                             library_dirs=["/usr/X11R6/lib"],
                             libraries=tui_libraries,
                             define_macros=tui_define_macros,
                             extra_objects=tui_extra_objects)
                   ]
      )


